import org.gradle.api.GradleException
import org.gradle.api.tasks.Input
import org.gradle.api.tasks.InputDirectory
import org.gradle.api.tasks.TaskAction
import java.io.File

abstract class BuildAndroidNativeTask : BuildNativeLibrariesTask() {
    
    @get:InputDirectory
    lateinit var ndkDir: File

    @get:Input
    var minSdk: Int? = null
    
    @get:Input
    var abiFilters: Set<String>? = null

    @TaskAction
    fun buildLibraries() {
        if (getDesktopOs() != "linux") {
            throw GradleException("Android native library build tasks are only supported on Linux right now.")
        }
        openBuildDirectory()
        check(abiFilters != null) { "No ABI filters specified - please set \"abiFilters\" under android.defaultConfig.ndk" }
        check(minSdk != null) { "No minimum Android SDK set" }
        abiFilters?.forEach { abi -> buildForAbi(abi) }
    }
    
    private fun buildForAbi(abi: String) {
        val target: String
        val vpxTarget: String
        when (abi) {
            "armeabi-v7a" -> {
                target = "armv7a-linux-androideabi"
                vpxTarget = "armv7-android-gcc"
            }
            "arm64-v8a" -> {
                target = "aarch64-linux-android"
                vpxTarget = "arm64-android-gcc"
            }
            "x86" -> {
                target = "i686-linux-android"
                vpxTarget = "x86-android-gcc"
            }
            "x86_64" -> {
                target = "x86_64-linux-android"
                vpxTarget = "x86_64-android-gcc"
            }
            else -> error("Unsupported ABI: $abi")
        }

        println("ABI: $abi | target = $target | vpxTarget = $vpxTarget")

        val prefixDir = File(buildDir, "_android_prefix_$abi")
        val toolchain = File(ndkDir, "toolchains/llvm/prebuilt/linux-x86_64")
        val toolchainBin = File(toolchain, "bin/llvm")
        val sysrootDir = File(toolchain, "sysroot")

        val clang = "$toolchain/bin/$target$minSdk-clang"
        val pkgConfig = "$prefixDir/lib/pkgconfig"
        environment("CC", clang)
        environment("CXX", "$toolchain/bin/$target$minSdk-clang++")
        environment("AC", "$toolchain-ar")
        environment("LD", clang)
        //environment("AS", "$toolchainBin-as")
        environment("STRIP", "$toolchainBin-strip")
        environment("NM", "$toolchainBin-nm")
        environment("AR", "$toolchainBin-ar")
        environment("LDFLAGS", "-static-libstdc++")
        environment("PKG_CONFIG_LIBDIR", pkgConfig)
        environment("PKG_CONFIG_PATH", pkgConfig)

        if (!prefixDir.exists()) prefixDir.mkdirs()

        val buildParams = NativeBuildParameters(
            prefixDir = prefixDir,
            target = target,
            sysrootDir = sysrootDir,
            vpxTarget = vpxTarget,
            abi = abi,
            cmakeToolchainFile = File("$ndkDir/build/cmake/android.toolchain.cmake"),
            androidPlatform = "android-$minSdk",
        )
        buildNativeLibraries(buildParams)
    }
}

