import org.gradle.api.tasks.InputDirectory
import org.gradle.api.tasks.Internal
import org.gradle.api.tasks.OutputDirectory
import org.gradle.api.tasks.TaskAction
import java.io.File
import java.io.FileFilter

private const val DIR_JNI = "jni"
private val LIBRARY_EXTENSIONS = setOf("so", "dll", "dylib")
private val LIBRARY_FILTER = FileFilter { it.extension in LIBRARY_EXTENSIONS }

private const val MINGW_GCC_DLL = "libgcc_s_seh-1.dll"
private const val MINGW_THREAD_DLL = "libwinpthread-1.dll"
private const val MINGW_OPENMP_DLL = "libgomp-1.dll"

abstract class BuildDesktopJniTask : NativeBuildExecTask() {

    @get:InputDirectory
    lateinit var sourceDir: File
    
    @get:OutputDirectory
    lateinit var outputDir: File

    @get:Internal
    var cmakeJni: Boolean = false

    @get:Internal
    var openMp: Boolean = false

    @TaskAction
    fun buildLibraries() {
        val target = getDesktopTarget()
        val prefixPath = "_${getDesktopOs()}_prefix_${getDesktopArch()}"
        val dir = File(buildDir, DIR_JNI)
        if (dir.exists()) dir.deleteRecursively()
        dir.mkdirs()
        dir.useDirectory {
            val cmakeArgs = mutableMapOf<String, String>()
            cmakeArgs += "PREFIX_DIR" to prefixPath
            if (cmakeJni) {
                val jniIncludePath = "$javaHome/include"
                cmakeArgs += "JAVA_INCLUDE_PATH" to jniIncludePath
                cmakeArgs += "JAVA_INCLUDE_PATH2" to "$jniIncludePath/win32"
                cmakeArgs += "JAVA_AWT_LIBRARY" to "NotNeeded"
                cmakeArgs += "JAVA_JVM_LIBRARY" to "NotNeeded"
                cmakeArgs += "JAVA_AWT_INCLUDE_PATH" to "NotNeeded"
            }
            if (openMp) {
                cmakeArgs += "OPEN_MP" to "true"
            }
            if (mingwSysRootPath.isNotEmpty()) {
                cmakeArgs += "MINGW_PATH" to convertToUnixPath(mingwSysRootPath)
            }
            cmake(
                relative = "../../",
                args = cmakeArgs,
            )

            // Copy freshly baked DLLs into our app resources folder
            val outputDir = File(outputDir, "$target/lib")
            dir.listFiles(LIBRARY_FILTER)?.forEach {
                val outName = if (it.name.startsWith("cyg")) it.name.substringAfter("cyg") else it.name
                copyFile(it.name, outputDir, outName)
            }

            // MinGW required libraries
            if (mingwSysRootPath.isNotEmpty()) {
                outputDir.run {
                    copyMinGwLibrary(MINGW_GCC_DLL)
                    copyMinGwLibrary(MINGW_THREAD_DLL)
                    copyMinGwLibrary(MINGW_OPENMP_DLL)
                }
            }
        }
    }

    private fun File.copyMinGwLibrary(fileName: String) {
        val libraryFile = File(mingwSysRootPath, "bin/$fileName")
        if (!libraryFile.exists()) {
            error("Failed to find required library file: $libraryFile")
        }
        libraryFile.copyTo(File(this, fileName), overwrite = true)
    }
}