import java.io.File

private const val DIR_TOXCORE = "toxcore"
private const val DIR_SODIUM = "libsodium"
private const val DIR_VPX = "libvpx"
private const val DIR_GIF = "giflib"

// TODO - Read from the version catalog here, if possible
private const val VERSION_LIBSODIUM = "1.0.20-RELEASE"
private const val VERSION_OPUS = "v1.4"
private const val VERSION_OPUSFILE = "v0.12"
private const val VERSION_OPUSENC = "v0.2.1"
private const val VERSION_OGG = "v1.3.5"
private const val VERSION_LIBVPX = "v1.14.1"
private const val VERSION_TOXCORE = "v0.2.19"
private const val VERSION_GIFLIB = "5.2.2"

abstract class BuildNativeLibrariesTask : NativeBuildExecTask() {

    protected fun openBuildDirectory() {
        if (!buildDir.exists()) buildDir.mkdirs()
        workingDir(buildDir)
    }
    
    protected fun buildNativeLibraries(params: NativeBuildParameters) {
        buildSodiumLibrary(params)
        buildOggLibrary(params)
        buildOpusLibrary(params)
        buildOpusFileLibrary(params)
        buildOpusEncoderLibrary(params)
        buildVpxLibrary(params)
        buildToxCore(params)
        buildGifLibrary()
    }

    private fun buildSodiumLibrary(params: NativeBuildParameters) {
        if (prefixOutputExists(params, "sodium")) return

        println("Building libsodium...")

        val dir = initBuildDirectory(DIR_SODIUM) {
            gitClone(
                repo = "https://github.com/jedisct1/libsodium",
                depth = 1,
                branch = VERSION_LIBSODIUM,
            )
        }
        dir.useDirectory {
            gitClean()
            autoReconfigure()
            configure(
                prefixDir = params.prefixDir,
                args = buildArgs(
                    "disable-pie" to null,
                    "disable-shared" to null,
                ) {
                    params.target?.let { this += "host" to it }
                    params.sysrootDir?.let { this += "with-sysroot" to it.absolutePath }
                },
            )
            makeInstall()
        }
    }
    
    private fun buildXiphLibrary(
        repoUrl: String,
        branch: String,
        libName: String,
        params: NativeBuildParameters,
        args: Map<String, String?> = emptyMap(),
    ) {
        if (prefixOutputExists(params, libName)) return

        println("Building lib$libName...")

        val dir = initBuildDirectory(libName) {
            gitClone(
                repo = repoUrl,
                depth = 1,
                branch = branch,
                dirName = libName,
            )
        }
        dir.useDirectory {
            gitClean()
            autoReconfigure()
            configure(
                prefixDir = params.prefixDir,
                forceFpic = params.forceFpic,
                args = buildArgs(
                    "disable-shared" to null,
                ) {
                    params.target?.let { this += "host" to it }
                    params.sysrootDir?.let { this += "with-sysroot" to it.absolutePath }
                } + args
            )
            params.target?.let { fixXiphMakefile(it) }
            makeInstall()
        }
    }

    private fun buildOpusLibrary(params: NativeBuildParameters) = buildXiphLibrary(
        repoUrl = "https://github.com/xiph/opus",
        branch = VERSION_OPUS,
        libName = "opus",
        params = params,
    )
    
    private fun buildOpusFileLibrary(params: NativeBuildParameters) = buildXiphLibrary(
        repoUrl = "https://github.com/xiph/opusfile",
        branch = VERSION_OPUSFILE,
        libName = "opusfile",
        params = params,
        args = mapOf(
            "enable-http" to "no",
            "disable-examples" to null,
        ),
    )

    private fun buildOpusEncoderLibrary(params: NativeBuildParameters) = buildXiphLibrary(
        repoUrl = "https://github.com/xiph/libopusenc",
        branch = VERSION_OPUSENC,
        libName = "opusenc",
        params = params,
    )
    
    private fun buildOggLibrary(params: NativeBuildParameters) = buildXiphLibrary(
        repoUrl = "https://github.com/xiph/ogg",
        branch = VERSION_OGG,
        libName = "ogg",
        params = params,
    )

    private fun buildVpxLibrary(params: NativeBuildParameters) {
        if (prefixOutputExists(params, "vpx")) return

        println("Building libvpx...")

        val dir = initBuildDirectory(DIR_VPX) {
            gitClone(
                repo = "https://github.com/webmproject/libvpx",
                depth = 1,
                branch = VERSION_LIBVPX,
            )
        }
        dir.useDirectory {
            gitClean()
            File(dir, "_toxbuild").useDirectory {
                configure(
                    relative = "..",
                    prefixDir = params.prefixDir,
                    args = buildArgs(
                        "disable-examples" to null,
                        "disable-unit-tests" to null,
                        "enable-pic" to null,
                    ) {
                        params.vpxTarget?.let { this += "target" to it }
                        params.sysrootDir?.let { this += "libc" to it.absolutePath }
                    }
                )
                makeInstall()
            }
        }
    }

    private fun buildToxCore(params: NativeBuildParameters) {
        val dir = initBuildDirectory(DIR_TOXCORE) {
            gitClone(
                repo = "https://github.com/Toktok/c-toxcore",
                dirName = DIR_TOXCORE,
                recurseSubmodules = true,
                depth = 1,
                branch = VERSION_TOXCORE,
            )
        }
        val buildDirName = params.abi?.let { "_build_$it" } ?: "_build"
        File(dir, buildDirName).useDirectory { buildDir ->
            if (File(buildDir, "libtoxcore.a").exists()) return@useDirectory
            cmake(
                relative = "..",
                args = buildArgs(
                    "BUILD_TOXAV" to "ON",
                    "BOOTSTRAP_DAEMON" to "OFF",
                    "DHT_BOOTSTRAP" to "OFF",
                    "ENABLE_SHARED" to "OFF",
                    "UNITTEST" to "OFF",
                    "CMAKE_POSITION_INDEPENDENT_CODE" to "ON",
                    "CMAKE_INSTALL_PREFIX" to params.prefixDir.absolutePath,
                    "CMAKE_PREFIX_PATH" to params.prefixDir.absolutePath,
                    "CMAKE_C_FLAGS" to "-ffile-prefix-map=${this.buildDir}=/src",
                ) {
                    params.cmakeToolchainFile?.let { this += "CMAKE_TOOLCHAIN_FILE" to it.absolutePath }
                    params.androidPlatform?.let { this += "ANDROID_PLATFORM" to it }
                    params.abi?.let { this += "ANDROID_ABI" to it }
                },
            )
        }
        dir.useDirectory {
            val includeDir = dir.setupIncludeDir("include/tox")
            copyFile("toxcore/tox.h", includeDir)
            copyFile("toxcore/tox_dispatch.h", includeDir)
            copyFile("toxcore/tox_events.h", includeDir)
            copyFile("toxav/toxav.h", includeDir)
            copyFile("toxencryptsave/toxencryptsave.h", includeDir)
        }
    }

    private fun buildGifLibrary() {
        val dir = initBuildDirectory(DIR_GIF) {
            gitClone(
                repo = "https://git.code.sf.net/p/giflib/code",
                depth = 1,
                branch = VERSION_GIFLIB,
                dirName = DIR_GIF,
            )
        }
        dir.useDirectory {
            val libFile = File(dir, "libgif.a")
            if (libFile.exists()) libFile.delete()
            make("libgif.a")

            val includeDir = dir.setupIncludeDir("include")
            copyFile("gif_lib.h", includeDir)
        }
    }

    protected fun prefixOutputExists(params: NativeBuildParameters, name: String) =
        File(params.prefixDir, "lib/lib$name.a").exists()

    private fun File.setupIncludeDir(dirName: String): File {
        val includeDir = File(this, dirName)
        if (includeDir.exists()) includeDir.deleteRecursively()
        includeDir.mkdirs()
        return includeDir
    }

    private fun buildArgs(
        vararg baseArgs: Pair<String, String?>,
        block: MutableMap<String, String?>.() -> Unit,
    ) = mutableMapOf(*baseArgs).apply(block)

    // Hacky fix for the prefix paths getting messed up after configure
    // This only happens for opusfile and opusenc for some reason
    private fun fixXiphMakefile(target: String) {
        if (!target.contains("mingw32")) return

        val makeFile = File(workingDir, "Makefile")
        if (!makeFile.exists()) return

        val removePrefix = "/usr/$target/sys-root/"
        val modified = makeFile.readText().replace(removePrefix, "/")
        makeFile.writeText(modified)
    }
}