import org.gradle.api.DefaultTask
import org.gradle.api.file.DirectoryProperty
import org.gradle.api.tasks.InputDirectory
import org.gradle.api.tasks.TaskAction
import java.io.File

private const val VALUES_DIR = "values"
private const val STRINGS_NAME = "strings.xml"
private const val ATTR_NAME = "name"
private const val ATTR_TRANSLATABLE = "translatable"
private val REGEX_STRING_ELEMENT = Regex("^<string(?:-array)?.*>")
private val REGEX_XML_ATTR = Regex("[\\w_]+=\"[\\w_]+\"")

abstract class LocalizationCoverageTask : DefaultTask() {

    @get:InputDirectory
    abstract val resourceDirectory: DirectoryProperty
    
    @TaskAction
    fun analyzeStringLocalization() {
        val baseFile = resourceDirectory.file("$VALUES_DIR/$STRINGS_NAME").get().asFile
        check(baseFile.exists()) { "Base strings.xml not found" }
        
        val localizedFiles = resourceDirectory.asFile.get().listFiles().orEmpty().filter { file ->
            file.isDirectory && file.name.startsWith("$VALUES_DIR-")
        }.mapNotNull { dir ->
            File(dir, STRINGS_NAME).takeIf { it.exists() }
        }
        if (localizedFiles.isEmpty()) return
        
        var hasErrors = false
        val baseDefs = baseFile.getStringDefinitions()
        for (localization in localizedFiles) {
            val dirName = localization.parentFile?.name ?: continue
            val defs = localization.getStringDefinitions()
            for (baseDef in baseDefs) {
                if (baseDef !in defs) {
                    System.err.println("$dirName: Missing string \"$baseDef\"")
                    hasErrors = true
                }
            }
        }
        if (hasErrors) {
            error("Missing localized string resources - please add the definitions shown above")
        }
    }
    
    private fun File.getStringDefinitions(): Set<String> {
        val definitions = mutableSetOf<String>()
        forEachLine {
            val line = it.trim()
            if (line.matches(REGEX_STRING_ELEMENT)) {
                val attributes = REGEX_XML_ATTR.findAll(line).mapNotNull { match ->
                    match.groupValues.firstOrNull()?.let { value ->
                        val tag = value.substringBefore('=')
                        tag to value.substringAfter('=').trim('\"')
                    }
                }.toMap()
                if (attributes.containsKey(ATTR_NAME) || attributes[ATTR_TRANSLATABLE] != "false") {
                    definitions += attributes[ATTR_NAME].orEmpty()
                }
            }
        }
        return definitions
    }
}