import org.gradle.api.tasks.Exec
import org.gradle.api.tasks.Internal
import java.io.File

private val NUM_THREADS = Runtime.getRuntime().availableProcessors() * 2
private val WINDOWS_DRIVE_REGEX = Regex("([A-Z]):\\\\")

abstract class NativeBuildExecTask : Exec() {

    @get:Internal
    lateinit var bashExePath: String

    @get:Internal
    lateinit var cygRoot: String

    @get:Internal
    lateinit var mingwSysRootPath: String

    @get:Internal
    lateinit var buildDir: File

    @get:Internal
    protected val javaHome: String get() {
        val javaHome = environment["JAVA_HOME"].toString()
        if (javaHome.isBlank()) error("Failed to find JAVA_HOME!")
        return when {
            cygRoot.isNotEmpty() -> convertToUnixPath(javaHome)
            else -> javaHome
        }
    }

    protected fun initBuildDirectory(path: String, onCreate: File.() -> Unit): File {
        val dir = File(buildDir, path)
        if (!dir.exists()) dir.onCreate()
        return dir
    }

    protected fun File.useDirectory(block: (File) -> Unit) {
        val currentDir = workingDir
        println("Moving from $currentDir to $this")
        if (!exists()) mkdirs()
        workingDir(this)
        block(this)
        println("Moving from $workingDir to $currentDir")
        workingDir(currentDir)
    }

    protected fun gitClone(
        repo: String,
        dirName: String? = null,
        depth: Int? = null,
        branch: String? = null,
        recurseSubmodules: Boolean = false,
    ) {
        val args = mutableListOf("git", "clone")
        if (recurseSubmodules) args += "--recurse-submodules"
        depth?.let { args += "--depth=$it" }
        branch?.let { args += "--branch=$it" }
        args += repo
        dirName?.let { args += dirName }
        exec(args)
    }

    protected fun gitClean() = exec("git", "clean", "-ffdx")
    
    protected fun wget(url: String) = exec("wget", url)
    
    protected fun extractTar(
        fileName: String,
        outputName: String = fileName,
    ) {
        val cmdArgs = mutableListOf("tar", "-xzf", fileName)
        if (outputName != fileName) {
            cmdArgs += listOf("--one-top-level=$outputName", "--strip-components", "1")
        }
        exec(cmdArgs)
    }

    protected fun autoReconfigure() = exec("autoreconf", "-fi")

    protected fun configure(
        prefixDir: File,
        relative: String = ".",
        args: Map<String, String?> = emptyMap(),
        forceFpic: Boolean = false,
    ) {
        val cmdArgs = mutableListOf("$relative/configure", "--prefix=${prefixDir.absolutePath}")
        args.forEach { (key, value) -> cmdArgs += kvArg("--", key, value) }
        if (forceFpic) cmdArgs += "CFLAGS=-fPIC"
        exec(cmdArgs)
    }
    
    protected fun make(target: String = "all") = exec("make", "-j$NUM_THREADS", target)

    protected fun makeInstall() = make("install")

    protected fun cmake(
        relative: String,
        args: Map<String, String?> = emptyMap(),
    ) {
        val cmdArgs = mutableListOf("cmake", relative)
        args.forEach { (key, value) -> cmdArgs += kvArg("-D", key, value) }
        exec(cmdArgs)
        exec("cmake", "--build", ".")
    }
    
    protected fun copyFile(
        from: String,
        dir: File,
        outName: String = from.substringAfterLast('/'),
    ) {
        if (!dir.exists() && !dir.mkdirs()) error("Failed to create directory: $dir")
        val srcFile = File(workingDir, from)
        val dstFile = File(dir, outName)
        if (!srcFile.exists()) error("Failed to copy missing file: $srcFile")
        srcFile.copyTo(dstFile, overwrite = true)
    }

    private fun exec(args: List<String>) {
        if (bashExePath.isNotEmpty()) {
            val unixWorkingDir = convertToUnixPath(workingDir.absolutePath)
            val unixArgs = args.joinToString(" ") { convertToUnixPath(it) }
            println("EXEC: $unixArgs")
            commandLine(bashExePath, "-lc", "cd $unixWorkingDir; $unixArgs")
        } else {
            commandLine(args)
            println("EXEC: ${args.joinToString(" ")}")
        }
        try {
            exec()
        } catch (e: Exception) {
            e.printStackTrace()
            throw e
        }
    }

    private fun exec(vararg args: String) = exec(args.toList())

    protected fun convertToUnixPath(winPath: String): String {
        val winDriveLetter = WINDOWS_DRIVE_REGEX.findAll(winPath).map { it.groupValues[1] }.firstOrNull()
        val unixDriveLetter = winDriveLetter?.lowercase()
        val unixRoot = when {
            cygRoot.contains("cygdrive") -> "/cygdrive/$unixDriveLetter"
            cygRoot.isNotEmpty() -> "/$unixDriveLetter"
            else -> ""
        }
        val unixPath = when {
            winDriveLetter != null -> winPath.replace("$winDriveLetter:", unixRoot)
            else -> winPath
        }
        return unixPath.replace('\\', '/')
    }

    private fun kvArg(prefix: String, key: String, value: String?) = buildString {
        append(prefix)
        append(key)
        value?.let { value ->
            append('=')
            if (value.contains(' ')) {
                append("'$value'")
            } else {
                append(value)
            }
        }
    }
}