import com.android.build.gradle.internal.api.BaseVariantOutputImpl
import org.gradle.api.tasks.testing.logging.TestExceptionFormat
import org.gradle.nativeplatform.platform.internal.DefaultNativePlatform
import org.jetbrains.compose.desktop.application.dsl.TargetFormat
import org.jetbrains.kotlin.gradle.ExperimentalKotlinGradlePluginApi
import org.jetbrains.kotlin.gradle.dsl.JvmTarget

plugins {
    alias(libs.plugins.kotlinMultiplatform)
    alias(libs.plugins.kotlinSerialization)
    alias(libs.plugins.androidApplication)
    alias(libs.plugins.jetbrainsCompose)
    alias(libs.plugins.jetbrainsComposeKotlin)
    //alias(libs.plugins.ktComplete) // Disabled until this supports Kotlin 2.0.20+
    alias(libs.plugins.buildConfig)
}

private val appName = "Seers Lodge"
private val appDesc = "A multi-platform chat application built on the Tox core library."
private val appDescShort = "P2P Chat Application"
private val appDirName = "seers-lodge"
private val appIconBasePath = "icons/seers"
private val appUuid = "44f74802-0a9f-4f7b-b2b0-5fa9735745d2"
private val appPackageName = "nand.apps.chat"
private val appNamespace = "chat.seers.app"
private val versionMajor = 1
private val versionMinor = 1
private val versionPatch = 1
private val appVersionCode = 4
version = "$versionMajor.$versionMinor.$versionPatch"

private val desktopAppResourcesDir = project.layout.projectDirectory.dir("src/desktopMain/appResources")

// Enable the mock chat engine
private val mockEngine: String? by project

// Native build arguments
private val nativeBuildTarget: String? by project
private val nativeBuildMingwPath: String? by project
private val nativeBuildBashExe: String? by project
private val nativeBuildCygRoot: String? by project
private val nativeBuildCmakeJni: String? by project
private val nativeBuildOpenMp: String? by project

buildConfig {
    packageName = appPackageName
    buildConfigField("APP_NAME", appName)
    buildConfigField("APP_DIR_NAME", appDirName)
    buildConfigField("APP_VERSION", provider { version.toString() })
    buildConfigField("APP_VERSION_MAJOR", versionMajor)
    buildConfigField("APP_VERSION_MINOR", versionMinor)
    buildConfigField("APP_VERSION_PATCH", versionPatch)
    buildConfigField("COMMIT_HASH", getGitCommitHash())
    buildConfigField("MOCK_ENGINE", mockEngine.toBoolean())
}

@OptIn(ExperimentalKotlinGradlePluginApi::class)
kotlin {
    androidTarget {
        compilerOptions {
            jvmTarget = JvmTarget.JVM_11
        }
    }
    
    jvm("desktop")
    
    listOf(
        iosX64(),
        iosArm64(),
        iosSimulatorArm64()
    ).forEach { iosTarget ->
        iosTarget.binaries.framework {
            baseName = "ComposeApp"
            isStatic = true
        }
    }
    
    sourceSets {
        val desktopMain by getting
        
        androidMain.dependencies {
            implementation(libs.compose.ui.tooling.preview)
            implementation(libs.androidx.activity.compose)
            implementation(libs.androidx.preference)
            implementation(libs.androidx.material3.adaptive)
            implementation(libs.koin.android)
            implementation(libs.ktor.client.okhttp)
            implementation(libs.bundles.slf4j)
            implementation(libs.bundles.androidx.camera)
            implementation(libs.zxing.cpp)
            implementation(libs.oboe)
        }
        commonMain.dependencies {
            implementation(compose.runtime)
            implementation(compose.foundation)
            implementation(compose.material)
            implementation(compose.ui)
            implementation(compose.components.resources)
            implementation(compose.components.uiToolingPreview)
            implementation(libs.bundles.koin.common)
            implementation(libs.bundles.kmp.viewmodel)
            implementation(libs.material.icons)
            implementation(libs.kotlinx.datetime)
            implementation(libs.kotlinx.serialization.json)
            implementation(libs.ktor.client.core)
            implementation(libs.bundles.coil)
            implementation(libs.ksoup.lite)
            implementation(libs.okio)
            implementation(libs.kim)
            implementation(libs.qrcode)
            implementation(libs.bundles.compottie)
        }
        desktopMain.dependencies {
            implementation(compose.desktop.currentOs)
            implementation(libs.kotlinx.coroutines.swing)
            implementation(libs.ktor.client.okhttp)
            implementation(libs.ktor.client.mock)
            implementation(libs.bundles.slf4j)
            implementation(libs.appdirs)
            implementation(libs.devdirs)
            implementation(libs.bundles.scrimage)
            implementation(libs.jlayer)
        }
        iosMain.dependencies {
            implementation(libs.ktor.client.darwin)
        }
        
        // Test dependencies
        commonTest.dependencies {
            implementation(libs.kotlin.test)
            implementation(libs.kotlinx.coroutines.test)
            implementation(libs.okio.fakefilesystem)
        }
    }
}

android {
    namespace = appNamespace
    compileSdk = libs.versions.android.compileSdk.get().toInt()
    ndkVersion = "26.3.11579264"

    sourceSets["main"].manifest.srcFile("src/androidMain/AndroidManifest.xml")
    sourceSets["main"].res.srcDirs("src/androidMain/res")
    sourceSets["main"].resources.srcDirs("src/commonMain/resources")

    defaultConfig {
        applicationId = appNamespace
        minSdk = libs.versions.android.minSdk.get().toInt()
        targetSdk = libs.versions.android.targetSdk.get().toInt()
        versionCode = appVersionCode
        versionName = version.toString()
        externalNativeBuild {
            cmake {
                arguments.addAll(listOf(
                    "-DJAVA_AWT_LIBRARY=NotNeeded",
                    "-DJAVA_JVM_LIBRARY=NotNeeded",
                    "-DJAVA_INCLUDE_PATH2=NotNeeded",
                    "-DJAVA_AWT_INCLUDE_PATH=NotNeeded",
                    "-DANDROID_STL=c++_shared",
                ))
            }
        }
        ndk {
            //noinspection ChromeOsAbiSupport
            abiFilters += listOf("arm64-v8a")
        }
        resourceConfigurations += arrayOf("en", "it")
    }
    packaging {
        resources {
            excludes += "/META-INF/{AL2.0,LGPL2.1}"
        }
    }
    applicationVariants.all {
        outputs.all {
            val output = this as BaseVariantOutputImpl
            output.outputFileName = "${appName.replace(" ", "")}-$version.apk"
        }
    }
    buildFeatures {
        prefab = true
    }
    buildTypes {
        getByName("release") {
            isMinifyEnabled = false
            proguardFiles(
                getDefaultProguardFile("proguard-android-optimize.txt"),
                "proguard/release.pro",
            )
        }
    }
    compileOptions {
        sourceCompatibility = JavaVersion.VERSION_11
        targetCompatibility = JavaVersion.VERSION_11
    }
    externalNativeBuild {
        cmake {
            path = File("native/CMakeLists.txt")
        }
    }
    dependencies {
        debugImplementation(libs.compose.ui.tooling)
    }
    dependenciesInfo {
        // Disables dependency metadata when building APKs.
        includeInApk = false
        // Disables dependency metadata when building Android App Bundles.
        includeInBundle = false
    }
}

compose.desktop {
    application {
        mainClass = "MainKt"

        nativeDistributions {
            packageName = appName
            packageVersion = version.toString()
            description = appDesc
            vendor = "NAND"
            appResourcesRootDir = desktopAppResourcesDir
            val os = DefaultNativePlatform.getCurrentOperatingSystem()
            when {
                os.isWindows -> {
                    targetFormats(TargetFormat.Msi, TargetFormat.Exe)
                    description = appName
                }
                os.isLinux -> {
                    targetFormats(TargetFormat.Deb, TargetFormat.Rpm)
                    description = appDescShort
                }
                os.isMacOsX -> {
                    targetFormats(TargetFormat.Dmg, TargetFormat.Pkg)
                }
            }
            macOS {
                iconFile = project.file("$appIconBasePath.icns")
            }
            windows {
                iconFile = project.file("$appIconBasePath.ico")
                upgradeUuid = appUuid
                shortcut = true
            }
            linux {
                iconFile = project.file("$appIconBasePath.png")
                menuGroup = "Network"
            }
            buildTypes.release.proguard {
                version = "7.4.0"
                obfuscate = false
                configurationFiles.from(project.file("proguard/release.pro"))
            }
        }
    }
}

private val nativeDir get() = project.layout.projectDirectory.file("native").asFile

private fun NativeBuildExecTask.setupNativeBuildDirs() {
    buildDir = File(nativeDir, "build")
    mingwSysRootPath = nativeBuildMingwPath.orEmpty()
    bashExePath = nativeBuildBashExe.orEmpty()
    cygRoot = nativeBuildCygRoot.orEmpty()
}

tasks.register<BuildAndroidNativeTask>("buildAndroidNativeLibraries") {
    ndkDir = android.ndkDirectory
    minSdk = android.defaultConfig.minSdk
    abiFilters = android.defaultConfig.ndk.abiFilters
    setupNativeBuildDirs()
    commandLine("echo", "Building native libraries for Android...")
}

tasks.register<BuildDesktopNativeTask>("buildDesktopNativeLibraries") {
    setupNativeBuildDirs()
    target = nativeBuildTarget.orEmpty()
    commandLine("echo", "Building native libraries for desktop...")
}

tasks.register<BuildDesktopJniTask>("buildDesktopJniLibraries") {
    setupNativeBuildDirs()
    sourceDir = File(nativeDir, "jni")
    outputDir = desktopAppResourcesDir.asFile
    cmakeJni = nativeBuildCmakeJni == "true"
    openMp = nativeBuildOpenMp == "true"
    commandLine("echo", "Building JNI libraries for desktop...")
}

tasks.register<LocalizationCoverageTask>("analyzeLocalizationCoverage") {
    resourceDirectory = project.layout.projectDirectory.file("src/commonMain/composeResources").asFile
}

tasks.withType(Test::class.java) {
    testLogging {
        exceptionFormat = TestExceptionFormat.FULL
        showStackTraces = true
    }
}

tasks.build {
    dependsOn("buildAndroidNativeLibraries")
    dependsOn("buildDesktopNativeLibraries")
    dependsOn("buildDesktopJniLibraries")
}

private fun getGitCommitHash(): String {
    val gitFolder = "$rootDir/.git/"
    val takeFromHash = 8
    val head = File(gitFolder + "HEAD").readText().split(":")
    val isCommit = head.size == 1
    if (isCommit) return head[0].trim().take(takeFromHash)
    val refHead = File(gitFolder + head[1].trim())
    return refHead.readText().trim().take(takeFromHash)
}