#include "AudioEngine.h"
#include "logging_macros.h"

using namespace oboe;

AudioEngine::AudioEngine(Direction direction) : direction(direction) {}

int32_t AudioEngine::start(int deviceId, int sampleRate, int channels, bool isLowLatency) {
    std::lock_guard<std::mutex> lock(mLock);
    
    this->channels = channels;
    this->sampleRate = sampleRate;
    int bufferSize = (int) (0.5 * (float) sampleRate); // TODO - Adjustable setting?
    sampleBuffer = std::make_shared<SampleBuffer>(bufferSize);
    
    AudioStreamBuilder builder;
    Result result = builder.setSharingMode(isLowLatency ? SharingMode::Exclusive : SharingMode::Shared)
            ->setPerformanceMode(isLowLatency ? PerformanceMode::LowLatency : PerformanceMode::PowerSaving)
            ->setDeviceId(deviceId)
            ->setDirection(direction)
            ->setChannelCount(channels)
            ->setSampleRate(sampleRate)
            ->setSampleRateConversionQuality(SampleRateConversionQuality::Medium)
            ->setFormat(AudioFormat::I16)
            ->setDataCallback(this)
            ->openStream(mStream);
    if (result == Result::OK) {
        result = mStream->requestStart();
    }
    return (int32_t) result;
}

int AudioEngine::writeSamples(const short *samples, int numSamples, long streamId) {
    std::lock_guard<std::mutex> lock(mLock);
    return sampleBuffer ? sampleBuffer->writeSamples(samples, numSamples, streamId) : 0;
}

int AudioEngine::readSamples(short *samples, int numSamples) {
    std::lock_guard<std::mutex> lock(mLock);
    return sampleBuffer ? sampleBuffer->readSamples(samples, numSamples) : 0;
}

int AudioEngine::getVacancy(long streamId) {
    if (streamId != -1) {
        return sampleBuffer->getCapacity() - sampleBuffer->getStreamPosition(streamId);
    } else {
        return sampleBuffer->getCapacity() - sampleBuffer->getBufferSize();
    }
}

int AudioEngine::getStreamPosition(long streamId) {
    return sampleBuffer->getStreamPosition(streamId);
}

int AudioEngine::getBufferSize() {
    return sampleBuffer->getBufferSize();
}

int32_t AudioEngine::stop() {
    std::lock_guard<std::mutex> lock(mLock);
    Result result = Result::OK;
    if (mStream) {
        mStream->stop();
        result = mStream->close();
        mStream.reset();
    }
    if (sampleBuffer) sampleBuffer.reset();
    return (int32_t) result;
}

DataCallbackResult AudioEngine::onAudioReady(AudioStream *oboeStream, void *audioData, int32_t numFrames) {
    auto *samples = static_cast<short *>(audioData);
    int numSamples = numFrames * channels;
    if (direction == Direction::Output) {
        readSamples(samples, numSamples);
    } else {
        writeSamples(samples, numSamples, -1);
    }
    return DataCallbackResult::Continue;
}