#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <opus.h>
#include <opus_multistream.h>
#include <opusenc.h>
#include <opusfile.h>
#include <ogg/ogg.h>

#include "opusjni.h"

#define BUFFER_SIZE 960
#define DEFAULT_SAMPLE_RATE 48000
#define DEFAULT_CHANNELS 2

// XXX - Ugly workaround to prevent certain library functions from being optimized out
void __attribute__((optimize("O0"))) require(void* list, ...) {
    va_list args;
    va_start(args, list);
    va_end(args);
}
#define REQUIRE(...) require(NULL, __VA_ARGS__)
#define REQUIRE_FUNCTIONS() REQUIRE(opus_get_version_string, opus_multistream_surround_encoder_create, ogg_sync_init, opus_multistream_decoder_destroy)

jboolean Java_nand_apps_opus_OpusNative_encodeToFile(JNIEnv *env, jclass, jobject wav, jint bitRate, jstring path) {
    REQUIRE_FUNCTIONS();

    jclass cls = env->FindClass("nand/apps/opus/OpusWavContainer");
    jmethodID getPcm = env->GetMethodID(cls, "getPcm", "()[S");
    jmethodID getSampleRate = env->GetMethodID(cls, "getSampleRate", "()I");
    jmethodID getChannels = env->GetMethodID(cls, "getChannels", "()I");
    jmethodID logMethodId = env->GetMethodID(cls, "log", "([B)V");
    
    jshortArray pcm = (jshortArray) env->CallObjectMethod(wav, getPcm);
    jint sampleRate = env->CallIntMethod(wav, getSampleRate);
    jint channels = env->CallIntMethod(wav, getChannels);
    
    const char *pathPtr = env->GetStringUTFChars(path, NULL);
    
    int err;
    OggOpusComments *comments = ope_comments_create();
    OggOpusEnc *enc = ope_encoder_create_file(pathPtr, comments, sampleRate, channels, channels > 2 ? 1 : 0, &err);
    if (!enc) {
        env->ReleaseStringUTFChars(path, pathPtr);
        ope_comments_destroy(comments);
        return false;
    }
    
    short *pcmPtr = env->GetShortArrayElements(pcm, NULL);
    jint numSamples = env->GetArrayLength(pcm);
    
    int srcIndex = 0;
    int fullBufferSize = channels * BUFFER_SIZE;
    bool success = true;
    while (srcIndex < numSamples) {
        int remaining = numSamples - srcIndex;
        int samplesPerChannel = BUFFER_SIZE < remaining ? BUFFER_SIZE : remaining;
        err = ope_encoder_write(enc, pcmPtr + srcIndex, samplesPerChannel);
        if (err != 0) {
            success = false;
            break;
        }
        srcIndex += fullBufferSize;
    }
    
    env->ReleaseStringUTFChars(path, pathPtr);
    env->ReleaseShortArrayElements(pcm, pcmPtr, 0);
    ope_encoder_drain(enc);
    ope_encoder_destroy(enc);
    ope_comments_destroy(comments);
    
    return success;
}

jlong Java_nand_apps_opus_OpusNative_openFileReader(JNIEnv *env, jclass, jstring path) {
    REQUIRE_FUNCTIONS();
    int err;
    const char *pathPtr = env->GetStringUTFChars(path, NULL);
    OggOpusFile *of = op_open_file(pathPtr, &err);
    env->ReleaseStringUTFChars(path, pathPtr);
    return (jlong) of;
}

jlong Java_nand_apps_opus_OpusNative_openMemoryReader(JNIEnv *env, jclass, jbyteArray data) {
    REQUIRE_FUNCTIONS();
    int err;
    jbyte *dataPtr = env->GetByteArrayElements(data, NULL);
    int dataSize = env->GetArrayLength(data);
    unsigned char* dataCopy = (unsigned char*) malloc(dataSize);
    memcpy(dataCopy, dataPtr, dataSize);
    OggOpusFile *of = op_open_memory(dataCopy, dataSize, &err);
    env->ReleaseByteArrayElements(data, dataPtr, 0);
    return (jlong) of;
}

jlong Java_nand_apps_opus_OpusNative_getSampleCount(JNIEnv *, jclass, jlong ptr) {
    return op_pcm_total((OggOpusFile*) ptr, -1);
}

jlong Java_nand_apps_opus_OpusNative_getSamplePosition(JNIEnv *, jclass, jlong ptr) {
    return op_pcm_tell((OggOpusFile*) ptr);
}

jboolean Java_nand_apps_opus_OpusNative_seekTo(JNIEnv *, jclass, jlong ptr, jlong position) {
    OggOpusFile *of = (OggOpusFile*) ptr;
    if (op_seekable(of)) {
        return op_pcm_seek(of, position) == 0;
    } else {
        return false;
    }
}

jint Java_nand_apps_opus_OpusNative_readSamples(JNIEnv *env, jclass, jlong ptr, jshortArray buffer, jint offset, jint length) {
    short *bufferPtr = env->GetShortArrayElements(buffer, NULL);
    int numRead = op_read_stereo((OggOpusFile*) ptr, bufferPtr + offset, length);
    env->ReleaseShortArrayElements(buffer, bufferPtr, 0);
    return numRead;
}

void Java_nand_apps_opus_OpusNative_closeReader(JNIEnv *, jclass, jlong ptr) {
    op_free((OggOpusFile*) ptr);
}

jlong Java_nand_apps_opus_OpusNative_openFileWriter(JNIEnv *env, jclass, jstring path, jint sampleRate, jint channels, jint bitRate) {
    REQUIRE_FUNCTIONS();
    int err;
    const char *pathPtr = env->GetStringUTFChars(path, NULL);
    OggOpusComments *comments = ope_comments_create();
    OggOpusEnc *enc = ope_encoder_create_file(pathPtr, comments, sampleRate, channels, channels > 2 ? 1 : 0, &err);
    if (enc) ope_encoder_ctl(enc, OPUS_SET_BITRATE(bitRate));
    env->ReleaseStringUTFChars(path, pathPtr);
    ope_comments_destroy(comments);
    return (jlong) enc;
}

jint Java_nand_apps_opus_OpusNative_writeSamples(JNIEnv *env, jclass, jlong ptr, jshortArray samples, jint offset, jint samplesPerChannel) {
    short *samplesPtr = env->GetShortArrayElements(samples, NULL);
    int err = ope_encoder_write((OggOpusEnc*) ptr, samplesPtr + offset, samplesPerChannel);
    env->ReleaseShortArrayElements(samples, samplesPtr, 0);
    return err == 0 ? samplesPerChannel : 0;
}

void Java_nand_apps_opus_OpusNative_closeWriter(JNIEnv *, jclass, jlong ptr) {
    OggOpusEnc *enc = (OggOpusEnc*) ptr;
    ope_encoder_drain(enc);
    ope_encoder_destroy(enc);
}