#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include "callbacks.h"
#include "groups.h"
#include "util.h"

#define CALL_METHOD(name, signature, ...)                                            \
    jmethodID methodId = getEnv()->GetMethodID(getCallbackClass(), name, signature); \
    getEnv()->CallVoidMethod(getCallback(), methodId, __VA_ARGS__)

/** FRIEND CALLBACKS **/

void self_connection_status_cb(Tox *tox, TOX_CONNECTION connection_status, void *user_data) {
    CALL_METHOD("onConnectionStatus", "(I)V", connection_status);
}

void friend_request_cb(Tox *tox, const uint8_t *public_key, const uint8_t *message, size_t length, void *user_data) {    
    jbyteArray publicKey = newByteArray(public_key, TOX_PUBLIC_KEY_SIZE);
    jbyteArray messageBytes = newByteArray(message, length);
    CALL_METHOD("onFriendRequest", "([B[B)V", publicKey, messageBytes);
}

void friend_message_cb(Tox *tox, Tox_Friend_Number friend_num, TOX_MESSAGE_TYPE type, const uint8_t *message, size_t length, void *user_data) {
    jbyteArray messageBytes = newByteArray(message, length);
    CALL_METHOD("onFriendMessage", "(II[B)V", friend_num, type, messageBytes);
}

void friend_name_cb(Tox *tox, Tox_Friend_Number friend_num, const uint8_t *name, size_t length, void *user_data) {
    jbyteArray nameBytes = newByteArray(name, length);
    CALL_METHOD("onFriendName", "(I[B)V", friend_num, nameBytes);
}

void friend_status_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_User_Status status, void *user_data) {
    CALL_METHOD("onFriendStatus", "(II)V", friend_num, status);
}

void friend_status_message_cb(Tox *tox, Tox_Friend_Number friend_num, const uint8_t *message, size_t length, void *user_data) {
    jbyteArray messageBytes = newByteArray(message, length);
    CALL_METHOD("onFriendStatusMessage", "(I[B)V", friend_num, messageBytes);
}

void friend_connection_status_cb(Tox *tox, Tox_Friend_Number friend_num, TOX_CONNECTION connection_status, void *user_data) {
    CALL_METHOD("onFriendConnectionStatus", "(II)V", friend_num, connection_status);
}

void friend_read_receipt_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_Friend_Message_Id message_id, void *user_data) {
    CALL_METHOD("onFriendReadReceipt", "(II)V", friend_num, message_id);
}

void friend_packet_cb(Tox *tox, Tox_Friend_Number friend_num, const uint8_t data[], size_t length, void *user_data) {
    jbyteArray packetBytes = newByteArray(data, length);
    CALL_METHOD("onFriendPacket", "(I[B)V", friend_num, packetBytes);
}

/** CONFERENCE CALLBACKS **/

void conference_invite_cb(Tox *tox, Tox_Friend_Number friend_num, TOX_CONFERENCE_TYPE type, const uint8_t *cookie, size_t length, void *user_data) {
    jbyteArray cookieBytes = newByteArray(cookie, length);
    CALL_METHOD("onConferenceInvite", "(II[B)V", friend_num, type, cookieBytes);
}

void conference_connected_cb(Tox *tox, Tox_Conference_Number conference_number, void *user_data) {
    CALL_METHOD("onConferenceConnected", "(I)V", conference_number);
}

void conference_title_cb(Tox *tox, uint32_t group_num, uint32_t peer_num, const uint8_t *title, size_t length, void *user_data) {
    jbyteArray titleBytes = newByteArray(title, length);
    CALL_METHOD("onConferenceTitle", "(II[B)V", group_num, peer_num, titleBytes);
}

void conference_message_cb(Tox *tox, uint32_t group_num, uint32_t peer_num, TOX_MESSAGE_TYPE type, const uint8_t *message, size_t length, void *user_data) {
    if (tox_conference_peer_number_is_ours(tox, group_num, peer_num, NULL)) return;
    jbyteArray messageBytes = newByteArray(message, length);
    CALL_METHOD("onConferenceMessage", "(III[B)V", group_num, peer_num, type, messageBytes);
}

void conference_audio_cb(void *tox, uint32_t group_num, uint32_t peer_num, const int16_t pcm[], uint32_t sample_count, uint8_t channels, uint32_t sample_rate, void *user_data) {
    jshortArray samples = getEnv()->NewShortArray(sample_count * channels);
    getEnv()->SetShortArrayRegion(samples, 0, sample_count * channels, (jshort*) pcm);
    CALL_METHOD("onConferenceAudioFrame", "(II[SII)V", group_num, peer_num, samples, channels, sample_rate);
}

void conference_peer_list_changed_cb(Tox *tox, uint32_t group_num, void *user_data) {
    CALL_METHOD("onConferenceUsersChanged", "(I)V", group_num);
}

void conference_peer_name_cb(Tox *tox, uint32_t group_num, uint32_t peer_num, const uint8_t *name, size_t length, void *user_data) {
    jbyteArray nameBytes = newByteArray(name, length);
    CALL_METHOD("onConferenceUserName", "(II[B)V", group_num, peer_num, nameBytes);
}

/** GROUP CALLBACKS **/

void group_peer_name_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, const uint8_t name[], size_t name_length, void *user_data) {
    jbyteArray nameBytes = newByteArray(name, name_length);
    CALL_METHOD("onGroupPeerName", "(II[B)V", group_num, peer_id, nameBytes);
}

void group_peer_status_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, Tox_User_Status status, void *user_data) {
    CALL_METHOD("onGroupPeerStatus", "(III)V", group_num, peer_id, status);
}

void group_topic_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, const uint8_t topic[], size_t topic_length, void *user_data) {
    jbyteArray topicBytes = newByteArray(topic, topic_length);
    CALL_METHOD("onGroupTopic", "(II[B)V", group_num, peer_id, topicBytes);
}

void group_privacy_state_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Privacy_State privacy_state, void *user_data) {
    CALL_METHOD("onGroupPrivacyState", "(II)V", group_num, privacy_state);
}

void group_voice_state_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Voice_State voice_state, void *user_data) {
    CALL_METHOD("onGroupVoiceState", "(II)V", group_num, voice_state);
}

void group_topic_lock_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Topic_Lock topic_lock, void *user_data) {
    CALL_METHOD("onGroupTopicLock", "(II)V", group_num, topic_lock);
}

void group_peer_limit_cb(Tox *tox, Tox_Group_Number group_num, uint32_t peer_limit, void *user_data) {
    CALL_METHOD("onGroupPeerLimit", "(II)V", group_num, peer_limit);
}

void group_password_cb(Tox *tox, Tox_Group_Number group_num, const uint8_t password[], size_t password_length, void *user_data) {
    jbyteArray passwordBytes = newByteArray(password, password_length);
    CALL_METHOD("onGroupPassword", "(I[B)V", group_num, passwordBytes);
}

void group_message_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, Tox_Message_Type message_type, const uint8_t message[], size_t message_length, Tox_Group_Message_Id message_id, void *user_data) {
    jbyteArray messageBytes = newByteArray(message, message_length);
    CALL_METHOD("onGroupMessage", "(III[BI)V", group_num, peer_id, message_type, messageBytes, message_id);
}

void group_private_message_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, Tox_Message_Type message_type, const uint8_t message[], size_t message_length, Tox_Group_Message_Id message_id, void *user_data) {
    jbyteArray messageBytes = newByteArray(message, message_length);
    CALL_METHOD("onGroupPrivateMessage", "(III[BI)V", group_num, peer_id, message_type, messageBytes, message_id);
}

void group_custom_packet_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, const uint8_t data[], size_t data_length, void *user_data) {
    jbyteArray dataBytes = newByteArray(data, data_length);
    CALL_METHOD("onGroupCustomPacket", "(II[B)V", group_num, peer_id, dataBytes);
}

void group_custom_private_packet_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, const uint8_t data[], size_t data_length, void *user_data) {
    jbyteArray dataBytes = newByteArray(data, data_length);
    CALL_METHOD("onGroupCustomPrivatePacket", "(II[B)V", group_num, peer_id, dataBytes);
}

void group_invite_cb(Tox *tox, Tox_Friend_Number friend_num, const uint8_t invite_data[], size_t invite_data_length, const uint8_t group_name[], size_t group_name_length, void *user_data) {
    jbyteArray inviteBytes = newByteArray(invite_data, invite_data_length);
    jbyteArray nameBytes = newByteArray(group_name, group_name_length);
    CALL_METHOD("onGroupInvite", "(I[B[B)V", friend_num, inviteBytes, nameBytes);
}

void group_peer_join_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, void *user_data) {
    jobject peerData = getGroupPeerData(getEnv(), tox, group_num, peer_id);
    CALL_METHOD("onGroupPeerJoin", "(ILnand/apps/tox/ToxUserData;)V", group_num, peerData);
}

void group_peer_exit_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number peer_id, Tox_Group_Exit_Type exit_type, const uint8_t name[], size_t name_length, const uint8_t part_message[], size_t part_message_length, void *user_data) {
    jbyteArray nameBytes = newByteArray(name, name_length);
    jbyteArray partMessageBytes = part_message != NULL ? newByteArray(part_message, part_message_length) : NULL;
    CALL_METHOD("onGroupPeerExit", "(III[B[B)V", group_num, peer_id, exit_type, nameBytes, partMessageBytes);
}

void group_self_join_cb(Tox *tox, Tox_Group_Number group_num, void *user_data) {
    jobject groupData = getGroupData(getEnv(), tox, group_num);
    jobject peerData = getGroupSelfPeerData(getEnv(), tox, group_num);
    CALL_METHOD("onGroupSelfJoin", "(Lnand/apps/tox/ToxGroupData;Lnand/apps/tox/ToxUserData;)V", groupData, peerData);
}

void group_join_fail_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Join_Fail fail_type, void *user_data) {
    CALL_METHOD("onGroupJoinFail", "(II)V", group_num, fail_type);
}

void group_moderation_cb(Tox *tox, Tox_Group_Number group_num, Tox_Group_Peer_Number source_peer_id, Tox_Group_Peer_Number target_peer_id, Tox_Group_Mod_Event mod_type, void *user_data) {
    CALL_METHOD("onGroupModeration", "(IIII)V", group_num, source_peer_id, target_peer_id, mod_type);
}

/** FILE CALLBACKS **/

void file_recv_control_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_File_Number file_num, Tox_File_Control control, void *user_data) {
    CALL_METHOD("onFileControl", "(III)V", friend_num, file_num, control);
}

void file_recv_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_File_Number file_num, uint32_t kind, uint64_t file_size, const uint8_t filename[], size_t filename_length, void *user_data) {
    uint8_t file_id[TOX_FILE_ID_LENGTH];
    if (!tox_file_get_file_id(tox, friend_num, file_num, file_id, NULL)) return;
    jbyteArray fileNameBytes = newByteArray(filename, filename_length);
    jbyteArray fileUidBytes = newByteArray(file_id, TOX_FILE_ID_LENGTH);
    CALL_METHOD("onFileReceived", "(II[BIJ[B)V", friend_num, file_num, fileUidBytes, kind, file_size, fileNameBytes);
}

void file_recv_chunk_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_File_Number file_num, uint64_t position, const uint8_t data[], size_t length, void *user_data) {
    jbyteArray dataBytes = newByteArray(data, length);
    CALL_METHOD("onFileChunkReceived", "(IIJ[B)V", friend_num, file_num, position, dataBytes);
}

void file_chunk_request_cb(Tox *tox, Tox_Friend_Number friend_num, Tox_File_Number file_num, uint64_t position, size_t length, void *user_data) {
    CALL_METHOD("onFileChunkRequest", "(IIJJ)V", friend_num, file_num, position, (jlong) length);
}

/** AV CALLBACKS */

void av_call_cb(ToxAV *av, uint32_t friend_num, bool audio_enabled, bool video_enabled, void *user_data) {
    CALL_METHOD("onCallStarted", "(IZZ)V", friend_num, audio_enabled, video_enabled);
}

void av_call_state_cb(ToxAV *av, uint32_t friend_num, uint32_t state, void *user_data) {
    CALL_METHOD("onCallState", "(II)V", friend_num, state);
}

void av_audio_rate_cb(ToxAV *av, uint32_t friend_num, uint32_t audio_bit_rate, void *user_data) {
    CALL_METHOD("onAudioBitrate", "(II)V", friend_num, audio_bit_rate);
}

void av_video_rate_cb(ToxAV *av, uint32_t friend_num, uint32_t video_bit_rate, void *user_data) {
    CALL_METHOD("onVideoBitrate", "(II)V", friend_num, video_bit_rate);
}

void av_audio_receive_frame_cb(ToxAV *av, uint32_t friend_num, const int16_t pcm[], size_t sample_count, uint8_t channels, uint32_t sampling_rate, void *user_data) {
    jshortArray samples = getEnv()->NewShortArray(sample_count * channels);
    getEnv()->SetShortArrayRegion(samples, 0, sample_count * channels, (jshort*) pcm);
    CALL_METHOD("onAudioFrame", "(I[SII)V", friend_num, samples, channels, sampling_rate);
}

void av_video_receive_frame_cb(ToxAV *av, uint32_t friend_num, uint16_t width, uint16_t height, const uint8_t y[], const uint8_t u[], const uint8_t v[], int32_t ystride, int32_t ustride, int32_t vstride, void *user_data) {
    jbyteArray yBytes = newByteArray(y, MAX(width, abs(ystride)) * height);
    jbyteArray uBytes = newByteArray(u, MAX(width / 2, abs(ustride)) * (height / 2));
    jbyteArray vBytes = newByteArray(v, MAX(width / 2, abs(vstride)) * (height / 2));
    CALL_METHOD("onVideoFrame", "(III[B[B[BIII)V", friend_num, width, height, yBytes, uBytes, vBytes, ystride, ustride, vstride);
}