#include <tox/tox.h>
#include <tox/toxav.h>
#include <stdlib.h>

#include "conferences.h"
#include "callbacks.h"
#include "util.h"
#include "logger.h"

jobjectArray Java_nand_apps_tox_ToxNative_getConferenceList(JNIEnv *env, jclass, jlong toxPtr) {
    Tox *tox = (Tox *) toxPtr;
    
    jclass cls = env->FindClass("nand/apps/tox/ToxGroupData");
    
    size_t numConferences = tox_conference_get_chatlist_size(tox);
    uint32_t *conferenceList = (uint32_t *) malloc(sizeof(uint32_t) * numConferences);
    tox_conference_get_chatlist(tox, conferenceList);
    
    jobjectArray groupArray = env->NewObjectArray(numConferences, cls, NULL);

    for (int i = 0; i < numConferences; i++) {
        jobject groupData = getConferenceData(env, tox, conferenceList[i]);
        if (groupData != NULL) {
            env->SetObjectArrayElement(groupArray, i, groupData);
        }
    }
    
    free(conferenceList);
    
    return groupArray;
}

jobject Java_nand_apps_tox_ToxNative_createConference(JNIEnv *env, jclass, jlong toxPtr, jbyteArray title, jint type) {
    Tox *tox = (Tox *) toxPtr;
    Tox_Err_Conference_New error = TOX_ERR_CONFERENCE_NEW_OK;
    uint8_t uid[TOX_CONFERENCE_ID_SIZE];
    uint8_t selfKey[TOX_PUBLIC_KEY_SIZE];
    
    jint groupNum = type == TOX_GROUP_TYPE_AV
        ? toxav_add_av_groupchat(tox, NULL, NULL)
        : tox_conference_new(tox, &error);
    
    if (groupNum != -1 && error == TOX_ERR_CONFERENCE_NEW_OK) {
        jbyte *titlePtr = GET_ARRAY_BYTES(title);
        int titleLength = GET_ARRAY_LEN(title);
        
        tox_conference_set_title(tox, groupNum, (const uint8_t*) titlePtr, titleLength, NULL);
        tox_conference_get_id(tox, groupNum, uid);
        tox_self_get_public_key(tox, selfKey);
        
        jobject groupData = getSharedGroupData(env, tox, groupNum, (uint8_t*) titlePtr, NULL, uid, type, TOX_GROUP_PRIVACY_STATE_PRIVATE, selfKey);
        
        RELEASE_ARRAY(title, titlePtr);
        
        if (type == TOX_GROUP_TYPE_AV) {
            // Audio is enabled by default for some reason - we don't really want this behavior
            toxav_groupchat_disable_av(tox, groupNum);
        }
        
        return groupData;
    } else {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to create new conference (error = %d)", error);
        return NULL;
    }
}

jobjectArray Java_nand_apps_tox_ToxNative_getConferencePeerList(JNIEnv *env, jclass, jlong toxPtr, jint groupNum) {
    Tox *tox = (Tox *) toxPtr;
    
    jclass cls = env->FindClass("nand/apps/tox/ToxUserData");
    
    Tox_Err_Conference_Peer_Query error;
    
    size_t numOnline = tox_conference_peer_count(tox, groupNum, &error);
    if (error != TOX_ERR_CONFERENCE_PEER_QUERY_OK) {
        return env->NewObjectArray(0, cls, NULL);
    }
    
    size_t numOffline = tox_conference_offline_peer_count(tox, groupNum, &error);
    if (error != TOX_ERR_CONFERENCE_PEER_QUERY_OK) {
        return env->NewObjectArray(0, cls, NULL);
    }
    
    jobjectArray peerArray = env->NewObjectArray(numOnline + numOffline, cls, NULL);

    uint8_t *name;
    uint8_t publicKey[TOX_PUBLIC_KEY_SIZE];
    uint64_t lastOnline;

    for (int i = 0; i < numOnline + numOffline; i++) {
    
        bool isOnline = i < numOnline;
        int peerNum = isOnline ? i : i - numOnline;
        
        if (isOnline) {
            tox_conference_peer_get_public_key(tox, groupNum, peerNum, publicKey, &error);
        } else {
            tox_conference_offline_peer_get_public_key(tox, groupNum, peerNum, publicKey, &error);
        }
        if (error != TOX_ERR_CONFERENCE_PEER_QUERY_OK) {
            LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get public key for %s peer #%d", isOnline ? "online" : "offline", peerNum);
            continue;
        }
        
        size_t len = isOnline
            ? tox_conference_peer_get_name_size(tox, groupNum, peerNum, &error)
            : tox_conference_offline_peer_get_name_size(tox, groupNum, peerNum, &error);
        if (error != TOX_ERR_CONFERENCE_PEER_QUERY_OK) {
            LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get name length for %s peer #%d", isOnline ? "online" : "offline", peerNum);
            continue;
        }
        
        name = (uint8_t *) calloc(1, len + 1);
        if (isOnline) {
            tox_conference_peer_get_name(tox, groupNum, peerNum, name, &error);
        } else {
            tox_conference_offline_peer_get_name(tox, groupNum, peerNum, name, &error);
        }
        if (error != TOX_ERR_CONFERENCE_PEER_QUERY_OK) {
            LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get name for %s peer #%d", isOnline ? "online" : "offline", peerNum);
            free(name);
            continue;
        }
        
        uint64_t lastOnline = isOnline ? 0 : tox_conference_offline_peer_get_last_active(tox, groupNum, peerNum, NULL);
        Tox_Connection protocol = isOnline ? TOX_CONNECTION_TCP : TOX_CONNECTION_NONE;
        
        jobject peerData = getUserData(env, tox, peerNum, name, TOX_USER_STATUS_NONE, NULL, publicKey, TOX_GROUP_ROLE_USER, protocol, lastOnline);
        env->SetObjectArrayElement(peerArray, i, peerData);
        
        free(name);
    }
    
    return peerArray;
}

jboolean Java_nand_apps_tox_ToxNative_setConferenceTitle(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray title) {
    Tox *tox = (Tox *) toxPtr;
    
    jbyte *titlePtr = GET_ARRAY_BYTES(title);
    int titleLength = GET_ARRAY_LEN(title);
    
    jboolean success = tox_conference_set_title(tox, groupNum, (const uint8_t*) titlePtr, titleLength, NULL);
    
    RELEASE_ARRAY(title, titlePtr);
    
    return success;
}

jboolean Java_nand_apps_tox_ToxNative_setConferenceAudioEnabled(JNIEnv *, jclass, jlong toxPtr, jint groupNum, jboolean isEnabled) {
    Tox *tox = (Tox *) toxPtr;
    int32_t result;
    if (isEnabled) {
        result = toxav_groupchat_enable_av(tox, groupNum, conference_audio_cb, NULL);
    } else {
        result = toxav_groupchat_disable_av(tox, groupNum);
    }
    return result == 0;
}

jboolean Java_nand_apps_tox_ToxNative_isConferenceAudioEnabled(JNIEnv *, jclass, jlong toxPtr, jint groupNum) {
    return toxav_groupchat_av_enabled((Tox *) toxPtr, groupNum);
}

void Java_nand_apps_tox_ToxNative_sendConferenceMessage(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray message, jint messageType) {
    Tox *tox = (Tox *) toxPtr;
    jbyte *messagePtr = GET_ARRAY_BYTES(message);
    int messageLength = GET_ARRAY_LEN(message);
    tox_conference_send_message(tox, groupNum, (Tox_Message_Type) messageType, (uint8_t*) messagePtr, messageLength, NULL);
    RELEASE_ARRAY(message, messagePtr);
}

jboolean Java_nand_apps_tox_ToxNative_sendConferenceAudioFrame(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jshortArray pcm, jint channels, jint sampleRate) {
    jshort *pcmPtr = env->GetShortArrayElements(pcm, NULL);
    int pcmLength = env->GetArrayLength(pcm) / channels;
    int32_t result = toxav_group_send_audio((Tox*) toxPtr, groupNum, (const int16_t*) pcmPtr, pcmLength, channels, sampleRate);
    env->ReleaseShortArrayElements(pcm, pcmPtr, 0);
    return result == 0;
}

jboolean Java_nand_apps_tox_ToxNative_sendConferenceInvite(JNIEnv *, jclass, jlong toxPtr, jint friendNum, jint groupNum) {
    return tox_conference_invite((Tox *) toxPtr, friendNum, groupNum, NULL);
}

jobject Java_nand_apps_tox_ToxNative_acceptConferenceInvite(JNIEnv *env, jclass, jlong toxPtr, jint friendNum, jint type, jbyteArray cookie) {
    Tox *tox = (Tox *) toxPtr;
    
    jbyte *cookiePtr = GET_ARRAY_BYTES(cookie);
    int cookieLength = GET_ARRAY_LEN(cookie);
    Tox_Err_Conference_Join error = TOX_ERR_CONFERENCE_JOIN_OK;
    
    jint groupNum = type == TOX_GROUP_TYPE_AV
        ? toxav_join_av_groupchat(tox, friendNum, (const uint8_t*) cookiePtr, cookieLength, NULL, NULL)
        : tox_conference_join(tox, friendNum, (const uint8_t*) cookiePtr, cookieLength, &error);
    
    RELEASE_ARRAY(cookie, cookiePtr);
    
    if (error == TOX_ERR_CONFERENCE_JOIN_OK) {
        if (type == TOX_GROUP_TYPE_AV) {
            // Audio is enabled by default - we don't really want this behavior
            toxav_groupchat_disable_av(tox, groupNum);
        }
        return getConferenceData(env, tox, groupNum);
    } else {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to accept conference invite from #%d (error = %d)", friendNum, error);
        return NULL;
    }
}

jboolean Java_nand_apps_tox_ToxNative_removeConference(JNIEnv *env, jclass, jlong toxPtr, jint groupNum) {
    return tox_conference_delete((Tox *) toxPtr, groupNum, NULL);
}