#include <tox/tox.h>
#include <tox/toxav.h>
#include <stdlib.h>

#include "core.h"
#include "util.h"
#include "callbacks.h"
#include "logger.h"

JavaVM *jvm = NULL;
int jniVersion = JNI_VERSION_1_6;
jclass loggerClass;
jobject logger;
jclass callbackClass;
jobject callback;

JavaVM *getJvm() {
    return jvm;
}

JNIEnv *getEnv() {
    JNIEnv* env;
    jvm->GetEnv((void**)&env, jniVersion);
    return env;
}

jclass getCallbackClass() {
    return callbackClass;
}

jobject getCallback() {
    return callback;
}

jclass getLoggerClass() {
    return loggerClass;
}

jobject getLogger() {
    return logger;
}

/** JNI **/

jobject Java_nand_apps_tox_ToxNative_getConstants(JNIEnv *env, jclass) {
    jclass cls = env->FindClass("nand/apps/tox/ToxConstants");
    jmethodID constructor = env->GetMethodID(cls, "<init>", "(IIIIIIIIIIIIIIII)V");
    return env->NewObject(
        cls,
        constructor,
        tox_version_major(),
        tox_version_minor(),
        tox_version_patch(),
        tox_max_name_length(),
        tox_max_status_message_length(),
        tox_max_friend_request_length(),
        tox_max_message_length(),
        tox_max_custom_packet_size(),
        tox_group_max_group_name_length(),
        tox_group_max_topic_length(),
        tox_group_max_password_size(),
        tox_group_max_part_length(),
        tox_max_filename_length(),
        tox_max_hostname_length(),
        tox_public_key_size(),
        tox_address_size()
    );
}

jobject Java_nand_apps_tox_ToxNative_initTox(JNIEnv *env, jclass, jbyteArray saveData, jobject initOptions, jobject toxLogger) {

    env->GetJavaVM(&jvm);
    jniVersion = env->GetVersion();
    logger = env->NewGlobalRef(toxLogger);
    loggerClass = (jclass) env->NewGlobalRef((jobject) env->GetObjectClass(logger));

    struct Tox_Options *options = tox_options_new(NULL);
    
    tox_options_set_log_callback(options, onLogMessage);

    // Load save data
    jbyte *savePtr = saveData != NULL ? env->GetByteArrayElements(saveData, NULL) : NULL;
    jsize saveLength = saveData != NULL ? env->GetArrayLength(saveData) : 0;
    if (saveLength > 0) {
        tox_options_set_savedata_type(options, TOX_SAVEDATA_TYPE_TOX_SAVE);
        tox_options_set_savedata_data(options, (uint8_t*)savePtr, saveLength);
    }
    
    tox_options_set_proxy_type(options, TOX_PROXY_TYPE_NONE);
    tox_options_set_proxy_host(options, NULL);
    tox_options_set_proxy_port(options, 0);
    tox_options_set_udp_enabled(options, true);
    tox_options_set_ipv6_enabled(options, true);
    tox_options_set_local_discovery_enabled(options, true);
    tox_options_set_start_port(options, 0);
    tox_options_set_end_port(options, 0);
    
    // Options from app layer
    if (initOptions != NULL) {
        jclass cls = env->FindClass("nand/apps/tox/ToxInitOptions");
        jboolean isUdpEnabled = env->CallBooleanMethod(initOptions, env->GetMethodID(cls, "isUdpEnabled", "()Z"));
        jboolean isIPv6Enabled = env->CallBooleanMethod(initOptions, env->GetMethodID(cls, "isIPv6Enabled", "()Z"));
        jboolean isLocalDiscoveryEnabled = env->CallBooleanMethod(initOptions, env->GetMethodID(cls, "isLocalDiscoveryEnabled", "()Z"));
        
        tox_options_set_udp_enabled(options, isUdpEnabled);
        tox_options_set_ipv6_enabled(options, isIPv6Enabled);
        tox_options_set_local_discovery_enabled(options, isLocalDiscoveryEnabled);
        
        jint proxyType = env->CallIntMethod(initOptions, env->GetMethodID(cls, "getProxyType", "()I"));
        jstring proxyHost = (jstring) env->CallObjectMethod(initOptions, env->GetMethodID(cls, "getProxyHost", "()Ljava/lang/String;"));
        jint proxyPort = env->CallIntMethod(initOptions, env->GetMethodID(cls, "getProxyPort", "()I"));
        
        if (proxyType != TOX_PROXY_TYPE_NONE && proxyHost != NULL && proxyPort > 0) {
            const char *hostPtr = env->GetStringUTFChars(proxyHost, NULL);
            tox_options_set_proxy_type(options, (Tox_Proxy_Type) proxyType);
            tox_options_set_proxy_host(options, hostPtr);
            tox_options_set_proxy_port(options, proxyPort);
            env->ReleaseStringUTFChars(proxyHost, hostPtr);
            
            if (isUdpEnabled) {
                LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "UDP + Proxy = Insecure! Disabling UDP...");
                tox_options_set_udp_enabled(options, false);
            }
        }
        
        // Persistence of DHT groups (a.k.a "Advanced" groups)
        // Toxcore advises against using this function since it's experimental, but there's currently no other (reliable)
        // way of reinitializing a DHT group between client sessions
        jboolean isDhtGroupPersistenceEnabled = env->CallBooleanMethod(initOptions, env->GetMethodID(cls, "isDhtGroupPersistenceEnabled", "()Z"));
        tox_options_set_experimental_groups_persistence(options, isDhtGroupPersistenceEnabled);
    }

    Tox_Err_New error;
    Tox *tox = tox_new(options, &error);
    tox_options_free(options);
    
    if (saveData != NULL) {
        env->ReleaseByteArrayElements(saveData, savePtr, 0);
    }
    
    if (error != TOX_ERR_NEW_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to instantiate Tox (error = %d)", error);
    }
    
    jclass resultClass = env->FindClass("nand/apps/tox/ToxInitResult");
    jmethodID resultCtor = env->GetMethodID(resultClass, "<init>", "(JI)V");

    return env->NewObject(
        resultClass,
        resultCtor,
        (jlong) (error == TOX_ERR_NEW_OK ? tox : NULL),
        error
    );
}

void Java_nand_apps_tox_ToxNative_registerCallbacks(JNIEnv *env, jclass cl, jlong toxPtr, jlong avPtr, jobject cb) {
    callback = env->NewGlobalRef(cb);
    callbackClass = (jclass) env->NewGlobalRef((jobject) env->GetObjectClass(cb));
    
    Tox *tox = (Tox *) toxPtr;
    
    tox_callback_self_connection_status(tox, self_connection_status_cb);

    tox_callback_friend_request(tox, friend_request_cb);
    tox_callback_friend_message(tox, friend_message_cb);
    tox_callback_friend_name(tox, friend_name_cb);
    tox_callback_friend_status(tox, friend_status_cb);
    tox_callback_friend_status_message(tox, friend_status_message_cb);
    tox_callback_friend_connection_status(tox, friend_connection_status_cb);
    tox_callback_friend_read_receipt(tox, friend_read_receipt_cb);
    tox_callback_friend_lossy_packet(tox, friend_packet_cb);
    tox_callback_friend_lossless_packet(tox, friend_packet_cb);

    tox_callback_conference_invite(tox, conference_invite_cb);
    tox_callback_conference_connected(tox, conference_connected_cb);
    tox_callback_conference_title(tox, conference_title_cb);
    tox_callback_conference_message(tox, conference_message_cb);
    tox_callback_conference_peer_list_changed(tox, conference_peer_list_changed_cb);
    tox_callback_conference_peer_name(tox, conference_peer_name_cb);
    
    tox_callback_group_peer_name(tox, group_peer_name_cb);
    tox_callback_group_peer_status(tox, group_peer_status_cb);
    tox_callback_group_topic(tox, group_topic_cb);
    tox_callback_group_privacy_state(tox, group_privacy_state_cb);
    tox_callback_group_voice_state(tox, group_voice_state_cb);
    tox_callback_group_topic_lock(tox, group_topic_lock_cb);
    tox_callback_group_peer_limit(tox, group_peer_limit_cb);
    tox_callback_group_password(tox, group_password_cb);
    tox_callback_group_message(tox, group_message_cb);
    tox_callback_group_private_message(tox, group_private_message_cb);
    tox_callback_group_custom_packet(tox, group_custom_packet_cb);
    tox_callback_group_custom_private_packet(tox, group_custom_private_packet_cb);
    tox_callback_group_invite(tox, group_invite_cb);
    tox_callback_group_peer_join(tox, group_peer_join_cb);
    tox_callback_group_peer_exit(tox, group_peer_exit_cb);
    tox_callback_group_self_join(tox, group_self_join_cb);
    tox_callback_group_join_fail(tox, group_join_fail_cb);
    tox_callback_group_moderation(tox, group_moderation_cb);
    
    tox_callback_file_recv_control(tox, file_recv_control_cb);
    tox_callback_file_recv(tox, file_recv_cb);
    tox_callback_file_recv_chunk(tox, file_recv_chunk_cb);
    tox_callback_file_chunk_request(tox, file_chunk_request_cb);
    
    ToxAV *av = (ToxAV*) avPtr;
    if (av) {
        toxav_callback_call(av, av_call_cb, NULL);
        toxav_callback_call_state(av, av_call_state_cb, NULL);
        toxav_callback_audio_bit_rate(av, av_audio_rate_cb, NULL);
        toxav_callback_video_bit_rate(av, av_video_rate_cb, NULL);
        toxav_callback_audio_receive_frame(av, av_audio_receive_frame_cb, NULL);
        toxav_callback_video_receive_frame(av, av_video_receive_frame_cb, NULL);
    }
}

jint Java_nand_apps_tox_ToxNative_iterate(JNIEnv *, jclass, jlong toxPtr) {
    Tox *tox = (Tox *) toxPtr;
    tox_iterate(tox, NULL);
    return tox_iteration_interval(tox);
}

void Java_nand_apps_tox_ToxNative_destroyTox(JNIEnv *env, jclass, jlong toxPtr) {
    tox_kill((Tox *) toxPtr);
    env->DeleteGlobalRef((jobject) loggerClass);
    env->DeleteGlobalRef(logger);
    env->DeleteGlobalRef((jobject) callbackClass);
    env->DeleteGlobalRef(callback);
}

jbyteArray Java_nand_apps_tox_ToxNative_getSaveData(JNIEnv *env, jclass, jlong toxPtr) {
    Tox *tox = (Tox *) toxPtr;
    size_t length = tox_get_savedata_size(tox);
    uint8_t *savedata = (uint8_t*) malloc(length);
    tox_get_savedata(tox, savedata);
    jbyteArray saveBytes = newByteArray(env, savedata, length);
    free(savedata);
    return saveBytes;
}