#include <tox/tox.h>
#include <tox/toxencryptsave.h>
#include <stdlib.h>

#include "encryption.h"
#include "util.h"

jbyteArray Java_nand_apps_tox_ToxNative_hash(JNIEnv *env, jclass, jbyteArray data) {
    jbyte *dataPtr = env->GetByteArrayElements(data, NULL);
    int dataLength = env->GetArrayLength(data);
    uint8_t hash[TOX_HASH_LENGTH];
    
    bool success = tox_hash(hash, (const uint8_t*)dataPtr, dataLength);
    env->ReleaseByteArrayElements(data, dataPtr, 0);
    
    return success ? newByteArray(env, hash, TOX_HASH_LENGTH) : NULL;
}

jlong Java_nand_apps_tox_ToxNative_createPassKey(JNIEnv *env, jclass, jbyteArray passPhrase, jbyteArray cipher) {
    jbyte *passPtr = env->GetByteArrayElements(passPhrase, NULL);
    int passLength = env->GetArrayLength(passPhrase);
    
    uint8_t salt[TOX_PASS_SALT_LENGTH];
    jbyte *cipherPtr = cipher != NULL ? env->GetByteArrayElements(cipher, NULL) : NULL;
    
    bool hasSalt = cipherPtr != NULL && tox_get_salt((const uint8_t *) cipherPtr, salt, NULL);
    
    Tox_Pass_Key *key = hasSalt
        ? tox_pass_key_derive_with_salt((const uint8_t*) passPtr, passLength, salt, NULL)
        : tox_pass_key_derive((const uint8_t*) passPtr, passLength, NULL);
    
    env->ReleaseByteArrayElements(passPhrase, passPtr, 0);
    if (cipherPtr != NULL) {
        env->ReleaseByteArrayElements(cipher, cipherPtr, 0);
    }
    
    return key != NULL ? (jlong) key : 0;
}

void Java_nand_apps_tox_ToxNative_freePassKey(JNIEnv *, jclass, jlong passKeyPtr) {
    tox_pass_key_free((Tox_Pass_Key *) passKeyPtr);
}

jbyteArray Java_nand_apps_tox_ToxNative_encryptWithPassKey(JNIEnv *env, jclass, jlong passKeyPtr, jbyteArray text) {
    Tox_Pass_Key *key = (Tox_Pass_Key *) passKeyPtr;
    jbyte *textPtr = env->GetByteArrayElements(text, NULL);
    int textLength = env->GetArrayLength(text);
    int cipherLength = textLength + TOX_PASS_ENCRYPTION_EXTRA_LENGTH;
    
    uint8_t *cipherPtr = (uint8_t *) malloc(cipherLength);
    
    bool success = tox_pass_key_encrypt(key, (const uint8_t*) textPtr, textLength, cipherPtr, NULL);
    
    jbyteArray cipher = success ? newByteArray(env, cipherPtr, cipherLength) : NULL;
    
    env->ReleaseByteArrayElements(text, textPtr, 0);
    free(cipherPtr);
    
    return cipher;
}

jbyteArray Java_nand_apps_tox_ToxNative_decryptWithPassKey(JNIEnv *env, jclass, jlong passKeyPtr, jbyteArray cipher) {
    Tox_Pass_Key *key = (Tox_Pass_Key *) passKeyPtr;
    jbyte *cipherPtr = env->GetByteArrayElements(cipher, NULL);
    int cipherLength = env->GetArrayLength(cipher);
    int textLength = cipherLength - TOX_PASS_ENCRYPTION_EXTRA_LENGTH;
    
    uint8_t *textPtr = (uint8_t *) malloc(textLength);
    
    bool success = tox_pass_key_decrypt(key, (const uint8_t*) cipherPtr, cipherLength, textPtr, NULL);
    
    jbyteArray text = success ? newByteArray(env, textPtr, textLength) : NULL;
    
    env->ReleaseByteArrayElements(cipher, cipherPtr, 0);
    free(textPtr);
    
    return text;
}

jboolean Java_nand_apps_tox_ToxNative_isEncrypted(JNIEnv *env, jclass, jbyteArray data) {
    jbyte *dataPtr = env->GetByteArrayElements(data, NULL);
    bool isEncrypted = tox_is_data_encrypted((const uint8_t*) dataPtr);
    env->ReleaseByteArrayElements(data, dataPtr, 0);
    return isEncrypted;
}