#include <tox/tox.h>
#include <tox/toxav.h>
#include <stdlib.h>
#include <string.h>

#include "groups.h"
#include "util.h"
#include "logger.h"

jobject getGroupPeerData(JNIEnv *env, Tox *tox, int groupNum, int peerId) {

    uint8_t *name;
    uint8_t publicKey[TOX_PUBLIC_KEY_SIZE];
    Tox_Err_Group_Peer_Query error;

    tox_group_peer_get_public_key(tox, groupNum, peerId, publicKey, &error);
    if (error != TOX_ERR_GROUP_PEER_QUERY_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get public key for peer #%d", peerId);
        return NULL;
    }
    
    size_t len = tox_group_peer_get_name_size(tox, groupNum, peerId, &error);
    if (error != TOX_ERR_GROUP_PEER_QUERY_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get name length for peer #%d", peerId);
        return NULL;
    }
    
    name = (uint8_t *) calloc(1, len + 1);
    tox_group_peer_get_name(tox, groupNum, peerId, name, &error);
    if (error != TOX_ERR_GROUP_PEER_QUERY_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to get name for peer #%d", peerId);
        free(name);
        return NULL;
    }
    
    Tox_User_Status status = tox_group_peer_get_status(tox, groupNum, peerId, NULL);
    Tox_Group_Role role = tox_group_peer_get_role(tox, groupNum, peerId, NULL);
    Tox_Connection protocol = tox_group_peer_get_connection_status(tox, groupNum, peerId, NULL);
    
    return getUserData(env, tox, peerId, name, status, NULL, publicKey, role, protocol, 0L);
}

jobject getGroupSelfPeerData(JNIEnv *env, Tox *tox, int groupNum) {

    Tox_Err_Group_Self_Query error;
    uint8_t publicKey[TOX_PUBLIC_KEY_SIZE];
    Tox_User_Status status;
    Tox_Group_Role role;
    Tox_Group_Peer_Number peerId;
    Tox_Connection protocol;
    jobject selfData;
    
    uint8_t *name;
    size_t nameLen = tox_group_self_get_name_size(tox, groupNum, &error);    
    if (error != TOX_ERR_GROUP_SELF_QUERY_OK) goto fail;
    
    name = (uint8_t *) calloc(1, nameLen + 1);
    tox_group_self_get_name(tox, groupNum, name, &error);
    if (error != TOX_ERR_GROUP_SELF_QUERY_OK) goto fail;
    
    tox_group_self_get_public_key(tox, groupNum, publicKey, &error);
    if (error != TOX_ERR_GROUP_SELF_QUERY_OK) goto fail;
    
    peerId = tox_group_self_get_peer_id(tox, groupNum, &error);
    if (error != TOX_ERR_GROUP_SELF_QUERY_OK) goto fail;
    
    status = tox_group_self_get_status(tox, groupNum, NULL);
    role = tox_group_self_get_role(tox, groupNum, NULL);    
    protocol = tox_group_peer_get_connection_status(tox, groupNum, peerId, NULL);
    selfData = getUserData(env, tox, peerId, name, status, NULL, publicKey, role, protocol, 0L);
    free(name);
    return selfData;
    
fail:
    free(name);
    return NULL;
}

jobjectArray Java_nand_apps_tox_ToxNative_getGroupList(JNIEnv *env, jclass, jlong toxPtr) {
    Tox *tox = (Tox *) toxPtr;
    jclass cls = env->FindClass("nand/apps/tox/ToxGroupData");
    uint32_t numGroups = tox_group_get_number_groups(tox);
    jobjectArray groupArray = env->NewObjectArray(numGroups, cls, NULL);
    for (int i = 0; i < numGroups; i++) {       
        jobject groupData = getGroupData(env, tox, i);
        if (groupData != NULL) {
            env->SetObjectArrayElement(groupArray, i, groupData);
        }
    }
    return groupArray;
}

jobject Java_nand_apps_tox_ToxNative_createGroup(JNIEnv *env, jclass, jlong toxPtr, jint visibility, jbyteArray title, jbyteArray selfNickname) {
    Tox *tox = (Tox *) toxPtr;
    
    jbyte *titlePtr = GET_ARRAY_BYTES(title);
    int titleLength = GET_ARRAY_LEN(title);
    uint8_t *titleStr = (uint8_t *) calloc(1, titleLength + 1);
    memcpy(titleStr, titlePtr, titleLength);
    
    jbyte *namePtr = GET_ARRAY_BYTES(selfNickname);
    int nameLength = GET_ARRAY_LEN(selfNickname);
    
    Tox_Err_Group_New error;
    jint groupNum = tox_group_new(tox, (Tox_Group_Privacy_State) visibility, (uint8_t*) titlePtr, titleLength, (uint8_t*) namePtr, nameLength, &error);
    
    jobject groupData = NULL;
    
    if (error == TOX_ERR_GROUP_NEW_OK) {
        uint8_t uid[TOX_GROUP_CHAT_ID_SIZE];
        uint8_t selfKey[TOX_PUBLIC_KEY_SIZE];
        if (tox_group_get_chat_id(tox, groupNum, uid, NULL) && tox_group_self_get_public_key(tox, groupNum, selfKey, NULL)) {
            groupData = getSharedGroupData(env, tox, groupNum, titleStr, NULL, uid, TOX_GROUP_TYPE_ADVANCED, visibility, selfKey);
        }
    } else {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to create new group (error = %d)", error);
    }
    
    free(titleStr);
    RELEASE_ARRAY(title, titlePtr);
    RELEASE_ARRAY(selfNickname, namePtr);
    
    return groupData;
}

jint Java_nand_apps_tox_ToxNative_joinGroup(JNIEnv *env, jclass, jlong toxPtr, jbyteArray uid, jbyteArray selfNickname, jbyteArray password) {
    Tox *tox = (Tox *) toxPtr;
    Tox_Err_Group_Join error;
    
    uint8_t chatId[TOX_GROUP_CHAT_ID_SIZE];
    jbyte *uidPtr = GET_ARRAY_BYTES(uid);
    
    jbyte *namePtr = GET_ARRAY_BYTES(selfNickname);
    int nameLen = namePtr != NULL ? GET_ARRAY_LEN(selfNickname) : 0;
    
    if (namePtr == NULL) {
        nameLen = tox_self_get_name_size(tox);
        namePtr = (jbyte*) calloc(1, nameLen);
        tox_self_get_name(tox, (uint8_t*) namePtr);
    }
        
    jbyte *passPtr = GET_ARRAY_BYTES(password);
    int passLen = GET_ARRAY_LEN(password);
    
    memcpy(chatId, uidPtr, TOX_GROUP_CHAT_ID_SIZE);
    
    jint groupNum = tox_group_join(tox, chatId, (uint8_t*)namePtr, nameLen, (uint8_t*)passPtr, passLen, &error);
    
    if (error != TOX_ERR_GROUP_JOIN_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to join group (error = %d)", error);
    }
    
    RELEASE_ARRAY(uid, uidPtr);
    RELEASE_ARRAY(selfNickname, namePtr);
    RELEASE_ARRAY(password, passPtr);
    
    return error == TOX_ERR_GROUP_JOIN_OK ? groupNum : -1;
}

jobject Java_nand_apps_tox_ToxNative_getGroupSelfPeerData(JNIEnv *env, jclass, jlong toxPtr, jint groupNum) {
    return getGroupSelfPeerData(env, (Tox *) toxPtr, groupNum);
}

jobjectArray Java_nand_apps_tox_ToxNative_getGroupPeerData(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jintArray peerIds) {
    Tox *tox = (Tox *) toxPtr;

    jclass cls = env->FindClass("nand/apps/tox/ToxUserData");
    
    jint numPeers = GET_ARRAY_LEN(peerIds);
    int *peerIdArray = (int *)env->GetIntArrayElements(peerIds, NULL);
    jobjectArray peerArray = env->NewObjectArray(numPeers, cls, NULL);

    for (int i = 0; i < numPeers; i++) {        
        jobject peerData = getGroupPeerData(env, tox, groupNum, peerIdArray[i]);
        env->SetObjectArrayElement(peerArray, i, peerData);
    }
    
    env->ReleaseIntArrayElements(peerIds, (jint *)peerIdArray, 0);
    
    return peerArray;
}

jboolean Java_nand_apps_tox_ToxNative_setGroupTopic(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray topic) {
    jbyte *topicPtr = GET_ARRAY_BYTES(topic);
    int topicLength = GET_ARRAY_LEN(topic);
    jboolean success = tox_group_set_topic((Tox *) toxPtr, groupNum, (const uint8_t*) topicPtr, topicLength, NULL);
    RELEASE_ARRAY(topic, topicPtr);
    return success;
}

jboolean Java_nand_apps_tox_ToxNative_setGroupPassword(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray password) {
    jbyte *passPtr = GET_ARRAY_BYTES(password);
    int passLength = GET_ARRAY_LEN(password);
    jboolean success = tox_group_set_password((Tox *) toxPtr, groupNum, (const uint8_t*) passPtr, passLength, NULL);
    RELEASE_ARRAY(password, passPtr);
    return success;
}

jboolean Java_nand_apps_tox_ToxNative_setGroupNickname(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray name) {
    jbyte *namePtr = GET_ARRAY_BYTES(name);
    int nameLength = GET_ARRAY_LEN(name);
    jboolean success = tox_group_self_set_name((Tox *) toxPtr, groupNum, (const uint8_t*) namePtr, nameLength, NULL);
    RELEASE_ARRAY(name, namePtr);
    return success;
}

void Java_nand_apps_tox_ToxNative_sendGroupMessage(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray message, jint messageType) {
    Tox *tox = (Tox *) toxPtr;
    jbyte *messagePtr = GET_ARRAY_BYTES(message);
    int messageLength = GET_ARRAY_LEN(message);
    tox_group_send_message(tox, groupNum, (Tox_Message_Type) messageType, (uint8_t*) messagePtr, messageLength, NULL);
    RELEASE_ARRAY(message, messagePtr);
}

void Java_nand_apps_tox_ToxNative_sendGroupPacket(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray packet) {
    Tox *tox = (Tox *) toxPtr;
    jbyte *packetPtr = GET_ARRAY_BYTES(packet);
    int packetLength = GET_ARRAY_LEN(packet);
    tox_group_send_custom_packet(tox, groupNum, true, (uint8_t*) packetPtr, packetLength, NULL);
    RELEASE_ARRAY(packet, packetPtr);
}

void Java_nand_apps_tox_ToxNative_sendPeerPacket(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jint peerId, jbyteArray packet) {
    Tox *tox = (Tox *) toxPtr;
    jbyte *packetPtr = GET_ARRAY_BYTES(packet);
    int packetLength = GET_ARRAY_LEN(packet);
    tox_group_send_custom_private_packet(tox, groupNum, peerId, true, (uint8_t*) packetPtr, packetLength, NULL);
    RELEASE_ARRAY(packet, packetPtr);
}

jboolean Java_nand_apps_tox_ToxNative_sendGroupInvite(JNIEnv *, jclass, jlong toxPtr, jint friendNum, jint groupNum) {
    Tox_Err_Group_Invite_Friend error;
    bool success = tox_group_invite_friend((Tox *) toxPtr, friendNum, groupNum, &error);
    if (!success) {
        LOGGER_WRITE(TOX_LOG_LEVEL_WARNING, "Failed to send group #%d invite to friend #%d (error = %d)", groupNum, friendNum, error);
    }
    return success;
}

jobject Java_nand_apps_tox_ToxNative_acceptGroupInvite(JNIEnv *env, jclass, jlong toxPtr, jint friendNum, jbyteArray inviteData, jbyteArray peerName, jbyteArray password) {
    Tox *tox = (Tox *) toxPtr;
    
    jbyte *invitePtr = GET_ARRAY_BYTES(inviteData);
    size_t inviteLength = GET_ARRAY_LEN(inviteData);
    
    jbyte *namePtr = GET_ARRAY_BYTES(peerName);
    size_t nameLength = GET_ARRAY_LEN(peerName);
    
    jbyte *passwordPtr = GET_ARRAY_BYTES(password);
    size_t passwordLength = GET_ARRAY_LEN(password);
    
    Tox_Err_Group_Invite_Accept error;
    
    jint groupNum = tox_group_invite_accept(tox, friendNum,
        (const uint8_t*) invitePtr, inviteLength,
        (const uint8_t*) namePtr, nameLength,
        (const uint8_t*) passwordPtr, passwordLength, &error);
    
    RELEASE_ARRAY(inviteData, invitePtr);
    RELEASE_ARRAY(peerName, namePtr);
    RELEASE_ARRAY(password, passwordPtr);
    
    return error == TOX_ERR_GROUP_INVITE_ACCEPT_OK ? getGroupData(env, tox, groupNum) : NULL;
}

jboolean Java_nand_apps_tox_ToxNative_setGroupRole(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jint peerId, jint role) {
    return tox_group_set_role((Tox *) toxPtr, (Tox_Group_Number) groupNum, (Tox_Group_Peer_Number) peerId, (Tox_Group_Role) role, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_kickGroupPeer(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jint peerId) {
    return tox_group_kick_peer((Tox *) toxPtr, (Tox_Group_Number) groupNum, (Tox_Group_Peer_Number) peerId, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_setGroupPeerIgnore(JNIEnv *, jclass, jlong toxPtr, jint groupNum, jint peerId, jboolean ignore) {
    return tox_group_set_ignore((Tox *) toxPtr, (Tox_Group_Number) groupNum, (Tox_Group_Peer_Number) peerId, ignore, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_reconnectGroup(JNIEnv *, jclass, jlong toxPtr, jint groupNum) {
    return tox_group_reconnect((Tox *) toxPtr, groupNum, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_disconnectGroup(JNIEnv *, jclass, jlong toxPtr, jint groupNum) {
    return tox_group_disconnect((Tox *) toxPtr, groupNum, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_isGroupConnected(JNIEnv *, jclass, jlong toxPtr, jint groupNum) {
    return tox_group_is_connected((Tox *) toxPtr, groupNum, NULL);
}

jboolean Java_nand_apps_tox_ToxNative_removeGroup(JNIEnv *env, jclass, jlong toxPtr, jint groupNum, jbyteArray partMsg) {
    jbyte *msgPtr = GET_ARRAY_BYTES(partMsg);
    int msgLength = GET_ARRAY_LEN(partMsg);
    bool success = tox_group_leave((Tox *) toxPtr, groupNum, (const uint8_t*) msgPtr, msgLength, NULL);
    RELEASE_ARRAY(partMsg, msgPtr);
    return success;
}