#include <string.h>
#include <stdlib.h>

#include "util.h"
#include "logger.h"

jbyteArray newByteArray(JNIEnv *env, const uint8_t *data, size_t length) {
    jbyteArray byteArray = env->NewByteArray(length);
    env->SetByteArrayRegion(byteArray, 0, length, (jbyte*) data);
    return byteArray;
}

jbyteArray newByteArray(const uint8_t *data, size_t length) {
    return newByteArray(getEnv(), data, length);
}

jobject getUserData(JNIEnv *env, Tox *tox, jint id, uint8_t *name, Tox_User_Status status, uint8_t *statusMessage,
    uint8_t *pubkey, Tox_Group_Role role, Tox_Connection protocol, uint64_t lastOnline) {
    
    jclass cls = env->FindClass("nand/apps/tox/ToxUserData");
    jmethodID constructor = env->GetMethodID(cls, "<init>", "(I[BI[B[BIIJ)V");
    jbyteArray nameStr = newByteArray(env, name, strlen((const char*) name));
    jbyteArray statusStr = statusMessage != NULL ? newByteArray(env, statusMessage, strlen((const char*) statusMessage)) : NULL;
    jbyteArray pubKeyBytes = newByteArray(env, pubkey, TOX_PUBLIC_KEY_SIZE);
    return env->NewObject(cls, constructor, id, nameStr, status, statusStr, pubKeyBytes, role, protocol, lastOnline);
}

jobject getConferenceData(JNIEnv *env, Tox *tox, Tox_Conference_Number conferenceNum) {
    uint8_t selfKey[TOX_PUBLIC_KEY_SIZE];
    uint8_t uid[TOX_CONFERENCE_ID_SIZE];
    Tox_Err_Conference_Title titleError;
    Tox_Err_Conference_Get_Type typeError;
    
    Tox_Conference_Type type = tox_conference_get_type(tox, conferenceNum, &typeError);
    if (typeError != TOX_ERR_CONFERENCE_GET_TYPE_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to get conference data for #%d (type error = %d)", conferenceNum, typeError);
        return NULL;
    }

    uint8_t *title = NULL;
    int titleLen = tox_conference_get_title_size(tox, conferenceNum, &titleError) + 1;
    if (titleError == TOX_ERR_CONFERENCE_TITLE_OK) {
        title = (uint8_t *) calloc(1, titleLen);
        tox_conference_get_title(tox, conferenceNum, title, &titleError);
        if (titleError != TOX_ERR_CONFERENCE_TITLE_OK) {
            LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to get conference data for #%d (title error = %d)", conferenceNum, titleError);
        }
    } else {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to get conference data for #%d (title size error = %d)", conferenceNum, titleError);
    }

    tox_conference_get_id(tox, conferenceNum, uid);
    tox_self_get_public_key(tox, selfKey);
    
    jobject groupData = getSharedGroupData(env, tox, conferenceNum, title, NULL, uid, type, TOX_GROUP_PRIVACY_STATE_PRIVATE, selfKey);
    if (title != NULL) free(title);
    return groupData;
}

jobject getGroupData(JNIEnv *env, Tox *tox, int groupNum) {

    uint8_t selfKey[TOX_PUBLIC_KEY_SIZE];
    uint8_t uid[TOX_GROUP_CHAT_ID_SIZE];
    Tox_Err_Group_State_Query queryError;
    Tox_Err_Group_Self_Query selfError;
    jobject groupData = NULL;
    
    int nameLen = tox_group_get_name_size(tox, groupNum, &queryError);
    if (queryError != TOX_ERR_GROUP_STATE_QUERY_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to get group data for #%d (name size error = %d)", groupNum, queryError);
        return NULL;
    }
    
    tox_group_self_get_public_key(tox, groupNum, selfKey, &selfError);
    if (selfError != TOX_ERR_GROUP_SELF_QUERY_OK) {
        LOGGER_WRITE(TOX_LOG_LEVEL_ERROR, "Failed to get group data for #%d (public key error = %d)", groupNum, queryError);
        return NULL;
    }
    
    int topicLen = tox_group_get_topic_size(tox, groupNum, NULL);
    
    uint8_t *title = (uint8_t *) calloc(1, nameLen);
    uint8_t *topic = (uint8_t *) calloc(1, topicLen);
    bool isNameValid = tox_group_get_name(tox, groupNum, title, NULL);
    bool isTopicValid = tox_group_get_topic(tox, groupNum, topic, NULL);
    bool isChatIdValid = tox_group_get_chat_id(tox, groupNum, uid, NULL);
    Tox_Group_Privacy_State visibility = tox_group_get_privacy_state(tox, groupNum, &queryError);
    
    if (isNameValid && isChatIdValid && queryError == TOX_ERR_GROUP_STATE_QUERY_OK) {
        groupData = getSharedGroupData(env, tox, groupNum, title, isTopicValid ? topic : NULL, uid, TOX_GROUP_TYPE_ADVANCED, visibility, selfKey);
    }
    
    free(title);
    return groupData;
}

jobject getSharedGroupData(JNIEnv *env, Tox *tox, jint group_num, uint8_t *title, uint8_t *topic, uint8_t *uid, jint type, jint visibility, uint8_t public_key[TOX_PUBLIC_KEY_SIZE]) {
    jclass cls = env->FindClass("nand/apps/tox/ToxGroupData");
    jmethodID constructor = env->GetMethodID(cls, "<init>", "(I[B[B[BII[B)V");
    jbyteArray titleStr = title != NULL ? newByteArray(env, title, strlen((const char*) title)) : NULL;
    jbyteArray topicStr = topic != NULL ? newByteArray(env, topic, strlen((const char*) topic)) : NULL;
    jbyteArray uidBytes = newByteArray(env, uid, TOX_CONFERENCE_ID_SIZE);
    jbyteArray pubKeyBytes = newByteArray(env, public_key, TOX_PUBLIC_KEY_SIZE);
    return env->NewObject(cls, constructor, group_num, titleStr, topicStr, uidBytes, type, visibility, pubKeyBytes);
}