package nand.apps.chat

import android.content.ComponentCallbacks
import android.content.Intent
import android.os.Bundle
import android.view.KeyEvent
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.lifecycle.lifecycleScope
import kotlinx.coroutines.launch
import nand.apps.chat.audio.*
import nand.apps.chat.camera.AndroidCameraLauncher
import nand.apps.chat.camera.CameraLauncher
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.intent.getUidExtra
import nand.apps.chat.io.AndroidFileBrowser
import nand.apps.chat.io.ChatFileBrowser
import nand.apps.chat.model.event.*
import nand.apps.chat.notification.AndroidNotificationController
import nand.apps.chat.notification.INTENT_KEY_CONVERSATION_UID
import nand.apps.chat.permission.PermissionManager
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.ui.app.App
import nand.apps.chat.ui.shortcut.AndroidKeyEventDispatcher
import org.koin.android.ext.android.get
import org.koin.android.ext.android.getKoin
import org.koin.android.ext.android.inject
import org.koin.compose.KoinContext
import org.koin.core.parameter.ParametersDefinition
import org.koin.core.qualifier.Qualifier

class MainActivity : ComponentActivity() {

    private val chatEngine: ChatEngine by inject()
    private val contactRepo: ContactRepo by inject()
    private val notificationRepo: NotificationRepo by inject()
    private val notificationController: AndroidNotificationController by inject()
    private val cameraLauncher by inject<CameraLauncher, AndroidCameraLauncher>()
    private val fileBrowser by inject<ChatFileBrowser, AndroidFileBrowser>()
    private val eventHandler: ChatEventHandler by inject()
    private val audioPlayer by inject<AudioPlayer, AndroidAudioPlayer>()
    private val audioRecorder by inject<AudioRecorder, AndroidAudioRecorder>()
    private val permissionManager by lazy { PermissionManager(this, getKoin()) }
    private val keyEventDispatcher: AndroidKeyEventDispatcher by inject()

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        notificationRepo.start(lifecycleScope)
        cameraLauncher.start(this)
        fileBrowser.start(this)
        audioPlayer.init(lifecycleScope)
        audioRecorder.init(lifecycleScope, permissionManager)
        setContent {
            KoinContext {
                keyEventDispatcher.CompositionLocalProvider {
                    App()
                }
            }
        }
        lifecycleScope.launch {
            eventHandler.onEvent.collect { event ->
                when (event) {
                    is LoginCompleteEvent -> {
                        onLoginComplete()
                        permissionManager.requestNotificationPermission()
                    }
                    is LogoutEvent -> chatEngine.stop()
                    is OpenFileBrowserEvent -> permissionManager.requestStoragePermissions()
                    else -> Unit
                }
            }
        }
        handleIntent(intent)
    }

    override fun onResume() {
        super.onResume()
        audioPlayer.start()
        audioRecorder.resume()
        eventHandler += AppFocusChangedEvent(isFocused = true)
    }

    override fun onPause() {
        eventHandler += AppFocusChangedEvent(isFocused = false)
        audioPlayer.stop()
        audioRecorder.pause()
        super.onPause()
    }
    
    override fun onDestroy() {
        chatEngine.stop()
        notificationController.stop()
        cameraLauncher.stop()
        fileBrowser.stop()
        super.onDestroy()
    }

    override fun onNewIntent(intent: Intent) {
        super.onNewIntent(intent)
        handleIntent(intent)
    }

    override fun onKeyDown(keyCode: Int, event: KeyEvent): Boolean {
        return keyEventDispatcher.onKeyDown(event) || super.onKeyDown(keyCode, event)
    }

    override fun onKeyUp(keyCode: Int, event: KeyEvent): Boolean {
        return keyEventDispatcher.onKeyUp(event) || super.onKeyUp(keyCode, event)
    }
    
    private fun onLoginComplete() {
        chatEngine.start(lifecycleScope)
    }
    
    private fun handleIntent(intent: Intent) {
        intent.getUidExtra(INTENT_KEY_CONVERSATION_UID)?.let { uid ->
            contactRepo.getContact(uid)?.let { contact ->
                lifecycleScope.launch {
                    eventHandler += OpenConversationEvent(contact)
                }
            }
        }
    }
}

private inline fun <reified T : Any,  reified U : Any> ComponentCallbacks.inject(
    qualifier: Qualifier? = null,
    mode: LazyThreadSafetyMode = LazyThreadSafetyMode.SYNCHRONIZED,
    noinline parameters: ParametersDefinition? = null,
) = lazy(mode) { get<T>(qualifier, parameters) as U }