package nand.apps.chat.audio

import android.content.Context
import android.media.AudioDeviceInfo
import android.media.AudioDeviceInfo.*
import android.media.AudioManager
import android.media.MediaPlayer
import android.net.Uri
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.runBlocking
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.util.removeIf
import okio.Path
import okio.Path.Companion.toOkioPath
import org.jetbrains.compose.resources.ExperimentalResourceApi
import seers.composeapp.generated.resources.Res
import java.io.File
import kotlin.math.roundToInt

class AndroidAudioPlayer(
    private val context: Context,
    fileSystem: ChatFileSystem,
) : AndroidAudioEngine(
    context = context,
    isPlaybackEngine = true,
), AudioPlayer {

    private var isRunning = false
    private val fileCache = mutableMapOf<String, File>()
    private val tmpDir = fileSystem.getTempDirectory("sounds")
    private val streams = mutableMapOf<Long, AndroidAudioPlaybackStream>()
    private val logger = getLogger("AndroidAudioPlayer")
    
    private lateinit var coroutineScope: CoroutineScope
    private var globalStreamId = 0L
    
    private val deviceTypes = mapOfDeviceTypes(
        TYPE_BLUETOOTH_A2DP,
        TYPE_WIRED_HEADSET,
        TYPE_WIRED_HEADPHONES,
        TYPE_USB_HEADSET,
        getBleHeadsetType(),
        getBleSpeakerType(),
        TYPE_BUILTIN_EARPIECE,
        TYPE_BUILTIN_SPEAKER,
    )
    
    fun init(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    fun start() {
        if (!isRunning) {
            isRunning = startEngine(
                deviceId = getDefaultOutputDevice()?.id ?: 0, // TODO - Output device setting
                channels = 2,
                isLowLatency = false,
            )
        }
    }

    fun stop() {
        if (isRunning) {
            isRunning = false
            streams.values.forEach { it.stop() }
            streams.clear()
            globalStreamId = 0L
            stopEngine()
        }
    }
    
    @OptIn(ExperimentalResourceApi::class)
    override fun queueSound(name: String): AudioHandle {
        val file = fileCache[name] ?: runBlocking {
            // MediaPlayer only accepts file paths and Android resources, but we only have the raw data
            // Solution: Save the raw sound data to a file first
            val fullName = "$name.mp3"
            val bytes = Res.readBytes("files/sounds/$fullName")
            tmpDir.writeToFile(fullName, bytes, safe = false)
            File(tmpDir.getFilePath(fullName).toString())
        }.also { file ->
            fileCache[name] = file
            file.deleteOnExit()
        }
        return queueFile(file.toOkioPath())
    }

    override fun queueFile(path: Path, startTime: Double): AudioHandle {
        val player = MediaPlayer.create(context, Uri.fromFile(path.toFile())).apply {
            start()
            if (startTime > 0) seekTo((startTime * 1000).roundToInt())
        }
        return AndroidSoundEffectHandle(
            fileName = path.name,
            player = player,
        )
    }

    override fun queueFrame(frame: AudioFrame, last: AudioHandle?): AudioHandle? {
        if (!this::coroutineScope.isInitialized) return null
        
        checkStaleStreams()
        
        val streamId = (last as? AndroidStreamHandle)?.streamId ?: globalStreamId++
        val stream = when (last) {
            is AndroidStreamHandle -> streams[streamId]
            else -> null
        } ?: AndroidAudioPlaybackStream(
            coroutineScope = coroutineScope,
            enginePtr = enginePtr,
            streamId = streamId,
        ).also { streams[streamId] = it }
        
        stream.queue(frame)
        
        return last ?: AndroidStreamHandle(streamId)
    }

    override fun stopSound(handle: AudioHandle) {
        when (handle) {
            is AndroidSoundEffectHandle -> handle.player.stop()
            is AndroidStreamHandle -> streams.remove(handle.streamId)?.stop()
        }
    }
    
    private fun checkStaleStreams() {
        streams.removeIf { _, stream -> !stream.isActive }
    }
    
    private fun getDefaultOutputDevice(): AudioDeviceInfo? {
        val devices = audioManager.getDevices(AudioManager.GET_DEVICES_OUTPUTS).sortedBy(deviceTypes)
        return devices.firstOrNull() ?: run {
            logger.error("No output devices found")
            null
        }
    }
}