package nand.apps.chat.audio

import android.Manifest
import android.content.Context
import android.media.AudioDeviceInfo
import android.media.AudioDeviceInfo.*
import android.media.AudioManager
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import nand.apps.chat.log.getLogger
import nand.apps.chat.permission.PermissionManager
import nand.apps.chat.util.fifoSharedFlow

// TODO - Adjustable based on settings?
private const val CHANNELS = 2
private const val BURST_MS = 50
private const val DELAY_MS = 10L

class AndroidAudioRecorder(context: Context) : AndroidAudioEngine(
    context = context,
    isPlaybackEngine = false,
), AudioRecorder {
    
    private var readJob: Job? = null
    override var isActive = false
    private var isRunning = false
    private var isDisabled = false
    private lateinit var permissionManager: PermissionManager
    private lateinit var coroutineScope: CoroutineScope
    
    private val logger = getLogger("AudioRecorder")

    private val deviceTypes = mapOfDeviceTypes(
        TYPE_WIRED_HEADSET,
        TYPE_USB_HEADSET,
        getBleHeadsetType(),
        TYPE_BUILTIN_MIC,
    )
    
    override val onFrame = fifoSharedFlow<AudioFrame>()

    fun init(coroutineScope: CoroutineScope, permissionManager: PermissionManager) {
        this.coroutineScope = coroutineScope
        this.permissionManager = permissionManager
    }
    
    override fun start() {
        if (!isActive) {
            isActive = true
            resume()
        }
    }
    
    override fun stop() {
        if (isActive) {
            isActive = false
            pause()
        }
    }
    
    fun resume() {
        if (isDisabled || !isActive || isRunning) return
        permissionManager.requestPermissions(Manifest.permission.RECORD_AUDIO) {
            startEngine(
                deviceId = getDefaultInputDevice()?.id ?: 0, // TODO - Input device setting
                channels = CHANNELS,
                isLowLatency = false,
            )
            isRunning = true
            readJob = coroutineScope.launch {
                val samples = ShortArray((defaultSampleRate / BURST_MS) * CHANNELS)
                while (isRunning) {
                    val samplesRead = AndroidAudioNative.readSamples(enginePtr, samples)
                    if (samplesRead > 0) {
                        onFrame.emit(
                            AudioFrame(
                                samples = samples,
                                channels = CHANNELS,
                                sampleRate = defaultSampleRate,
                            )
                        )
                    }
                    delay(DELAY_MS)
                }
            }
        }
    }
    
    fun pause() {
        if (isRunning) {
            isRunning = false
            readJob?.cancel()
            readJob = null
            stopEngine()
        }
    }
    
    fun disable() {
        isDisabled = true
    }
    
    private fun getDefaultInputDevice(): AudioDeviceInfo? {
        val devices = audioManager.getDevices(AudioManager.GET_DEVICES_INPUTS).sortedBy(deviceTypes)
        return devices.firstOrNull() ?: run {
            logger.error("No input devices found")
            null
        }
    }
}