package nand.apps.chat.camera

import android.content.Context
import androidx.camera.core.CameraSelector
import androidx.camera.core.ImageAnalysis
import androidx.camera.core.Preview
import androidx.camera.lifecycle.ProcessCameraProvider
import androidx.camera.view.PreviewView
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.padding
import androidx.compose.runtime.Composable
import androidx.compose.runtime.DisposableEffect
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.LocalLifecycleOwner
import androidx.compose.ui.viewinterop.AndroidView
import androidx.core.content.ContextCompat
import de.markusfisch.android.zxingcpp.ZxingCpp
import de.markusfisch.android.zxingcpp.ZxingCpp.ReaderOptions
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import kotlin.coroutines.resume
import kotlin.coroutines.suspendCoroutine
import seers.composeapp.generated.resources.*

@Composable
fun QrCodeCameraScreen(
    onResult: (ZxingCpp.Result) -> Unit,
    onCancel: () -> Unit,
) {
    val lensFacing = CameraSelector.LENS_FACING_BACK
    val lifecycleOwner = LocalLifecycleOwner.current
    val context = LocalContext.current
    val preview = Preview.Builder().build()
    val previewView = remember { PreviewView(context) }
    val cameraxSelector = CameraSelector.Builder().requireLensFacing(lensFacing).build()
    val useCase = remember {
        ImageAnalysis.Builder()
            .setOutputImageFormat(ImageAnalysis.OUTPUT_IMAGE_FORMAT_YUV_420_888)
            .setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST)
            .build()
    }
    LaunchedEffect(lensFacing) {
        val cameraProvider = context.getCameraProvider()
        cameraProvider.unbindAll()
        cameraProvider.bindToLifecycle(lifecycleOwner, cameraxSelector, preview, useCase)
        preview.setSurfaceProvider(previewView.surfaceProvider)
    }
    DisposableEffect(useCase) {
        var decoding = true
        useCase.setAnalyzer(ContextCompat.getMainExecutor(context)) { image ->
            image.use {
                if (decoding) {
                    ZxingCpp.readBitmap(
                        bitmap = image.toBitmap(),
                        left = 0,
                        top = 0,
                        width = previewView.width,
                        height = previewView.height,
                        rotation = 0,
                        options = ReaderOptions(
                            tryHarder = false,
                            tryRotate = true,
                            tryInvert = true,
                            tryDownscale = true,
                            returnCodabarStartEnd = true,
                            maxNumberOfSymbols = 1
                        ),
                    )?.let { results ->
                        results.find { it.format == ZxingCpp.BarcodeFormat.QR_CODE }?.let { result ->
                            decoding = false
                            onResult(result)
                        }
                    }
                }
            }
        }
        onDispose {
            useCase.clearAnalyzer()
        }
    }
    Box(
        contentAlignment = Alignment.BottomCenter,
        modifier = Modifier.fillMaxSize(),
    ) {
        AndroidView(
            factory = { previewView },
            modifier = Modifier.fillMaxSize(),
        )
        SimpleTextButton(
            text = stringResource(Res.string.cancel),
            onClick = onCancel,
            modifier = Modifier.padding(ChatTheme.dimens.spacingLarge),
        )
    }
}

private suspend fun Context.getCameraProvider(): ProcessCameraProvider = suspendCoroutine { continuation ->
    ProcessCameraProvider.getInstance(this).also { cameraProvider ->
        cameraProvider.addListener(
            { continuation.resume(cameraProvider.get()) },
            ContextCompat.getMainExecutor(this),
        )
    }
}