package nand.apps.chat.notification

import android.app.Notification
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.os.Bundle
import androidx.core.app.NotificationCompat
import chat.seers.app.R
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.runBlocking
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.intent.*
import nand.apps.chat.model.call.CallAction
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.call.CallStatus
import nand.apps.chat.model.call.getActions
import nand.apps.chat.model.notification.CallNotification
import nand.apps.chat.model.notification.ChatNotificationGroup
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.repo.CallRepo
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.getString
import org.koin.core.Koin
import seers.composeapp.generated.resources.*

private const val INTENT_ACTION_CALL = "$APP_NAME Call Action"

class AndroidCallNotificationReceiver(
    private val context: Context,
    private val controller: AndroidNotificationController,
    koin: Koin,
) : AndroidNotificationReceiver<CallNotification>(context, koin) {

    private val callRepo: CallRepo by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()
    private var collectJob: Job? = null

    override fun start(coroutineScope: CoroutineScope) {
        registerReceiver(IntentFilter(INTENT_ACTION_CALL))
        collectJob = coroutineScope.collect(callRepo.callObserver, ::onCallsUpdated)
    }

    override fun stop() {
        unregisterReceiver()
        collectJob?.cancel()
        collectJob = null
    }
    
    override fun getAppNotification(notification: CallNotification): Notification? {
        val call = callRepo.getCall(notification.contactUid) ?: return null
        val avatar = getAvatarBitmap(notification.avatar)
        val title = getNotificationTitle(call, notification.sender)
        val extras = Bundle().apply {
            putExtra(INTENT_KEY_CONVERSATION_UID, notification.contactUid)
            putString(INTENT_KEY_SENDER_NAME, notification.sender)
        }
        return NotificationCompat.Builder(context, NOTIFICATION_CHANNEL_ID)
            .setSmallIcon(R.drawable.ic_launcher_monochrome)
            .setLargeIcon(avatar)
            .setContentTitle(title)
            .setContentIntent(getOpenConversationIntent(notification.contactUid))
            .setCategory(Notification.CATEGORY_CALL)
            .setExtras(extras)
            .setAutoCancel(true)
            .setOnlyAlertOnce(true)
            .addCallActions(call)
            .build()
    }

    override fun onIntent(intent: Intent) {
        val action = intent.getCustomExtra<CallAction>(INTENT_KEY_CALL_ACTION) ?: return
        val contactUid = intent.getUidExtra(INTENT_KEY_CONVERSATION_UID) ?: return
        eventHandler += action.getEvent(contactUid)
    }
    
    private fun onCallsUpdated(calls: Map<ContactUid, CallData>) {
        val notification = controller.getActiveNotification(ChatNotificationGroup.CALLS) ?: return
        val contactUid = notification.extras?.getUid(INTENT_KEY_CONVERSATION_UID) ?: return
        val sender = notification.extras?.getString(INTENT_KEY_SENDER_NAME) ?: return
        val call = calls[contactUid] ?: return
        if (!call.isInProgress) {
            controller.dismiss(ChatNotificationGroup.CALLS)
            return
        }
        controller.notify(
            group = ChatNotificationGroup.CALLS,
            notification = NotificationCompat.Builder(context, notification)
                .setContentTitle(getNotificationTitle(call, sender))
                .addCallActions(call)
                .build(),
        )
    }
    
    private fun getNotificationTitle(call: CallData, sender: String) = runBlocking {
        when (call.status) {
            CallStatus.INCOMING -> getString(Res.string.incoming_call_message, sender)
            CallStatus.OUTGOING -> getString(Res.string.outgoing_call_message, sender)
            CallStatus.ACTIVE -> getString(Res.string.active_call_message, sender)
            CallStatus.PAUSED -> getString(Res.string.paused_call_message, sender)
            CallStatus.VOICEMAIL -> getString(Res.string.voicemail_call_message, sender)
            else -> ""
        }
    }
    
    private fun NotificationCompat.Builder.addCallActions(call: CallData): NotificationCompat.Builder {
        var b = clearActions()
        call.status.getActions().forEach { action ->
            val intentAction = NotificationCompat.Action.Builder(
                R.drawable.ic_launcher_monochrome,
                getStringBlocking(action.shortLabel),
                getBroadcastIntent(
                    request = action.toNotificationRequest(),
                    intent = Intent(INTENT_ACTION_CALL)
                        .setPackage(context.packageName)
                        .putExtra(INTENT_KEY_CONVERSATION_UID, call.contactUid)
                        .putExtra(INTENT_KEY_CALL_ACTION, action)
                )
            )
            b = b.addAction(intentAction.build())
        }
        return b
    }
}