package nand.apps.chat.notification

import android.app.Notification
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import androidx.core.app.NotificationCompat
import androidx.core.app.RemoteInput
import chat.seers.app.R
import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.intent.getUidExtra
import nand.apps.chat.intent.putExtra
import nand.apps.chat.model.event.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.notification.ChatNotificationGroup
import nand.apps.chat.model.notification.MessageNotification
import nand.apps.chat.event.ChatEventHandler
import org.koin.core.Koin

private const val INTENT_ACTION_REPLY = "$APP_NAME Reply Intent"
private const val INPUT_KEY_REPLY = "$APP_NAME Reply"

class AndroidMessageNotificationReceiver(
    private val context: Context,
    private val controller: AndroidNotificationController,
    koin: Koin,
) : AndroidNotificationReceiver<MessageNotification>(context, koin) {

    private val eventHandler: ChatEventHandler by koin.inject()

    override fun start(coroutineScope: CoroutineScope) {
        registerReceiver(IntentFilter(INTENT_ACTION_REPLY))
    }

    override fun stop() {
        unregisterReceiver()
    }
    
    override fun getAppNotification(notification: MessageNotification): Notification? {
        val message = notification.message
        val avatar = getAvatarBitmap(notification.avatar)
        val remoteInput = RemoteInput.Builder(INPUT_KEY_REPLY).run {
            setLabel(context.getString(R.string.reply))
            build()
        }
        val replyPendingIntent = getBroadcastIntent(
            request = AndroidNotificationRequest.REPLY,
            intent = getReplyIntent(message),
        )
        val replyAction = NotificationCompat.Action.Builder(
            R.drawable.ic_launcher_monochrome,
            context.getString(R.string.reply),
            replyPendingIntent
        ).addRemoteInput(remoteInput).build()
        return NotificationCompat.Builder(context, NOTIFICATION_CHANNEL_ID)
            .setSmallIcon(R.drawable.ic_launcher_monochrome)
            .setLargeIcon(avatar)
            .setContentTitle(notification.title)
            .setContentText(notification.getContentText())
            .setContentIntent(getOpenConversationIntent(message.conversationUid))
            .setCategory(Notification.CATEGORY_MESSAGE)
            .addAction(replyAction)
            .setAutoCancel(true)
            .build()
    }

    override fun onIntent(intent: Intent) {
        RemoteInput.getResultsFromIntent(intent)?.getCharSequence(INPUT_KEY_REPLY)?.let { text ->
            val conversationUid = intent.getUidExtra(INTENT_KEY_CONVERSATION_UID) ?: return@let
            eventHandler += SendMessageEvent(
                conversationUid = conversationUid,
                text = text.toString(),
            )
        }
        controller.dismiss(ChatNotificationGroup.MESSAGES)
    }

    private fun MessageNotification.getContentText(): String {
        return when {
            message.senderUid == message.conversationUid -> message.text
            else -> "$sender: ${message.text}"
        }
    }

    private fun getReplyIntent(message: ChatMessage): Intent {
        return Intent(INTENT_ACTION_REPLY)
            .setPackage(context.packageName)
            .putExtra(INTENT_KEY_CONVERSATION_UID, message.conversationUid)
    }
}