package nand.apps.chat.notification

import android.Manifest
import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.content.Context
import android.content.pm.PackageManager
import androidx.core.app.ActivityCompat
import androidx.core.app.NotificationManagerCompat
import androidx.core.content.ContextCompat.getSystemService
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.model.notification.CallNotification
import nand.apps.chat.model.notification.ChatNotification
import nand.apps.chat.model.notification.ChatNotificationGroup
import nand.apps.chat.model.notification.MessageNotification
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.ui.util.collect
import org.koin.core.Koin

class AndroidNotificationController(
    private val context: Context,
    koin: Koin,
) {
    private val channel = NotificationChannel(NOTIFICATION_CHANNEL_ID, APP_NAME, NotificationManager.IMPORTANCE_DEFAULT)
    private val manager = NotificationManagerCompat.from(context)
    private val notificationRepo: NotificationRepo by koin.inject()
    
    private val messageController by lazy { AndroidMessageNotificationReceiver(context, this, koin) }
    private val callController by lazy { AndroidCallNotificationReceiver(context, this, koin) }
    private val controllers = listOf(messageController, callController)
    
    private var collectorJob: Job? = null
    private val notificationGroups = mutableSetOf<ChatNotificationGroup>()
    private var isActive = false
    
    fun start(coroutineScope: CoroutineScope) {
        if (!isActive) {
            getSystemService(context, NotificationManager::class.java)?.createNotificationChannel(channel)
            controllers.forEach { it.start(coroutineScope) }
            isActive = true
            collectorJob = coroutineScope.collect(notificationRepo.notificationObserver) { notifications ->
                notifications.forEach { (_, notification) -> show(notification) }
                notificationGroups.forEach { group ->
                    if (group !in notifications) dismiss(group)
                }
                notificationGroups.clear()
                notificationGroups.addAll(notifications.keys)
            }
        }
    }
    
    fun stop() {
        if (isActive) {
            getSystemService(context, NotificationManager::class.java)?.deleteNotificationChannel(NOTIFICATION_CHANNEL_ID)
            controllers.forEach { it.stop() }
            collectorJob?.cancel()
            collectorJob = null
            isActive = false
        }
    }

    fun dismiss(group: ChatNotificationGroup) {
        manager.cancel(group.notificationId)
        notificationRepo.dismiss(group)
    }
    
    fun getActiveNotification(group: ChatNotificationGroup): Notification? {
        return manager.activeNotifications.find { it.id == group.notificationId }?.notification
    }

    fun notify(group: ChatNotificationGroup, notification: Notification) {
        if (ActivityCompat.checkSelfPermission(context, Manifest.permission.POST_NOTIFICATIONS) == PackageManager.PERMISSION_GRANTED) {
            manager.notify(group.notificationId, notification)
        }
    }

    private fun show(notification: ChatNotification) {
        val appNotification = when (notification) {
            is MessageNotification -> messageController.getAppNotification(notification)
            is CallNotification -> callController.getAppNotification(notification)
        } ?: return
        notify(notification.group, appNotification)
    }
}