package nand.apps.chat.notification

import android.app.Notification
import android.app.PendingIntent
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.os.Build
import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.MainActivity
import nand.apps.chat.intent.putExtra
import nand.apps.chat.model.avatar.ChatAvatar
import nand.apps.chat.model.avatar.ChatUserAvatar
import nand.apps.chat.model.notification.ChatNotification
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.repo.ChatAvatarRepo
import org.koin.core.Koin

private const val PENDING_INTENT_FLAGS = PendingIntent.FLAG_UPDATE_CURRENT or PendingIntent.FLAG_MUTABLE

abstract class AndroidNotificationReceiver<T : ChatNotification>(
    private val context: Context,
    koin: Koin,
) {

    private val avatarRepo: ChatAvatarRepo by koin.inject()
    
    private val receiver = object : BroadcastReceiver() {
        override fun onReceive(context: Context?, intent: Intent?) {
            intent?.let { onIntent(it) }
        }
    }
    
    protected fun registerReceiver(filter: IntentFilter) {
        val receiverFlags = when {
            Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU -> Context.RECEIVER_NOT_EXPORTED
            else -> 0
        }
        context.registerReceiver(receiver, filter, receiverFlags)
    }
    
    protected fun unregisterReceiver() = context.unregisterReceiver(receiver)
    
    abstract fun start(coroutineScope: CoroutineScope)
    
    abstract fun stop()
    
    abstract fun getAppNotification(notification: T): Notification?
    
    abstract fun onIntent(intent: Intent)

    protected fun getAvatarBitmap(avatar: ChatAvatar): Bitmap? {
        return (avatar as? ChatUserAvatar)?.path?.let { path ->
            BitmapFactory.decodeFile(path.toString())
        }
    }
    
    protected fun getBroadcastIntent(request: AndroidNotificationRequest, intent: Intent): PendingIntent {
        return PendingIntent.getBroadcast(context, request.code, intent, PENDING_INTENT_FLAGS)
    }
    
    protected fun getActivityIntent(request: AndroidNotificationRequest, intent: Intent): PendingIntent {
        return PendingIntent.getActivity(context, request.code, intent, PENDING_INTENT_FLAGS)
    }

    protected fun getOpenConversationIntent(conversationUid: ContactUid): PendingIntent {
        return getActivityIntent(
            request = AndroidNotificationRequest.OPEN_CONVERSATION,
            intent = Intent(context, MainActivity::class.java)
                .putExtra(INTENT_KEY_CONVERSATION_UID, conversationUid),
        )
    }
}