package nand.apps.chat.permission

import android.Manifest
import android.content.pm.PackageManager
import android.os.Build
import androidx.activity.result.contract.ActivityResultContracts.RequestMultiplePermissions
import androidx.core.app.ActivityCompat
import androidx.lifecycle.lifecycleScope
import nand.apps.chat.MainActivity
import nand.apps.chat.audio.AndroidAudioRecorder
import nand.apps.chat.audio.AudioRecorder
import nand.apps.chat.model.event.*
import nand.apps.chat.notification.AndroidNotificationController
import nand.apps.chat.event.ChatEventHandler
import org.koin.core.Koin

class PermissionManager(
    private val activity: MainActivity,
    koin: Koin,
) {

    private val eventHandler: ChatEventHandler by koin.inject()
    private val notificationController: AndroidNotificationController by koin.inject()
    private val audioRecorder: AudioRecorder by koin.inject()
    
    private val lifecycleScope get() = activity.lifecycleScope

    private val requestPermission = activity.registerForActivityResult(RequestMultiplePermissions()) { permissions ->
        permissions.forEach { (permission, isGranted) ->
            when (permission) {
                Manifest.permission.POST_NOTIFICATIONS -> {
                    if (isGranted) {
                        notificationController.start(lifecycleScope)
                    } else {
                        notificationController.stop()
                    }
                }
                Manifest.permission.READ_EXTERNAL_STORAGE,
                Manifest.permission.READ_MEDIA_IMAGES,
                Manifest.permission.READ_MEDIA_VIDEO,
                Manifest.permission.READ_MEDIA_AUDIO -> {
                    if (isGranted) eventHandler += ReloadFileBrowserEvent
                }
                Manifest.permission.RECORD_AUDIO -> {
                    (audioRecorder as? AndroidAudioRecorder)?.run {
                        if (isGranted) resume() else disable()
                    }
                }
            }
        }
    }

    fun requestPermissions(
        vararg permissions: String,
        onAlreadyGranted: (String) -> Unit = {},
    ) {
        val toRequest = mutableListOf<String>()
        permissions.forEach { permission ->
            if (ActivityCompat.checkSelfPermission(activity, permission) != PackageManager.PERMISSION_GRANTED) {
                toRequest += permission
            } else {
                onAlreadyGranted(permission)
            }
        }
        if (toRequest.isNotEmpty()) {
            requestPermission.launch(toRequest.toTypedArray())
        }
    }

    fun requestNotificationPermission() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
            requestPermissions(Manifest.permission.POST_NOTIFICATIONS) {
                notificationController.start(lifecycleScope)
            }
        } else {
            notificationController.start(lifecycleScope)
        }
    }

    fun requestStoragePermissions() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
            requestPermissions(
                Manifest.permission.READ_MEDIA_IMAGES,
                Manifest.permission.READ_MEDIA_VIDEO,
                Manifest.permission.READ_MEDIA_AUDIO,
            )
        } else {
            requestPermissions(Manifest.permission.READ_EXTERNAL_STORAGE)
        }
    }
}