package nand.apps.chat.profile

import android.content.Context
import android.security.keystore.KeyGenParameterSpec
import android.security.keystore.KeyProperties
import androidx.preference.PreferenceManager
import nand.apps.chat.log.getLogger
import java.security.KeyStore
import javax.crypto.Cipher
import javax.crypto.KeyGenerator
import javax.crypto.SecretKey
import javax.crypto.spec.IvParameterSpec
import kotlin.io.encoding.Base64
import kotlin.io.encoding.ExperimentalEncodingApi

private const val KEYSTORE_PROVIDER = "AndroidKeyStore"
private const val KEYSTORE_ALIAS = "login_credentials"
private const val KEYSTORE_PURPOSES = KeyProperties.PURPOSE_ENCRYPT or KeyProperties.PURPOSE_DECRYPT
private const val KEYSTORE_ALGORITHM = KeyProperties.KEY_ALGORITHM_AES
private const val KEYSTORE_BLOCK_MODE = KeyProperties.BLOCK_MODE_CBC
private const val KEYSTORE_PADDING = KeyProperties.ENCRYPTION_PADDING_PKCS7
private const val KEYSTORE_TRANSFORMATION = "$KEYSTORE_ALGORITHM/$KEYSTORE_BLOCK_MODE/$KEYSTORE_PADDING"

private const val PASSWORD_KEY = "login_password"

@OptIn(ExperimentalEncodingApi::class)
class AndroidCredentialsStore(context: Context) : ProfileCredentialsStore {
    
    private val keystore by lazy { KeyStore.getInstance(KEYSTORE_PROVIDER).apply { load(null) } }
    private val preferences = PreferenceManager.getDefaultSharedPreferences(context)
    private val logger = getLogger("ChatLoginKeystore")
    
    override fun savePassword(password: String) {
        try {
            preferences.edit().putString(PASSWORD_KEY, encrypt(password)).apply()
        } catch (e: Exception) {
            logger.warn("Failed to cache login password", e)
        }
    }
    
    override fun loadPassword(): String? {
        return try {
            preferences.getString(PASSWORD_KEY, null)?.let { decrypt(it) }
        } catch (e: Exception) {
            logger.error("Failed to decrypt login password", e)
            null
        }
    }

    override fun reset() {
        preferences.edit().remove(PASSWORD_KEY).apply()
    }

    private fun getOrCreateKey(): SecretKey {
        val existingKey = keystore.getEntry(KEYSTORE_ALIAS, null) as? KeyStore.SecretKeyEntry
        return existingKey?.secretKey ?: createKey()
    }
    
    private fun createKey(): SecretKey {
        val parameterSpec = KeyGenParameterSpec.Builder(KEYSTORE_ALIAS, KEYSTORE_PURPOSES).run {
            setBlockModes(KEYSTORE_BLOCK_MODE)
            setEncryptionPaddings(KEYSTORE_PADDING)
            setUserAuthenticationRequired(false)
            setRandomizedEncryptionRequired(true)
            build()
        }
        return KeyGenerator.getInstance(KEYSTORE_ALGORITHM, KEYSTORE_PROVIDER).run {
            init(parameterSpec)
            generateKey()
        }
    }

    private fun encrypt(inputText: String): String {
        val cipher = Cipher.getInstance(KEYSTORE_TRANSFORMATION)
        cipher.init(Cipher.ENCRYPT_MODE, getOrCreateKey())

        val encryptedBytes = cipher.doFinal(inputText.encodeToByteArray())
        val iv = cipher.iv

        val encryptedDataWithIV = ByteArray(iv.size + encryptedBytes.size)
        System.arraycopy(iv, 0, encryptedDataWithIV, 0, iv.size)
        System.arraycopy(encryptedBytes, 0, encryptedDataWithIV, iv.size, encryptedBytes.size)
        return Base64.encode(encryptedDataWithIV)
    }

    private fun decrypt(data: String): String {
        val encryptedDataWithIV = Base64.decode(data)
        val cipher = Cipher.getInstance(KEYSTORE_TRANSFORMATION)
        val iv = encryptedDataWithIV.copyOfRange(0, cipher.blockSize)
        cipher.init(Cipher.DECRYPT_MODE, getOrCreateKey(), IvParameterSpec(iv))

        val encryptedData = encryptedDataWithIV.copyOfRange(cipher.blockSize, encryptedDataWithIV.size)
        val decryptedBytes = cipher.doFinal(encryptedData)
        return decryptedBytes.decodeToString()
    }
}