package nand.apps.chat.di

import io.ktor.client.*
import nand.apps.chat.BuildConfig.MOCK_ENGINE
import nand.apps.chat.dev.DevCommandDefaultHandlers
import nand.apps.chat.embed.EmbedPreviewFactory
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.engine.DefaultChatEngineCallback
import nand.apps.chat.engine.mock.getMockKoinModule
import nand.apps.chat.engine.tox.getToxKoinModule
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.net.TempFileCache
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.DefaultBackgroundWorker
import nand.apps.chat.monitor.AutoAwayMonitor
import nand.apps.chat.net.FileUploader
import nand.apps.chat.net.packet.ChatPacketConverter
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.repo.*
import nand.apps.chat.text.emoji.EmojiRepo
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.ui.app.AppViewModel
import nand.apps.chat.ui.chat.screen.ChatScreenViewModel
import nand.apps.chat.ui.contacts.ContactListViewModel
import nand.apps.chat.ui.dialog.image.ImageCropperViewModel
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.net.packet.MessagePacketHandler
import nand.apps.chat.profile.ProfileBrowser
import nand.apps.chat.ui.file.ChatFilePickerViewModel
import nand.apps.chat.ui.friends.AddFriendDialogViewModel
import nand.apps.chat.ui.groups.CreateGroupDialogViewModel
import nand.apps.chat.ui.login.LoginScreenViewModel
import nand.apps.chat.ui.profile.SelfProfileDialogViewModel
import nand.apps.chat.ui.profile.SelfProfileRowViewModel
import nand.apps.chat.ui.profile.auth.ProfileAuthenticationDialogViewModel
import nand.apps.chat.ui.profile.export.ProfileExportDialogViewModel
import nand.apps.chat.ui.profile.password.LoginPasswordDialogViewModel
import nand.apps.chat.ui.profile.selector.ProfileSelectorViewModel
import nand.apps.chat.ui.profile.username.LoginUsernameDialogViewModel
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.tutorial.TutorialViewModel
import nand.apps.chat.util.createDefaultHttpClient
import okio.FileSystem
import okio.SYSTEM
import org.koin.core.Koin
import org.koin.core.KoinApplication
import org.koin.core.context.startKoin
import org.koin.core.module.dsl.factoryOf
import org.koin.core.module.dsl.singleOf
import org.koin.dsl.KoinAppDeclaration
import org.koin.dsl.module

fun initKoin(
    mockEngine: Boolean = MOCK_ENGINE,
    appDeclaration: KoinAppDeclaration = {},
): KoinApplication {
    val registeredModules = module {
        // Global
        single<Koin> { getKoin() }
        factory<HttpClient> { createDefaultHttpClient() }
        single<FileSystem> { FileSystem.SYSTEM }
        single<ChatBackgroundWorker> { DefaultBackgroundWorker() }
        singleOf(::ChatFileSystem)
        singleOf(::TempFileCache)
        singleOf(::FileUploader)
        singleOf(::AutoAwayMonitor)
        singleOf(::ChatPacketConverter)
        singleOf(::DateTimeFormatter)
        singleOf(::ProfileBrowser)
        
        // Core chat engine
        single<ChatEngineConstraints> {
            val engine: ChatEngine by getKoin().inject()
            engine.constraints
        }
        single<ChatEngineCallback> { DefaultChatEngineCallback(get()) }
        
        // Repos
        singleOf(::SettingsRepo)
        singleOf(::FileTransferRepo)
        singleOf(::ChatAvatarRepo)
        singleOf(::SelfProfileRepo)
        singleOf(::CallRepo)
        singleOf(::VoicemailRepo)
        singleOf(::ChatMessageRepo)
        singleOf(::ContactRepo)
        singleOf(::ChatPacketRepo)
        singleOf(::NotificationRepo)
        singleOf(::EmojiRepo)
        
        // Event handlers
        singleOf(::ChatEventHandler)
        singleOf(::MessagePacketHandler)
        
        // UI + view models
        singleOf(::EmbedPreviewFactory)
        factoryOf(::AppViewModel)
        factoryOf(::LoginScreenViewModel)
        factoryOf(::ChatScreenViewModel)
        factoryOf(::ContactListViewModel)
        factoryOf(::SelfProfileRowViewModel)
        factoryOf(::SelfProfileDialogViewModel)
        factoryOf(::ProfileExportDialogViewModel)
        factoryOf(::ProfileAuthenticationDialogViewModel)
        factoryOf(::ProfileSelectorViewModel)
        factoryOf(::LoginUsernameDialogViewModel)
        factoryOf(::LoginPasswordDialogViewModel)
        factoryOf(::ImageCropperViewModel)
        factoryOf(::AddFriendDialogViewModel)
        factoryOf(::CreateGroupDialogViewModel)
        factoryOf(::ChatFilePickerViewModel)
        factoryOf(::SettingsViewModel)
        factoryOf(::TutorialViewModel)
        
        // Developer tools
        singleOf(::DevCommandDefaultHandlers)
    }
    return startKoin {
        modules(registeredModules)
        // TODO - Make this more modular
        //  Not really necessary for now since Tox is the only engine
        val engineModules = when {
            mockEngine -> getMockKoinModule()
            else -> getToxKoinModule()
        }
        modules(engineModules)
        appDeclaration()
    }
}