package nand.apps.chat.embed

import io.ktor.client.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import nand.apps.chat.embed.amazon.AmazonUrlParser
import nand.apps.chat.embed.ebay.EbayListingUrlParser
import nand.apps.chat.embed.fourchan.FourChanUrlParser
import nand.apps.chat.embed.mediawiki.MediaWikiUrlParser
import nand.apps.chat.embed.mediawiki.WikipediaUrlParser
import nand.apps.chat.embed.stackexchange.StackExchangeUrlParser
import nand.apps.chat.embed.youtube.channel.YoutubeChannelUrlParser
import nand.apps.chat.embed.youtube.video.YoutubeVideoUrlParser
import nand.apps.chat.log.getLogger
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.util.getHostName
import kotlin.coroutines.cancellation.CancellationException

class EmbedPreviewFactory(
    val client: HttpClient,
    private val settingsRepo: SettingsRepo,
) {

    private val parsers = listOf(
        AmazonUrlParser(),
        EbayListingUrlParser(),
        FourChanUrlParser(),
        YoutubeVideoUrlParser(),
        YoutubeChannelUrlParser(),
        WikipediaUrlParser(),
        StackExchangeUrlParser(),
        MediaWikiUrlParser(),
        DefaultUrlParser(),
    )
    
    private val cache = mutableMapOf<String, EmbedPreviewData?>()
    private val logger = getLogger("EmbedPreviewFactory")
    
    fun isHostAllowed(url: String): Boolean {
        val enabledHosts = settingsRepo.settings.embedSettings.enabledHosts
        return enabledHosts.isEmpty() || enabledHosts.contains(url.getHostName())
    }
    
    fun getCachedEmbedData(url: String) = cache[url]

    suspend fun getEmbedData(url: String) = cache[url] ?: run {
        if (isHostAllowed(url)) {
            generateEmbedData(url).also { cache[url] = it }
        } else {
            null
        }
    }
    
    private suspend fun generateEmbedData(url: String): EmbedPreviewData? {
        try {
            for (parser in parsers) {
                if (parser.matches(url)) {
                    withContext(Dispatchers.IO) {
                        parser.parse(url, client)
                    }?.let {
                        return it
                    }
                }
            }
        } catch (ignore: CancellationException) {
            // Coroutine cancelled before task was finished - no need to log this
        } catch (e: Exception) {
            logger.error("Exception when generating embed preview for $url", e)
        }
        return null
    }
}