package nand.apps.chat.embed.ebay

import com.fleeksoft.ksoup.nodes.Document
import io.ktor.client.*
import nand.apps.chat.embed.EmbedUrlParser
import nand.apps.chat.util.getHtmlDocument

private val LISTING_URL_REGEX = Regex("(?:https?://)?(?:www)?\\.ebay\\.[a-z.]+(?:/.*)?/itm/[0-9]+(?:\\?.*)?")

class EbayListingUrlParser : EmbedUrlParser {

    override fun matches(url: String) = LISTING_URL_REGEX.matches(url)

    override suspend fun parse(url: String, client: HttpClient): EbayListingMetadata? {
        val doc = client.getHtmlDocument(url) ?: return null
        val mainContent = doc.selectFirst("#mainContent") ?: return null
        val name = mainContent.selectFirst(".x-item-title__mainTitle")?.text()
        val price = mainContent.selectFirst(".x-price-primary")?.text()
        val condition = mainContent.selectFirst(".x-item-condition-text .ux-textspans")?.text()
        val imageUrl = doc.selectFirst("#PicturePanel div.ux-image-carousel-item.active.image img")?.attr("src")
        val sellerMetadata = doc.parseSellerCard() ?: doc.parseSellerSection()
        return if (name != null && price != null && imageUrl != null) {
            EbayListingMetadata(
                url = url,
                name = name,
                price = price,
                condition = condition,
                imageUrl = imageUrl,
                seller = sellerMetadata,
            )
        } else {
            null
        }
    }

    /**
     * New-style card that contains the seller's avatar
     */
    private fun Document.parseSellerCard(): EbaySellerMetadata? {
        val card = selectFirst("#mainContent .x-sellercard-atf") ?: return null
        val info = card.selectFirst(".x-sellercard-atf__info__about-seller")
        val name = info?.selectFirst("a")?.text() ?: return null
        val reviewCount = info.selectFirst(".x-sellercard-atf__about-seller")?.text()?.trim('(', ')')
        val rating = card.selectFirst("ul.x-sellercard-atf__data-item-wrapper > li > a")?.text()?.substringBefore(' ')
        val thumbnailUrl = card.selectFirst(".x-sellercard-atf__image img")?.attr("src")
        return EbaySellerMetadata(
            name = name,
            reviewCount = reviewCount,
            rating = rating,
            thumbnailUrl = thumbnailUrl,
        )
    }

    /**
     * Old-style seller info that does not contain an avatar + slightly different copy
     */
    private fun Document.parseSellerSection(): EbaySellerMetadata? {
        val sectionLinks = select(".x-about-this-seller ul.ux-seller-section__content li")
        val actionLinks = sectionLinks.firstOrNull()?.select(".ux-seller-section__item--seller a.ux-action")
        val name = actionLinks?.firstOrNull()?.text() ?: return null
        val reviewCount = actionLinks.getOrNull(1)?.text()
        val rating = sectionLinks.getOrNull(1)?.text()?.substringBefore(' ')
        val thumbnailUrl = selectFirst(".d-stores-info-categories__container__info__image img")?.attr("src")
        return EbaySellerMetadata(
            name = name,
            reviewCount = reviewCount,
            rating = rating,
            thumbnailUrl = thumbnailUrl,
        )
    }
}