package nand.apps.chat.embed.mediawiki

import androidx.compose.ui.util.fastForEach
import com.fleeksoft.ksoup.Ksoup
import com.fleeksoft.ksoup.nodes.Element
import com.fleeksoft.ksoup.nodes.TextNode
import io.ktor.client.*
import kotlinx.serialization.json.JsonObject
import nand.apps.chat.embed.EmbedUrlParser
import nand.apps.chat.util.*

private val PAGE_URL_REGEX = Regex("(?:https?://)?((?:[\\w-]+\\.)?[\\w-]+\\.[a-z]{1,32})/(?:wiki|w)/([^/#?&]+)(?:[#?&].*)?")
private const val THUMBNAIL_SIZE = 128

class MediaWikiUrlParser : EmbedUrlParser {

    override fun matches(url: String) = PAGE_URL_REGEX.matches(url) ||
            url.getHostName()?.contains("wiki") == true

    override suspend fun parse(url: String, client: HttpClient): MediaWikiPageMetadata? {
        return parseActionApi(url, client) ?: parseRawHtml(url, client)
    }
    
    private suspend fun parseActionApi(url: String, client: HttpClient): MediaWikiPageMetadata? {
        val (_, hostName, pageName) = PAGE_URL_REGEX.matchEntire(url)?.groupValues ?: return null
        if (pageName.startsWith("File:") || pageName.startsWith("Special:")) return null
        val json = client.getJsonObject("https://$hostName/w/api.php?action=parse&format=json&page=$pageName") ?: return null
        val parse = json["parse"] as? JsonObject ?: return null
        val title = parse["title"].getString() ?: return null
        var summary: String? = null
        var thumbnail = parse["images"].jsonArraySafe().firstOrNull()?.getString()?.let { fileName ->
            MediaWikiImageMetadata(
                url = "",
                fileName = fileName,
                width = 1,
                height = 1,
            )
        }
        parse.find("text", "*").getString()?.let { html ->
            try {
                val doc = Ksoup.parse(html, url)
                summary = doc.parseSummary()
                doc.parseArticleImage()?.let { thumbnail = it }
            } catch (ignore: Exception) {
            }
        }
        return MediaWikiPageMetadata(
            url = url,
            title = title,
            summary = summary,
            thumbnail = thumbnail?.fixUrl(url),
        )
    }
    
    private suspend fun parseRawHtml(url: String, client: HttpClient): MediaWikiPageMetadata? {
        val doc = client.getHtmlDocument(url)
        val root = doc?.selectFirst("#mw-content-text") ?: return null
        val metaTitle = doc.selectFirst("meta[property=\"og:title\"]")?.attr("content")
        val pageTitle = doc.selectFirst(".mw-page-title-main")?.text()?.takeIf { it.isNotBlank() }
        val firstHeading = doc.selectFirst(".mw-first-heading")?.text()?.takeIf { it.isNotBlank() }
        val title = pageTitle ?: metaTitle ?: firstHeading ?: return null
        return MediaWikiPageMetadata(
            url = url,
            title = title,
            summary = root.parseSummary(),
            thumbnail = root.parseArticleImage()?.fixUrl(url),
        )
    }
    
    private fun Element.parseSummary(): String? {
        val root = selectFirst("div.mw-parser-output") ?: return null
        val paragraph = root.findFirstParagraph() ?: selectNonEmpty("div.mw-parser-output p") ?: return null
        paragraph.forEachNode { node ->
            if (node is Element) {
                val tagName = node.tagName()
                if (tagName == "sup" && node.hasClass("reference") ||
                    tagName == "a" && node.parent()?.tagName() == "sup" ||
                    tagName == "br" && node.siblingIndex() == 0 && node.parentNode() == paragraph
                ) {
                    node.remove()
                }
            }
        }
        return paragraph.html()
    }
    
    private fun Element.findFirstParagraph(): Element? {
        val builtParagraph = Element("p")
        childNodes().fastForEach { node ->
            when (node) {
                is TextNode -> builtParagraph.addChildren(node)
                is Element -> when {
                    node.tagName() == "p" && node.hasText() -> {
                        return if (builtParagraph.hasText()) builtParagraph else node
                    }
                    node.isBlock() -> if (builtParagraph.hasText()) {
                        return builtParagraph
                    }
                    !node.isBlock() -> {
                        builtParagraph.addChildren(node)
                    }
                }
            }
        }
        return if (builtParagraph.hasText()) builtParagraph else null
    }
    
    private fun Element.parseArticleImage(): MediaWikiImageMetadata? {
        val firstImage = selectFirst(".mw-parser-output a.mw-file-description img.mw-file-element")
            ?: selectFirst(".mw-parser-output a.image img")
            ?: selectFirst("a.image[href*=\"File:\"] img")
        val firstImageLink = firstImage?.parentNode() as? Element
        return firstImageLink?.attr("href")?.let {
            val fileName = it.substringAfter("File:")
            MediaWikiImageMetadata(
                url = fileName,
                fileName = fileName,
                width = firstImage.attr("width").toIntOrNull() ?: 1,
                height = firstImage.attr("height").toIntOrNull() ?: 1,
            )
        }
    }
    
    private fun Element.selectNonEmpty(cssQuery: String) =
        select(cssQuery).find { it.text().isNotBlank() }
    
    private fun MediaWikiImageMetadata.fixUrl(pageUrl: String) = copy(
        url = when {
            fileName.isUrl() -> fileName
            else -> {
                val basePath = pageUrl.substringBeforeLast('/')
                "$basePath/Special:Filepath/$fileName?width=$THUMBNAIL_SIZE&height=$THUMBNAIL_SIZE"
            }
        },
    )
}