package nand.apps.chat.embed.youtube.channel

import io.ktor.client.*
import kotlinx.serialization.json.*
import nand.apps.chat.embed.EmbedUrlParser
import nand.apps.chat.log.getLogger
import nand.apps.chat.util.*

private val YOUTUBE_URL_REGEX = Regex("(?:https?://)?(?:www\\.|m\\.)?youtube\\.com/(?:@\\w+|channel/[\\w-]+|user/\\w+)(?:/.*)?")
private const val JS_PREFIX = "var ytInitialData = {"

class YoutubeChannelUrlParser : EmbedUrlParser {

    private val logger = getLogger("Youtube")

    override fun matches(url: String) = YOUTUBE_URL_REGEX.matches(url)

    override suspend fun parse(url: String, client: HttpClient): YoutubeChannelMetadata? {
        val document = client.getHtmlDocument(url) ?: return null
        val jsMetadata = document.select("script").map { it.data() }.find { it.startsWith(JS_PREFIX) }?.let { jsCode ->
            try {
                parseJs(url, jsCode)
            } catch (e: Exception) {
                logger.error("Failed to parse JSON metadata: $url", e)
                null
            }
        }
        return jsMetadata
    }

    private fun parseJs(url: String, jsCode: String): YoutubeChannelMetadata {
        val jsonString = jsCode.substring(jsCode.indexOf('{'), jsCode.indexOf("};") + 1)
        val json: JsonObject = Json.decodeFromString(jsonString)
        val metadata = json.find("metadata", "channelMetadataRenderer").jsonObjectSafe()
        val microformat = json.find("microformat", "microformatDataRenderer").jsonObjectSafe()
        val thumbnails = (metadata["avatar"] ?: microformat["thumbnail"]).jsonObjectSafe()["thumbnails"]?.jsonArray ?: JsonArray(emptyList())
        val headerRenderer = json.find("header", "c4TabbedHeaderRenderer").jsonObjectSafe()
        return YoutubeChannelMetadata(
            url = url,
            id = metadata["externalId"].getString(),
            name = (metadata["title"] ?: microformat["title"]).getString(),
            handle = headerRenderer["channelHandleText"].jsonObjectSafe().parseRunningText(),
            description = (metadata["description"] ?: microformat["description"]).getString(),
            thumbnailUrl = thumbnails.firstOrNull()?.jsonObject?.get("url").getString(),
            subscriberCount = headerRenderer.find("subscriberCountText", "simpleText")?.getString(),
            videoCount = headerRenderer["videosCountText"].jsonObjectSafe().parseRunningText(),
        )
    }
    
    private fun JsonObject.parseRunningText() = this["runs"]?.jsonArraySafe()?.joinToString("") {
        it.jsonObject["text"].getString().orEmpty()
    }
}