package nand.apps.chat.embed.youtube.video

import com.fleeksoft.ksoup.nodes.DataNode
import com.fleeksoft.ksoup.nodes.Document
import com.fleeksoft.ksoup.nodes.Element
import io.ktor.client.*
import kotlinx.datetime.Instant
import kotlinx.datetime.format.DateTimeComponents
import kotlinx.datetime.format.FormatStringsInDatetimeFormats
import kotlinx.datetime.format.byUnicodePattern
import kotlinx.serialization.json.*
import nand.apps.chat.embed.EmbedUrlParser
import nand.apps.chat.embed.youtube.channel.YoutubeChannelMetadata
import nand.apps.chat.log.getLogger
import nand.apps.chat.util.*

private val YOUTUBE_URL_REGEX = Regex("(?:https?://)?(?:youtu\\.be/|(?:www\\.|m\\.)?youtube(?:-nocookie)?\\.com/(?:watch|v|embed|live|shorts)(?:\\.php)?(?:\\?.*v=|/))([\\w\\-]{7,15})(?:[?&][\\w\\-]+=[\\w\\-]+)*")
private const val VIDEO_URL = "https://www.youtube.com/watch?v=$1"
private const val THUMBNAIL_URL = "https://i.ytimg.com/vi/$1/hqdefault.jpg"

@OptIn(FormatStringsInDatetimeFormats::class)
private val DATE_FORMAT = DateTimeComponents.Format {
    byUnicodePattern("yyyy-MM-dd'T'HH:mm:ssxxx")
}

class YoutubeVideoUrlParser : EmbedUrlParser {
    
    private val logger = getLogger("Youtube")

    override fun matches(url: String) = YOUTUBE_URL_REGEX.matches(url)
    
    override suspend fun parse(url: String, client: HttpClient): YoutubeVideoMetadata? {
        val id = YOUTUBE_URL_REGEX.matchEntire(url)?.groupValues?.find { it.length == 11 } ?: return null
        val document = client.getHtmlDocument(VIDEO_URL.replace("$1", id)) ?: return null
        
        // Top-level basic metadata
        val title = document.title().substringBeforeLast(" - YouTube")
        val basicMetadata = YoutubeVideoMetadata(
            id = id,
            url = url,
            title = document.getMetaContent("title") ?: title,
            description = document.getMetaContent("description"),
            thumbnailUrl = getDefaultThumbnail(id),
        )
        
        // The first content block contains more metadata
        val voMetadata = document.selectFirst("body > div[itemtype=\"http://schema.org/VideoObject\"]")?.let { metadata ->
            try {
                parseVideoObject(url, metadata)
            } catch (e: Exception) {
                logger.error("Failed to parse VideoObject metadata", e)
                null
            }
        }
        
        // Javascript contains the most video metadata
        val dataJson = document.findJsonData()
        val jsMetadata = dataJson["InitialPlayerResponse"]?.let { initialPlayerResponse ->
            try {
                parseJson(
                    url = url,
                    id = id,
                    initialPlayerResponse = initialPlayerResponse,
                    initialData = dataJson["InitialData"],
                )
            } catch (e: Exception) {
                logger.error("Failed to parse JSON metadata: $url", e)
                null
            }
        }
        
        return jsMetadata ?: voMetadata ?: basicMetadata.takeIf { it.title.isNotEmpty() }
    }
    
    private fun parseVideoObject(url: String, div: Element): YoutubeVideoMetadata? {
        val metadata = div.select("[itemprop]").mapNotNull { element ->
            val parentKey = if (element.parent() != div) element.parent()?.attr("itemprop") + "." else ""
            val key = parentKey + element.attr("itemprop")
            val value = when (element.tagName()) {
                "meta" -> "content"
                "link" -> "href"
                else -> null
            }?.let { element.attr(it) }
            if (key.isNotBlank() && !value.isNullOrBlank()) {
                key to value
            } else {
                null
            }
        }.toMap()
        val id = metadata["identifier"] ?: return null
        return YoutubeVideoMetadata(
            url = url,
            id = id,
            title = metadata["name"].orEmpty(),
            description = metadata["description"],
            lengthSecs = metadata["duration"]?.parsePtDuration() ?: 0,
            viewCount = metadata["interactionCount"].toLongOrZero(),
            uploadDate = metadata["uploadDate"].parseDate(),
            publishDate = metadata["datePublished"].parseDate(),
            thumbnailUrl = getDefaultThumbnail(id),
            channel = YoutubeChannelMetadata(
                url = metadata["author.url"],
                name = metadata["author.name"],
            ),
            isFamilySafe = metadata["isFamilyFriendly"].toBoolean(),
        )
    }
    
    private fun parseJson(
        url: String,
        id: String,
        initialPlayerResponse: String,
        initialData: String?,
    ): YoutubeVideoMetadata {
        val playerJson: JsonObject = Json.decodeFromString(initialPlayerResponse)
        val dataJson: JsonObject? = initialData?.let { Json.decodeFromString(it) }
        val videoDetails = playerJson["videoDetails"].jsonObjectSafe()
        val microFormat = playerJson.find("microformat", "playerMicroformatRenderer").jsonObjectSafe()
        val playabilityStatus = playerJson["playabilityStatus"].jsonObjectSafe()
        val status = YoutubeVideoStatus[playabilityStatus["status"].getString()]
        val contentsJson = dataJson?.find("contents", "twoColumnWatchNextResults", "results", "results", "contents") as? JsonArray
        val errorMetadata = parseError(playabilityStatus)
        val liveMetadata = parseLiveMetadata(microFormat, contentsJson)
        return YoutubeVideoMetadata(
            url = url,
            id = id,
            title = videoDetails["title"].getString().orEmpty(),
            description = videoDetails["shortDescription"].getString()?.replace("\r", ""),
            status = status,
            lengthSecs = videoDetails["lengthSeconds"].getInt(),
            viewCount = videoDetails["viewCount"].getLong(),
            uploadDate = microFormat["uploadDate"].parseDate(),
            publishDate = microFormat["publishDate"].parseDate(),
            thumbnailUrl = getDefaultThumbnail(id),
            channel = YoutubeChannelMetadata(
                url = microFormat["ownerProfileUrl"].getString(),
                name = videoDetails["author"].getString(),
                id = videoDetails["channelId"].getString(),
                thumbnailUrl = parseChannelThumbnail(contentsJson),
            ),
            isFamilySafe = microFormat["isFamilySafe"].getBoolean(),
            isUnlisted = microFormat["isUnlisted"].getBoolean(),
            isShortsEligible = microFormat["isShortsEligible"].getBoolean(),
            isUpcoming = videoDetails["isUpcoming"].getBoolean(),
            isMembersOnly = parseMembersOnly(contentsJson),
            liveBroadcast = liveMetadata,
            error = errorMetadata,
        )
    }
    
    private fun parseError(playabilityStatus: JsonObject): YoutubeVideoErrorMetadata? {
        val errorScreen = playabilityStatus["errorScreen"]?.jsonObject ?: return null
        val errorMessage = errorScreen["playerErrorMessageRenderer"]?.jsonObject
        val offerRenderer = errorScreen["playerLegacyDesktopYpcOfferRenderer"]?.jsonObject
        val reason = errorMessage?.get("reason").jsonObjectSafe().parseErrorText() ?: offerRenderer?.get("itemTitle").getString()
        val description = errorMessage?.get("subreason").jsonObjectSafe().parseErrorText() ?: offerRenderer?.get("offerDescription").getString()
        return YoutubeVideoErrorMetadata(
            reason = reason,
            description = description,
        )
    }
    
    private fun parseLiveMetadata(
        microFormat: JsonObject,
        contentsJson: JsonArray?,
    ): YoutubeLiveMetadata? {
        return microFormat["liveBroadcastDetails"]?.jsonObject?.let { liveDetails ->
            val isLiveNow = liveDetails["isLiveNow"].getBoolean()
            val viewers = contentsJson?.find(
                "videoPrimaryInfoRenderer",
                "viewCount",
                "videoViewCountRenderer",
                "originalViewCount",
            ).getString()?.toLongOrZero() ?: 0
            YoutubeLiveMetadata(
                isLiveNow = isLiveNow,
                viewers = viewers,
                startDate = liveDetails["startTimestamp"].parseDate(),
                endDate = liveDetails["endTimestamp"].parseDate(),
            )
        }
    }
    
    private fun parseMembersOnly(contentsJson: JsonArray?): Boolean {
        val badgeStyle = contentsJson?.find(
            "videoPrimaryInfoRenderer",
            "badges",
            "metadataBadgeRenderer",
            "style",
        ).getString()
        return badgeStyle == "BADGE_STYLE_TYPE_MEMBERS_ONLY"
    }
    
    private fun parseChannelThumbnail(contentsJson: JsonArray?) = contentsJson?.find(
        "videoSecondaryInfoRenderer",
        "owner",
        "videoOwnerRenderer",
        "thumbnail",
        "thumbnails",
    )?.jsonArraySafe()?.firstOrNull()?.jsonObjectSafe()?.get("url").getString()
    
    private fun Document.getMetaContent(name: String) =
        selectFirst("meta[name=\"$name\"]")?.attr("content")
    
    private fun Document.findJsonData() = select("body > script").mapNotNull {
        if (it.childNodeSize() == 1 && it.childNode(0) is DataNode) {
            val data = it.data()
            if (data.startsWith("var yt")) {
                val key = data.substringBefore(" = {").substringAfter("var yt")
                val value = data.substring(data.indexOf('{'), data.indexOf("};") + 1)
                key to value
            } else {
                null
            }
        } else {
            null
        }
    }.toMap()
    
    private fun JsonObject.parseErrorText(): String? {
        return when {
            containsKey("simpleText") -> get("simpleText").getString()
            containsKey("runs") -> get("runs")?.jsonArray?.mapNotNull {
                if (it is JsonObject) {
                    it["text"].getString()
                } else {
                    null
                }
            }?.joinToString("")
            else -> null
        }
    }
    
    private fun JsonElement?.parseDate() = this.getString().parseDate()
    
    private fun String?.parseDate() = this?.let {
        try {
            Instant.parse(it, DATE_FORMAT).toEpochMilliseconds()
        } catch (ignore: Exception) {
            null
        }
    } ?: 0L

    private fun getDefaultThumbnail(id: String?) = id?.let { THUMBNAIL_URL.replace("$1", id) } ?: THUMBNAIL_URL_ERROR
    
    private fun String.parsePtDuration() = if (startsWith("PT")) {
        val (mins, secs) = substring(2).split('M')
        mins.toIntOrZero() * 60 + secs.substringBefore('S').toIntOrZero()
    } else {
        toIntOrZero()
    }
}