package nand.apps.chat.engine

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.file.ChatFileControl
import nand.apps.chat.model.file.ChatFileControlResponse
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus

interface ChatEngine {

    val version: String
    val constraints: ChatEngineConstraints
    
    suspend fun initialize(profile: ChatProfileData): ChatEngineInitResult
    
    fun start(coroutineScope: CoroutineScope)
    
    fun stop()
    
    /* Self profile */
    
    fun getSelfUserData(): UserData
    
    fun setSelfUserData(name: String, statusMessage: String): Boolean
    
    fun setSelfStatus(status: UserStatus)
    
    fun getSelfAddress(): String
    
    fun resetSelfAddress()
    
    /* Friends */
    
    suspend fun getFriendList(): List<UserData>

    fun sendFriendRequest(address: String, message: String): UserUid?

    fun acceptFriendRequest(requestUid: FriendRequestUid): UserUid?

    fun removeFriend(friendUid: UserUid): Boolean
    
    /* Messaging and packets */

    fun sendMessage(message: ChatMessage): Boolean
    
    fun editMessage(old: ChatMessage, new: ChatMessage): Boolean
    
    fun setMessageReactions(message: ChatMessage, reactions: Set<String>): Boolean
    
    fun removeMessage(message: ChatMessage): Boolean

    fun sendPacket(uid: ContactUid, packet: ByteArray): Boolean

    fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray): Boolean
    
    /* Groups */

    suspend fun getGroupList(): List<ChatGroupData>

    fun getGroupPeerList(groupUid: GroupUid): Map<UserUid, UserData>
    
    fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData?
    
    fun setGroupTitle(groupUid: GroupUid, title: String): Boolean
    
    fun setGroupDescription(groupUid: GroupUid, description: String): Boolean
    
    fun setGroupPassword(groupUid: GroupUid, password: String?): Boolean
    
    fun setGroupNickname(groupUid: GroupUid, name: String): Boolean
    
    fun setGroupAvatar(groupUid: GroupUid, avatarHash: String): Boolean
    
    fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean): Boolean
    
    fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid): Boolean
    
    fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData?
    
    fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole): Boolean
    
    fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid): Boolean
    
    fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean): Boolean
    
    fun joinGroup(groupUid: GroupUid, selfNickname: String, password: String?): Boolean
    
    fun removeGroup(groupUid: GroupUid, partingMessage: String?): Boolean

    /* Contacts (both friends and groups) */

    fun setContactMuted(uid: ContactUid, isMuted: Boolean): Boolean

    /* File transfers */

    fun sendFile(friendUid: UserUid, kind: ChatFileType, fileSize: Long, fileName: String, fileUid: FileUid): Boolean

    fun sendFileChunk(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray): Boolean

    fun sendFileControl(friendUid: UserUid, fileUid: FileUid, control: ChatFileControl): ChatFileControlResponse
    
    /* Audio/video calls */
    
    fun startCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean
    
    fun answerCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean
    
    fun sendCallControl(uid: ContactUid, control: CallControl): Boolean
    
    fun sendAudioFrame(uid: ContactUid, frame: AudioFrame): Boolean
    
    fun sendAudioBitRate(uid: ContactUid, bitRate: Int): Boolean
    
    /* Hashing and encryption */
    
    fun generateHash(data: ByteArray): ByteArray?
}