package nand.apps.chat.engine.mock

import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import nand.apps.chat.BuildConfig.APP_VERSION
import nand.apps.chat.engine.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.file.*
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.repo.*
import nand.apps.chat.text.emoji.isEmojiString
import nand.apps.chat.text.emoji.splitByEmoji
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.CHARACTER_RANGE_HEX
import nand.apps.chat.util.replaceAll
import nand.apps.chat.util.sha256
import nand.apps.chat.util.trimToLength
import okio.ByteString.Companion.toByteString
import org.koin.core.Koin
import kotlin.random.Random

private const val STARTUP_DELAY = 1000L
private const val ITERATE_DELAY = 250L
private const val MOCK_FILE_SIZE = 1_000_000L
private const val MOCK_CHUNK_SIZE = 4096
private const val SELF_UID = "29B5C4BA34C412D230DF374B2840B27DE647561948B9C227BD15E113091DB6B9"

class MockChatEngine(koin: Koin) : ChatEngine {

    private var isRunning = false
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val callRepo: CallRepo by koin.inject()
    private val engineCallback: ChatEngineCallback by koin.inject()
    private var self = SELF_USER
    private var selfNoSpam = 0
    private val friends = MOCK_FRIENDS.associateBy { it.uid }.toMutableMap()
    private val groups = MOCK_GROUPS.associateBy { it.uid }.toMutableMap()
    private val queuedEvents = mutableListOf<MockChatEvent>()
    private val logger = getLogger("MockChatEngine")

    override val version = "MockChatEngine v$APP_VERSION"
    override val constraints = ChatEngineConstraints(
        maxNameLength = 128,
        maxStatusMessageLength = 1007,
        maxFriendRequestLength = 1016,
        maxMessageLength = 1372,
        maxPacketLength = 1372,
        maxGroupNameLength = 48,
        maxGroupDescLength = 512,
        maxGroupPasswordLength = 32,
        maxGroupPartingMessageLength = 128,
        maxFileNameLength = 255,
        maxHostnameLength = 255,
        uidLength = 64,
        addressLength = 76,
        uidCharacterRange = CHARACTER_RANGE_HEX,
    )
    
    override suspend fun initialize(profile: ChatProfileData): ChatEngineInitResult {
        engineCallback.onEngineInit(profile)
        return ChatEngineInitResult.SUCCESS
    }

    override fun start(coroutineScope: CoroutineScope) {
        if (isRunning) return

        isRunning = true
        
        coroutineScope.launch {
            engineCallback.onEngineStart(coroutineScope)
            mockContactConnections()
            while (isRunning) {
                handleMockEvents()
                delay(ITERATE_DELAY)
            }
        }
    }

    override fun stop() {
        isRunning = false
        engineCallback.onEngineStop()
    }

    override fun getSelfUserData() = self.copy(
        avatarPath = avatarRepo.getSelfAvatarPath(),
    )

    override fun setSelfUserData(name: String, statusMessage: String): Boolean {
        self = self.copy(name = name, statusMessage = statusMessage)
        return true
    }

    override fun setSelfStatus(status: UserStatus) {
        self = self.copy(status = status)
    }

    override fun getSelfAddress() = generateUserAddress("${self.name}_$selfNoSpam")
    
    override fun resetSelfAddress() {
        selfNoSpam = Random.nextInt()
    }

    override suspend fun getFriendList() = friends.values.toList()

    override suspend fun getGroupList() = groups.values.toList()

    override fun getGroupPeerList(groupUid: GroupUid) = groups[groupUid]?.peers.orEmpty()

    override fun sendMessage(message: ChatMessage): Boolean {
        logger.debug("sendMessage($message)")
        if (!(message.text.startsWith('/') && handleCommand(message))) {
            if (message.conversationUid is UserUid) {
                queuedEvents += FriendReadReceiptEvent(
                    timestamp = delayedTimestamp(),
                    friendUid = message.conversationUid,
                    messageUid = message.uid,
                )
            }
            return true
        }
        return false
    }

    override fun editMessage(old: ChatMessage, new: ChatMessage): Boolean {
        return true
    }

    override fun setMessageReactions(message: ChatMessage, reactions: Set<String>): Boolean {
        return true
    }

    override fun removeMessage(message: ChatMessage): Boolean {
        return true
    }

    override fun sendPacket(uid: ContactUid, packet: ByteArray): Boolean {
        logger.debug("sendPacket($uid, B[${packet.size}])")
        return true
    }

    override fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray): Boolean {
        logger.debug("sendPeerPacket($groupUid, $peerUid, B[${packet.size}])")
        return true
    }

    override fun sendFile(friendUid: UserUid, kind: ChatFileType, fileSize: Long, fileName: String, fileUid: FileUid): Boolean {
        return true
    }

    override fun sendFileChunk(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray): Boolean {
        return true
    }

    override fun sendFileControl(friendUid: UserUid, fileUid: FileUid, control: ChatFileControl): ChatFileControlResponse {
        logger.debug("SEND: $control to friend $friendUid for file $fileUid")
        if (control == ChatFileControl.RESUME) {
            val fileTransfer = ChatFileTransfer(
                fileUid = fileUid,
                friendUid = friendUid,
                name = "",
                path = "",
                size = MOCK_FILE_SIZE,
                direction = ChatFileTransferDirection.INCOMING,
            )
            var timestamp = currentTimeMillis()
            var offset = 0L
            while (offset < MOCK_FILE_SIZE) {
                queuedEvents += FileChunkEvent(
                    timestamp = timestamp,
                    fileTransfer = fileTransfer,
                    position = offset,
                )
                offset += MOCK_CHUNK_SIZE
                timestamp += 20L
            }
            queuedEvents += FileChunkEvent(
                timestamp = timestamp,
                fileTransfer = fileTransfer,
                position = MOCK_FILE_SIZE,
            )
        }
        return ChatFileControlResponse.OK
    }

    override fun sendFriendRequest(address: String, message: String): UserUid? {
        val uid = UserUid(address.trimToLength(64))
        if (uid.value == SELF_UID) return null
        friends[uid] = createMockUser(
            name = address,
            statusMessage = message,
        )
        return uid
    }

    override fun acceptFriendRequest(requestUid: FriendRequestUid): UserUid {
        logger.debug("ACCEPT: $requestUid")
        val newFriend = createMockUser(
            name = "New Friend #${friends.size}",
            statusMessage = "Some status",
            uid = requestUid.value,
        )
        friends[newFriend.uid] = newFriend
        mockFriendJoinEvents(newFriend)
        return newFriend.uid
    }

    override fun removeFriend(friendUid: UserUid): Boolean {
        return friends.remove(friendUid) != null
    }

    override fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData {
        val newGroup = createMockGroup(
            title = title,
            type = type,
            peers = listOf(SELF_USER),
            visibility = visibility,
            isOnline = true,
        )
        groups[newGroup.uid] = newGroup
        return newGroup
    }

    override fun setGroupTitle(groupUid: GroupUid, title: String): Boolean {
        val group = groups[groupUid] ?: return false
        groups[groupUid] = group.copy(title = title)
        return true
    }

    override fun setGroupDescription(groupUid: GroupUid, description: String): Boolean {
        val group = groups[groupUid] ?: return false
        groups[groupUid] = group.copy(description = description)
        return true
    }

    override fun setGroupPassword(groupUid: GroupUid, password: String?): Boolean {
        val group = groups[groupUid] ?: return false
        groups[groupUid] = group.copy(password = password)
        return true
    }

    override fun setGroupNickname(groupUid: GroupUid, name: String): Boolean {
        val group = groups[groupUid] ?: return false
        if (ChatGroupFeature.EDIT_NICKNAME !in group.features) return false
        val self = group.findSelf(self).copy(name = name)
        group.updatePeer(self)
        return true
    }

    override fun setGroupAvatar(groupUid: GroupUid, avatarHash: String): Boolean {
        val group = groups[groupUid] ?: return false
        val self = group.findSelf(self).copy(avatarPath = avatarRepo.getSelfAvatarPath())
        group.updatePeer(self)
        return true
    }

    override fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean): Boolean {
        return true
    }

    override fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole): Boolean {
        val group = groups[groupUid] ?: return false
        val peer = group.peers[peerUid] ?: return false
        groups[groupUid] = group.copy(
            peers = group.peers.toMutableMap().apply {
                this[peerUid] = peer.copy(role = role)
            }
        )
        val timestamp = currentTimeMillis() + 100
        queuedEvents += GroupPeerRoleChangedEvent(
            timestamp = timestamp,
            groupUid = groupUid,
            sourceUid = group.selfUid,
            targetUid = peerUid,
            role = role,
        )
        return true
    }

    override fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid): Boolean {
        val group = groups[groupUid] ?: return false
        groups[groupUid] = group.copy(
            peers = group.peers.toMutableMap().apply { remove(peerUid) }
        )
        val timestamp = currentTimeMillis() + 100
        queuedEvents += GroupPeerKickedEvent(
            timestamp = timestamp,
            groupUid = groupUid,
            sourceUid = group.selfUid,
            targetUid = peerUid,
        )
        return true
    }

    override fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean): Boolean {
        return true
    }

    override fun joinGroup(groupUid: GroupUid, selfNickname: String, password: String?): Boolean {
        return true
    }

    override fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid): Boolean {
        val friend = friends[friendUid] ?: return false
        val group = groups[groupUid] ?: return false
        if (group.peers.containsKey(friendUid)) return true
        
        val timestamp = delayedTimestamp()
        group.updatePeer(friend)
        queuedEvents += GroupUsersChangedEvent(
            timestamp = timestamp,
            groupUid = groupUid,
        )
        queuedEvents += GroupUserNameEvent(
            timestamp = timestamp,
            groupUid = groupUid,
            peerUid = friendUid,
            name = friend.displayName,
        )
        return true
    }

    override fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData {
        logger.debug("ACCEPT: ${invite.cookie}")
        val newGroup = createMockGroup(
            title = "New Group #${groups.size}",
            type = invite.groupType,
            uid = invite.cookie,
            peers = listOfNotNull(SELF_USER, friends[invite.friendUid]),
        )
        groups[newGroup.uid] = newGroup
        mockGroupJoinEvents(newGroup)
        return newGroup
    }

    override fun removeGroup(groupUid: GroupUid, partingMessage: String?): Boolean {
        return groups.remove(groupUid) != null
    }

    override fun setContactMuted(uid: ContactUid, isMuted: Boolean): Boolean {
        when (uid) {
            is UserUid -> friends[uid] = friends[uid]?.copy(isMuted = isMuted) ?: return false
            is GroupUid -> groups[uid] = groups[uid]?.copy(isMuted = isMuted) ?: return false
            else -> return false
        }
        return true
    }

    override fun startCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        return true
    }

    override fun answerCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        return true
    }

    override fun sendCallControl(uid: ContactUid, control: CallControl): Boolean {
        return true
    }

    override fun sendAudioFrame(uid: ContactUid, frame: AudioFrame): Boolean {
        return false
    }

    override fun sendAudioBitRate(uid: ContactUid, bitRate: Int): Boolean {
        return false
    }

    override fun generateHash(data: ByteArray): ByteArray {
        return data.toByteString().sha256().toByteArray()
    }
    
    private suspend fun mockContactConnections() {
        delay(STARTUP_DELAY)
        selfProfileRepo.onSelfConnectionProtocol(ConnectionProtocol.UDP)

        for (friend in MOCK_FRIENDS) {
            delay(ITERATE_DELAY)
            val status = UserStatus.entries.random()
            val protocol = ConnectionProtocol.entries.random()
            friends[friend.uid] = friend.copy(
                status = status,
                protocol = protocol,
            )
            contactRepo.onFriendConnectionProtocol(friend.uid, protocol)
            contactRepo.onFriendStatus(friend.uid, status)
        }

        for (group in MOCK_GROUPS) {
            delay(ITERATE_DELAY)
            val peers = group.peers.values.toMutableList()
            peers.replaceAll {
                val friend = friends[it.uid]
                it.copy(protocol = if (friend?.isOnline == true) friend.protocol else it.protocol)
            }
            val newGroup = group.copy(peers = peers.associateBy { it.uid })
            groups[group.uid] = newGroup
            contactRepo.onGroupConnected(
                groupUid = group.uid,
                fetchPeers = true,
            )
            if (newGroup.type == ChatGroupType.ADVANCED) {
                delay(ITERATE_DELAY)
                peers.forEach { peer ->
                    contactRepo.onGroupPeerJoin(
                        groupUid = group.uid,
                        peer = peer,
                        isSelf = peer.uid == SELF_USER_ADV_GROUP.uid,
                    )
                }
            }
        }
    }

    private fun handleCommand(message: ChatMessage): Boolean {
        val timestamp = delayedTimestamp()
        val defaultKey = generatePublicKey(message.text)
        val friendUid = message.conversationUid as? UserUid ?: UserUid(defaultKey)
        val groupUid = message.conversationUid as? GroupUid ?: GroupUid(defaultKey)
        val group = if (message.isGroupConversation) groups[groupUid] else null
        val peerUid = group?.peers?.values?.randomOrNull()?.uid ?: UserUid(defaultKey)
        val command = message.text.substringAfter('/').substringBefore(' ').lowercase()
        val args = message.text.substringAfter(' ').trim()
        val lastMessage = messageRepo.getMessages(message.conversationUid).lastOrNull()
        queuedEvents += when (command) {
            "name" -> if (message.isGroupConversation) {
                GroupTitleEvent(
                    timestamp = timestamp,
                    groupUid = groupUid,
                    peerUid = peerUid,
                    title = args,
                )
                
            } else {
                FriendNameEvent(
                    timestamp = timestamp,
                    friendUid = friendUid,
                    name = args
                )
            }
            "status" -> FriendStatusEvent(
                timestamp = timestamp,
                friendUid = friendUid,
                status = UserStatus.entries.find { it.name.equals(args, ignoreCase = true) } ?: return true
            )
            "statusmsg" -> FriendStatusMessageEvent(
                timestamp = timestamp,
                friendUid = friendUid,
                message = args,
            )
            "echo" -> NewMessageEvent(
                timestamp = timestamp,
                message = message.copy(
                    senderUid = if (message.isGroupConversation) peerUid else friendUid,
                    text = args,
                )
            )
            "edit" -> lastMessage?.let { existing ->
                EditMessageEvent(
                    timestamp = timestamp,
                    message = existing.copy(text = args),
                )
            } ?: return false
            "react" -> if (lastMessage != null && args.isEmojiString()) {
                ReactMessageEvent(
                    timestamp = timestamp,
                    senderUid = if (message.isGroupConversation) peerUid else friendUid,
                    message = lastMessage,
                    reactions = args.splitByEmoji(),
                )
            } else {
                return false
            }
            "filerequest" -> FileRequestEvent(
                timestamp = timestamp,
                fileTransfer = ChatFileTransfer(
                    fileUid = FileUid(generatePublicKey(args)),
                    friendUid = friendUid,
                    name = args,
                    path = args,
                    size = MOCK_FILE_SIZE,
                    direction = ChatFileTransferDirection.INCOMING,
                ),
            )
            "friendrequest" -> FriendRequestEvent(
                timestamp = timestamp,
                request = FriendRequest(
                    uid = FriendRequestUid(generatePublicKey(args)),
                    message = args,
                ),
            )
            "groupinvite" -> GroupInviteEvent(
                timestamp = timestamp,
                invite = ChatGroupInvite(
                    friendUid = friendUid,
                    groupName = null,
                    groupType = ChatGroupType.AV,
                    cookie = generatePublicKey(args),
                    password = null,
                ),
            )
            "newpeer" -> group?.let {
                val newPeer = createMockUser(
                    name = args,
                    protocol = if (Random.nextBoolean()) ConnectionProtocol.TCP else ConnectionProtocol.NONE,
                    lastOnline = timestamp - 1000000L,
                )
                group.updatePeer(newPeer)
                GroupUsersChangedEvent(
                    timestamp = timestamp,
                    groupUid = groupUid,
                )
            } ?: return false
            "peername" -> group?.let {
                GroupUserNameEvent(
                    timestamp = timestamp,
                    groupUid = groupUid,
                    peerUid = peerUid,
                    name = args,
                )
            } ?: return false
            "peerconnect" -> group?.let {
                val peer = group.peers.values.random()
                group.updatePeer(
                    peer.copy(
                        protocol = when (peer.isOnline) {
                            true -> ConnectionProtocol.NONE
                            false -> ConnectionProtocol.TCP
                        }
                    )
                )
                GroupUsersChangedEvent(
                    timestamp = timestamp,
                    groupUid = groupUid,
                )
            } ?: return false
            "peermention" -> group?.let {
                NewMessageEvent(
                    timestamp = timestamp,
                    message = message.copy(
                        senderUid = peerUid,
                        text = "@${group.selfUid.value} $args",
                    )
                )
            } ?: return false
            "call" -> ReceiveCallEvent(
                timestamp = timestamp,
                contactUid = group?.uid ?: friendUid,
            )
            "clear" -> ClearEvent(
                timestamp = 0L,
                conversationUid = message.conversationUid,
            )
            else -> return false
        }
        return true
    }
    
    private fun handleMockEvents() {
        val curTime = currentTimeMillis()
        val events = mutableListOf<MockChatEvent>()
        queuedEvents.removeAll { event ->
            if (event.timestamp < curTime) {
                events += event
                true
            } else {
                false
            }
        }
        events.forEach { event ->
            try {
                fireMockEvent(event)
            } catch (e: Exception) {
                logger.error("Failed to fire mock event: $event", e)
            }
        }
    }
    
    private fun fireMockEvent(event: MockChatEvent) {
        when (event) {
            is NewMessageEvent -> if (event.message.isGroupConversation) {
                messageRepo.onGroupMessage(
                    groupUid = event.message.conversationUid as GroupUid,
                    peerUid = event.message.senderUid as UserUid,
                    messageType = event.message.type,
                    message = event.message.text,
                )
            } else {
                messageRepo.onFriendMessage(
                    friendUid = event.message.senderUid as UserUid,
                    messageType = event.message.type,
                    message = event.message.text,
                )
            }
            is EditMessageEvent -> messageRepo.updateMessage(
                message = event.message.copy(isEdited = true),
                localOnly = true,
            )
            is ReactMessageEvent -> messageRepo.setReactions(
                message = event.message,
                senderUid = event.senderUid,
                reactions = event.reactions.toSet(),
                localOnly = true,
            )
            is FriendReadReceiptEvent -> messageRepo.onFriendReadReceipt(
                friendUid = event.friendUid,
                messageUid = event.messageUid,
            )
            is FriendNameEvent -> contactRepo.onFriendName(
                friendUid = event.friendUid,
                name = event.name,
            )
            is FriendStatusEvent -> contactRepo.onFriendStatus(
                friendUid = event.friendUid,
                status = event.status,
            )
            is FriendStatusMessageEvent -> contactRepo.onFriendStatusMessage(
                friendUid = event.friendUid,
                message = event.message,
            )
            is FriendProtocolEvent -> contactRepo.onFriendConnectionProtocol(
                friendUid = event.friendUid,
                protocol = event.protocol,
            )
            is FriendRequestEvent -> {
                contactRepo.onFriendRequest(event.request)
                messageRepo.onFriendRequest(event.request)
            }
            is GroupTitleEvent -> contactRepo.onGroupTitle(
                groupUid = event.groupUid,
                title = event.title,
            )
            is GroupInviteEvent -> {
                contactRepo.onGroupInvite(event.invite)
                messageRepo.onGroupInvite(event.invite)
            }
            is GroupConnectedEvent -> contactRepo.onGroupConnected(
                groupUid = event.groupUid,
                fetchPeers = true,
            )
            is GroupUserNameEvent -> contactRepo.onGroupPeerName(
                groupUid = event.groupUid,
                peerUid = event.peerUid,
                name = event.name,
            )
            is GroupUsersChangedEvent -> contactRepo.onGroupUsersChanged(
                groupUid = event.groupUid,
            )
            is GroupPeerRoleChangedEvent -> {
                contactRepo.onGroupPeerRoleChanged(event.groupUid, event.sourceUid, event.targetUid, event.role)
                messageRepo.onGroupPeerRoleChanged(event.groupUid, event.sourceUid, event.targetUid, event.role)
            }
            is GroupPeerKickedEvent -> {
                contactRepo.onGroupPeerKicked(event.groupUid, event.sourceUid, event.targetUid)
                messageRepo.onGroupPeerKicked(event.groupUid, event.sourceUid, event.targetUid)
            }
            is FileRequestEvent -> fileTransferRepo.onFileReceived(
                friendUid = event.fileTransfer.friendUid,
                fileUid = event.fileTransfer.fileUid,
                fileSize = event.fileTransfer.size,
                fileName = event.fileTransfer.name,
            )
            is FileChunkEvent -> fileTransferRepo.onFileChunkReceived(
                friendUid = event.fileTransfer.friendUid,
                fileUid = event.fileTransfer.fileUid,
                position = event.position,
                data = ByteArray(if (event.position == MOCK_FILE_SIZE) 0 else MOCK_CHUNK_SIZE),
            )
            is ReceiveCallEvent -> callRepo.onCallReceived(
                contactUid = event.contactUid,
                isAudioEnabled = true,
                isVideoEnabled = false,
            )
            is ClearEvent -> messageRepo.removeMessages(
                conversationUid = event.conversationUid,
                isRemovingContact = false,
            )
        }
    }
    
    private fun mockFriendJoinEvents(friend: UserData) {
        val timestamp = delayedTimestamp()
        queuedEvents += FriendNameEvent(
            timestamp = timestamp,
            friendUid = friend.uid,
            name = friend.name,
        )
        queuedEvents += FriendStatusMessageEvent(
            timestamp = timestamp,
            friendUid = friend.uid,
            message = friend.statusMessage,
        )
        queuedEvents += FriendProtocolEvent(
            timestamp = timestamp,
            friendUid = friend.uid,
            protocol = ConnectionProtocol.UDP,
        )
    }
    
    private fun mockGroupJoinEvents(group: ChatGroupData) {
        val timestamp = delayedTimestamp()
        group.peers.forEach { (_, peer) ->
            queuedEvents += GroupUsersChangedEvent(
                timestamp = timestamp,
                groupUid = group.uid,
            )
            queuedEvents += GroupUserNameEvent(
                timestamp = timestamp,
                groupUid = group.uid,
                peerUid = peer.uid,
                name = peer.name,
            )
        }
        queuedEvents += GroupConnectedEvent(
            timestamp = timestamp,
            groupUid = group.uid,
        )
        queuedEvents += GroupTitleEvent(
            timestamp = timestamp,
            groupUid = group.uid,
            peerUid = SELF_USER.uid,
            title = group.title,
        )
    }
    
    private fun delayedTimestamp(
        delayMs: Long = Random.nextLong(1000L, 3000L),
    ) = currentTimeMillis() + delayMs

    private fun ChatGroupData.updatePeer(peer: UserData) {
        val newPeers = peers.toMutableMap()
        newPeers[peer.uid] = peer
        groups[uid] = copy(peers = newPeers)
    }
}

private val SELF_USER = createMockUser(
    name = "NAND",
    statusMessage = "Likes to rollerskate",
    uid = SELF_UID,
)

private val SELF_USER_ADV_GROUP = SELF_USER.copy(
    uid = SELF_USER.getGroupUserUid(),
    protocol = ConnectionProtocol.UDP,
    role = ChatGroupRole.FOUNDER,
)

private val MOCK_FRIENDS = listOf(
    createMockUser(
        name = "Uiik",
        statusMessage = "Fake doctor",
        lastOnline = 1555611634000L,
    ),
    createMockUser(
        name = "psimagnet",
        statusMessage = "hanging out at seers",
        lastOnline = 1715745600000L,
    ),
    createMockUser(
        name = "PSK",
        statusMessage = "Go Sexy Girl",
        lastOnline = 1715779293000L,
    ),
    createMockUser(
        name = "dirac",
        statusMessage = "come out with your cheeks spread",
        lastOnline = 1715739694300L,
    )
)

private val MOCK_BASIC_GROUP = createMockGroup(
    title = "Joe's Crab Shack",
    type = ChatGroupType.AV,
)

private val MOCK_ADVANCED_GROUP = createMockGroup(
    title = "Gameboy Advanced",
    type = ChatGroupType.ADVANCED,
    selfUid = SELF_USER_ADV_GROUP.uid,
    peers = listOf(
        SELF_USER_ADV_GROUP,
        MOCK_FRIENDS[0].copy(
            uid = MOCK_FRIENDS[0].getGroupUserUid(),
            protocol = ConnectionProtocol.UDP,
            role = ChatGroupRole.MODERATOR,
        ),
        MOCK_FRIENDS[1].copy(
            uid = MOCK_FRIENDS[1].getGroupUserUid(),
            protocol = ConnectionProtocol.TCP,
            role = ChatGroupRole.USER,
        ),
        MOCK_FRIENDS[2].copy(
            uid =  MOCK_FRIENDS[2].getGroupUserUid(),
            protocol = ConnectionProtocol.TCP,
            status = UserStatus.AWAY,
            role = ChatGroupRole.OBSERVER,
        ),
    ),
    features = setOf(
        ChatGroupFeature.EDIT_DESCRIPTION,
        ChatGroupFeature.EDIT_NICKNAME,
        ChatGroupFeature.EDIT_MESSAGES,
        ChatGroupFeature.REACT_MESSAGES,
        ChatGroupFeature.PASSWORD,
        ChatGroupFeature.ROLES,
    ),
)

private val MOCK_GROUPS = listOf(MOCK_BASIC_GROUP, MOCK_ADVANCED_GROUP)

private fun createMockUser(
    name: String,
    statusMessage: String = "",
    protocol: ConnectionProtocol = ConnectionProtocol.NONE,
    lastOnline: Long = -1L,
    uid: String = generatePublicKey(name),
) = UserData(
    uid = UserUid(uid),
    name = name,
    protocol = protocol,
    statusMessage = statusMessage,
    lastOnline = lastOnline,
)

private fun createMockGroup(
    title: String,
    type: ChatGroupType,
    uid: String = generatePublicKey("$title $type"),
    visibility: ChatGroupVisibility = ChatGroupVisibility.PRIVATE,
    isOnline: Boolean = false,
    selfUid: UserUid = SELF_USER.uid,
    peers: List<UserData> = MOCK_FRIENDS + SELF_USER,
    features: Set<ChatGroupFeature> = setOf(ChatGroupFeature.EDIT_TITLE),
) = ChatGroupData(
    title = title,
    uid = GroupUid(uid),
    type = type,
    visibility = visibility,
    isOnline = isOnline,
    selfUid = selfUid,
    peers = peers.associate { friend ->
        friend.uid to UserData(
            uid = friend.uid,
            name = friend.name,
            status = friend.status,
            protocol = if (friend.isOnline) friend.protocol else ConnectionProtocol.NONE,
            lastOnline = if (friend.isOnline) 0 else currentTimeMillis() - 1_000_000,
            role = friend.role,
        )
    },
    features = features,
)

@OptIn(ExperimentalStdlibApi::class)
private fun generatePublicKey(name: String) =
    Random(name.hashCode()).nextBytes(32).toHexString(HexFormat.UpperCase)

@OptIn(ExperimentalStdlibApi::class)
private fun generateUserAddress(name: String) =
    Random(name.hashCode()).nextBytes(38).toHexString(HexFormat.UpperCase)

private fun UserData.getGroupUserUid() = UserUid((uid.value + "-GROUP").sha256())