package nand.apps.chat.engine.tox

import kotlinx.coroutines.*
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.event.SaveProfileEvent
import nand.apps.chat.engine.tox.callback.*
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.log.getLogger
import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.engine.*
import nand.apps.chat.engine.event.JoinGroupEvent
import nand.apps.chat.engine.event.SaveFriendEvent
import nand.apps.chat.engine.event.SaveGroupEvent
import nand.apps.chat.engine.tox.dev.ToxDevCommandHandler
import nand.apps.chat.engine.tox.friend.ToxFriendSaver
import nand.apps.chat.engine.tox.group.ToxGroupSaver
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.file.ChatFileControl
import nand.apps.chat.model.file.ChatFileControlResponse
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.model.net.packet.EditMessagePacketData
import nand.apps.chat.model.net.packet.GroupPeerAvatarPacketData
import nand.apps.chat.model.net.packet.ReactMessagePacketData
import nand.apps.chat.model.net.packet.RemoveMessagePacketData
import nand.apps.chat.profile.ProfileAuthenticator
import nand.apps.chat.repo.*
import nand.apps.chat.util.*
import nand.apps.tox.*
import nand.apps.tox.callback.*
import nand.apps.tox.net.ToxBootstrapData
import nand.apps.tox.net.ToxBootstrapNode
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.ExperimentalResourceApi
import org.jetbrains.compose.resources.getStringArray
import org.koin.core.Koin
import seers.composeapp.generated.resources.*
import kotlin.random.Random

const val NULL_PTR = 0L
const val FRIEND_PACKET_OPCODE = 169.toByte()

private const val DEFAULT_MAIN_ITERATE_DELAY = 20L
private const val DEFAULT_AV_ITERATE_DELAY = 200L
private const val DEFAULT_AUDIO_RATE = 64
private const val DEFAULT_VIDEO_RATE = 1024 * 1024

private const val BOOTSTRAP_NODES_FILE = "toxBootstrapNodes.json"
private const val BOOTSTRAP_NODE_COUNT = 3

private const val PUBLIC_KEY_BYTE_LENGTH = 32
private const val PUBLIC_KEY_HEX_LENGTH = PUBLIC_KEY_BYTE_LENGTH * 2

internal const val DHT_GROUP_PERSISTENCE_ENABLED = true

@OptIn(ExperimentalStdlibApi::class, ExperimentalResourceApi::class)
class ToxChatEngine(koin: Koin) : ChatEngine, ToxCallback,
    ToxFileCallback by ToxFileCallbackImpl(koin),
    ToxAvCallback by ToxAvCallbackImpl(koin),
    ToxFriendCallback by ToxFriendCallbackImpl(koin),
    ToxConferenceCallback by ToxConferenceCallbackImpl(koin),
    ToxGroupCallback by ToxGroupCallbackImpl(koin)
{
    private var toxPtr = NULL_PTR
    private var avPtr = NULL_PTR
    private val isToxInitialized get() = toxPtr != NULL_PTR
    private val isAvInitialized get() = avPtr != NULL_PTR
    private var isEngineStarted = false
    
    private val repo: ToxChatEngineRepo by koin.inject()
    private val saver: ToxChatEngineSaver by koin.inject()
    private val friendSaver: ToxFriendSaver by koin.inject()
    private val groupSaver: ToxGroupSaver by koin.inject()
    private val eventHandler: ToxChatEventHandler by koin.inject()
    private val authenticator: ProfileAuthenticator by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val packetRepo: ChatPacketRepo by koin.inject()
    private val fileSystem: ChatFileSystem by koin.inject()
    private val engineCallback: ChatEngineCallback by koin.inject()

    private val encryption: ToxChatEngineEncryption by lazy {
        val encryption: ChatEngineEncryption by koin.inject()
        encryption as ToxChatEngineEncryption
    }

    private val devCommandHandler: ToxDevCommandHandler by lazy {
        val commandHandler: ChatEngineCommandHandler by koin.inject()
        commandHandler as ToxDevCommandHandler
    }
    
    private val configDir by lazy { fileSystem.getConfigDirectory() }
    private val json = Json { ignoreUnknownKeys = true }
    
    private val logger = getLogger("ToxChatEngine")
    private val toxConstants by lazy { ToxNative.getConstants() }

    override val version: String by lazy {
        toxConstants.run { "toxcore v$versionMajor.$versionMinor.$versionPatch" }
    }

    override val constraints: ChatEngineConstraints by lazy {
        toxConstants.run {
            ChatEngineConstraints(
                maxNameLength = maxNameLength,
                maxStatusMessageLength = maxStatusMessageLength,
                maxFriendRequestLength = maxFriendRequestLength,
                maxMessageLength = maxMessageLength,
                maxPacketLength = maxCustomPacketLength - 1,
                maxGroupNameLength = maxGroupNameLength,
                maxGroupDescLength = maxGroupTopicLength,
                maxGroupPasswordLength = maxGroupPasswordLength,
                maxGroupPartingMessageLength = maxGroupPartingMessageLength,
                maxFileNameLength = maxFileNameLength,
                maxHostnameLength = maxHostnameLength,
                uidLength = publicKeyLength * 2,
                addressLength = addressLength * 2,
                uidCharacterRange = CHARACTER_RANGE_HEX,
            )
        }
    }

    override suspend fun initialize(profile: ChatProfileData): ChatEngineInitResult {
        if (isToxInitialized) return ChatEngineInitResult.SUCCESS

        val dir = fileSystem.getDirectoryForFile(profile.filePath) ?: return ChatEngineInitResult.INVALID_PROFILE
        val profileName = profile.filePath.toPath().name
        
        var profileData = if (profileName in dir) {
            dir.readFromFile(profileName)
        } else {
            null
        }
        
        // Setup encryption and decrypt existing profile data
        val isProfileEncrypted = profileData != null && ToxNative.isEncrypted(profileData)
        val encryption = profile.password.takeIf { !it.isNullOrBlank() }?.let { password ->
            ToxFileEncryption(
                password = password,
                cipher = if (isProfileEncrypted) profileData else null,
            )
        }
        if (isProfileEncrypted && profileData != null) {
            profileData = encryption?.decryptData(profileData) ?: return ChatEngineInitResult.INVALID_CREDENTIALS
        }

        val result = ToxNative.initTox(
            saveData = profileData,
            options = profile.getInitOptions(),
            logger = DefaultToxLogger(),
        )
        if (result.ptr == NULL_PTR) {
            encryption?.close()
            return convertToxInitResult(result.error)
        }
        toxPtr = result.ptr

        // Audio/video calls
        avPtr = ToxNative.initAv(toxPtr)

        this.encryption.init(encryption)
        saver.init(
            toxPtr = toxPtr,
            encryption = encryption,
            profileDir = dir,
            profileName = profileName,
        )
        friendSaver.init(
            dirName = profile.username,
            encryption = encryption,
        )
        groupSaver.init(
            dirName = profile.username,
            encryption = encryption,
        )
        (authenticator as? ToxProfileAuthenticator)?.init(
            isAuthRequired = encryption != null,
            profilePath = profile.filePath,
        )

        // Initialize and save new profile
        if (profileData == null && profile.username.isNotBlank()) {
            setSelfUserData(
                name = profile.username,
                statusMessage = getStringArray(Res.array.default_status_messages).random(),
            )
        }

        // Contact and message loading
        engineCallback.onEngineInit(profile)

        ToxNative.registerCallbacks(toxPtr, avPtr, this)

        // Parse nodes from the configuration and choose 3 random nodes to bootstrap
        val nodes = withContext(Dispatchers.IO) {
            parseBootstrapData().nodes.filter { it.isOnline }.shuffled()
        }
        var successCount = 0
        for (node in nodes) {
            if (node.bootstrap() && ++successCount == BOOTSTRAP_NODE_COUNT) break
        }
        
        devCommandHandler.onEngineInit(toxPtr)
        
        return ChatEngineInitResult.SUCCESS
    }
    
    override fun start(coroutineScope: CoroutineScope) {
        if (!isToxInitialized || isEngineStarted) return

        isEngineStarted = true

        // Main Tox iterator (run on main thread)
        coroutineScope.launch {
            engineCallback.onEngineStart(coroutineScope)

            logger.verbose("Started main iterator loop")
            while (isToxInitialized && isEngineStarted) {
                val sleepMs = try {
                    ToxNative.iterate(toxPtr).toLong().also { eventHandler.handleEvents() }
                } catch (e: Exception) {
                    logger.error("Main iterator error", e)
                    DEFAULT_MAIN_ITERATE_DELAY
                }
                delay(sleepMs)
            }
        }
        
        // AV iterator
        // TODO - Run this on a dedicated thread (IO coroutines cause a crash)
        if (isAvInitialized) {
            coroutineScope.launch {
                logger.verbose("Started audio iterator loop")
                while (isAvInitialized && isEngineStarted) {
                    val sleepMs = try {
                        ToxNative.iterateAudio(avPtr).toLong()
                    } catch (e: Exception) {
                        logger.error("Audio iterator error", e)
                        DEFAULT_AV_ITERATE_DELAY
                    }
                    delay(sleepMs)
                }
            }
        }
    }

    override fun stop() {
        if (isEngineStarted) {
            isEngineStarted = false
            engineCallback.onEngineStop()
        }
        if (isAvInitialized) {
            ToxNative.destroyAv(avPtr)
            avPtr = NULL_PTR
        }
        if (isToxInitialized) {
            devCommandHandler.onEngineStop()
            ToxNative.destroyTox(toxPtr)
            toxPtr = NULL_PTR
        }
        encryption.destroy()
        repo.clear()
    }

    override fun getSelfUserData() = ToxNative.getSelfUserData(toxPtr).convert()

    override fun setSelfUserData(name: String, statusMessage: String): Boolean {
        logger.debug("setSelfUserData($name, $statusMessage)")
        return ToxNative.setSelfUserData(
            toxPtr = toxPtr,
            name = name.encodeToByteArray(),
            statusMessage = statusMessage.encodeToByteArray(),
        ).also(::saveIfModified)
    }

    override fun setSelfStatus(status: UserStatus) {
        logger.debug("setSelfStatus($status)")
        ToxNative.setSelfStatus(toxPtr, status.toToxValue())
        saveProfile()
    }

    override fun getSelfAddress() = ToxNative.getSelfAddress(toxPtr).toHexString(HexFormat.UpperCase)

    override fun resetSelfAddress() {
        ToxNative.setSelfNoSpam(toxPtr, Random.Default.nextInt())
        saveProfile()
    }
    
    override suspend fun getFriendList(): List<UserData> {
        val savedFriends = withContext(Dispatchers.IO) {
            friendSaver.loadFriends().associateBy { it.uid }
        }
        return ToxNative.getFriendList(toxPtr).map { toxUser ->
            toxUser.convert().run {
                repo.addFriend(toxUser.id, uid)
                copy(
                    isMuted = savedFriends[uid]?.isMuted ?: false
                )
            }
        }
    }

    override suspend fun getGroupList(): List<ChatGroupData> {
        val savedGroups = withContext(Dispatchers.IO) {
            groupSaver.loadGroups().associateBy { it.uid }
        }
        val conferences = getConferenceGroups(savedGroups)
        val dhtGroups = when {
            DHT_GROUP_PERSISTENCE_ENABLED -> getDhtGroups(savedGroups)
            else -> savedGroups.values
        }
        return dhtGroups + conferences
    }

    override fun getGroupPeerList(groupUid: GroupUid): Map<UserUid, UserData> {
        val groupId = repo.getGroupId(groupUid) ?: return emptyMap()
        val peerUids = mutableMapOf<Int, UserUid>()
        val toxPeers = if (groupId.isAdvanced) {
            val peerIds = repo.getPeerIds(groupId).values.toIntArray()
            if (peerIds.isEmpty()) {
                ToxNative.getGroupSelfPeerData(toxPtr, groupId.number)?.let { arrayOf(it) }.orEmpty()
            } else {
                ToxNative.getGroupPeerData(toxPtr, groupId.number, peerIds)
            }
        } else {
            ToxNative.getConferencePeerList(toxPtr, groupId.number)
        }
        return toxPeers.mapNotNull { toxPeer ->
            toxPeer?.convert()?.also { peer ->
                if (peer.isOnline) {
                    peerUids[toxPeer.id] = peer.uid
                }
            }
        }.associateBy { it.uid }.also { peers ->
            if (toxPeers.size != peers.size) {
                logger.warn("getGroupPeerList($groupUid) conversion mismatch (${peers.size}/${toxPeers.size})")
            }
            repo.setPeerUids(groupId, peerUids)
        }
    }

    override fun sendMessage(message: ChatMessage): Boolean {
        logger.debug("sendMessage($message)")
        val messageBytes = message.text.encodeToByteArray()
        val messageType = message.type.toToxValue()
        when (val uid = message.conversationUid) {
            is UserUid -> repo.getFriendId(uid)?.let { friendId ->
                val messageId = ToxNative.sendFriendMessage(
                    toxPtr = toxPtr,
                    friendNum = friendId,
                    message = messageBytes,
                    messageType = messageType,
                )
                if (messageId != -1) {
                    repo.addMessage(messageId, message.uid)
                }
            }
            is GroupUid -> repo.getGroupId(uid)?.let { groupId ->
                if (groupId.isAdvanced) {
                    ToxNative.sendGroupMessage(
                        toxPtr = toxPtr,
                        groupNum = groupId.number,
                        message = messageBytes,
                        messageType = messageType,
                    )
                } else {
                    ToxNative.sendConferenceMessage(
                        toxPtr = toxPtr,
                        groupNum = groupId.number,
                        message = messageBytes,
                        messageType = messageType,
                    )
                }
            }
            else -> {}
        }
        return true
    }

    override fun editMessage(old: ChatMessage, new: ChatMessage): Boolean {
        logger.debug("editMessage(${old.uid})")
        return packetRepo.sendPacket(
            uid = old.conversationUid,
            data = EditMessagePacketData(
                textHash = old.text.sha256(),
                timestamp = old.timestamp,
                text = new.text,
                type = new.type,
            ),
        )
    }

    override fun setMessageReactions(message: ChatMessage, reactions: Set<String>): Boolean {
        logger.debug("setMessageReactions(${message.uid}, ${reactions.size})")
        return packetRepo.sendPacket(
            uid = message.conversationUid,
            data = ReactMessagePacketData(
                messageSenderUid = message.senderUid.value,
                textHash = message.text.sha256(),
                timestamp = message.timestamp,
                reactions = reactions.toList(),
            ),
        )
    }

    override fun removeMessage(message: ChatMessage): Boolean {
        logger.debug("removeMessage(${message.uid})")
        repo.removeMessage(message.uid)
        return packetRepo.sendPacket(
            uid = message.conversationUid,
            data = RemoveMessagePacketData(
                textHash = message.text.sha256(),
                timestamp = message.timestamp,
            ),
        )
    }

    override fun sendPacket(uid: ContactUid, packet: ByteArray): Boolean {
        logger.debug("sendPacket($uid, B[${packet.size}])")
        when (uid) {
            is UserUid -> repo.getFriendId(uid)?.let { friendId ->
                val compatPacket = ByteArray(packet.size + 1)
                compatPacket[0] = FRIEND_PACKET_OPCODE
                packet.copyInto(compatPacket, destinationOffset = 1)
                return ToxNative.sendFriendPacket(toxPtr, friendId, compatPacket)
            }
            is GroupUid -> repo.getAdvancedGroupId(uid)?.let { groupId ->
                return ToxNative.sendGroupPacket(toxPtr, groupId.number, packet)
            }
            else -> {}
        }
        return false
    }

    override fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray): Boolean {
        logger.debug("sendPeerPacket($groupUid, $peerUid, B[${packet.size}])")
        val groupId = repo.getAdvancedGroupId(groupUid) ?: return false
        return ToxNative.sendPeerPacket(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            peerId = repo.getPeerId(groupId, peerUid) ?: return false,
            packet = packet,
        )
    }

    override fun sendFile(friendUid: UserUid, kind: ChatFileType, fileSize: Long, fileName: String, fileUid: FileUid): Boolean {
        logger.debug("sendFile($friendUid, $kind, $fileSize, $fileUid, $fileName)")
        val friendId = repo.getFriendId(friendUid) ?: return false
        val fileNum = ToxNative.sendFile(
            toxPtr = toxPtr,
            friendNum = friendId,
            kind = kind.toToxValue(),
            fileSize = fileSize,
            fileId = fileUid.value.decodeHex(),
            fileName = when (kind) {
                ChatFileType.GENERIC -> fileName.encodeToByteArray()
                ChatFileType.AVATAR -> fileName.decodeHex()
            },
        )
        if (fileNum == -1) return false
        
        repo.addFileTransfer(fileNum, fileUid)
        return true
    }

    override fun sendFileChunk(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray): Boolean {
        if (position == 0L) {
            logger.debug("sendFileChunk($friendUid, $fileUid, $position, B[${data.size}])")
        }
        val friendId = repo.getFriendId(friendUid) ?: return false
        val fileId = repo.getFileId(fileUid) ?: return false
        return ToxNative.sendFileChunk(toxPtr, friendId, fileId, position, data)
    }

    override fun sendFileControl(friendUid: UserUid, fileUid: FileUid, control: ChatFileControl): ChatFileControlResponse {
        val friendId = repo.getFriendId(friendUid) ?: return ChatFileControlResponse.FRIEND_NOT_FOUND
        val fileId = repo.getFileId(fileUid) ?: return ChatFileControlResponse.NOT_FOUND
        logger.debug("SEND: $control to friend $friendUid for file $fileUid")
        val responseCode = ToxNative.sendFileControl(
            toxPtr = toxPtr,
            friendNum = friendId,
            fileNum = fileId,
            control = control.toToxValue(),
        )
        return convertToxFileControlResponse(responseCode)
    }

    override fun sendFriendRequest(address: String, message: String): UserUid? {
        val newFriendId = ToxNative.sendFriendRequest(
            toxPtr = toxPtr,
            address = address.hexToByteArray(),
            message = message.encodeToByteArray(),
        )
        return repo.addFriend(newFriendId, address.trimToLength(PUBLIC_KEY_HEX_LENGTH))?.also { saveProfile() }
    }

    override fun acceptFriendRequest(requestUid: FriendRequestUid): UserUid? {
        val newFriendId = ToxNative.acceptFriendRequest(toxPtr, requestUid.value.decodeHex())
        return repo.addFriend(newFriendId, requestUid.value)?.also { saveProfile() }
    }

    override fun removeFriend(friendUid: UserUid): Boolean {
        val friendId = repo.getFriendId(friendUid) ?: return false
        return ToxNative.removeFriend(toxPtr, friendId).also {
            friendSaver.remove(friendUid)
            repo.removeFriend(friendId, friendUid)
            saveProfile()
        }
    }

    override fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData? {
        val toxGroup = if (type == ChatGroupType.ADVANCED) {
            ToxNative.createGroup(
                toxPtr = toxPtr,
                privacyState = visibility.toToxValue(),
                title = title.encodeToByteArray(),
                selfNickname = selfProfileRepo.self.name.encodeToByteArray(),
            )
        } else {
            ToxNative.createConference(
                toxPtr = toxPtr,
                title = title.encodeToByteArray(),
                type = type.toToxValue(),
            )
        } ?: return null
            
        val group = toxGroup.register()
        val peers = if (type == ChatGroupType.ADVANCED) {
            val selfData = ToxNative.getGroupSelfPeerData(toxPtr, toxGroup.id)
            selfData?.convert()?.let { self ->
                repo.addPeer(ToxGroupId(toxGroup.id, isAdvanced = true), self.uid, selfData.id)
                mutableMapOf(self.uid to self)
            } ?: emptyMap()
        } else {
            getGroupPeerList(group.uid)
        }
        saveGroup(group.uid)
        return group.copy(peers = peers)
    }

    override fun setGroupTitle(groupUid: GroupUid, title: String): Boolean {
        val groupId = repo.getAdvancedGroupId(groupUid) ?: return false
        return ToxNative.setConferenceTitle(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            title = title.encodeToByteArray(),
        ).also(::saveIfModified)
    }

    override fun setGroupDescription(groupUid: GroupUid, description: String): Boolean {
        val groupId = repo.getAdvancedGroupId(groupUid) ?: return false
        return ToxNative.setGroupTopic(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            topic = description.encodeToByteArray(),
        ).saveIfModified(groupUid)
    }

    override fun setGroupPassword(groupUid: GroupUid, password: String?): Boolean {
        val groupId = repo.getAdvancedGroupId(groupUid) ?: return false
        return ToxNative.setGroupPassword(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            password = password?.encodeToByteArray(),
        ).saveIfModified(groupUid)
    }

    override fun setGroupNickname(groupUid: GroupUid, name: String): Boolean {
        val groupId = repo.getAdvancedGroupId(groupUid) ?: return false
        return ToxNative.setGroupNickname(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            name = name.encodeToByteArray(),
        ).saveIfModified(groupUid)
    }

    override fun setGroupAvatar(groupUid: GroupUid, avatarHash: String): Boolean {
        return packetRepo.sendPacket(
            data = GroupPeerAvatarPacketData(avatarHash),
            uid = groupUid,
        )
    }
    
    override fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean): Boolean {
        val groupId = repo.getGroupId(groupUid) ?: return false
        return if (ToxNative.isConferenceAudioEnabled(toxPtr, groupId.number) != isEnabled) {
            ToxNative.setConferenceAudioEnabled(toxPtr, groupId.number, isEnabled)
        } else {
            true
        }
    }

    override fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole): Boolean {
        val groupId = repo.getGroupId(groupUid) ?: return false
        val peerId = repo.getPeerId(groupId, peerUid) ?: return false
        return ToxNative.setGroupRole(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            peerId = peerId,
            role = role.toToxValue(),
        ).saveIfModified(groupUid)
    }

    override fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid): Boolean {
        val groupId = repo.getGroupId(groupUid) ?: return false
        val peerId = repo.getPeerId(groupId, peerUid) ?: return false
        return ToxNative.kickGroupPeer(
            toxPtr = toxPtr,
            groupNum = groupId.number,
            peerId = peerId,
        ).saveIfModified(groupUid)
    }

    override fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean): Boolean {
        val groupId = repo.getGroupId(groupUid) ?: return false
        val peerId = repo.getPeerId(groupId, peerUid)
        if (groupId.isAdvanced && peerId != null) {
            // XXX - Seems kind of redundant since we already handle this on the app layer, but I guess it doesn't hurt
            ToxNative.setGroupPeerIgnore(
                toxPtr = toxPtr,
                groupNum = groupId.number,
                peerId = peerId,
                ignore = block,
            )
        }
        saveGroup(groupUid)
        return true
    }

    override fun joinGroup(groupUid: GroupUid, selfNickname: String, password: String?): Boolean {
        if (repo.getGroupId(groupUid) != null) {
            removeGroup(groupUid, null)
            // Need to wait a frame...
            eventHandler += JoinGroupEvent(groupUid, selfNickname, password)
            return true
        }
        val groupNum = ToxNative.joinGroup(
            toxPtr = toxPtr,
            uid = groupUid.value.decodeHex(),
            selfNickname = selfNickname.encodeToByteArray(),
            password = password?.encodeToByteArray(),
        )
        if (groupNum == -1) return false
        val newId = ToxGroupId(groupNum, isAdvanced = true)
        repo.addGroup(newId, groupUid)
        saveGroup(groupUid)
        return true
    }
    
    override fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid): Boolean {
        val friendId = repo.getFriendId(friendUid) ?: return false
        val groupId = repo.getGroupId(groupUid) ?: return false
        return if (groupId.isAdvanced) {
            ToxNative.sendGroupInvite(toxPtr, friendId, groupId.number)
        } else {
            ToxNative.sendConferenceInvite(toxPtr, friendId, groupId.number)
        }
    }
    
    override fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData? {
        val friendId = repo.getFriendId(invite.friendUid) ?: return null
        val inviteData = invite.cookie.decodeHex()
        val toxGroup = when (invite.groupType) {
            ChatGroupType.ADVANCED -> {
                val peerName = selfProfileRepo.self.name.encodeToByteArray()
                val passwordBytes = invite.password?.encodeToByteArray()
                ToxNative.acceptGroupInvite(toxPtr, friendId, inviteData, peerName, passwordBytes)
            }
            else -> ToxNative.acceptConferenceInvite(toxPtr, friendId, invite.groupType.toToxValue(), inviteData)
        } ?: return null
        saveProfile()
        return toxGroup.register().copy(
            title = invite.groupName.orEmpty(),
            password = invite.password,
        ).also { saveGroup(it.uid) }
    }

    override fun removeGroup(groupUid: GroupUid, partingMessage: String?): Boolean {
        groupSaver.remove(groupUid)
        val groupId = repo.getGroupId(groupUid) ?: return false
        val messageBytes = partingMessage?.encodeToByteArray()
        return if (groupId.isAdvanced) {
            ToxNative.removeGroup(toxPtr, groupId.number, messageBytes)
        } else {
            messageBytes?.let {
                ToxNative.sendConferenceMessage(toxPtr, groupId.number, it, ChatMessageType.NORMAL.toToxValue())
            }
            ToxNative.removeConference(toxPtr, groupId.number)
        }.also {
            repo.removeGroup(groupId, groupUid)
            saveProfile()
        }
    }

    override fun setContactMuted(uid: ContactUid, isMuted: Boolean): Boolean {
        when (uid) {
            is UserUid -> saveFriend(uid)
            is GroupUid -> saveGroup(uid)
            else -> return false
        }
        return true
    }

    override fun startCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        return if (uid is UserUid) {
            ToxNative.startCall(
                avPtr = avPtr,
                friendNum = repo.getFriendId(uid) ?: return false,
                audioBitRate = if (isAudioEnabled) DEFAULT_AUDIO_RATE else 0,
                videoBitRate = if (isVideoEnabled) DEFAULT_VIDEO_RATE else 0,
            )
        } else {
            TODO("Not yet implemented")
        }
    }

    override fun answerCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        return if (uid is UserUid) {
            val success = ToxNative.answerCall(
                avPtr = avPtr,
                friendNum = repo.getFriendId(uid) ?: return false,
                audioBitRate = if (isAudioEnabled) DEFAULT_AUDIO_RATE else 0,
                videoBitRate = if (isVideoEnabled) DEFAULT_VIDEO_RATE else 0,
            )
            logger.debug("answerCall($uid) = $success")
            success
        } else {
            TODO("Not yet implemented")
        }
    }

    override fun sendCallControl(uid: ContactUid, control: CallControl): Boolean {
        return if (uid is UserUid) {
            ToxNative.sendCallControl(
                avPtr = avPtr,
                friendNum = repo.getFriendId(uid) ?: return false,
                control = control.toToxValue(),
            )
        } else {
            TODO("Not yet implemented")
        }
    }

    override fun sendAudioFrame(uid: ContactUid, frame: AudioFrame): Boolean {
        return when (uid) {
            is UserUid -> ToxNative.sendAudioFrame(
                avPtr = avPtr,
                friendNum = repo.getFriendId(uid) ?: return false,
                samples = frame.samples,
                channels = frame.channels,
                sampleRate = frame.sampleRate,
            )
            is GroupUid -> ToxNative.sendConferenceAudioFrame(
                toxPtr = toxPtr,
                groupNum = repo.getGroupId(uid)?.number ?: return false,
                samples = frame.samples,
                channels = frame.channels,
                sampleRate = frame.sampleRate,
            )
            else -> return false
        }
    }

    override fun sendAudioBitRate(uid: ContactUid, bitRate: Int): Boolean {
        return if (uid is UserUid) {
            ToxNative.sendAudioBitRate(
                avPtr = avPtr,
                friendNum = repo.getFriendId(uid) ?: return false,
                bitRate = bitRate,
            )
        } else {
            TODO("Not yet implemented")
        }
    }

    override fun generateHash(data: ByteArray) = ToxNative.hash(data)

    override fun onConnectionStatus(status: Int) {
        val protocol = convertToxConnectionProtocol(status)
        selfProfileRepo.onSelfConnectionProtocol(protocol = protocol)
        if (protocol != ConnectionProtocol.NONE) connectToGroups()
    }
    
    private fun connectToGroups() {
        contactRepo.getOfflineGroups().forEach { it.join() }
    }

    // JSON pulled from https://nodes.tox.chat/json
    // TODO - Auto-update and store this in .config
    private suspend fun parseBootstrapData(): ToxBootstrapData {
        val bootstrapNodesJson = if (BOOTSTRAP_NODES_FILE in configDir) {
            configDir.readFromFile(BOOTSTRAP_NODES_FILE)
        } else {
            null
        } ?: Res.readBytes("files/$BOOTSTRAP_NODES_FILE")
        return json.decodeFromString(bootstrapNodesJson.decodeToString())
    }
    
    private fun ToxBootstrapNode.bootstrap(): Boolean {
        val udpPort = if (isUdpOnline) udpPort else -1
        val tcpPort = if (isTcpOnline) tcpPorts.random() else -1
        logger.verbose("Bootstrapping node: $ip [UDP = $udpPort, TCP = $tcpPort]")
        return ToxNative.bootstrap(
            toxPtr = toxPtr,
            host = ip,
            udpPort = udpPort,
            tcpPort = tcpPort,
            publicKey = publicKey.decodeHex(),
        )
    }
    
    private fun ToxGroupData.register() = convert().also { group ->
        val isAdvanced = convertToxGroupType(type) == ChatGroupType.ADVANCED
        repo.addGroup(ToxGroupId(id, isAdvanced = isAdvanced), group.uid)
    }
    
    private fun ToxUserData.convert() = convert(avatarRepo)
    
    private fun getConferenceGroups(savedGroups: Map<GroupUid, ChatGroupData>): List<ChatGroupData> {
        return ToxNative.getConferenceList(toxPtr).mapNotNull { toxGroup ->
            toxGroup?.register()?.let { group ->
                group.copy(
                    peers = getGroupPeerList(group.uid),
                    blocked = savedGroups[group.uid]?.blocked.orEmpty(),
                )
            }
        }
    }
    
    private fun getDhtGroups(savedGroups: Map<GroupUid, ChatGroupData>): List<ChatGroupData> {
        return ToxNative.getGroupList(toxPtr).mapNotNull { toxGroup ->
            toxGroup?.register()?.let { group ->
                val saved = savedGroups[group.uid]
                group.copy(
                    peers = saved?.peers.orEmpty(),
                    blocked = saved?.blocked.orEmpty(),
                )
            }
        }
    }
    
    private fun ChatGroupData.join() {
        if (type != ChatGroupType.ADVANCED || DHT_GROUP_PERSISTENCE_ENABLED) {
            // Only applicable to advanced groups
            return
        }
        val groupNum = ToxNative.joinGroup(
            toxPtr = toxPtr,
            uid = uid.value.decodeHex(),
            selfNickname = peers[selfUid]?.name?.encodeToByteArray(),
            password = password?.encodeToByteArray(),
        )
        if (groupNum == -1) {
            logger.error("Failed to register advanced group $uid")
            return
        }
        val groupId = ToxGroupId(
            number = groupNum,
            isAdvanced = type == ChatGroupType.ADVANCED,
        )
        repo.addGroup(groupId, uid)
        var peerId = 0
        val peerMap = peers.map { peerId++ to it.key }.toMutableMap()
        repo.setPeerUids(groupId, peerMap)

        val isConnected = ToxNative.reconnectGroup(
            toxPtr = toxPtr,
            groupNum = groupNum,
        )
        if (isConnected) {
            contactRepo.onGroupConnected(
                groupUid = uid,
                groupData = this,
                fetchPeers = false,
            )
            logger.debug("Joined and connected to group $uid")
        } else {
            logger.error("Failed to connect to group $uid")
        }
    }
    
    private fun saveProfile() {
        eventHandler += SaveProfileEvent
    }

    private fun saveFriend(uid: UserUid) {
        eventHandler += SaveFriendEvent(uid)
    }
    
    private fun saveGroup(uid: GroupUid) {
        if (DHT_GROUP_PERSISTENCE_ENABLED) saveProfile()
        eventHandler += SaveGroupEvent(uid)
    }
    
    private fun saveIfModified(modified: Boolean) {
        if (modified) saveProfile()
    }

    private fun Boolean.saveIfModified(groupUid: GroupUid): Boolean {
        if (this) saveGroup(groupUid)
        return this
    }
}