package nand.apps.chat.engine.tox

import nand.apps.chat.io.ChatAvatarProvider
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.file.ChatFileControl
import nand.apps.chat.model.file.ChatFileControlResponse
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.net.NetworkProxyType
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.tox.ToxGroupData
import nand.apps.chat.model.group.ChatGroupErrorType
import nand.apps.tox.ToxPeerExitType
import nand.apps.tox.ToxUserData
import nand.apps.tox.ToxInitOptions

private val TEXT_GROUP_FEATURES = setOf(ChatGroupFeature.EDIT_TITLE)

private val AV_GROUP_FEATURES = setOf(
    ChatGroupFeature.EDIT_TITLE,
    ChatGroupFeature.AUDIO_CALLS,
)

private val ADVANCED_GROUP_FEATURES = setOf(
    ChatGroupFeature.EDIT_DESCRIPTION,
    ChatGroupFeature.EDIT_NICKNAME,
    ChatGroupFeature.EDIT_MESSAGES,
    ChatGroupFeature.REACT_MESSAGES,
    ChatGroupFeature.PASSWORD,
    ChatGroupFeature.ROLES,
)

fun convertToxInitResult(error: Int) = when (error) {
    0 -> ChatEngineInitResult.SUCCESS
    5 -> ChatEngineInitResult.PROXY_BAD_HOST
    6 -> ChatEngineInitResult.PROXY_BAD_PORT
    7 -> ChatEngineInitResult.PROXY_NOT_FOUND
    9 -> ChatEngineInitResult.INVALID_PROFILE
    else -> ChatEngineInitResult.INTERNAL_ERROR
}

fun convertToxConnectionProtocol(status: Int) = when (status) {
    1 -> ConnectionProtocol.TCP
    2 -> ConnectionProtocol.UDP
    else -> ConnectionProtocol.NONE
}

fun convertToxUserStatus(status: Int) = when (status) {
    1 -> UserStatus.AWAY
    2 -> UserStatus.BUSY
    else -> UserStatus.DEFAULT
}

fun convertToxMessageType(type: Int) = when (type) {
    1 -> ChatMessageType.ACTION
    else -> ChatMessageType.NORMAL
}

fun convertToxGroupType(type: Int) = when (type) {
    1 -> ChatGroupType.AV
    2 -> ChatGroupType.ADVANCED
    else -> ChatGroupType.TEXT_ONLY
}

fun convertToxGroupVisibility(visibility: Int) = when (visibility) {
    0 -> ChatGroupVisibility.PUBLIC
    else -> ChatGroupVisibility.PRIVATE
}

fun convertToxGroupRole(role: Int) = when (role) {
    0 -> ChatGroupRole.FOUNDER
    1 -> ChatGroupRole.MODERATOR
    3 -> ChatGroupRole.OBSERVER
    else -> ChatGroupRole.USER
}

fun convertToxExitType(type: Int) = when (type) {
    0 -> ToxPeerExitType.QUIT
    1 -> ToxPeerExitType.TIMEOUT
    3 -> ToxPeerExitType.SELF_DISCONNECTED
    4 -> ToxPeerExitType.KICK
    5 -> ToxPeerExitType.SYNC_ERROR
    else -> ToxPeerExitType.DISCONNECTED
}

fun convertToxGroupJoinError(type: Int) = when (type) {
    0 -> ChatGroupErrorType.PEER_LIMIT
    1 -> ChatGroupErrorType.INVALID_PASSWORD
    else -> ChatGroupErrorType.UNKNOWN
}

fun convertToxGroupModerationEventRole(type: Int) = when (type) {
    1 -> ChatGroupRole.OBSERVER
    2 -> ChatGroupRole.USER
    3 -> ChatGroupRole.MODERATOR
    else -> null
}

fun convertToxFileType(type: Int) = when (type) {
    1 -> ChatFileType.AVATAR
    else -> ChatFileType.GENERIC
}

fun convertToxFileControl(control: Int) = when (control) {
    0 -> ChatFileControl.RESUME
    1 -> ChatFileControl.PAUSE
    else -> ChatFileControl.CANCEL
}

fun convertToxFileControlResponse(code: Int) = when (code) {
    0 -> ChatFileControlResponse.OK
    1 -> ChatFileControlResponse.FRIEND_NOT_FOUND
    2 -> ChatFileControlResponse.FRIEND_NOT_CONNECTED
    3 -> ChatFileControlResponse.NOT_FOUND
    4 -> ChatFileControlResponse.NOT_PAUSED
    5 -> ChatFileControlResponse.DENIED
    6 -> ChatFileControlResponse.ALREADY_PAUSED
    7 -> ChatFileControlResponse.SENDQ
    else -> ChatFileControlResponse.UNKNOWN
}

fun UserStatus.toToxValue() = when (this) {
    UserStatus.AWAY -> 1
    UserStatus.BUSY -> 2
    else -> 0
}

fun ChatMessageType.toToxValue() = when (this) {
    ChatMessageType.ACTION -> 1
    else -> 0
}

fun ChatFileType.toToxValue() = when (this) {
    ChatFileType.AVATAR -> 1
    else -> 0
}

fun ChatFileControl.toToxValue() = when (this) {
    ChatFileControl.RESUME -> 0
    ChatFileControl.PAUSE -> 1
    ChatFileControl.CANCEL -> 2
}

fun ChatGroupType.toToxValue() = when (this) {
    ChatGroupType.TEXT_ONLY -> 0
    ChatGroupType.AV -> 1
    ChatGroupType.ADVANCED -> 2 // This doesn't correspond to a Tox_Conference_Type
}

fun ChatGroupVisibility.toToxValue() = when (this) {
    ChatGroupVisibility.PUBLIC -> 0
    ChatGroupVisibility.PRIVATE -> 1
}

fun ChatGroupRole.toToxValue() = when (this) {
    ChatGroupRole.FOUNDER -> 0
    ChatGroupRole.MODERATOR -> 1
    ChatGroupRole.USER -> 2
    ChatGroupRole.OBSERVER -> 3
}

fun CallControl.toToxValue() = when (this) {
    CallControl.RESUME -> 0
    CallControl.PAUSE -> 1
    CallControl.CANCEL -> 2
    CallControl.MUTE_AUDIO -> 3
    CallControl.UNMUTE_AUDIO -> 4
    CallControl.HIDE_VIDEO -> 5
    CallControl.SHOW_VIDEO -> 6
}

fun NetworkProxyType.toToxValue() = when (this) {
    NetworkProxyType.NONE -> 0
    NetworkProxyType.HTTP -> 1
    NetworkProxyType.SOCKS -> 2
}

fun ToxUserData.convert(avatarProvider: ChatAvatarProvider? = null): UserData {
    val uid = publicKey.toUserUid()
    return UserData(
        uid = uid,
        name = name.decodeToString(),
        status = convertToxUserStatus(status),
        statusMessage = statusMessage?.decodeToString().orEmpty(),
        avatarPath = avatarProvider?.getAvatarPath(uid),
        protocol = convertToxConnectionProtocol(protocol),
        role = convertToxGroupRole(role),
        lastOnline = lastOnline * 1000L,
    )
}

fun ToxGroupData.convert(): ChatGroupData {
    val type = convertToxGroupType(type)
    return ChatGroupData(
        uid = uid.toGroupUid(),
        title = title?.decodeToString().orEmpty(),
        description = topic?.decodeToString().orEmpty(),
        type = type,
        visibility = convertToxGroupVisibility(visibility),
        features = type.getToxGroupFeatures(),
        selfUid = selfPublicKey.toUserUid(),
    )
}

fun ChatGroupType.getToxGroupFeatures() = when (this) {
    ChatGroupType.TEXT_ONLY -> TEXT_GROUP_FEATURES
    ChatGroupType.AV -> AV_GROUP_FEATURES
    ChatGroupType.ADVANCED -> ADVANCED_GROUP_FEATURES
}

fun ChatProfileData.getInitOptions() = ToxInitOptions(
    isUdpEnabled = networkSettings?.isUdpEnabled ?: true,
    isIPv6Enabled = networkSettings?.isIPv6Enabled ?: true,
    isLocalDiscoveryEnabled = networkSettings?.isLanEnabled ?: true,
    proxyType = networkSettings?.proxy?.type?.toToxValue() ?: 0,
    proxyHost = networkSettings?.proxy?.host,
    proxyPort = networkSettings?.proxy?.port ?: 0,
    isDhtGroupPersistenceEnabled = DHT_GROUP_PERSISTENCE_ENABLED,
)

@OptIn(ExperimentalStdlibApi::class)
private fun ByteArray.toUserUid() = UserUid(toHexString(HexFormat.UpperCase))

@OptIn(ExperimentalStdlibApi::class)
private fun ByteArray.toGroupUid() = GroupUid(toHexString(HexFormat.UpperCase))