package nand.apps.chat.engine.tox

import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.MessageUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.util.toMutableMap
import nand.apps.tox.ToxGroupId

class ToxChatEngineRepo {

    private val friendUids = mutableMapOf<Int, UserUid>()
    private val friendIds = mutableMapOf<UserUid, Int>()
    private val connectedFriends = mutableSetOf<UserUid>()
    private val groupUids = mutableMapOf<ToxGroupId, GroupUid>()
    private val groupIds = mutableMapOf<GroupUid, ToxGroupId>()
    private val connectedGroups = mutableSetOf<GroupUid>()
    private val peerUids = mutableMapOf<ToxGroupId, MutableMap<Int, UserUid>>()
    private val peerIds = mutableMapOf<ToxGroupId, MutableMap<UserUid, Int>>()
    private val fileUids = mutableMapOf<Int, FileUid>()
    private val fileIds = mutableMapOf<FileUid, Int>()
    private val messageUids = mutableMapOf<Int, MessageUid>()
    private val messageIds = mutableMapOf<MessageUid, Int>()
    
    fun clear() {
        friendUids.clear()
        friendIds.clear()
        connectedFriends.clear()
        groupUids.clear()
        groupIds.clear()
        connectedGroups.clear()
        peerUids.clear()
        peerIds.clear()
        fileUids.clear()
        fileIds.clear()
        messageUids.clear()
        messageIds.clear()
    }

    fun getFriendUid(id: Int) = friendUids[id]
    
    fun getFriendId(uid: UserUid) = friendIds[uid]

    fun isFriendConnected(uid: UserUid) = uid in connectedFriends
    
    fun getGroupUid(id: ToxGroupId) = groupUids[id]
    
    fun getGroupId(uid: GroupUid) = groupIds[uid]
    
    fun getAdvancedGroupId(uid: GroupUid) = groupIds[uid]?.takeIf { it.isAdvanced }
    
    fun isGroupConnected(uid: GroupUid) = uid in connectedGroups
    
    fun getPeerUid(groupId: ToxGroupId, peerId: Int) = peerUids[groupId]?.get(peerId)
    
    fun getPeerId(groupId: ToxGroupId, peerUid: UserUid) = peerIds[groupId]?.get(peerUid)
    
    fun getPeerIds(groupId: ToxGroupId) = peerIds[groupId].orEmpty()
    
    fun getFileUid(id: Int) = fileUids[id]
    
    fun getFileId(uid: FileUid) = fileIds[uid]
    
    fun getMessageUid(id: Int) = messageUids[id]
    
    fun addFriend(friendId: Int, key: String): UserUid? {
        if (friendId == -1) return null
        val uid = UserUid(key)
        addFriend(friendId, uid)
        return uid
    }
    
    fun addFriend(friendId: Int, friendUid: UserUid) {
        friendUids[friendId] = friendUid
        friendIds[friendUid] = friendId
    }
    
    fun setFriendConnected(friendUid: UserUid, isConnected: Boolean) {
        if (isConnected) {
            connectedFriends += friendUid
        } else {
            connectedFriends -= friendUid
        }
    }
    
    fun addGroup(groupId: ToxGroupId, groupUid: GroupUid) {
        groupUids[groupId] = groupUid
        groupIds[groupUid] = groupId
    }
    
    fun setGroupConnected(groupUid: GroupUid, isConnected: Boolean) {
        if (isConnected) {
            connectedGroups += groupUid
        } else {
            connectedGroups -= groupUid
        }
    }

    fun setPeerUids(groupId: ToxGroupId, peers: MutableMap<Int, UserUid>) {
        peerUids[groupId] = peers
        peerIds[groupId] = peers.map { it.value to it.key }.toMutableMap()
    }
    
    fun addPeer(groupId: ToxGroupId, uid: UserUid, peerId: Int) {
        peerUids.getOrPut(groupId) { mutableMapOf() }[peerId] = uid
        peerIds.getOrPut(groupId) { mutableMapOf() }[uid] = peerId
    }
    
    fun addFileTransfer(fileNum: Int, fileUid: FileUid) {
        fileUids[fileNum] = fileUid
        fileIds[fileUid] = fileNum
    }
    
    fun addMessage(messageId: Int, uid: MessageUid) {
        messageUids[messageId] = uid
        messageIds[uid] = messageId
    }
    
    fun removeMessage(uid: MessageUid) {
        messageIds.remove(uid)?.let { messageUids.remove(it) }
    }
    
    fun removeFriend(friendId: Int, friendUid: UserUid) {
        friendIds.remove(friendUid)
        friendUids.remove(friendId)
        connectedFriends.remove(friendUid)
    }
    
    fun removeGroup(groupId: ToxGroupId, groupUid: GroupUid) {
        groupIds.remove(groupUid)
        groupUids.remove(groupId)
        peerIds.remove(groupId)
        peerUids.remove(groupId)
        connectedGroups.remove(groupUid)
    }
}