package nand.apps.chat.engine.tox

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.profile.ProfileAuthenticator
import nand.apps.chat.time.currentTimeMillis
import nand.apps.tox.ToxNative
import okio.Path.Companion.toPath
import okio.use

private const val AUTH_TIMEOUT = 5 * 60 * 1000L

class ToxProfileAuthenticator(
    private val fileSystem: ChatFileSystem,
) : ProfileAuthenticator {
    
    private var isAuthRequired: Boolean = false
    private var profilePath: String? = null
    private var lastAuthTime = 0L
    
    fun init(
        isAuthRequired: Boolean,
        profilePath: String,
    ) {
        this.isAuthRequired = isAuthRequired
        this.profilePath = profilePath
    }

    override fun isAuthenticationRequired(): Boolean {
        return isAuthRequired && profilePath != null && currentTimeMillis() - lastAuthTime >= AUTH_TIMEOUT
    }

    override suspend fun isLoginCorrect(profile: ChatProfileData): Boolean {
        val dir = fileSystem.getDirectoryForFile(profile.filePath) ?: run {
            // Invalid file path
            return false
        }
        val profileName = profile.filePath.toPath().name
        
        if (profileName !in dir) {
            // Profile doesn't exist, so the login is valid
            return true
        }

        val profileData = dir.readFromFile(profileName) ?: run {
            // Failed to read data, so the profile exists but couldn't be read for some reason
            return false
        }
        
        if (!ToxNative.isEncrypted(profileData)) {
            // Data not encrypted, so we're good
            return true
        }

        // Check profile data can be decrypted using the provided password
        val encryption = profile.password?.let { password ->
            ToxFileEncryption(
                password = password,
                cipher = profileData,
            )
        }
        return encryption.use { it?.decryptData(profileData) != null }
    }

    override suspend fun isPasswordCorrect(password: String): Boolean {
        if (!isAuthRequired) return true
        val filePath = profilePath?.toPath() ?: return false
        val dir = fileSystem.getDirectoryForFile(filePath.toString()) ?: return false
        val isCorrect = withContext(Dispatchers.IO) {
            dir.readFromFile(filePath.name)?.let { data ->
                ToxFileEncryption(
                    password = password,
                    cipher = data,
                ).use { it.decryptData(data) }
            } != null
        }
        if (isCorrect) {
            lastAuthTime = currentTimeMillis()
        }
        return isCorrect
    }
}