package nand.apps.chat.engine.tox.callback

import nand.apps.chat.engine.tox.ToxChatEngineRepo
import nand.apps.chat.log.getLogger
import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.repo.CallRepo
import nand.apps.chat.util.hasBits
import nand.apps.tox.ToxAvCallState
import nand.apps.tox.callback.ToxAvCallback
import org.koin.core.Koin

class ToxAvCallbackImpl(koin: Koin) : ToxAvCallback {

    private val toxRepo: ToxChatEngineRepo by koin.inject()
    private val callRepo: CallRepo by koin.inject()
    
    private val logger = getLogger("ToxChatEngine")
    
    override fun onCallStarted(friendNum: Int, isAudioEnabled: Boolean, isVideoEnabled: Boolean) {
        logger.debug("onCallStarted($friendNum, $isAudioEnabled, $isVideoEnabled)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        callRepo.onCallReceived(
            contactUid = friendUid,
            isAudioEnabled = isAudioEnabled,
            isVideoEnabled = isVideoEnabled,
        )
    }

    override fun onCallState(friendNum: Int, state: Int) {
        logger.debug("onCallState($friendNum, $state)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        when (state) {
            ToxAvCallState.ERROR -> {
                logger.error("Received error for call from $friendUid")
                callRepo.onCallFinished(friendUid)
            }
            ToxAvCallState.FINISHED -> callRepo.onCallFinished(friendUid)
            else -> callRepo.onCallAnswered(
                contactUid = friendUid,
                isSendingAudio = state.hasBits(ToxAvCallState.SENDING_A),
                isSendingVideo = state.hasBits(ToxAvCallState.SENDING_V),
                isAcceptingAudio = state.hasBits(ToxAvCallState.ACCEPTING_A),
                isAcceptingVideo = state.hasBits(ToxAvCallState.ACCEPTING_V),
            )
        }
    }

    override fun onAudioBitrate(friendNum: Int, bitRate: Int) {
        logger.debug("onAudioBitrate($friendNum, $bitRate)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        callRepo.onAudioBitRate(friendUid, bitRate)
    }

    override fun onVideoBitrate(friendNum: Int, bitRate: Int) {
        logger.debug("onVideoBitrate($friendNum, $bitRate)")
        // TODO - https://gitlab.com/nand-apps/seers-lodge/-/issues/15
    }

    override fun onAudioFrame(friendNum: Int, samples: ShortArray, channels: Int, sampleRate: Int) {
        logger.verbose("onAudioFrame($friendNum, S[${samples.size}], $channels, $sampleRate)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val frame = AudioFrame(
            samples = samples,
            channels = channels,
            sampleRate = sampleRate,
        )
        callRepo.onAudioFrame(friendUid, frame)
    }

    override fun onVideoFrame(friendNum: Int, width: Int, height: Int, y: ByteArray, u: ByteArray, v: ByteArray, yStride: Int, uStride: Int, vStride: Int) {
        // TODO - https://gitlab.com/nand-apps/seers-lodge/-/issues/15
    }
}