package nand.apps.chat.engine.tox.callback

import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.engine.event.GroupUsersChangedEvent
import nand.apps.chat.engine.event.SaveGroupEvent
import nand.apps.chat.engine.event.SaveProfileEvent
import nand.apps.chat.engine.tox.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.ChatGroupInvite
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.repo.CallRepo
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.tox.ToxGroupId
import nand.apps.tox.callback.ToxConferenceCallback
import org.koin.core.Koin

@OptIn(ExperimentalStdlibApi::class)
class ToxConferenceCallbackImpl(koin: Koin) : ToxConferenceCallback {

    private val toxRepo: ToxChatEngineRepo by koin.inject()
    private val eventHandler: ToxChatEventHandler by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val callRepo: CallRepo by koin.inject()
    private val logger = getLogger("ToxChatEngine")
    
    override fun onConferenceInvite(friendNum: Int, groupType: Int, cookie: ByteArray) {
        logger.debug("onConferenceInvite($friendNum, $groupType)")
        val groupInvite = ChatGroupInvite(
            friendUid = toxRepo.getFriendUid(friendNum) ?: return,
            groupName = null,
            groupType = convertToxGroupType(groupType),
            cookie = cookie.toHexString(HexFormat.UpperCase),
            password = null,
        )
        contactRepo.onGroupInvite(groupInvite)
        messageRepo.onGroupInvite(groupInvite)
    }

    override fun onConferenceConnected(groupNum: Int) {
        logger.debug("onConferenceConnected($groupNum)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        contactRepo.onGroupConnected(
            groupUid = groupUid,
            fetchPeers = true,
        )
        saveGroup(groupUid)
    }

    override fun onConferenceTitle(groupNum: Int, peerId: Int, title: ByteArray) {
        val titleStr = title.decodeToString()
        logger.debug("onConferenceTitle($groupNum, $peerId, $titleStr)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        contactRepo.onGroupTitle(
            groupUid = groupUid,
            title = titleStr,
        )
        saveGroup(groupUid)
    }

    override fun onConferenceMessage(groupNum: Int, peerId: Int, messageType: Int, message: ByteArray) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        if (contactRepo.isPeerBlocked(groupUid, peerUid)) return
        val messageStr = message.decodeToString()
        logger.debug("onConferenceMessage($groupNum, $peerId, $messageType, $messageStr)")
        messageRepo.onGroupMessage(
            groupUid = groupUid,
            peerUid = peerUid,
            messageType = convertToxMessageType(messageType),
            message = messageStr,
        )
    }
    
    override fun onConferenceAudioFrame(groupNum: Int, peerId: Int, samples: ShortArray, channels: Int, sampleRate: Int) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        if (contactRepo.isPeerBlocked(groupUid, peerUid)) return
        logger.verbose("onConferenceAudioFrame($groupNum, $peerId, S[${samples.size}], $channels, $sampleRate)")
        callRepo.onGroupAudioFrame(
            groupUid = groupUid,
            peerUid = peerUid,
            frame = AudioFrame(
                samples = samples,
                channels = channels,
                sampleRate = sampleRate,
            ),
        )
    }

    override fun onConferenceUsersChanged(groupNum: Int) {
        logger.debug("onConferenceUsersChanged($groupNum)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        if (!toxRepo.isGroupConnected(groupUid)) {
            toxRepo.setGroupConnected(groupUid, true)
            contactRepo.onGroupConnected(
                groupUid = groupUid,
                fetchPeers = false,
            )
        }
        // This event should be deferred to the next iterate call since it's invoked before
        // the group state is finished being modified
        eventHandler += GroupUsersChangedEvent(groupUid)
        saveGroup(groupUid)
    }

    override fun onConferenceUserName(groupNum: Int, peerId: Int, name: ByteArray) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        val nameStr = name.decodeToString()
        logger.debug("onConferenceUserName($groupNum, $peerId, $nameStr)")
        contactRepo.onGroupPeerName(
            groupUid = groupUid,
            peerUid = peerUid,
            name = nameStr,
        )
        saveGroup(groupUid)
    }
    
    private fun id(groupNum: Int) = ToxGroupId(groupNum, isAdvanced = false)

    private fun getGroupAndPeerUids(groupNum: Int, peerId: Int): Pair<GroupUid, UserUid>? {
        val groupId = id(groupNum)
        val groupUid = toxRepo.getGroupUid(groupId) ?: return null
        val peerUid = toxRepo.getPeerUid(groupId, peerId) ?: return null
        return groupUid to peerUid
    }
    
    private fun saveGroup(uid: GroupUid) {
        eventHandler += SaveProfileEvent
        eventHandler += SaveGroupEvent(uid)
    }
}