package nand.apps.chat.engine.tox.callback

import nand.apps.chat.engine.tox.ToxChatEngineRepo
import nand.apps.chat.engine.tox.convertToxFileControl
import nand.apps.chat.engine.tox.convertToxFileType
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.FileTransferRepo
import nand.apps.tox.callback.ToxFileCallback
import org.koin.core.Koin

@OptIn(ExperimentalStdlibApi::class)
class ToxFileCallbackImpl(koin: Koin) : ToxFileCallback {

    private val toxRepo: ToxChatEngineRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    
    private val logger = getLogger("ToxChatEngine")

    override fun onFileControl(friendNum: Int, fileNum: Int, control: Int) {
        logger.debug("onFileControl($friendNum, $fileNum, $control)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val fileUid = toxRepo.getFileUid(fileNum) ?: return
        fileTransferRepo.onFileControl(friendUid, fileUid, convertToxFileControl(control))
    }
    
    override fun onFileReceived(friendNum: Int, fileNum: Int, fileUid: ByteArray, fileType: Int, fileSize: Long, fileName: ByteArray) {
        val uid = FileUid(fileUid.toHexString(HexFormat.UpperCase))
        logger.debug("onFileReceived($friendNum, $fileNum, $uid, $fileType, $fileSize, B[${fileName.size}]")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        toxRepo.addFileTransfer(fileNum, uid)
        when (convertToxFileType(fileType)) {
            ChatFileType.GENERIC -> fileTransferRepo.onFileReceived(friendUid, uid, fileSize, fileName.decodeToString())
            ChatFileType.AVATAR -> avatarRepo.onAvatarReceived(friendUid, uid, fileSize, fileName.toHexString(HexFormat.UpperCase))
        }
    }

    override fun onFileChunkReceived(friendNum: Int, fileNum: Int, position: Long, data: ByteArray) {
        if (position == 0L) {
            logger.debug("onFileChunkReceived($friendNum, $fileNum, $position, B[${data.size}])")
        }
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val fileUid = toxRepo.getFileUid(fileNum) ?: return
        fileTransferRepo.onFileChunkReceived(friendUid, fileUid, position, data)
    }

    override fun onFileChunkRequest(friendNum: Int, fileNum: Int, position: Long, length: Long) {
        if (position == 0L) {
            logger.debug("onFileChunkRequest($friendNum, $fileNum, $position, $length)")
        }
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val fileUid = toxRepo.getFileUid(fileNum) ?: return
        fileTransferRepo.onFileChunkRequest(friendUid, fileUid, position, length)
    }
}