package nand.apps.chat.engine.tox.callback

import nand.apps.chat.engine.event.SaveFriendEvent
import nand.apps.chat.engine.tox.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatClientVersion
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.tox.callback.ToxFriendCallback
import org.koin.core.Koin

@OptIn(ExperimentalStdlibApi::class)
class ToxFriendCallbackImpl(koin: Koin) : ToxFriendCallback {

    private val toxRepo: ToxChatEngineRepo by koin.inject()
    private val toxSaver: ToxChatEngineSaver by koin.inject()
    private val eventHandler: ToxChatEventHandler by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val packetRepo: ChatPacketRepo by koin.inject()

    private val logger = getLogger("ToxChatEngine")
    
    override fun onFriendRequest(publicKey: ByteArray, message: ByteArray) {
        val publicKeyStr = publicKey.toHexString(HexFormat.UpperCase)
        val messageStr = message.decodeToString()
        logger.debug("onFriendRequest(B[${publicKey.size}], $messageStr)")
        val friendRequest = FriendRequest(
            uid = FriendRequestUid(publicKeyStr),
            message = messageStr,
        )
        contactRepo.onFriendRequest(friendRequest)
        messageRepo.onFriendRequest(friendRequest)
        toxSaver.save()
        // TODO - Persist friend request here
    }

    override fun onFriendMessage(friendNum: Int, messageType: Int, message: ByteArray) {
        val messageStr = message.decodeToString()
        logger.debug("onFriendMessage($friendNum, $messageType, $messageStr)")
        messageRepo.onFriendMessage(
            friendUid = toxRepo.getFriendUid(friendNum) ?: return,
            message = messageStr,
            messageType = convertToxMessageType(messageType),
        )
    }

    override fun onFriendName(friendNum: Int, name: ByteArray) {
        val nameStr = name.decodeToString()
        logger.debug("onFriendName($friendNum, $nameStr)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        contactRepo.onFriendName(
            friendUid = friendUid,
            name = nameStr,
        )
        saveFriend(friendUid)
    }

    override fun onFriendStatus(friendNum: Int, status: Int) {
        logger.debug("onFriendStatus($friendNum, $status)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        contactRepo.onFriendStatus(
            friendUid = friendUid,
            status = convertToxUserStatus(status),
        )
    }

    override fun onFriendStatusMessage(friendNum: Int, message: ByteArray) {
        val messageStr = message.decodeToString()
        logger.debug("onFriendStatusMessage($friendNum, $messageStr)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        contactRepo.onFriendStatusMessage(
            friendUid = friendUid,
            message = messageStr,
        )
        saveFriend(friendUid)
    }

    override fun onFriendConnectionStatus(friendNum: Int, status: Int) {
        logger.debug("onFriendConnectionStatus($friendNum, $status)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val protocol = convertToxConnectionProtocol(status)
        val isConnected = protocol != ConnectionProtocol.NONE
        contactRepo.onFriendConnectionProtocol(
            friendUid = friendUid,
            protocol = protocol,
        )
        if (!toxRepo.isFriendConnected(friendUid) && isConnected) {
            packetRepo.sendPacket(friendUid, ChatClientVersion().toPacketData())
        }
        toxRepo.setFriendConnected(friendUid, isConnected)
    }

    override fun onFriendReadReceipt(friendNum: Int, messageId: Int) {
        logger.debug("onFriendReadReceipt($friendNum, $messageId)")
        val friendUid = toxRepo.getFriendUid(friendNum) ?: return
        val messageUid = toxRepo.getMessageUid(messageId) ?: return
        messageRepo.onFriendReadReceipt(
            friendUid = friendUid,
            messageUid = messageUid,
        )
    }

    override fun onFriendPacket(friendNum: Int, packet: ByteArray) {
        logger.debug("onFriendPacket($friendNum, B[${packet.size}]")
        val compat = if (packet[0] == FRIEND_PACKET_OPCODE) {
            packet.copyOfRange(1, packet.size)
        } else {
            packet
        }
        packetRepo.onFriendPacketReceived(
            friendUid = toxRepo.getFriendUid(friendNum) ?: return,
            packet = compat,
        )
    }

    private fun saveFriend(uid: UserUid) {
        eventHandler += SaveFriendEvent(uid)
        toxSaver.save()
    }
}