package nand.apps.chat.engine.tox.callback

import nand.apps.chat.engine.event.SaveGroupEvent
import nand.apps.chat.engine.event.SaveProfileEvent
import nand.apps.chat.engine.tox.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.ChatGroupInvite
import nand.apps.chat.model.group.ChatGroupType
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.net.packet.GroupPeerAvatarPacketData
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatClientVersion
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.tox.*
import nand.apps.tox.callback.ToxGroupCallback
import org.koin.core.Koin

@OptIn(ExperimentalStdlibApi::class)
class ToxGroupCallbackImpl(koin: Koin) : ToxGroupCallback {

    private val toxRepo: ToxChatEngineRepo by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val packetRepo: ChatPacketRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val eventHandler: ToxChatEventHandler by koin.inject()
    private val logger = getLogger("ToxChatEngine")

    override fun onGroupTopic(groupNum: Int, peerId: Int, topic: ByteArray) {
        val topicStr = topic.decodeToString()
        logger.debug("onGroupTopic($groupNum, $peerId, $topicStr)")
        val uid = toxRepo.getGroupUid(id(groupNum)) ?: return
        contactRepo.onGroupDescription(
            groupUid = uid,
            description = topicStr,
        )
        saveGroup(uid)
    }

    override fun onGroupPrivacyState(groupNum: Int, privacyState: Int) {
        logger.debug("onGroupPrivacyState($groupNum, $privacyState)")
        val uid = toxRepo.getGroupUid(id(groupNum)) ?: return
        contactRepo.onGroupVisibility(
            groupUid = uid,
            visibility = convertToxGroupVisibility(privacyState),
        )
        saveGroup(uid)
    }

    override fun onGroupVoiceState(groupNum: Int, voiceState: Int) {
        logger.debug("onGroupVoiceState($groupNum, $voiceState)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        // TODO - Support this
        saveGroup(groupUid)
    }

    override fun onGroupTopicLock(groupNum: Int, topicLock: Int) {
        logger.debug("onGroupTopicLock($groupNum, $topicLock)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        // TODO - Support this
        saveGroup(groupUid)
    }

    override fun onGroupPeerLimit(groupNum: Int, peerLimit: Int) {
        logger.debug("onGroupPeerLimit($groupNum, $peerLimit)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        // TODO - Support this
        saveGroup(groupUid)
    }

    override fun onGroupPassword(groupNum: Int, password: ByteArray) {
        logger.debug("onGroupPassword($groupNum, *****)")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        contactRepo.onGroupPassword(
            groupUid = groupUid,
            password = password.decodeToString(),
        )
        saveGroup(groupUid)
    }

    override fun onGroupMessage(groupNum: Int, peerId: Int, messageType: Int, message: ByteArray, messageId: Int) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        if (contactRepo.isPeerBlocked(groupUid, peerUid)) return
        val messageStr = message.decodeToString()
        logger.debug("onGroupMessage($groupNum, $peerId, $messageType, $messageStr, $messageId)")
        messageRepo.onGroupMessage(
            groupUid = groupUid,
            peerUid = peerUid,
            messageType = convertToxMessageType(messageType),
            message = messageStr,
        )
    }

    override fun onGroupPrivateMessage(groupNum: Int, peerId: Int, messageType: Int, message: ByteArray, messageId: Int) {
        val messageStr = message.decodeToString()
        logger.debug("onGroupPrivateMessage($groupNum, $peerId, $messageType, $messageStr, $messageId)")
        // TODO - Support this
    }

    override fun onGroupCustomPacket(groupNum: Int, peerId: Int, packet: ByteArray) {
        logger.debug("onGroupCustomPacket($groupNum, $peerId, B[${packet.size}])")
        onPacketReceived(
            groupNum = groupNum,
            peerId = peerId,
            packet = packet,
            isBroadcast = true,
        )
    }

    override fun onGroupCustomPrivatePacket(groupNum: Int, peerId: Int, packet: ByteArray) {
        logger.debug("onGroupCustomPrivatePacket($groupNum, $peerId, B[${packet.size}])")
        onPacketReceived(
            groupNum = groupNum,
            peerId = peerId,
            packet = packet,
            isBroadcast = false,
        )
    }
    
    override fun onGroupInvite(friendNum: Int, inviteData: ByteArray, groupName: ByteArray) {
        val groupStr = groupName.decodeToString()
        logger.debug("onGroupInvite($friendNum, B[${inviteData.size}], $groupStr)")
        val groupInvite = ChatGroupInvite(
            friendUid = toxRepo.getFriendUid(friendNum) ?: return,
            groupName = groupStr,
            groupType = ChatGroupType.ADVANCED,
            cookie = inviteData.toHexString(HexFormat.UpperCase),
            password = null,
        )
        contactRepo.onGroupInvite(groupInvite)
        messageRepo.onGroupInvite(groupInvite)
    }

    override fun onGroupPeerJoin(groupNum: Int, peerData: ToxUserData) {
        logger.debug("onGroupPeerJoin($groupNum, ${peerData.id})")
        val groupUid = toxRepo.getGroupUid(id(groupNum)) ?: return
        if (!toxRepo.isGroupConnected(groupUid)) {
            toxRepo.setGroupConnected(groupUid, true)
            contactRepo.onGroupConnected(
                groupUid = groupUid,
                fetchPeers = false,
            )
        }
        addPeer(groupNum, peerData, isSelf = false)
    }

    override fun onGroupPeerExit(groupNum: Int, peerId: Int, exitType: Int, name: ByteArray, partingMessage: ByteArray?) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        val nameStr = name.decodeToString()
        val partStr = partingMessage?.decodeToString()
        logger.debug("onGroupPeerExit($groupNum, $peerId, $exitType, $nameStr, $partStr)")
        val exit = convertToxExitType(exitType)
        if (!partStr.isNullOrBlank() && exit == ToxPeerExitType.QUIT && !contactRepo.isPeerBlocked(groupUid, peerUid)) {
            // TODO - Special message type for parting messages
            messageRepo.onGroupMessage(
                groupUid = groupUid,
                peerUid = peerUid,
                messageType = ChatMessageType.NORMAL,
                message = partStr,
            )
        }
        contactRepo.onGroupPeerExit(
            groupUid = groupUid,
            peerUid = peerUid,
        )
        saveGroup(groupUid)
    }

    override fun onGroupPeerName(groupNum: Int, peerId: Int, name: ByteArray) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        val nameStr = name.decodeToString()
        logger.debug("onGroupPeerName($groupNum, $peerId, $nameStr)")
        contactRepo.onGroupPeerName(
            groupUid = groupUid,
            peerUid = peerUid,
            name = nameStr,
        )
        saveGroup(groupUid)
    }

    override fun onGroupPeerStatus(groupNum: Int, peerId: Int, status: Int) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        logger.debug("onGroupPeerStatus($groupNum, $peerId, $status)")
        contactRepo.onGroupPeerStatus(
            groupUid = groupUid,
            peerUid = peerUid,
            status = convertToxUserStatus(status),
        )
    }

    override fun onGroupSelfJoin(groupData: ToxGroupData, peerData: ToxUserData) {
        val group = groupData.convert()
        logger.debug("onGroupSelfJoin(${groupData.id}, ${peerData.id})")
        toxRepo.setGroupConnected(group.uid, true)
        contactRepo.onGroupConnected(
            groupUid = group.uid,
            groupData = group,
            fetchPeers = false,
        )
        addPeer(groupData.id, peerData, isSelf = true)
    }

    override fun onGroupJoinFail(groupNum: Int, failType: Int) {
        logger.debug("onGroupJoinFail($groupNum, $failType)")
        val groupId = id(groupNum)
        val groupUid = toxRepo.getGroupUid(groupId) ?: return
        contactRepo.onGroupErrorState(
            groupUid = groupUid,
            error = convertToxGroupJoinError(failType),
        )
        saveGroup(groupUid)
    }

    override fun onGroupModeration(groupNum: Int, sourcePeerId: Int, targetPeerId: Int, modType: Int) {
        logger.debug("onGroupModeration($groupNum, $sourcePeerId, $targetPeerId, $modType)")
        val groupId = id(groupNum)
        val groupUid = toxRepo.getGroupUid(groupId) ?: return
        val sourceUid = toxRepo.getPeerUid(groupId, sourcePeerId) ?: return
        val targetUid = toxRepo.getPeerUid(groupId, targetPeerId) ?: return
        val role = convertToxGroupModerationEventRole(modType)
        when {
            modType == 0 -> {
                contactRepo.onGroupPeerKicked(groupUid, sourceUid, targetUid)
                messageRepo.onGroupPeerKicked(groupUid, sourceUid, targetUid)
            }
            role != null -> {
                contactRepo.onGroupPeerRoleChanged(groupUid, sourceUid, targetUid, role)
                messageRepo.onGroupPeerRoleChanged(groupUid, sourceUid, targetUid, role)
            }
        }
        saveGroup(groupUid)
    }
    
    private fun id(groupNum: Int) = ToxGroupId(groupNum, isAdvanced = true)

    private fun getGroupAndPeerUids(groupNum: Int, peerId: Int): Pair<GroupUid, UserUid>? {
        val groupId = id(groupNum)
        val groupUid = toxRepo.getGroupUid(groupId) ?: return null
        val peerUid = toxRepo.getPeerUid(groupId, peerId) ?: return null
        return groupUid to peerUid
    }
    
    private fun addPeer(groupNum: Int, peerData: ToxUserData, isSelf: Boolean) {
        val groupId = id(groupNum)
        val groupUid = toxRepo.getGroupUid(groupId) ?: return
        val peer = peerData.convert()
        toxRepo.addPeer(groupId, peer.uid, peerData.id)
        contactRepo.onGroupPeerJoin(groupUid, peer, isSelf)
        if (!isSelf) {
            // Send our client version and avatar to this peer
            packetRepo.sendPeerPacket(groupUid, peer.uid, ChatClientVersion().toPacketData())
            avatarRepo.getAvatarPath(contactRepo.self.uid)?.let { path ->
                packetRepo.sendPeerPacket(groupUid, peer.uid, GroupPeerAvatarPacketData(path.name))
            }
        }
        saveGroup(groupUid)
    }
    
    private fun onPacketReceived(groupNum: Int, peerId: Int, packet: ByteArray, isBroadcast: Boolean) {
        val (groupUid, peerUid) = getGroupAndPeerUids(groupNum, peerId) ?: return
        if (contactRepo.isPeerBlocked(groupUid, peerUid)) return
        packetRepo.onGroupPacketReceived(
            groupUid = groupUid,
            peerUid = peerUid,
            packet = packet,
            isBroadcast = isBroadcast,
        )
    }
    
    private fun saveGroup(uid: GroupUid) {
        if (DHT_GROUP_PERSISTENCE_ENABLED) {
            eventHandler += SaveProfileEvent
        }
        eventHandler += SaveGroupEvent(uid)
    }
}