package nand.apps.chat.engine.tox.group

import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.tox.ToxFileEncryption
import nand.apps.chat.engine.tox.getToxGroupFeatures
import nand.apps.chat.io.ChatDirectory
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.PersistedChatGroupData
import nand.apps.chat.model.group.toGroupData
import nand.apps.chat.model.group.toPersistedGroupData
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.repo.ChatAvatarRepo
import okio.Path

class ToxGroupSaver(
    private val fileSystem: ChatFileSystem,
    private val avatarRepo: ChatAvatarRepo,
) {
    
    private lateinit var directory: ChatDirectory
    private var encryption: ToxFileEncryption? = null
    private val json = Json { ignoreUnknownKeys = true }
    private val logger = getLogger("ToxGroupSaver")
    
    fun init(
        dirName: String,
        encryption: ToxFileEncryption?,
    ) {
        directory = fileSystem.getProfileDirectory("$dirName/groups")
        this.encryption = encryption
    }
    
    fun save(group: ChatGroupData): Boolean {
        val data = try {
            json.encodeToString(group.toPersistedGroupData()).encodeToByteArray()
        } catch (e: Exception) {
            logger.error("Failed to save group data", e)
            return false
        }
        return directory.writeToFile(
            fileName = group.uid.value,
            data = data,
            encryption = encryption,
        )
    }
    
    fun remove(groupUid: GroupUid) {
        directory.removeFile(groupUid.value)
    }
    
    fun loadGroups() = directory.listFiles().mapNotNull { load(it) }
    
    private fun load(groupPath: Path): ChatGroupData? {
        val data = directory.readFromFile(
            fileName = groupPath.name,
            encryption = encryption,
        ) ?: return null
        val group = try {
            json.decodeFromString<PersistedChatGroupData>(data.decodeToString())
        } catch (e: Exception) {
            logger.error("Failed to read group data: $groupPath", e)
            return null
        }
        return group.toGroupData(avatarRepo).copy(
            features = group.type.getToxGroupFeatures(),
        )
    }
}