package nand.apps.chat.event

import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.event.*
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.profile.ProfileCredentialsStore
import nand.apps.chat.repo.*
import nand.apps.chat.util.fifoSharedFlow

class ChatEventHandler(
    private val engine: ChatEngine,
    private val contactRepo: ContactRepo,
    private val messageRepo: ChatMessageRepo,
    private val fileTransferRepo: FileTransferRepo,
    private val callRepo: CallRepo,
    private val voicemailRepo: VoicemailRepo,
    private val notificationRepo: NotificationRepo,
    private val settingsRepo: SettingsRepo,
    private val credentialsStore: ProfileCredentialsStore,
) {
    
    val onEvent = fifoSharedFlow<ChatEvent>()
    var isAppLoaded = false
        private set
    
    operator fun plusAssign(event: ChatEvent) = push(event)
    
    fun push(event: ChatEvent) {
        when (event) {
            is AppLoadedEvent -> {
                isAppLoaded = true
            }
            is AppFocusChangedEvent -> {
                if (event.isFocused) {
                    notificationRepo.dismissAll()
                    messageRepo.onAppFocused()
                }
            }
            is LogoutEvent -> {
                engine.stop()
                credentialsStore.reset()
            }
            is SendMessageEvent -> {
                sendMessage(
                    text = event.text,
                    conversationUid = event.conversationUid,
                )
            }
            is AcceptFriendRequestEvent -> {
                contactRepo.acceptFriendRequest(event.request)?.let { newFriend ->
                    messageRepo.onFriendRequestAccepted(event.request, newFriend)
                }
            }
            is DeclineFriendRequestEvent -> {
                contactRepo.declineFriendRequest(event.request)
            }
            is RemoveFriendEvent -> {
                contactRepo.removeFriend(event.friend.uid)
                messageRepo.removeMessages(event.friend.uid, isRemovingContact = true)
                voicemailRepo.removeMessages(event.friend.uid)
            }
            is RenameGroupEvent -> {
                contactRepo.setGroupTitle(event.group.uid, event.title)
            }
            is SetGroupDescriptionEvent -> {
                contactRepo.setGroupDescription(event.group.uid, event.description)
            }
            is SetGroupPasswordEvent -> {
                contactRepo.setGroupPassword(event.group.uid, event.password)
            }
            is SetGroupNicknameEvent -> {
                contactRepo.setGroupNickname(event.group.uid, event.nickname)
            }
            is SetGroupRoleEvent -> {
                contactRepo.setGroupRole(event.group.uid, event.peer.uid, event.role)
            }
            is KickGroupPeerEvent -> {
                contactRepo.kickGroupPeer(event.group.uid, event.peer.uid)
            }
            is SetGroupPeerBlockedEvent -> {
                contactRepo.setGroupPeerBlocked(event.group.uid, event.peer.uid, event.block)
            }
            is JoinGroupEvent -> {
                contactRepo.joinGroup(
                    groupUid = event.group.uid,
                    selfNickname = event.group.peers[event.group.selfUid]?.name,
                    password = event.group.password,
                )
            }
            is SendGroupInviteEvent -> {
                contactRepo.sendGroupInvite(event.friend.uid, event.group.uid)
            }
            is AcceptGroupInviteEvent -> {
                contactRepo.acceptGroupInvite(event.invite)
                messageRepo.removeGroupInvite(event.invite)
            }
            is DeclineGroupInviteEvent -> {
                contactRepo.removeGroupInvite(event.invite)
                messageRepo.removeGroupInvite(event.invite)
            }
            is RemoveGroupEvent -> {
                contactRepo.removeGroup(event.group.uid, event.partingMessage)
                messageRepo.removeMessages(event.group.uid, isRemovingContact = true)
            }
            is FileTransferControlEvent -> {
                fileTransferRepo.sendFileControl(event.fileTransfer, event.control)
            }
            is OpenFileEvent -> {
                fileTransferRepo.openFile(event.fileTransfer)
            }
            is AnswerCallEvent -> event.run {
                callRepo.answerCall(
                    contactUid = contactUid,
                    isAudioEnabled = isAudioEnabled,
                    isVideoEnabled = isVideoEnabled,
                )
            }
            is EndCallEvent -> {
                callRepo.sendCallControl(event.contactUid, CallControl.CANCEL)
            }
            is PauseCallEvent -> {
                callRepo.sendCallControl(event.contactUid, CallControl.PAUSE)
            }
            is ResumeCallEvent -> {
                callRepo.sendCallControl(event.contactUid, CallControl.RESUME)
            }
            is UpdateMessageEvent -> messageRepo.updateMessage(
                message = event.message,
                localOnly = event.localOnly,
            )
            is ReactMessageEvent -> messageRepo.addReaction(
                message = event.message,
                reaction = event.reaction,
            )
            is DeleteMessageEvent -> messageRepo.removeMessage(
                message = event.message,
                localOnly = event.localOnly,
            )
            is ClearMessagesEvent -> messageRepo.removeMessages(
                conversationUid = event.conversationUid,
                isRemovingContact = false,
            )
            is MuteContactEvent -> contactRepo.setContactMuted(
                uid = event.contactUid,
                isMuted = event.isMuted,
            )
            is UpdateSettingsEvent -> {
                settingsRepo.settings = event.settings
            }
            else -> Unit
        }
        onEvent.tryEmit(event)
    }
    
    private fun sendMessage(
        text: String,
        conversationUid: ContactUid,
    ) = messageRepo.sendMessage(messageRepo.createMessage(text, conversationUid))
}