package nand.apps.chat.io

import nand.apps.chat.log.getLogger
import okio.*

class ChatDirectory(
    val path: Path,
    private val fileSystem: FileSystem,
) {
    
    private val logger = getLogger("ChatDirectory")

    fun exists() = fileSystem.exists(path)
    
    fun isEmpty() = !exists() || listFiles().isEmpty()
    
    fun getSubDirectory(subPath: String) = ChatDirectory(
        path = path.resolve(subPath),
        fileSystem = fileSystem,
    )
    
    fun getFilePath(fileName: String) = path.resolve(fileName)

    fun getFileSize(fileName: String) = fileSystem.metadata(getFilePath(fileName)).size ?: 0L
    
    fun getLastModified(fileName: String) = fileSystem.metadata(getFilePath(fileName)).lastModifiedAtMillis ?: 0L
    
    operator fun contains(fileName: String) = fileSystem.exists(getFilePath(fileName))
    
    operator fun contains(subPath: Path) = subPath.toString().startsWith(path.toString() + Path.DIRECTORY_SEPARATOR)
    
    fun listFiles() = fileSystem.listOrNull(path).orEmpty()
    
    fun writeToFile(
        fileName: String,
        data: ByteArray,
        safe: Boolean = true,
        encryption: ChatFileEncryption? = null,
    ): Boolean {
        val filePath = getFilePath(fileName)
        return try {
            createPathIfNeeded()
            val bytes = encryption?.run { if (isEncryptionSupported()) encryptData(data) else data } ?: data
            if (safe) {
                val tmpPath = getTempPath(fileName)
                fileSystem.write(tmpPath) { write(bytes) }
                if (fileSystem.exists(tmpPath) && fileSystem.metadata(tmpPath).size == bytes.size.toLong()) {
                    fileSystem.copy(tmpPath, filePath)
                    fileSystem.delete(tmpPath)
                } else{
                    error("Failed to write to temp file $tmpPath")
                }
            } else {
                fileSystem.write(filePath) { write(bytes) }
            }
            fileSystem.exists(filePath)
        } catch (e: Exception) {
            logger.error("Failed to write data to file: $filePath", e)
            false
        }
    }
    
    fun writeLines(
        fileName: String,
        lines: List<String>,
        safe: Boolean = true,
    ): Boolean {
        val filePath = getFilePath(fileName)
        return try {
            createPathIfNeeded()
            if (safe) {
                val tmpPath = getTempPath(fileName)
                writeLines(tmpPath, lines)
                if (fileSystem.exists(tmpPath)) {
                    fileSystem.copy(tmpPath, filePath)
                    fileSystem.delete(tmpPath)
                } else{
                    error("Failed to write lines to temp file $tmpPath")
                }
            } else {
                writeLines(fileName, lines)
            }
            fileSystem.exists(filePath)
        } catch (e: Exception) {
            logger.error("Failed to write lines to file: $filePath", e)
            false
        }
    }
    
    private fun writeLines(path: Path, lines: List<String>) {
        fileSystem.sink(path).use { fileSink ->
            fileSink.buffer().use { bufferedSink ->
                lines.forEach { line ->
                    bufferedSink.write((line + "\n").encodeToByteArray())
                }
            }
        }
    }
    
    fun appendToFile(fileName: String, data: ByteArray): Boolean {
        val filePath = getFilePath(fileName)
        return try {
            createPathIfNeeded()
            fileSystem.appendingSink(filePath).buffer().use { it.write(data) }
            fileSystem.exists(filePath)
        } catch (e: Exception) {
            logger.error("Failed to append data to file: $filePath", e)
            false
        }
    }
    
    fun readFromFile(fileName: String, encryption: ChatFileEncryption? = null): ByteArray? {
        val filePath = getFilePath(fileName)
        return try {
            val bytes = fileSystem.read(filePath) { readByteArray() }
            encryption?.run {
                if (isDecryptionSupported() && isEncrypted(bytes)) decryptData(bytes) else bytes
            } ?: bytes
        } catch (e: Exception) {
            logger.error("Failed to read data from file: $filePath", e)
            null
        }
    }

    fun readFromFile(fileName: String, position: Long, length: Long): ByteArray? {
        val filePath = getFilePath(fileName)
        return try {
            val handle = fileSystem.openReadOnly(filePath)
            val buffer = Buffer()
            handle.read(
                fileOffset = position,
                sink = buffer,
                byteCount = length,
            )
            buffer.readByteArray()
        } catch (e: Exception) {
            logger.error("Failed to read data from file: $filePath", e)
            null
        }
    }
    
    fun readLines(fileName: String, block: (String) -> Boolean) {
        val filePath = getFilePath(fileName)
        return try {
            fileSystem.source(filePath).buffer().use { source ->
                while (true) {
                    val line = source.readUtf8Line() ?: break
                    if (!block(line)) break
                }
            }
        } catch (e: Exception) {
            logger.error("Failed to read lines from file: $filePath", e)
        }
    }

    fun editLines(fileName: String, block: (String) -> String?) {
        val filePath = getFilePath(fileName)
        val tmpPath = getTempPath(fileName)
        return try {
            fileSystem.source(filePath).buffer().use { source ->
                fileSystem.sink(tmpPath).buffer().use { sink ->
                    while (true) {
                        val inputLine = source.readUtf8Line() ?: break
                        block(inputLine)?.let { outputLine ->
                            sink.writeUtf8("$outputLine\n")
                        }
                    }
                }
            }
            fileSystem.copy(tmpPath, filePath)
            fileSystem.delete(tmpPath)
        } catch (e: Exception) {
            logger.error("Failed to edit lines for file: $filePath", e)
        }
    }
    
    fun copyFromFile(path: Path): Path? {
        if (!fileSystem.exists(path)) return null
        return try {
            val filePath = getFilePath(path.name)
            createPathIfNeeded()
            fileSystem.copy(path, filePath)
            filePath
        } catch (e: Exception) {
            logger.error("Failed to copy from file: $path", e)
            null
        }
    }
    
    fun copyTo(
        outputDir: ChatDirectory,
        deleteOutputOnFailure: Boolean = false,
    ): Boolean {
        return try {
            outputDir.createPathIfNeeded()
            listFiles().forEach { file ->
                if (fileSystem.isValidDirectory(file.toString())) {
                    getSubDirectory(file.name).copyTo(outputDir.getSubDirectory(file.name))
                } else {
                    fileSystem.copy(file, outputDir.getFilePath(file.name))
                }
            }
            true
        } catch (e: Exception) {
            logger.error("Error while copying directory $path to ${outputDir.path}", e)
            if (deleteOutputOnFailure) outputDir.deleteRecursively()
            false
        }
    }
    
    fun moveTo(
        outputDir: ChatDirectory,
        deleteOutputOnFailure: Boolean = false,
    ): Boolean {
        return copyTo(
            outputDir = outputDir,
            deleteOutputOnFailure = deleteOutputOnFailure,
        ) && deleteRecursively()
    }
    
    fun encryptFile(
        fileName: String,
        encryption: ChatFileEncryption,
    ): Boolean {
        val bytes = readFromFile(
            fileName = fileName,
            encryption = encryption,
        ) ?: return false
        return writeToFile(
            fileName = fileName,
            data = bytes,
            encryption = encryption,
        )
    }
    
    fun encrypt(encryption: ChatFileEncryption): Boolean {
        return try {
            listFiles().forEach { file ->
                if (fileSystem.isValidDirectory(file.toString())) {
                    getSubDirectory(file.name).encrypt(encryption)
                } else {
                    if (!encryptFile(file.name, encryption)) {
                        error("Failed to encrypt file: $file")
                    }
                }
            }
            true
        } catch (e: Exception) {
            logger.error("Failed to encrypt directory: $path", e)
            false
        }
    }
    
    fun removeFile(fileName: String): Boolean {
        return try {
            fileSystem.delete(getFilePath(fileName))
            true
        } catch (e: Exception) {
            logger.error("Failed to remove file: $fileName", e)
            false
        }
    }
    
    fun deleteRecursively(): Boolean {
        return try {
            fileSystem.deleteRecursively(path)
            true
        } catch (e: Exception) {
            logger.error("Failed to delete directory: $path", e)
            false
        }
    }
    
    fun createPathIfNeeded() {
        if (!fileSystem.exists(path)) fileSystem.createDirectories(path)
    }
    
    private fun getTempPath(fileName: String) = getFilePath(".$fileName.tmp")

    override fun toString() = path.toString()
}