package nand.apps.chat.io

import coil3.Uri
import nand.apps.chat.platform.ChatPlatform
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.modifier.DragData
import okio.FileSystem
import okio.Path
import okio.Path.Companion.toPath

const val FULL_PROFILE_EXTENSION = "seers"
val IMAGE_EXTENSIONS = setOf("png", "jpg", "jpeg", "webp", "bmp", "gif")
val MUSIC_EXTENSIONS = setOf("mp2", "mp3", "wav", "ogg", "aiff", "flac", "ra")
val VIDEO_EXTENSIONS = setOf("mp4", "mkv", "mov", "avi", "wmv", "webm", "mpeg")
val CODE_SOURCE_EXTENSIONS = setOf("kt", "kts", "java", "c", "h", "cpp", "hpp", "py")
val EXECUTABLE_EXTENSIONS = setOf("exe", "bat", "sh")
val SETTINGS_EXTENSIONS = setOf("ini", "conf")
val PROFILE_EXTENSIONS = setOf("tox", FULL_PROFILE_EXTENSION)
val ZIP_EXTENSIONS = setOf("zip", FULL_PROFILE_EXTENSION)

val REGEX_FILE_PATH = Regex("(?:[A-Z]:[\\\\|/]|/).*")
val REGEX_SPACES = Regex(" +")
val REGEX_NON_ALPHA_NUMERIC = Regex("[^a-zA-Z0-9_]")

private val RESERVED_CHARS_WINDOWS = Regex("[<>:\"|?*/]")
private val RESERVED_NAMES_WINDOWS = Regex("(CON|PRN|AUX|NUL|COM[0-9¹²³]|LPT[0-9¹²³])", RegexOption.IGNORE_CASE)
private val REGEX_BASE_NAME_NUMBER = Regex(".*_[0-9]+")
const val UNTITLED_FILE = "Untitled"

fun String.isFilePath() = REGEX_FILE_PATH.matches(this)

fun String.toSanitizedFileName() = replace(REGEX_SPACES, "_")
    .replace(REGEX_NON_ALPHA_NUMERIC, "")
    .run {
        when {
            isEmpty() || this == "_" -> "file"
            length > 22 -> substring(0, 22)
            else -> this
        }
    }

val Path.extension get() = name.substringAfterLast('.', "").lowercase()

val Path.nameWithoutExtension get() = name.substringBeforeLast('.')

val Path.isImage get() = extension in IMAGE_EXTENSIONS

fun FileSystem.isValidDirectory(directory: String) =
    metadataOrNull(directory.toPath())?.isDirectory == true

fun DragData.getFilePath(): String? {
    val fileUri = (this as? DragData.FilesList)?.readFiles()?.firstOrNull() ?: return null
    val filePath = fileUri
        .substringAfter("file:")
        .replace("%20", " ")
        .replace("%5B", "[")
        .replace("%5D", "]")
        .replace("%5E", "^")
        .replace("%23", "#")
        .replace("%25", "%")
    return filePath.run {
        if (getPlatform() == ChatPlatform.WINDOWS) {
            trimStart('/')
        } else {
            this
        }
    }
}

fun Uri.toFilePath(): Path? {
    return when {
        scheme == "file" || scheme == null -> path
        getPlatform() == ChatPlatform.WINDOWS && scheme?.length == 1 -> toString()
        else -> null
    }?.replace("/", Path.DIRECTORY_SEPARATOR)?.toPath()
}

fun String.isValidFileName(
    platform: ChatPlatform = getPlatform(),
): Boolean {
    val isWindows = platform == ChatPlatform.WINDOWS
    val directorySeparator = if (isWindows) '\\' else '/'
    return when {
        // Cannot be blank
        isBlank() -> false

        // Cannot end with period or whitespace
        last() == '.' || last().isWhitespace() -> false

        // Cannot contain file separator
        contains(directorySeparator) -> false

        // Windows reserved file names
        isWindows && substringBefore('.').matches(RESERVED_NAMES_WINDOWS) -> false

        // Windows reserved characters
        isWindows && contains(RESERVED_CHARS_WINDOWS) -> false

        else -> true
    }
}

fun String.sanitizeFileName(
    platform: ChatPlatform = getPlatform(),
): String {
    if (isBlank()) return UNTITLED_FILE

    val isWindows = platform == ChatPlatform.WINDOWS
    val directorySeparator = if (isWindows) '\\' else '/'
    
    // Replace file separators with underscores
    var fileName = replace(directorySeparator, '_')

    // Remove trailing periods and whitespace
    var lastChar = fileName.last()
    while (lastChar == '.' || lastChar.isWhitespace()) {
        fileName = fileName.trimEnd().trimEnd('.')
        lastChar = fileName.lastOrNull() ?: return UNTITLED_FILE
    }

    // Windows reserved names and characters are replaced with underscores
    if (isWindows) {
        fileName = fileName.replace(RESERVED_CHARS_WINDOWS, "_")

        val (baseName, extension) = fileName.splitFileName()
        if (baseName.matches(RESERVED_NAMES_WINDOWS)) {
            fileName = UNTITLED_FILE + if (extension.isNotBlank()) ".$extension" else ""
        }
    }

    return fileName
}

fun String.splitFileName(): Pair<String, String> {
    return if (contains('.')) {
        substringBefore('.') to substringAfter('.')
    } else {
        this to ""
    }
}

fun ChatDirectory.resolveDuplicateFileName(fileName: String): String {
    if (fileName !in this) return fileName
    
    val (baseName, extension) = fileName.splitFileName()
    var baseNoNumber = baseName
    var baseNumber = 0
    if (baseName.matches(REGEX_BASE_NAME_NUMBER)) {
        baseNoNumber = baseName.substringBeforeLast('_')
        baseNumber = baseName.substringAfterLast('_').toIntOrNull() ?: 0
    }
    
    var modified = fileName
    while (modified in this) {
        baseNumber++
        modified = "${baseNoNumber}_${baseNumber}.$extension"
    }
    
    return modified
}

expect fun zipDirectory(directory: Path, zipPath: Path): Boolean

expect fun unzipDirectory(zipPath: Path, directory: Path): Boolean