package nand.apps.chat.model.message

import nand.apps.chat.embed.EmbedPreviewData
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupInvite
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.findUrls
import kotlin.math.abs

private const val MESSAGE_EDIT_THRESHOLD = 5 * 60 * 1000L // 5 minutes to edit/remove
const val MAX_REACTIONS_PER_USER = 10

data class ChatMessage(
    val conversationUid: ContactUid,
    val senderUid: ContactUid,
    val text: String,
    val uid: MessageUid = MessageUid(),
    val timestamp: Long = currentTimeMillis(),
    val type: ChatMessageType = ChatMessageType.NORMAL,
    val isDelivered: Boolean = false,
    val isEdited: Boolean = false,
    val isUnread: Boolean = false,
    val reactions: Map<UserUid, Set<String>> = emptyMap(),
    val fileTransferUid: FileTransferUid? = null,
    val friendRequest: FriendRequest? = null,
    val groupInvite: ChatGroupInvite? = null,
    val embedData: EmbedPreviewData? = null,
) {

    val isAction = type == ChatMessageType.ACTION
    val isGroupMessage = type == ChatMessageType.GROUP
    val isGroupConnectionEvent = type == ChatMessageType.PEER_ONLINE || type == ChatMessageType.PEER_OFFLINE
    val isGroupConversation = conversationUid is GroupUid
    val urls = text.findUrls()
    
    fun shouldMerge(other: ChatMessage?, thresholdMs: Long): Boolean {
        return other != null &&
                other.senderUid == senderUid &&
                abs(other.timestamp - timestamp) < thresholdMs
    }
    
    fun isEditable() = fileTransferUid == null &&
            friendRequest == null &&
            groupInvite == null &&
            (type == ChatMessageType.NORMAL || type == ChatMessageType.ACTION) &&
            currentTimeMillis() - timestamp <= MESSAGE_EDIT_THRESHOLD
    
    fun setReactions(senderUid: UserUid, reactions: Set<String>): ChatMessage {
        if (reactions.size > MAX_REACTIONS_PER_USER) return this
        val reactionMap = this.reactions.toMutableMap().apply {
            this[senderUid] = reactions
        }
        return copy(reactions = reactionMap)
    }

    fun containsSelfMention(
        group: ChatGroupData,
        mentionsRegex: Regex,
    ) = mentionsRegex.findAll(text).find {
        val uid = it.value.substring(1)
        uid == group.uid.value || uid == group.selfUid.value
    } != null
}