package nand.apps.chat.model.settings.shortcut

import androidx.compose.ui.input.key.Key
import nand.apps.chat.platform.ChatPlatform
import nand.apps.chat.platform.getPlatform

private const val CTRL = "Ctrl+"
private const val ALT = "Alt+"
private const val SHIFT = "Shift+"
private const val META = "Meta+"

data class ShortcutKeyCombo(
    val key: String,
    val ctrl: Boolean = false,
    val alt: Boolean = false,
    val shift: Boolean = false,
    val meta: Boolean = false,
) {

    fun isValid(): Boolean {
        val key = LABEL_TO_KEY[key] ?: return false
        return key in NO_MODIFIERS_REQUIRED || ctrl || alt || shift || meta
    }

    fun toPlatformString(
        platform: ChatPlatform = getPlatform(),
    ) = when (platform) {
        ChatPlatform.MAC,
        ChatPlatform.IOS -> toMacString()
        else -> toString()
    }
    
    override fun toString() = buildString {
        if (ctrl) append(CTRL)
        if (alt) append(ALT)
        if (shift) append(SHIFT)
        if (meta) append(META)
        append(key)
    }

    private fun toMacString() = buildString {
        if (ctrl) append("⌃")
        if (alt) append("⌥")
        if (shift) append("⇧")
        if (meta) append("⌘")
        append(MAC_LABELS[LABEL_TO_KEY[key]] ?: key)
    }
}

fun String.toShortcutKeyCombo() = ShortcutKeyCombo(
    key = substringAfterLast('+'),
    ctrl = contains(CTRL, ignoreCase = true),
    alt = contains(ALT, ignoreCase = true),
    shift = contains(SHIFT, ignoreCase = true),
)

val Key.label get() = KEY_TO_LABEL[this]

// Used internally for settings serialization and as human-readable fallback if no localization is provided
private val KEY_TO_LABEL = mapOf(
    
    // Letters
    Key.A to "A",
    Key.B to "B",
    Key.C to "C",
    Key.D to "D",
    Key.E to "E",
    Key.F to "F",
    Key.G to "G",
    Key.H to "H",
    Key.I to "I",
    Key.J to "J",
    Key.K to "K",
    Key.L to "L",
    Key.M to "M",
    Key.N to "N",
    Key.O to "O",
    Key.P to "P",
    Key.Q to "Q",
    Key.R to "R",
    Key.S to "S",
    Key.T to "T",
    Key.U to "U",
    Key.V to "V",
    Key.W to "W",
    Key.X to "X",
    Key.Y to "Y",
    Key.Z to "Z",
    
    // Numbers
    Key.Zero to "0",
    Key.One to "1",
    Key.Two to "2",
    Key.Three to "3",
    Key.Four to "4",
    Key.Five to "5",
    Key.Six to "6",
    Key.Seven to "7",
    Key.Eight to "8",
    Key.Nine to "9",
    
    // Punctuation and math
    Key.Minus to "-",
    Key.Equals to "=",
    Key.Comma to ",",
    Key.Period to ".",
    Key.Slash to "/",
    Key.Backslash to "\\",
    Key.Semicolon to ";",
    Key.Apostrophe to "'",
    Key.LeftBracket to "[",
    Key.RightBracket to "]",
    Key.Grave to "`",
    
    // Formatting and actions
    Key.Tab to "Tab",
    Key.Backspace to "Back",
    Key.Enter to "Enter",
    Key.Escape to "Escape",
    Key.Insert to "Insert",
    Key.Delete to "Delete",
    Key.PageUp to "PgUp",
    Key.PageDown to "PgDown",
    Key.MoveHome to "Home",
    Key.MoveEnd to "End",
    
    // Function keys
    Key.F1 to "F1",
    Key.F2 to "F2",
    Key.F3 to "F3",
    Key.F4 to "F4",
    Key.F5 to "F5",
    Key.F6 to "F6",
    Key.F7 to "F7",
    Key.F8 to "F8",
    Key.F9 to "F9",
    Key.F10 to "F10",
    Key.F11 to "F11",
    Key.F12 to "F12",
    
    // Arrow keys
    Key.DirectionLeft to "Left",
    Key.DirectionRight to "Right",
    Key.DirectionUp to "Up",
    Key.DirectionDown to "Down",
    
    // Num pad
    Key.NumPad0 to "Np0",
    Key.NumPad1 to "Np1",
    Key.NumPad2 to "Np2",
    Key.NumPad3 to "Np3",
    Key.NumPad4 to "Np4",
    Key.NumPad5 to "Np5",
    Key.NumPad6 to "Np6",
    Key.NumPad7 to "Np7",
    Key.NumPad8 to "Np8",
    Key.NumPad9 to "Np9",
    Key.NumPadEnter to "NpEnter",
    Key.NumPadAdd to "NpAdd",
    Key.NumPadSubtract to "NpSub",
    Key.NumPadComma to "NpComma",
    Key.NumPadDot to "NpDot",
    Key.NumPadDivide to "NpDiv",
    Key.NumPadMultiply to "NpMultiply",
    Key.NumPadLeftParenthesis to "NpLeftPar",
    Key.NumPadRightParenthesis to "NpRightPar",
    Key.NumPadEquals to "NpEquals",
    
    // Special and other obscure keys
    Key.Break to "Pause",
    Key.ZoomOut to "ZoomOut",
    Key.ZoomIn to "ZoomIn",
)

private val LABEL_TO_KEY = KEY_TO_LABEL.map { it.value to it.key }.toMap()

// Special characters for Mac shortcuts
private val MAC_LABELS = mapOf(
    Key.Delete to "⌦",
    Key.Backspace to "⌫",
    Key.Tab to "⇥",
    Key.Enter to "↩",
    Key.Escape to "⎋",
    Key.DirectionLeft to "←",
    Key.DirectionRight to "→",
    Key.DirectionUp to "↑",
    Key.DirectionDown to "↓",
    Key.PageUp to "⇞",
    Key.PageDown to "⇟",
    Key.MoveHome to "↖",
    Key.MoveEnd to "↘",
    Key.Insert to "Help",
    Key.PrintScreen to "F13",
    Key.ScrollLock to "F14",
    Key.Break to "F15",
    Key.CapsLock to "⇪",
    Key.NumLock to "⌧",
    Key.NumPad0 to "⌨0",
    Key.NumPad1 to "⌨1",
    Key.NumPad2 to "⌨2",
    Key.NumPad3 to "⌨3",
    Key.NumPad4 to "⌨4",
    Key.NumPad5 to "⌨5",
    Key.NumPad6 to "⌨6",
    Key.NumPad7 to "⌨7",
    Key.NumPad8 to "⌨8",
    Key.NumPad9 to "⌨9",
    Key.NumPadDivide to "⌨/",
    Key.NumPadMultiply to "⌨*",
)

// List of keys that don't require modifiers to be considered valid key combos
private val NO_MODIFIERS_REQUIRED = setOf(
    Key.F1,
    Key.F2,
    Key.F3,
    Key.F4,
    Key.F5,
    Key.F6,
    Key.F7,
    Key.F8,
    Key.F9,
    Key.F10,
    Key.F11,
    Key.F12,
    Key.Escape,
    Key.Break,
    Key.ZoomOut,
    Key.ZoomIn,
)