package nand.apps.chat.monitor

import androidx.compose.foundation.gestures.awaitEachGesture
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.input.key.onPreviewKeyEvent
import androidx.compose.ui.input.pointer.pointerInput
import kotlinx.coroutines.delay
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.time.currentTimeMillis
import org.koin.core.Koin

private const val CHECK_INTERVAL = 500L

class AutoAwayMonitor(koin: Koin) {
    
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private var lastActivity = currentTimeMillis()
    private var state by mutableStateOf(AutoAwayMonitorState())
    private val logger = getLogger("AutoAwayMonitor")
        
    val observer = snapshotFlow { state }
    val autoAwayTime get() = settingsRepo.settings.generalSettings.autoAwayTime
    
    fun onUserActivity() {
        lastActivity = currentTimeMillis()
    }

    fun forceAutoAway() {
        lastActivity = currentTimeMillis() - autoAwayTime
    }
    
    suspend fun checkLoop() {
        while (true) {
            val autoAwayTime = autoAwayTime
            val isAway = autoAwayTime > 0L && currentTimeMillis() - lastActivity >= autoAwayTime
            updateState(isAway)
            delay(CHECK_INTERVAL)
        }
    }
    
    private fun updateState(isAway: Boolean) {
        if (isAway != state.isAway) {
            state = state.copy(
                isAway = isAway,
                lastStatus = if (isAway) selfProfileRepo.self.status else state.lastStatus,
            )
            logger.debug("Auto-away " + if (isAway) "activated" else "deactivated")
            selfProfileRepo.setAutoAway(isAway, state.lastStatus)
        }
    }
}

data class AutoAwayMonitorState(
    val isAway: Boolean = false,
    val lastStatus: UserStatus = UserStatus.DEFAULT,
)

fun Modifier.onUserActivity(block: () -> Unit): Modifier {
    return pointerInput(Unit) {
        awaitEachGesture {
            awaitPointerEvent()
            block()
        }
    }.onPreviewKeyEvent {
        block()
        false
    }
}