package nand.apps.chat.net

import io.ktor.client.*
import io.ktor.client.request.*
import io.ktor.client.request.forms.*
import io.ktor.client.statement.*
import io.ktor.http.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.settings.filetransfer.FileTransferUploadMethod
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.util.isUrl
import okio.FileSystem
import okio.Path.Companion.toPath

class FileUploader(
    private val client: HttpClient,
    private val fileSystem: FileSystem,
    private val settingsRepo: SettingsRepo,
) {
    
    private val fileSettings get() = settingsRepo.settings.fileSettings
    private val logger = getLogger("FileUploader")
    
    suspend fun upload(filePath: String): String? {
        val config = fileSettings.uploadConfig
        val path = filePath.toPath()
        if (config.url.isBlank() || !fileSystem.exists(path)) {
            return null
        }
        
        try {
            val metadata = fileSystem.metadata(path)
            val size = metadata.size
            check(!metadata.isDirectory && size != null && size > 0L && size <= config.maxFileSize)
        } catch (ignore: Exception) {
            return null
        }
        
        return withContext(Dispatchers.IO) {            
            try {
                val fileBytes = fileSystem.read(path) { readByteArray() }
                val response = when (config.method) {
                    FileTransferUploadMethod.FORM -> client.submitFormWithBinaryData(
                        url = config.url,
                        formData = formData {
                            config.formFields.forEach { (key, value) -> append(key, value) }
                            append(
                                key = config.formFileNameKey,
                                value = fileBytes,
                                headers = Headers.build {
                                    append(HttpHeaders.ContentType, ContentType.defaultForFilePath(filePath))
                                    append(HttpHeaders.ContentDisposition, "filename=\"${path.name}\"")
                                }
                            )
                        },
                    )
                    FileTransferUploadMethod.PUT -> client.put(urlString = config.url) {
                        // TODO - Verify this works with real endpoint
                        setBody(fileBytes)
                    }
                    FileTransferUploadMethod.POST -> client.post(urlString = config.url) {
                        // TODO - Verify this works with real endpoint
                        setBody(fileBytes)
                    }
                }
                response.bodyAsText().takeIf { it.isUrl() }
            } catch (e: Exception) {
                logger.error("Error uploading file: $filePath", e)
                null
            }
        }
    }
}