package nand.apps.chat.net

import androidx.compose.ui.util.fastForEach
import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.http.*
import io.ktor.utils.io.*
import io.ktor.utils.io.core.*
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.time.DAY_MILLIS
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.sha256

// Max download size of 10MB for temp files
// TODO - Tie this to a preference
private const val MAX_DOWNLOAD_SIZE = 10_000_000L
private const val DEFAULT_BUFFER_SIZE = 16384
private const val CACHE_RETENTION_MS = DAY_MILLIS

class TempFileCache(
    private val client: HttpClient,
    private val fileSystem: ChatFileSystem,
) {
    
    private val dir = fileSystem.getTempDirectory("remote")
    private val queried = mutableSetOf<String>()
    private var isCacheCleaned = false
    
    suspend fun readFromUrl(
        url: String,
        maxSize: Long = MAX_DOWNLOAD_SIZE,
    ): ByteArray? {
        if (!isCacheCleaned) {
            cleanCache()
            isCacheCleaned = true
        }
        
        val fileName = url.sha256()
        if (fileName in dir) return dir.readFromFile(fileName)
        
        if (url in queried) {
            // We've already tried this URL this session - maybe next time
            return null
        }

        queried += url

        try {
            val response = client.get(url)
            if (response.status.isSuccess()) {
                val channel: ByteReadChannel = response.body()
                if (channel.availableForRead > maxSize) {
                    // File too large
                    channel.cancel()
                    return null
                }
                val packet = channel.readRemaining(maxSize)
                if (channel.isClosedForRead) {
                    val bytes = packet.readBytes()
                    dir.writeToFile(fileName, bytes)
                    return bytes
                } else {
                    channel.cancel()
                }
            }
        } catch (ignore: Exception) {
        }
        return null
    }
    
    private fun cleanCache() {
        val curTime = currentTimeMillis()
        dir.listFiles().fastForEach { path ->
            val lastModified = dir.getLastModified(path.name)
            if (curTime - lastModified >= CACHE_RETENTION_MS) {
                dir.removeFile(path.name)
            }
        }
    }
}