package nand.apps.chat.net.packet

import io.ktor.util.*
import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.*
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.net.packet.*
import kotlin.math.min
import kotlin.random.Random

private const val HEADER_START_BYTE = '{'.code.toByte()
private const val HEADER_END_BYTE = '}'.code.toByte()
private const val DELIMITER_BYTE = 0.toByte()

@OptIn(ExperimentalSerializationApi::class)
class ChatPacketConverter(
    engineConstraints: ChatEngineConstraints,
) {
    private val maxLength = engineConstraints.maxPacketLength

    private val json = Json {
        ignoreUnknownKeys = true
        classDiscriminatorMode = ClassDiscriminatorMode.NONE
    }

    private val logger = getLogger("ChatPacketConverter")

    fun serializePackets(data: ChatPacketData): List<ByteArray> {
        val uid = generateUid()
        val type = when (data) {
            is ClientVersionPacketData -> ChatPacketType.CLIENT_VERSION
            is GroupPeerAvatarPacketData -> ChatPacketType.GROUP_PEER_AVATAR
            is EditMessagePacketData -> ChatPacketType.EDIT_MESSAGE
            is ReactMessagePacketData -> ChatPacketType.REACT_MESSAGE
            is RemoveMessagePacketData -> ChatPacketType.REMOVE_MESSAGE
        }
        val payloadBytes = json.encodeToString(data).encodeToByteArray()
        val packets = mutableListOf<ByteArray>()
        var position = 0
        while (position < payloadBytes.size) {
            val header = ChatPacketHeader(
                uid = uid,
                type = type,
                pos = position,
                len = payloadBytes.size,
            )
            val headerBytes = header.toByteArray()
            val partSize = min(payloadBytes.size - position, headerBytes.payloadSpace)
            packets += headerBytes + payloadBytes.sliceArray(position..<partSize)
            position += partSize
        }
        return packets
    }

    fun getHeader(packet: ByteArray): ChatPacketHeader? {
        val delimiterIndex = packet.indexOf(DELIMITER_BYTE)
        if (packet.first() != HEADER_START_BYTE || delimiterIndex <= 0 || packet[delimiterIndex - 1] != HEADER_END_BYTE) {
            return null
        }
        return try {
            json.decodeFromString<ChatPacketHeader>(packet.decodeToString(0, delimiterIndex))
        } catch (e: Exception) {
            logger.error("Failed to decode packet header", e)
            return null
        }
    }
    
    fun getPayloadBytes(packet: ByteArray): ByteArray? {
        val delimiterIndex = packet.indexOf(DELIMITER_BYTE)
        if (delimiterIndex == -1) return null
        return packet.sliceArray(delimiterIndex + 1..<packet.size)
    }
    
    fun getPayloadComplete(type: ChatPacketType, payloadBytes: ByteArray): ChatPacketData? {
        return try {
            val payload = payloadBytes.decodeToString()
            when (type) {
                ChatPacketType.CLIENT_VERSION -> payload.fromJson<ClientVersionPacketData>()
                ChatPacketType.GROUP_PEER_AVATAR -> payload.fromJson<GroupPeerAvatarPacketData>()
                ChatPacketType.EDIT_MESSAGE -> payload.fromJson<EditMessagePacketData>()
                ChatPacketType.REACT_MESSAGE -> payload.fromJson<ReactMessagePacketData>()
                ChatPacketType.REMOVE_MESSAGE -> payload.fromJson<RemoveMessagePacketData>()
            }
        } catch (e: Exception) {
            logger.error("Failed to decode packet payload", e)
            null
        }
    }
    
    private fun ChatPacketHeader.toByteArray(): ByteArray {
        val headerBytes = json.encodeToString(this).encodeToByteArray()
        return headerBytes.copyOf(headerBytes.size + 1)
    }
    
    private val ByteArray.payloadSpace get() = maxLength - size

    // Generates a 32 character base64 string w/out padding
    private fun generateUid() = Random.Default.nextBytes(24).encodeBase64()
    
    private inline fun <reified T : ChatPacketData> String.fromJson() = json.decodeFromString<T>(this)
}