package nand.apps.chat.net.packet

import androidx.compose.ui.util.fastForEach
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.net.packet.*
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.removeIf
import org.koin.core.Koin

private const val MAX_PAYLOAD_SIZE = 65536
private const val MAX_ONGOING_PACKETS_PER_SENDER = 5
private const val PACKET_EXPIRATION_TIME = 10 * 60 * 1000L

private class OngoingPacketBuffer(
    val bytes: ByteArray,
    var lastPacketTime: Long = currentTimeMillis(),
) {
    
    val size get() = bytes.size
}

class ChatPacketRepo(
    private val engine: ChatEngine,
    private val converter: ChatPacketConverter,
    private val backgroundWorker: ChatBackgroundWorker,
    koin: Koin,
) : ChatEngineCallback {
    
    private val contactRepo: ContactRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val messagePacketHandler: MessagePacketHandler by koin.inject()
    private val packetHandles by lazy { listOf(messagePacketHandler) }
    private val ongoing = mutableMapOf<UserUid, MutableMap<String, OngoingPacketBuffer>>()
    
    private var lastOngoingClean = currentTimeMillis()
    private lateinit var coroutineScope: CoroutineScope
    
    private val logger = getLogger("ChatPacketRepo")

    override suspend fun onEngineInit(profile: ChatProfileData) {
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
    }
    
    fun sendPacket(uid: ContactUid, data: ChatPacketData): Boolean {
        for (packet in converter.serializePackets(data)) {
            if (!engine.sendPacket(uid, packet)) return false
        }
        return true
    }
    
    fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, data: ChatPacketData): Boolean {
        for (packet in converter.serializePackets(data)) {
            if (!engine.sendPeerPacket(groupUid, peerUid, packet)) return false
        }
        return true
    }
    
    fun onFriendPacketReceived(friendUid: UserUid, packet: ByteArray): Job? {
        val sender = contactRepo.getFriend(friendUid) ?: return null
        return coroutineScope.launch {
            handlePacket(
                packet = packet,
                sender = sender,
                group = null,
                isBroadcast = false,
            )
        }
    }
    
    fun onGroupPacketReceived(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray, isBroadcast: Boolean) {
        val group = contactRepo.getContact(groupUid) as? ChatGroupData ?: return
        val sender = group.peers[peerUid] ?: return
        coroutineScope.launch {
            handlePacket(
                packet = packet,
                sender = sender,
                group = group,
                isBroadcast = isBroadcast,
            )
        }
    }
    
    private suspend fun handlePacket(packet: ByteArray, sender: UserData, group: ChatGroupData?, isBroadcast: Boolean) {
        val header = converter.getHeader(packet)?.takeIf { it.len <= MAX_PAYLOAD_SIZE } ?: return
        
        cleanOngoingPackets()
        
        val ongoing = ongoing.getOrPut(sender.uid) { mutableMapOf() }
        if (!ongoing.containsKey(header.uid) && ongoing.size >= MAX_ONGOING_PACKETS_PER_SENDER) {
            logger.warn("Dropped packet from ${sender.uid} (exceeded max ongoing packets $MAX_ONGOING_PACKETS_PER_SENDER)")
            return
        }
        
        val buffer = ongoing.getOrPut(header.uid) { OngoingPacketBuffer(ByteArray(header.len)) }
        if (buffer.size != header.len) {
            logger.warn("Incoming packet from ${sender.uid} size mismatch (expected = ${buffer.size}, actual = ${header.len})")
            return
        }
        
        val payloadBytes = converter.getPayloadBytes(packet) ?: return
        payloadBytes.copyInto(buffer.bytes, destinationOffset = header.pos)
        if (header.pos + payloadBytes.size >= header.len) {
            ongoing.remove(header.uid)
            withContext(backgroundWorker.dispatcher) {
                converter.getPayloadComplete(header.type, buffer.bytes)
            }?.let {
                handlePacket(it, sender, group, isBroadcast)
            }
        } else {
            buffer.lastPacketTime = currentTimeMillis()
        }
    }
    
    private suspend fun handlePacket(data: ChatPacketData, sender: UserData, group: ChatGroupData?, isBroadcast: Boolean) {
        when (data) {
            is ClientVersionPacketData -> {
                group?.let {
                    contactRepo.onGroupPeerClientVersion(
                        groupUid = group.uid,
                        peerUid = sender.uid,
                        version = data.clientVersion,
                    )
                } ?: contactRepo.onFriendClientVersion(
                    friendUid = sender.uid,
                    version = data.clientVersion,
                )
            }
            is GroupPeerAvatarPacketData -> {
                group ?: return
                avatarRepo.onGroupPeerAvatarReceived(
                    group = group,
                    peer = sender,
                    hash = data.hash,
                )
            }
            else -> packetHandles.fastForEach { if (it.handlePacket(data, sender, group, isBroadcast)) return }
        }
    }

    private fun cleanOngoingPackets() {
        val curTime = currentTimeMillis()
        if (curTime - lastOngoingClean < PACKET_EXPIRATION_TIME) return
        ongoing.forEach { (_, ongoing) ->
            ongoing.removeIf { _, buffer -> curTime - buffer.lastPacketTime > PACKET_EXPIRATION_TIME }
        }
        lastOngoingClean = curTime
    }
}