package nand.apps.chat.net.packet

import kotlinx.coroutines.withContext
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.message.MAX_REACTIONS_PER_USER
import nand.apps.chat.model.net.packet.*
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.text.emoji.EmojiRepo

class MessagePacketHandler(
    private val contactRepo: ContactRepo,
    private val messageRepo: ChatMessageRepo,
    private val emojiRepo: EmojiRepo,
    private val settingsRepo: SettingsRepo,
    private val backgroundWorker: ChatBackgroundWorker,
) : ChatPacketHandler {

    private val settings get() = settingsRepo.settings.chatSettings
    
    override suspend fun handlePacket(
        data: ChatPacketData,
        sender: UserData,
        group: ChatGroupData?,
        isBroadcast: Boolean
    ): Boolean {
        val conversationUid = group?.uid ?: sender.uid
        when (data) {
            is ReactMessagePacketData -> data.handle(conversationUid, sender, group)
            is EditMessagePacketData -> data.handle(conversationUid, sender)
            is RemoveMessagePacketData -> data.handle(conversationUid, sender)
            else -> return false
        }
        return true
    }
    
    private suspend fun ReactMessagePacketData.handle(conversationUid: ContactUid, sender: UserData, group: ChatGroupData?) {
        if (!settings.isMessageReactionsEnabled || reactions.size > MAX_REACTIONS_PER_USER) return
        val messageSender = contactRepo.getContact(messageSenderUid, group) ?: return
        val reactions = withContext(backgroundWorker.dispatcher) {
            reactions.mapNotNull { emojiRepo.getEmojiData(it)?.value }.toSet()
        }
        messageRepo.setReactions(
            senderUid = sender.uid,
            message = messageRepo.findMessage(
                conversationUid = conversationUid,
                senderUid = messageSender.uid,
                textHash = textHash,
                timestamp = timestamp,
            ) ?: return,
            reactions = reactions,
            localOnly = true,
        )
    }
    
    private fun EditMessagePacketData.handle(conversationUid: ContactUid, sender: UserData) {
        if (!sender.canEditMessages() || type != ChatMessageType.NORMAL && type != ChatMessageType.ACTION) return
        val message = messageRepo.findMessage(
            conversationUid = conversationUid,
            senderUid = sender.uid,
            textHash = textHash,
            timestamp = timestamp,
        )?.takeIf { it.isEditable() } ?: return
        messageRepo.updateMessage(
            message = message.copy(
                text = text,
                type = type,
                isEdited = true,
            ),
            localOnly = true,
        )
    }
    
    private fun RemoveMessagePacketData.handle(conversationUid: ContactUid, sender: UserData) {
        if (!sender.canEditMessages()) return
        messageRepo.removeMessage(
            message = messageRepo.findMessage(
                conversationUid = conversationUid,
                senderUid = sender.uid,
                textHash = textHash,
                timestamp = timestamp,
            )?.takeIf { it.isEditable() } ?: return,
            localOnly = true,
        )
    }
    
    private fun UserData.canEditMessages() =
        role != ChatGroupRole.OBSERVER && settings.isMessageEditsEnabled
}