package nand.apps.chat.profile

import nand.apps.chat.io.*
import okio.Path
import okio.Path.Companion.toPath

const val DEFAULT_PROFILE_FILE_NAME = "profile.tox"

fun ChatFileSystem.isProfileZip(path: Path): Boolean {
    if (path.extension in ZIP_EXTENSIONS) {
        val dir = getDirectoryForZip(path.toString()) ?: return false
        return dir.getProfiles().isNotEmpty()
    }
    return false
}

fun ChatFileSystem.isProfileDirectory(path: Path): Boolean {
    return isDirectory(path.toString()) && getDirectory(path.toString()).getPrimaryProfile() != null
}

fun ChatFileSystem.directoryHasProfiles(directory: Path): Boolean {
    return getDirectory(directory.toString()).getProfiles().isNotEmpty()
}

fun ChatDirectory.getProfiles() = listFiles().filter { it.extension in PROFILE_EXTENSIONS }

fun ChatDirectory.getPrimaryProfile() = getProfiles().firstOrNull()

fun ChatFileSystem.selectProfile(path: String): ProfileSelectionData? {
    val properPath = path.toPath()
    val isDirectory = isDirectory(path)
    val isProfileExtension = properPath.extension in PROFILE_EXTENSIONS
    val username: String
    val isImportRequired: Boolean
    return when {
        (isDirectory || isProfileExtension) && properPath in getProfileDirectory() -> {
            isImportRequired = false
            if (isDirectory) {
                username = properPath.name
                getDirectory(path).getPrimaryProfile()
            } else {
                username = properPath.parent?.name.orEmpty()
                properPath
            }
        }
        isProfileExtension && !isDirectory || isProfileZip(properPath) -> {
            isImportRequired = true
            username = properPath.nameWithoutExtension
            properPath
        }
        else -> {
            username = ""
            isImportRequired = false
            null
        }
    }?.let { filePath ->
        ProfileSelectionData(
            filePath = filePath,
            username = username,
            isImportRequired = isImportRequired,
        )
    }
}

fun ChatFileSystem.importProfile(
    username: String,
    filePath: String,
): Pair<ProfileImportResult, Path?> {
    if (username.isBlank() || filePath.isBlank()) {
        return ProfileImportResult.INVALID_FILE to null
    }
    val dir = getProfileDirectory(username)
    if (dir.exists()) {
        return ProfileImportResult.PROFILE_EXISTS to null
    }
    val inputPath = filePath.toPath()
    val outputPath = when {
        isProfileZip(inputPath) -> {
            if (unzipDirectory(inputPath, dir.path)) {
                dir.getPrimaryProfile()
            } else {
                null
            }
        }

        else -> dir.copyFromFile(inputPath)
    }
    return (outputPath?.let { ProfileImportResult.IMPORT_SUCCESSFUL } ?: ProfileImportResult.INVALID_FILE) to outputPath
}