package nand.apps.chat.repo

import androidx.compose.runtime.*
import kotlinx.coroutines.*
import nand.apps.chat.audio.*
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.call.*
import nand.apps.chat.model.uid.CallUid
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.removeIf
import org.koin.core.Koin

private const val RING_SHAKE_DURATION = 2000L
private const val RING_PAUSE_DURATION = 3000L
private const val RING_FULL_DURATION = RING_SHAKE_DURATION + RING_PAUSE_DURATION
private const val CALL_ACTIVITY_MONITOR_INTERVAL = 500L
private const val SOUND_RING_INCOMING = "phone_ring_incoming"
private const val SOUND_RING_OUTGOING = "phone_ring_outgoing"
private const val SOUND_PICK_UP = "phone_pick_up"
private const val SOUND_HANG_UP = "phone_hang_up"
private const val SOUND_DECLINE = "phone_decline"
private const val SOUND_PAUSE_RESUME = "phone_pause_resume"

class CallRepo(
    private val audioPlayer: AudioPlayer,
    private val audioRecorder: AudioRecorder,
    koin: Koin,
) : ChatEngineCallback {
    
    private val engine: ChatEngine by koin.inject()
    private val voicemailRepo: VoicemailRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val settings get() = settingsRepo.settings.callSettings

    private val calls = mutableStateMapOf<ContactUid, CallData>()
    private val callActivity = mutableStateMapOf<ContactUid, Boolean>()
    private val lastCallActivityTimes = mutableMapOf<ContactUid, Long>()
    private var callActivityMonitor: Job? = null
    private var isCallRinging by mutableStateOf(false)
    private var isInActiveCall = false
    val callObserver = snapshotFlow { calls.toMap() }
    val callActivityObserver = snapshotFlow { callActivity.toMap() }
    val isCallRingingObserver = snapshotFlow { isCallRinging }
    val onCallReceived = fifoSharedFlow<CallData>()

    private var coroutineScope: CoroutineScope? = null
    private val frameHandles = mutableMapOf<CallUid, AudioHandle>()
    private val groupFrameHandles = mutableMapOf<GroupUid, MutableMap<UserUid, AudioHandle>>()

    private val incomingCallRepeater = AudioRepeater(
        audioPlayer = audioPlayer,
        soundName = SOUND_RING_INCOMING,
        onPlay = {
            isCallRinging = true
            delay(RING_SHAKE_DURATION)
            isCallRinging = false
            delay(RING_PAUSE_DURATION)
            onCallRing()
        },
        onReset = { isCallRinging = false },
    )
    private val outgoingCallRepeater = AudioRepeater(
        audioPlayer = audioPlayer,
        soundName = SOUND_RING_OUTGOING,
        onPlay = { delay(RING_FULL_DURATION) },
    )
    
    private val logger = getLogger("CallRepo")

    override suspend fun onEngineInit(profile: ChatProfileData) {
        // Nothing to do here
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
        coroutineScope.collect(audioRecorder.onFrame) { frame ->
            sendAudioFrame(frame)
        }
        coroutineScope.collect(settingsRepo.observer) { settings ->
            if (!settings.callSettings.isEnabled) endAllCalls()
        }
        callActivityMonitor = coroutineScope.launch {
            while (true) {
                if (lastCallActivityTimes.isNotEmpty()) {
                    val curTime = currentTimeMillis()
                    lastCallActivityTimes.removeIf { uid, time ->
                        if (curTime - time >= CALL_ACTIVITY_MONITOR_INTERVAL) {
                            setCallActivity(uid, isSpeaking = false)
                            true
                        } else {
                            false
                        }
                    }
                }
                delay(CALL_ACTIVITY_MONITOR_INTERVAL)
            }
        }
    }

    override fun onEngineStop() {
        calls.clear()
        callActivity.clear()
        lastCallActivityTimes.clear()
        incomingCallRepeater.reset()
        outgoingCallRepeater.reset()
        frameHandles.clear()
        groupFrameHandles.clear()
        callActivityMonitor?.cancel()
        callActivityMonitor = null
        coroutineScope = null
    }

    fun getCall(contactUid: ContactUid) = calls[contactUid]
    
    fun onCallReceived(contactUid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean) {
        logger.debug("onCallReceived($contactUid, $isAudioEnabled, $isVideoEnabled)")
        if (!settings.isEnabled && !settings.isVoicemailEnabled) {
            // Immediately reject
            engine.sendCallControl(contactUid, CallControl.CANCEL)
            return
        }
        val existing = calls[contactUid]
        if (existing != null && existing.isInProgress) {
            logger.warn("Received a duplicate call from $contactUid - ignoring...")
            return
        }
        val call = CallData(
            contactUid = contactUid,
            status = CallStatus.INCOMING,
            audioState = CallAudioState(isReceiverSending = isAudioEnabled),
            videoState = CallVideoState(isReceiverSending = isVideoEnabled),
        )
        calls[contactUid] = call
        when {
            settings.isEnabled -> {
                updatePendingCalls()
                onCallReceived.tryEmit(call)
            }
            settings.isVoicemailEnabled -> {
                // If you have calls disabled but voicemail enabled, go straight to voicemail
                answerCall(contactUid, isVoicemail = true)
            }
        }
    }

    fun onCallAnswered(contactUid: ContactUid, isSendingAudio: Boolean, isSendingVideo: Boolean, isAcceptingAudio: Boolean, isAcceptingVideo: Boolean) {
        logger.debug("onCallAnswered($contactUid, $isSendingAudio, $isSendingVideo, $isAcceptingAudio, $isAcceptingVideo)")
        val call = calls[contactUid] ?: return
        if (!call.isInProgress) {
            logger.warn("Call was unexpectedly answered with status ${call.status}")
            sendCallControl(contactUid, CallControl.CANCEL)
            return
        }
        updateCall(contactUid) {
            it.copy(
                status = if (it.isPending) CallStatus.ACTIVE else it.status,
                audioState = it.audioState.copy(
                    isReceiverSending = isSendingAudio,
                    isReceiverAccepting = isAcceptingAudio,
                ),
                videoState = it.videoState.copy(
                    isReceiverSending = isSendingVideo,
                    isReceiverAccepting = isAcceptingVideo,
                ),
            )
        }
    }
    
    fun onCallFinished(contactUid: ContactUid) {
        logger.debug("onCallFinished($contactUid)")
        val isCallInProgress = calls[contactUid]?.isInProgress == true
        if (updateCall(contactUid) { it.copy(status = CallStatus.FINISHED) }) {
            if (isCallInProgress) onContactHungUp(contactUid)
        }
    }
    
    fun onAudioBitRate(contactUid: ContactUid, bitRate: Int) {
        logger.debug("onAudioBitRate($contactUid, $bitRate)")
        updateCall(contactUid) { it.copy(audioState = it.audioState.copy(bitRate = bitRate)) }
    }
    
    fun onAudioFrame(contactUid: ContactUid, frame: AudioFrame) {
        val call = calls[contactUid] ?: return
        when (call.status) {
            CallStatus.ACTIVE -> {
                audioPlayer.queueFrame(frame, frameHandles[call.uid])?.also {
                    frameHandles[call.uid] = it
                }
            }
            CallStatus.VOICEMAIL -> {
                voicemailRepo.onAudioFrame(contactUid, frame)
            }
            else -> return
        }
    }
    
    fun onGroupAudioFrame(groupUid: GroupUid, peerUid: UserUid, frame: AudioFrame) {
        val frameHandlers = groupFrameHandles.getOrPut(groupUid) { mutableMapOf() }
        audioPlayer.queueFrame(frame, frameHandlers[peerUid])?.also {
            frameHandlers[peerUid] = it
            setCallActivity(peerUid, isSpeaking = true)
        }
    }
    
    fun startCall(contactUid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean) {
        if (contactUid is GroupUid) {
            if (setGroupAudioEnabled(contactUid, isEnabled = true)) {
                audioPlayer.queueSound(SOUND_PICK_UP)
            }
            return
        }
        val isStarted = engine.startCall(
            uid = contactUid,
            isAudioEnabled = isAudioEnabled,
            isVideoEnabled = isVideoEnabled,
        )
        if (isStarted) {
            calls[contactUid] = CallData(
                contactUid = contactUid,
                status = CallStatus.OUTGOING,
                audioState = CallAudioState(isSelfSending = isAudioEnabled),
                videoState = CallVideoState(isSelfSending = isVideoEnabled),
            )
            updatePendingCalls()
        }
    }
    
    fun answerCall(
        contactUid: ContactUid,
        isAudioEnabled: Boolean = true,
        isVideoEnabled: Boolean = false,
        isVoicemail: Boolean = false,
    ) {
        val isStarted = engine.answerCall(
            uid = contactUid,
            isAudioEnabled = isAudioEnabled,
            isVideoEnabled = isVideoEnabled,
        )
        if (isStarted) {
            updateCall(contactUid) {
                it.copy(status = if (isVoicemail) CallStatus.VOICEMAIL else CallStatus.ACTIVE)
            }
            if (isVoicemail) {
                voicemailRepo.sendVoicemail(contactUid)
            } else {
                audioPlayer.queueSound(SOUND_PICK_UP)
            }
        }
    }
    
    fun sendCallControl(contactUid: ContactUid, control: CallControl) {
        val call = calls[contactUid] ?: return
        val success = when (contactUid) {
            is UserUid -> engine.sendCallControl(contactUid, control)
            is GroupUid -> when (control) {
                CallControl.CANCEL -> setGroupAudioEnabled(contactUid, isEnabled = false)
                CallControl.PAUSE,
                CallControl.RESUME -> true
                else -> false
            }
            else -> false
        }
        if (success) {
            val newStatus = when (control) {
                CallControl.CANCEL -> CallStatus.FINISHED
                CallControl.RESUME -> CallStatus.ACTIVE
                CallControl.PAUSE -> CallStatus.PAUSED
                else -> call.status
            }
            updateCall(contactUid) { it.copy(status = newStatus) }
            when {
                call.status == CallStatus.ACTIVE && control == CallControl.CANCEL -> {
                    onSelfHungUp(contactUid)
                }
                call.isPending && control == CallControl.CANCEL -> {
                    if (call.status == CallStatus.OUTGOING) {
                        onSelfHungUp(contactUid)
                    } else {
                        onSelfRejectCall(contactUid)
                    }
                }
                control == CallControl.PAUSE -> onCallPaused(contactUid)
                control == CallControl.RESUME -> onCallResumed(contactUid)
            }
        }
    }
    
    fun sendAudioFrame(contactUid: ContactUid, frame: AudioFrame) {
        engine.sendAudioFrame(contactUid, frame)
    }
    
    fun sendAudioFrame(frame: AudioFrame) {
        calls.forEach { (uid, call) ->
            if (call.isActive) sendAudioFrame(uid, frame)
        }
    }
    
    /* EVENTS */
    
    private fun onSelfHungUp(contactUid: ContactUid) {
        audioPlayer.queueSound(SOUND_HANG_UP)
    }
    
    private fun onSelfRejectCall(contactUid: ContactUid) {
        audioPlayer.queueSound(SOUND_DECLINE)
    }
    
    private fun onContactHungUp(contactUid: ContactUid) {
        audioPlayer.queueSound(SOUND_DECLINE)
    }
    
    private fun onCallPaused(contactUid: ContactUid) {
        audioPlayer.queueSound(SOUND_PAUSE_RESUME)
    }

    private fun onCallResumed(contactUid: ContactUid) {
        audioPlayer.queueSound(SOUND_PAUSE_RESUME)
    }
    
    private fun updateCall(contactUid: ContactUid, block: (CallData) -> CallData): Boolean {
        val call = calls[contactUid] ?: run {
            logger.warn("Failed to find existing call from $contactUid")
            return false
        }
        val modified = block(call)
        calls[contactUid] = modified
        if (call.isPending != modified.isPending) {
            updatePendingCalls()
        }
        if (call.isActive != modified.isActive) {
            frameHandles.remove(call.uid)
            updateActiveCalls()
        }
        return call != modified
    }
    
    private fun updatePendingCalls() {
        val coroutineScope = coroutineScope ?: return
        val isReceivingCall = calls.values.find { it.status == CallStatus.INCOMING } != null
        val isSendingCall = calls.values.find { it.status == CallStatus.OUTGOING } != null
        outgoingCallRepeater.update(isSendingCall, coroutineScope)
        incomingCallRepeater.update(isReceivingCall, coroutineScope)
    }
    
    private fun updateActiveCalls() {
        val isInActiveCall = calls.values.find { it.isActive } != null
        if (isInActiveCall != this.isInActiveCall) {
            this.isInActiveCall = isInActiveCall
            if (isInActiveCall) audioRecorder.start() else audioRecorder.stop()
        }
    }
    
    private fun endAllCalls() {
        calls.values.filter { it.isInProgress && it.status != CallStatus.VOICEMAIL }.forEach {
            sendCallControl(it.contactUid, CallControl.CANCEL)
        }
    }
    
    private fun onCallRing() {
        val curTime = currentTimeMillis()
        val maxRingDuration = settings.maxRings * RING_FULL_DURATION
        val expiredCalls = calls.values.filter {
            it.status == CallStatus.INCOMING && curTime - it.startTime >= maxRingDuration
        }.map { it.contactUid }
        expiredCalls.forEach {
            if (settings.isVoicemailEnabled) {
                answerCall(it, isVoicemail = true)
            } else {
                sendCallControl(it, CallControl.CANCEL)
            }
        }
    }

    private fun setCallActivity(contactUid: ContactUid, isSpeaking: Boolean) {
        if (isSpeaking) {
            callActivity[contactUid] = isSpeaking
            lastCallActivityTimes[contactUid] = currentTimeMillis()
        } else {
            callActivity.remove(contactUid)
        }
    }

    private fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean): Boolean {
        if (engine.setGroupAudioEnabled(groupUid, isEnabled)) {
            groupFrameHandles[groupUid]?.clear()
            if (isEnabled) {
                calls[groupUid] = CallData(
                    contactUid = groupUid,
                    status = CallStatus.ACTIVE,
                )
            }
            updateActiveCalls()
            return true
        }
        return false
    }
}