package nand.apps.chat.repo

import androidx.compose.ui.util.fastForEach
import kotlinx.coroutines.*
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.image.SimpleBitmap
import nand.apps.chat.image.toPngByteArray
import nand.apps.chat.io.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.avatar.ChatAvatarCache
import nand.apps.chat.model.avatar.ChatSentAvatarData
import nand.apps.chat.model.file.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.time.currentTimeMillis
import okio.Path
import org.koin.core.Koin

const val AVATAR_CACHE_DIRECTORY = "avatars"
private const val AVATAR_CACHE_FILE = "cache.json"
private const val MAX_AVATAR_SIZE = 64 * 1024 // Per TCS 2.2.4

class ChatAvatarRepo(koin: Koin) : ChatEngineCallback, ChatAvatarProvider {

    private val engine: ChatEngine by koin.inject()
    private val encryption: ChatEngineEncryption by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val fileSystem: ChatFileSystem by koin.inject()

    private val avatarHashes = mutableMapOf<UserUid, String>()
    private val sentAvatarHashes = mutableMapOf<UserUid, ChatSentAvatarData>()

    lateinit var avatarCacheDirectory: ChatDirectory
        private set

    lateinit var incomingDirectory: ChatDirectory
        private set

    private lateinit var selfAvatarDirectory: ChatDirectory
    private lateinit var tempDirectory: ChatDirectory
    private lateinit var outgoingDirectory: ChatDirectory
    private lateinit var coroutineScope: CoroutineScope
    
    private val settings get() = settingsRepo.settings.fileSettings
    
    private val logger = getLogger("ChatAvatarRepo")
    
    override suspend fun onEngineInit(profile: ChatProfileData) {
        selfAvatarDirectory = fileSystem.getProfileDirectory(profile, "avatars")
        avatarCacheDirectory = fileSystem.getCacheDirectory("$AVATAR_CACHE_DIRECTORY/${profile.username}")
        tempDirectory = avatarCacheDirectory.getSubDirectory(".tmp")
        incomingDirectory = tempDirectory.getSubDirectory("incoming")
        outgoingDirectory = tempDirectory.getSubDirectory("outgoing")
        loadAvatarHashes()
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
        avatarHashes.clear()
        sentAvatarHashes.clear()
        tempDirectory.deleteRecursively()
    }

    override fun getAvatarPath(uid: UserUid): Path? {
        val hash = avatarHashes[uid] ?: return null
        return if (hash.isValidFileName() && hash in avatarCacheDirectory) {
            avatarCacheDirectory.getFilePath(hash)
        } else {
            null
        }
    }

    fun getSelfAvatarPath(status: UserStatus): Path? {
        val fileName = status.fileName
        return if (fileName in selfAvatarDirectory) {
            selfAvatarDirectory.getFilePath(fileName)
        } else {
            null
        }
    }

    fun getSelfAvatarPath(): Path? {
        val settings = selfProfileRepo.settings
        val status = when {
            settings.isLockedToDefaultState -> UserStatus.DEFAULT
            else -> settings.status
        }
        return getSelfAvatarPath(status)
    }

    suspend fun setSelfAvatar(bitmap: SimpleBitmap, status: UserStatus): String? {
        return withContext(Dispatchers.IO) {
            val pngBytes = try {
                bitmap.toPngByteArray()
            } catch (e: Exception) {
                logger.error("Failed to write PNG data to file", e)
                ByteArray(0)
            }
            if (pngBytes.isNotEmpty()) {
                selfAvatarDirectory.writeToFile(
                    fileName = status.fileName,
                    data = pngBytes,
                    encryption = encryption,
                )
                saveAvatarToCache(pngBytes)
            } else {
                null
            }
        }
    }

    fun checkSendSelfAvatar(selfUid: UserUid, friendUid: UserUid) {
        if (!settings.isAvatarTransferEnabled) return
        val selfHash = avatarHashes[selfUid] ?: return
        if (shouldSendAvatar(selfHash, friendUid)) {
            copyToOutgoingDirectory(selfHash)
            fileTransferRepo.sendFile(
                ChatFileTransfer(
                    fileUid = FileUid(selfHash),
                    friendUid = friendUid,
                    name = selfHash,
                    path = outgoingDirectory.getFilePath(selfHash).toString(),
                    size = outgoingDirectory.getFileSize(selfHash),
                    type = ChatFileType.AVATAR,
                    direction = ChatFileTransferDirection.OUTGOING,
                )
            )
        }
    }

    fun onSelfStatusChanged(self: UserData) {
        val avatarFileName = self.status.fileName.takeIf { it in selfAvatarDirectory } ?: return
        coroutineScope.launch {
            val hash = withContext(Dispatchers.IO) {
                selfAvatarDirectory.readFromFile(
                    fileName = avatarFileName,
                    encryption = encryption,
                )?.let { saveAvatarToCache(it) }
            }
            hash?.let { broadcastSelfAvatar(it) }
        }
    }

    fun onAvatarReceived(friendUid: UserUid, fileUid: FileUid, fileSize: Long, hash: String) {
        logger.debug("onAvatarReceived($friendUid, $fileUid, $fileSize, $hash)")
        if (!settings.isAvatarTransferEnabled) {
            logger.debug("Rejecting avatar for $friendUid - Avatar transfer is disabled")
            fileTransferRepo.rejectFileTransfer(friendUid, fileUid)
            return
        }
        if (fileSize > MAX_AVATAR_SIZE) {
            logger.debug("Rejecting avatar for $friendUid - Avatar exceeds max size of $MAX_AVATAR_SIZE ($fileSize)")
            fileTransferRepo.rejectFileTransfer(friendUid, fileUid)
            return
        }
        val hasAvatar = avatarCacheDirectory.contains(hash)
        val res = engine.sendFileControl(
            friendUid = friendUid,
            fileUid = fileUid,
            control = if (hasAvatar) ChatFileControl.CANCEL else ChatFileControl.RESUME,
        )
        when {
            hasAvatar -> updateAvatar(
                hash = hash,
                uid = friendUid,
            )
            res == ChatFileControlResponse.OK -> fileTransferRepo.updateTransfer(
                ChatFileTransfer(
                    fileUid = fileUid,
                    friendUid = friendUid,
                    name = hash,
                    path = incomingDirectory.getFilePath(hash).toString(),
                    size = fileSize,
                    type = ChatFileType.AVATAR,
                    state = ChatFileTransferState.ACTIVE,
                    direction = ChatFileTransferDirection.INCOMING,
                )
            )
        }
    }

    fun onAvatarSent(transfer: ChatFileTransfer) {
        fileTransferRepo.removeTransfer(transfer)
        sentAvatarHashes[transfer.friendUid] = ChatSentAvatarData(
            hash = transfer.name,
            lastSent = currentTimeMillis(),
        )
        saveAvatarHashes()
    }

    fun onGroupPeerAvatarReceived(group: ChatGroupData, peer: UserData, hash: String) {
        logger.debug("onGroupPeerAvatarReceived(${group.uid}, ${peer.uid}, $hash)")
        if (!settings.isAvatarTransferEnabled) return
        if (hash in avatarCacheDirectory) {
            updateAvatar(
                hash = hash,
                uid = peer.uid,
                groupUid = group.uid,
            )
        }
    }

    fun updateAvatar(hash: String, uid: UserUid, groupUid: GroupUid? = null): Boolean {
        if (avatarHashes[uid] != hash) {
            if (hash in incomingDirectory) moveFromIncomingDirectory(hash)
            avatarHashes[uid] = hash
            contactRepo.onUserAvatarChanged(
                uid = uid,
                groupUid = groupUid,
                avatarPath = getAvatarPath(uid),
            )
            saveAvatarHashes()
            return true
        }
        return false
    }

    private fun shouldSendAvatar(hash: String, friendUid: UserUid): Boolean {
        return sentAvatarHashes[friendUid]?.let { data ->
            data.hash != hash || currentTimeMillis() - data.lastSent >= settings.avatarExpireTime
        } ?: true
    }

    private fun broadcastSelfAvatar(hash: String) {
        if (hash !in avatarCacheDirectory) {
            logger.warn("Failed to broadcast non-existent avatar: $hash")
            return
        }

        if (!updateAvatar(hash, selfProfileRepo.getSelfUid()) || !settings.isAvatarTransferEnabled) return

        copyToOutgoingDirectory(hash)
        val path = outgoingDirectory.getFilePath(hash).toString()
        val fileSize = outgoingDirectory.getFileSize(hash)
        val fileUid = FileUid(hash)
        contactRepo.getOnlineFriends().fastForEach { friend ->
            if (shouldSendAvatar(hash, friend.uid)) {
                fileTransferRepo.sendFile(
                    ChatFileTransfer(
                        fileUid = fileUid,
                        friendUid = friend.uid,
                        name = hash,
                        path = path,
                        size = fileSize,
                        type = ChatFileType.AVATAR,
                        direction = ChatFileTransferDirection.OUTGOING,
                    )
                )
            }
        }
        val avatarPath = avatarCacheDirectory.getFilePath(hash)
        contactRepo.getOnlineGroups().fastForEach { group ->
            if (engine.setGroupAvatar(group.uid, hash)) {
                contactRepo.onUserAvatarChanged(
                    uid = group.selfUid,
                    groupUid = group.uid,
                    avatarPath = avatarPath,
                )
            }
        }
    }

    private val UserStatus.fileName get() = name.lowercase()

    @OptIn(ExperimentalStdlibApi::class)
    private fun saveAvatarToCache(pngData: ByteArray): String? {
        val hash = engine.generateHash(pngData)?.toHexString(HexFormat.UpperCase) ?: return null
        if (hash !in avatarCacheDirectory) {
            avatarCacheDirectory.writeToFile(
                fileName = hash,
                data = pngData,
                encryption = encryption,
            )
        }
        return hash
    }
    
    private fun copyToOutgoingDirectory(hash: String) {
        val data = avatarCacheDirectory.readFromFile(
            fileName = hash,
            encryption = encryption,
        ) ?: return
        outgoingDirectory.writeToFile(
            fileName = hash,
            data = data,
        )
    }
    
    private fun moveFromIncomingDirectory(hash: String) {
        val data = incomingDirectory.readFromFile(hash) ?: return
        incomingDirectory.removeFile(hash)
        avatarCacheDirectory.writeToFile(
            fileName = hash,
            data = data,
            encryption = encryption,
        )
    }

    private fun loadAvatarHashes() {
        if (AVATAR_CACHE_FILE !in avatarCacheDirectory) return

        val avatarCache: ChatAvatarCache = avatarCacheDirectory.readFromFile(
            fileName = AVATAR_CACHE_FILE,
            encryption = encryption,
        )?.let {
            try {
                Json.decodeFromString(it.decodeToString())
            } catch (e: Exception) {
                logger.error("Failed to decode avatar hashes", e)
                null
            }
        } ?: run {
            avatarCacheDirectory.removeFile(AVATAR_CACHE_FILE)
            return
        }

        avatarHashes.putAll(avatarCache.received)
        sentAvatarHashes.putAll(avatarCache.sent)
    }

    private fun saveAvatarHashes() {
        if (avatarHashes.isEmpty() && sentAvatarHashes.isEmpty()) return

        val cache = ChatAvatarCache(
            received = avatarHashes,
            sent = sentAvatarHashes,
        )
        
        val data = try {
            Json.encodeToString(cache).encodeToByteArray()
        } catch (e: Exception) {
            logger.error("Failed to encode avatar hashes", e)
            return
        }

        avatarCacheDirectory.writeToFile(
            fileName = AVATAR_CACHE_FILE,
            data = data,
            encryption = encryption,
        )
    }
}