package nand.apps.chat.repo

import androidx.compose.runtime.mutableStateMapOf
import androidx.compose.runtime.snapshotFlow
import io.ktor.util.*
import kotlinx.coroutines.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.io.ChatDirectory
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.event.ErrorDialogEvent
import nand.apps.chat.model.file.ChatFileTransfer
import nand.apps.chat.model.file.ChatFileTransferDirection
import nand.apps.chat.model.file.ChatFileTransferState
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.*
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.platform.AppState
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.ui.navigation.NavRoute
import nand.apps.chat.util.*
import org.koin.core.Koin
import seers.composeapp.generated.resources.*
import kotlin.math.abs

private const val FIND_MESSAGE_TIMESTAMP_THRESHOLD = 5 * 1000L // 5 seconds

class ChatMessageRepo(
    private val engine: ChatEngine,
    private val encryption: ChatEngineEncryption,
    private val fileSystem: ChatFileSystem,
    private val backgroundWorker: ChatBackgroundWorker,
    private val appStateProvider: AppStateProvider,
    koin: Koin,
) : ChatEngineCallback {

    private val messages = mutableMapOf<ContactUid, MutableList<ChatMessage>>()
    private val unreadMessages = mutableStateMapOf<ContactUid, UnreadMessageCounts>()
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()
    private val json = Json { ignoreUnknownKeys = true }
    private val persistedMutex = Mutex()
    private val logger = getLogger("ChatMessageRepo")

    private lateinit var messagesDirectory: ChatDirectory
    private lateinit var coroutineScope: CoroutineScope

    private var conversationUid: ContactUid? = null
    private var currentNavRoute: NavRoute? = null
    private var firstVisibleMessageIndex = 0

    val onNewMessage = fifoSharedFlow<ChatMessage>()
    val onMessagesChanged = fifoSharedFlow<ContactUid>()
    val unreadObserver = snapshotFlow { unreadMessages.toMap() }
    
    private val settings get() = settingsRepo.settings.chatSettings
    private val maxMessages get() = settings.maxMessages

    fun isConversationVisible(conversationUid: ContactUid) =
        appStateProvider.currentState == AppState.FOREGROUND &&
                currentNavRoute == MainRoute &&
                conversationUid == this.conversationUid

    fun onConversationChanged(conversationUid: ContactUid?) {
        if (this.conversationUid != conversationUid) {
            this.conversationUid = conversationUid
            markConversationRead()
        }
    }

    fun onAppFocused() {
        if (firstVisibleMessageIndex == 0) markConversationRead()
    }

    fun onFirstVisibleMessageIndex(index: Int) {
        if (index == firstVisibleMessageIndex) return

        val oldIndex = firstVisibleMessageIndex
        firstVisibleMessageIndex = index
        val conversationUid = conversationUid ?: return
        val messages = messages[conversationUid] ?: return
        if (index < oldIndex) {
            val unreadMessages = mutableMapOf<String, ChatMessage>()
            for (i in messages.lastIndex - oldIndex..messages.lastIndex - index) {
                if (messages[i].isUnread) {
                    messages[i] = messages[i].copy(isUnread = false)
                    unreadMessages[messages[i].uid.value] = messages[i]
                }
            }
            if (unreadMessages.isNotEmpty()) {
                this.unreadMessages[conversationUid] = getUnreadMessageCounts(conversationUid)
                markPersistedMessagesRead(
                    conversationUid = conversationUid,
                    unreadMessages = unreadMessages,
                )
            }
        }
    }

    fun onNavRoute(route: NavRoute) {
        if (currentNavRoute != route) {
            currentNavRoute = route
            if (route == MainRoute) markConversationRead()
        }
    }
    
    fun createMessage(text: String, conversationUid: ContactUid): ChatMessage {
        val messageType = if (text.startsWith("/me ")) ChatMessageType.ACTION else ChatMessageType.NORMAL
        val formattedMessage = if (messageType == ChatMessageType.ACTION) text.substringAfter(' ') else text
        return ChatMessage(
            conversationUid = conversationUid,
            senderUid = selfProfileRepo.getSelfUid(conversationUid),
            text = formattedMessage,
            type = messageType,
        )
    }

    fun getMessages(conversationUid: ContactUid) =
        messages[conversationUid]?.toList().orEmpty()

    fun getUnreadMessageCounts(conversationUid: ContactUid): UnreadMessageCounts {
        var directMessages = 0
        var groupMessages = 0
        val contact = contactRepo.getContact(conversationUid) ?: return UnreadMessageCounts()
        messages[conversationUid]?.forEach { message ->
            if (message.isUnread) {
                when (contact) {
                    is ChatGroupData -> {
                        if (message.containsSelfMention(contact)) directMessages++ else groupMessages++
                    }
                    else -> directMessages++
                }
            }
        }
        return UnreadMessageCounts(
            directMessages = directMessages,
            groupMessages = groupMessages,
        )
    }

    fun sendMessage(message: ChatMessage) {
        if (engine.sendMessage(message)) {
            addMessage(message)
        }
    }

    fun addMessage(message: ChatMessage) {
        val conversation = contactRepo.getContact(message.conversationUid) ?: return
        val selfUid = selfProfileRepo.getSelfUid(message.conversationUid)
        val isUnread = message.senderUid != selfUid && (firstVisibleMessageIndex > 0 || !isConversationVisible(message.conversationUid))
        val modified = if (isUnread) {
            message.copy(isUnread = true)
        } else {
            message
        }
        val messages = messages.getOrPut(message.conversationUid) { mutableListOf() }
        messages += modified
        if (messages.size > maxMessages) messages.removeAt(0)
        onNewMessage.tryEmit(modified)
        if (shouldPersistMessage(modified)) {
            persistedMessageLock { persistMessage(modified) }
        }
        if (isUnread) {
            val unreadMessages = unreadMessages[message.conversationUid] ?: UnreadMessageCounts()
            this.unreadMessages[message.conversationUid] = unreadMessages.run {
                if (conversation is UserData || conversation is ChatGroupData && modified.containsSelfMention(conversation)) {
                    copy(directMessages = directMessages + 1)
                } else {
                    copy(groupMessages = groupMessages + 1)
                }
            }
        }
    }
    
    fun updateMessage(message: ChatMessage, localOnly: Boolean) {
        var existing: ChatMessage? = null
        messages[message.conversationUid]?.replaceAll {
            if (message.uid == it.uid) {
                existing = it
                message
            } else {
                it
            }
        }
        onMessagesChanged.tryEmit(message.conversationUid)
        existing?.let {
            if (!localOnly) engine.editMessage(old = it, new = message)
        }
        if (shouldPersistMessage(message)) {
            persistedMessageLock { editPersistedMessage(message) }
        }
    }
    
    fun addReaction(
        message: ChatMessage,
        reaction: String,
    ) {
        val selfUid = selfProfileRepo.getSelfUid(message.conversationUid)
        val reactions = message.reactions[selfUid]?.toMutableSet() ?: mutableSetOf()
        if (reaction in reactions) {
            reactions -= reaction
        } else {
            reactions += reaction
        }
        if (reactions.size <= MAX_REACTIONS_PER_USER) {
            setReactions(
                senderUid = selfUid,
                message = message,
                reactions = reactions,
                localOnly = false,
            )
        } else {
            eventHandler += ErrorDialogEvent(getStringBlocking(Res.string.message_reaction_limit_message))
        }
    }
    
    fun setReactions(
        senderUid: UserUid,
        message: ChatMessage,
        reactions: Set<String>,
        localOnly: Boolean,
    ) {
        if (reactions.size > MAX_REACTIONS_PER_USER) return
        updateMessage(
            message = message.setReactions(
                senderUid = senderUid,
                reactions = reactions,
            ),
            localOnly = true,
        )
        if (!localOnly) {
            engine.setMessageReactions(
                message = message,
                reactions = reactions,
            )
        }
    }
    
    fun removeMessage(message: ChatMessage, localOnly: Boolean) {
        messages[message.conversationUid]?.removeAll { it.uid == message.uid }
        onMessagesChanged.tryEmit(message.conversationUid)
        if (!localOnly) engine.removeMessage(message)
        if (shouldPersistMessage(message)) {
            persistedMessageLock { removePersistedMessage(message) }
        }
    }

    fun removeMessages(conversationUid: ContactUid, isRemovingContact: Boolean) {
        messages.remove(conversationUid)
        unreadMessages.remove(conversationUid)
        deletePersistedMessages(conversationUid)
        if (!isRemovingContact) {
            onMessagesChanged.tryEmit(conversationUid)
        }
    }

    fun removeGroupInvite(groupInvite: ChatGroupInvite) {
        messages[groupInvite.friendUid]?.replaceAll { message ->
            if (groupInvite.cookie == message.groupInvite?.cookie) {
                message.copy(groupInvite = null)
            } else {
                message
            }
        }
        onMessagesChanged.tryEmit(groupInvite.friendUid)
    }

    fun findMessage(
        conversationUid: ContactUid,
        senderUid: ContactUid,
        textHash: String,
        timestamp: Long,
    ) = messages[conversationUid]?.find {
        it.senderUid == senderUid &&
                abs(it.timestamp - timestamp) <= FIND_MESSAGE_TIMESTAMP_THRESHOLD &&
                it.text.sha256() == textHash
    }

    override suspend fun onEngineInit(profile: ChatProfileData) {
        messagesDirectory = fileSystem.getProfileDirectory(profile, "messages")
        loadPersistedMessages()
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
        messages.clear()
        conversationUid = null
        firstVisibleMessageIndex = 0
    }

    fun onFriendRequest(request: FriendRequest) {
        logger.debug("onFriendRequest($request)")
        addMessage(
            ChatMessage(
                conversationUid = request.uid,
                senderUid = request.uid,
                text = request.message,
                friendRequest = request,
            )
        )
    }
    
    fun onFriendRequestAccepted(request: FriendRequest, newFriend: UserData) {
        messages.remove(request.uid)?.let { existingMessages ->
            messages[newFriend.uid] = existingMessages.map {
                it.copy(
                    conversationUid = request.uid,
                    senderUid = newFriend.uid,
                    friendRequest = null,
                )
            }.toMutableList()
        }
        addMessage(
            ChatMessage(
                conversationUid = newFriend.uid,
                senderUid = newFriend.uid,
                text = getStringBlocking(Res.string.user_added_message),
                type = ChatMessageType.ACTION,
            )
        )
    }

    fun onFriendMessage(friendUid: UserUid, messageType: ChatMessageType, message: String) {
        logger.debug("onFriendMessage($friendUid, $messageType, $message)")
        addMessage(
            ChatMessage(
                conversationUid = friendUid,
                senderUid = friendUid,
                text = message,
                type = messageType,
            )
        )
    }

    fun onFriendReadReceipt(friendUid: UserUid, messageUid: MessageUid) {
        logger.debug("onFriendReadReceipt($friendUid, $messageUid)")
        messages[friendUid]?.replaceAll { message ->
            if (messageUid == message.uid) {
                message.copy(
                    timestamp = currentTimeMillis(),
                    isDelivered = true,
                )
            } else {
                message
            }
        }
        onMessagesChanged.tryEmit(friendUid)
    }

    fun onGroupInvite(invite: ChatGroupInvite) {
        logger.debug("onGroupInvite($invite)")
        addMessage(
            ChatMessage(
                conversationUid = invite.friendUid,
                senderUid = invite.friendUid,
                text = getStringBlocking(Res.string.group_invite_peer_message, "\n\$tiny\$${invite.cookie}"),
                groupInvite = invite,
                type = ChatMessageType.ACTION,
            )
        )
    }

    fun onGroupMessage(groupUid: GroupUid, peerUid: UserUid, messageType: ChatMessageType, message: String) {
        logger.debug("onGroupMessage($groupUid, $peerUid, $messageType, $message)")
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = peerUid,
                text = message,
                type = messageType,
            )
        )
    }

    fun onGroupUserName(groupUid: GroupUid, peerUid: UserUid, oldName: String, newName: String) {
        logger.debug("onGroupUserName($groupUid, $peerUid, $oldName, $newName)")
        if (settings.isPeerNameMessagesEnabled) {
            addMessage(
                ChatMessage(
                    conversationUid = groupUid,
                    senderUid = groupUid,
                    text = if (oldName.isBlank() || oldName == newName) {
                        getStringBlocking(Res.string.peer_joined_group_message, newName)
                    } else {
                        getStringBlocking(Res.string.peer_name_changed_message, oldName, newName)
                    },
                    type = ChatMessageType.GROUP,
                )
            )
        }
    }
    
    fun onGroupPeerConnection(groupUid: GroupUid, peerUid: UserUid, isOnline: Boolean) {
        logger.debug("onGroupPeerConnection($groupUid, $peerUid, $isOnline)")
        if (settings.isPeerConnectMessagesEnabled) {
            addMessage(
                ChatMessage(
                    conversationUid = groupUid,
                    senderUid = peerUid,
                    text = "",
                    type = if (isOnline) ChatMessageType.PEER_ONLINE else ChatMessageType.PEER_OFFLINE,
                )
            )
        }
    }
    
    fun onGroupPeerKicked(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid) {
        logger.debug("onGroupPeerKicked($groupUid, $sourceUid, $targetUid)")
        val group = contactRepo.getGroup(groupUid) ?: return
        val source = group.peers[sourceUid] ?: return
        val target = group.peers[targetUid] ?: return
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = groupUid,
                text = getStringBlocking(Res.string.group_event_peer_kicked_message, source.displayName, target.displayName),
                type = ChatMessageType.GROUP,
            )
        )
    }

    fun onGroupPeerRoleChanged(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid, role: ChatGroupRole) {
        logger.debug("onGroupPeerRoleChanged($groupUid, $sourceUid, $targetUid, $role)")
        val group = contactRepo.getGroup(groupUid) ?: return
        val source = group.peers[sourceUid] ?: return
        val target = group.peers[targetUid] ?: return
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = groupUid,
                text = getStringBlocking(Res.string.group_event_peer_role_changed_message, source.displayName, target.displayName, role.getDisplayString()),
                type = ChatMessageType.GROUP,
            )
        )
    }
    
    fun addFileTransferMessage(
        transfer: ChatFileTransfer,
        isAutoAccepted: Boolean,
    ) = with(transfer) {
        if (transfer.type == ChatFileType.AVATAR) return
        val senderUid = when (transfer.direction) {
            ChatFileTransferDirection.OUTGOING -> selfProfileRepo.getSelfUid()
            ChatFileTransferDirection.INCOMING -> friendUid
        }
        val text = when {
            isAutoAccepted || transfer.direction == ChatFileTransferDirection.OUTGOING -> ""
            else -> getStringBlocking(Res.string.user_sent_file_message)
        }
        addMessage(
            ChatMessage(
                conversationUid = friendUid,
                senderUid = senderUid,
                text = text,
                type = if (isAutoAccepted) ChatMessageType.NORMAL else ChatMessageType.ACTION,
                fileTransferUid = uid,
            )
        )
    }

    private fun markConversationRead() {
        unreadMessages.remove(conversationUid)
        val conversationUid = conversationUid ?: return
        val unreadMessages = messages[conversationUid].orEmpty().filter { it.isUnread }.associateBy { it.uid.value }
        if (unreadMessages.isEmpty()) return
        messages[conversationUid]?.replaceAll { message ->
            if (message.isUnread) {
                message.copy(isUnread = false)
            } else {
                message
            }
        }
        markPersistedMessagesRead(
            conversationUid = conversationUid,
            unreadMessages = unreadMessages,
        )
    }

    private fun markPersistedMessagesRead(
        conversationUid: ContactUid,
        unreadMessages: Map<String, ChatMessage>,
    ) = persistedMessageLock {
        messagesDirectory.editLines(conversationUid.value) { line ->
            val uid = line.substringBefore(':')
            val message = unreadMessages[uid]
            if (message != null) {
                message.copy(isUnread = false).toPersistedMessageLine()
            } else {
                line
            }
        }
    }

    private fun ChatMessage.containsSelfMention(group: ChatGroupData) = containsSelfMention(
        group = group,
        mentionsRegex = engine.constraints.mentionRegex,
    )
    
    private fun shouldPersistMessage(message: ChatMessage): Boolean {
        // TODO - Persist pending friend requests
        return this::messagesDirectory.isInitialized &&
                settings.isMessageHistoryEnabled &&
                message.conversationUid !is FriendRequestUid
    }
    
    private fun persistedMessageLock(block: () -> Unit) {
        coroutineScope.launch(backgroundWorker.dispatcher) { persistedMutex.withLock(action = block) }
    }
    
    private fun persistMessage(message: ChatMessage) {
        val line = message.toPersistedMessageLine()
        messagesDirectory.appendToFile(
            fileName = message.conversationUid.value,
            data = "$line\n".encodeToByteArray(),
        )
    }

    private fun editPersistedMessage(message: ChatMessage) {
        messagesDirectory.editLines(message.conversationUid.value) { line ->
            val uid = line.substringBefore(':')
            if (uid == message.uid.value) {
                message.toPersistedMessageLine()
            } else {
                line
            }
        }
    }
    
    private fun removePersistedMessage(message: ChatMessage) {
        messagesDirectory.editLines(message.conversationUid.value) { line ->
            val uid = line.substringBefore(':')
            if (uid == message.uid.value) null else line
        }
    }
    
    private fun ChatMessage.toPersistedMessageLine(): String? {
        val fileTransfer = fileTransferUid?.let { fileTransferRepo[it] }
        val persistedMessage = toPersistedMessage(fileTransfer)
        val messageJson = json.encodeToString(persistedMessage)
        val bytes = messageJson.encodeToByteArray().let {
            if (encryption.isEncryptionSupported()) encryption.encryptData(it) else it
        } ?: run {
            logger.error("Failed to encrypt persisted message: $uid")
            return null
        }
        return uid.value + ":" + bytes.encodeBase64()
    }
    
    private suspend fun loadPersistedMessages() {
        if (!this::messagesDirectory.isInitialized || !settings.isMessageHistoryEnabled) return
        
        val start = currentTimeMillis()
        val messageLimit = maxMessages
        withContext(backgroundWorker.dispatcher) {
            messagesDirectory.listFiles().forEach { path ->
                var directMessages = 0
                var groupMessages = 0
                val conversation = contactRepo.getContact(path.name) ?: run {
                    messagesDirectory.removeFile(path.name)
                    return@forEach
                }
                var lineCount = 0
                val loadedMessages = mutableListOf<ChatMessage>()
                val rawLines = mutableListOf<String>()
                messagesDirectory.readLines(path.name) { line ->
                    lineCount++
                    decodePersistedMessage(line, conversation)?.let { message ->
                        loadedMessages += message
                        rawLines += line
                        if (message.isUnread) {
                            when (conversation) {
                                is ChatGroupData -> {
                                    if (message.containsSelfMention(conversation)) directMessages++ else groupMessages++
                                }
                                else -> directMessages++
                            }
                        }
                    }
                    if (loadedMessages.size > messageLimit) {
                        loadedMessages.removeAt(0)
                        rawLines.removeAt(0)
                    }
                    true
                }
                if (rawLines.size < lineCount) {
                    // Time to trim the fat
                    messagesDirectory.writeLines(
                        fileName = path.name,
                        lines = rawLines,
                    )
                }
                messages[conversation.uid] = loadedMessages
                unreadMessages[conversation.uid] = UnreadMessageCounts(
                    directMessages = directMessages,
                    groupMessages = groupMessages,
                )
            }
        }
        val loadTime = currentTimeMillis() - start
        if (loadTime > 1000L) {
            logger.warn("Loading persisted messages took $loadTime ms!")
        } else {
            logger.verbose("Loading persisted messages took $loadTime ms")
        }
    }
    
    private fun decodePersistedMessage(line: String, conversation: ChatContact): ChatMessage? {
        val uid = line.substringBefore(':', randomUid())
        val message = line.substringAfter(':').decodeBase64Bytes().let {
            if (encryption.isEncryptionSupported() && encryption.isEncrypted(it)) encryption.decryptData(it) else it
        }?.let {
            try {
                json.decodeFromString<PersistedChatMessage>(it.decodeToString())
            } catch (e: Exception) {
                null
            }
        } ?: return null
        val sender = contactRepo.getContact(
            uid = message.senderUid,
            userGroup = if (conversation is ChatGroupData) conversation else null,
        ) ?: return null
        val fileTransfer = message.fileTransfer?.run {
            if (fileSystem.exists(path) && fileSystem.getSize(path) == size) {
                ChatFileTransfer(
                    fileUid = FileUid(fileUid),
                    friendUid = UserUid(message.senderUid),
                    name = name,
                    path = path,
                    size = size,
                    direction = direction,
                    state = ChatFileTransferState.COMPLETE,
                ).also {
                    fileTransferRepo.addHistoricalTransfer(it)
                }
            } else {
                null
            }
        }
        return ChatMessage(
            conversationUid = conversation.uid,
            senderUid = sender.uid,
            text = message.text,
            uid = MessageUid(uid),
            timestamp = message.timestamp,
            type = message.type,
            isDelivered = message.isDelivered,
            isEdited = message.isEdited,
            isUnread = message.isUnread,
            reactions = message.reactions.orEmpty(),
            fileTransferUid = fileTransfer?.uid,
        )
    }
    
    private fun deletePersistedMessages(conversationUid: ContactUid) {
        if (this::messagesDirectory.isInitialized && conversationUid.value in messagesDirectory) {
            messagesDirectory.removeFile(conversationUid.value)
        }
    }
}