package nand.apps.chat.repo

import androidx.compose.runtime.*
import androidx.compose.ui.util.fastForEach
import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.*
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.*
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.model.event.*
import nand.apps.chat.util.getStringBlocking
import okio.Path
import org.koin.core.Koin
import seers.composeapp.generated.resources.*

class ContactRepo(
    private val engine: ChatEngine,
    koin: Koin,
) : ChatEngineCallback {

    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()
    
    private val friends = mutableStateMapOf<UserUid, UserData>()
    private val groups = mutableStateMapOf<GroupUid, ChatGroupData>()
    private val friendRequests = mutableStateMapOf<FriendRequestUid, FriendRequest>()
    private val groupInvites = mutableStateMapOf<String, ChatGroupInvite>()
    
    private val logger = getLogger("ContactRepo")

    val friendObserver = snapshotFlow { friends.toMap() }
    val groupObserver = snapshotFlow { groups.toMap() }
    val friendRequestObserver = snapshotFlow { friendRequests.toMap() }
    val self get() = selfProfileRepo.self
    
    fun getOnlineFriends() = friends.values.filter { it.isOnline }
    
    fun getOnlineGroups() = groups.values.filter { it.isOnline }

    fun getOfflineGroups() = groups.values.filter { !it.isOnline }
    
    fun getFriend(uid: UserUid) = friends[uid]
    
    fun getGroup(uid: GroupUid) = groups[uid]
    
    fun getContact(uid: ContactUid, userGroup: ChatGroupData? = null) = when (uid) {
        is UserUid -> when {
            uid == self.uid -> self
            userGroup?.peers?.containsKey(uid) == true -> userGroup.peers[uid]
            else -> friends[uid]
        }
        is FriendRequestUid -> friendRequests[uid]
        is GroupUid -> groups[uid]
    }
    
    fun getContact(uid: String, userGroup: ChatGroupData? = null): ChatContact? {
        return getContact(UserUid(uid), userGroup) ?: groups[GroupUid(uid)]
    }
    
    fun isPeerBlocked(groupUid: GroupUid, peerUid: UserUid) =
        getGroup(groupUid)?.isPeerBlocked(peerUid) == true

    override suspend fun onEngineInit(profile: ChatProfileData) {
        friends.putAll(engine.getFriendList().associateBy { it.uid })
        engine.getGroupList().fastForEach { groups[it.uid] = it }
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        // Nothing to do here
    }

    override fun onEngineStop() {
        friends.clear()
        groups.clear()
        friendRequests.clear()
        groupInvites.clear()
    }

    fun onFriendName(friendUid: UserUid, name: String): Boolean {
        logger.debug("onFriendName($friendUid, $name)")
        return updateFriend(friendUid) { it.copy(name = name) }
    }

    fun onFriendStatus(friendUid: UserUid, status: UserStatus) {
        logger.debug("onFriendStatus($friendUid, $status)")
        updateFriend(friendUid) { it.copy(status = status) }
    }

    fun onFriendStatusMessage(friendUid: UserUid, message: String): Boolean {
        logger.debug("onFriendStatusMessage($friendUid, $message)")
        return updateFriend(friendUid) { it.copy(statusMessage = message) }
    }

    fun onFriendClientVersion(friendUid: UserUid, version: ChatClientVersion): Boolean {
        logger.debug("onFriendClientVersion($friendUid, $version)")
        return updateFriend(friendUid) { it.copy(clientVersion = version) }
    }

    fun onFriendConnectionProtocol(friendUid: UserUid, protocol: ConnectionProtocol) {
        logger.debug("onFriendConnectionProtocol($friendUid, $protocol)")
        val wasOnline = friends[friendUid]?.isOnline == true
        updateFriend(friendUid) { it.copy(protocol = protocol) }
        if (!wasOnline && protocol != ConnectionProtocol.NONE) {
            avatarRepo.checkSendSelfAvatar(self.uid, friendUid)
        }
    }

    fun onUserAvatarChanged(uid: UserUid, groupUid: GroupUid?, avatarPath: Path?) {
        logger.debug("onUserAvatarChanged($uid, $groupUid, ${avatarPath?.name})")
        when {
            self.uid == uid -> selfProfileRepo.setCurrentAvatarPath(avatarPath)
            groupUid != null -> groups[groupUid]?.let { group ->
                group.peers[uid]?.let { peer ->
                    updateGroupPeer(groupUid, peer.uid) { it.copy(avatarPath = avatarPath) }
                }
            }
            else -> friends[uid]?.let {
                updateFriend(uid) { it.copy(avatarPath = avatarPath) }
            }
        }
    }

    fun onFriendRequest(request: FriendRequest) {
        logger.debug("onFriendRequest($request)")
        friendRequests[request.uid] = request
    }

    fun onGroupInvite(invite: ChatGroupInvite) {
        logger.debug("onGroupInvite($invite)")
        groupInvites[invite.cookie] = invite
    }
    
    fun onGroupConnected(
        groupUid: GroupUid,
        groupData: ChatGroupData? = null,
        fetchPeers: Boolean = false,
    ) {
        logger.debug("onGroupConnected($groupUid)")
        updateGroup(groupUid) {
            (groupData ?: it).copy(
                isOnline = true,
                peers = if (fetchPeers) engine.getGroupPeerList(groupUid) else it.peers,
            )
        }
    }
    
    fun onGroupErrorState(groupUid: GroupUid, error: ChatGroupErrorType) {
        logger.debug("onGroupErrorState($groupUid, $error)")
        updateGroup(groupUid) { it.copy(error = error, isOnline = false) }
    }

    fun onGroupTitle(groupUid: GroupUid, title: String): Boolean {
        logger.debug("onGroupTitle($groupUid, $title)")
        return updateGroup(groupUid) { it.copy(title = title) }
    }

    fun onGroupDescription(groupUid: GroupUid, description: String): Boolean {
        logger.debug("onGroupDescription($groupUid, $description)")
        return updateGroup(groupUid) { it.copy(description = description) }
    }
    
    fun onGroupPassword(groupUid: GroupUid, password: String?): Boolean {
        logger.debug("onGroupPassword($groupUid, *****)")
        return updateGroup(groupUid) { it.copy(password = password) }
    }

    fun onGroupVisibility(groupUid: GroupUid, visibility: ChatGroupVisibility): Boolean {
        logger.debug("onGroupVisibility($groupUid, $visibility)")
        return updateGroup(groupUid) { it.copy(visibility = visibility) }
    }

    fun onGroupUsersChanged(groupUid: GroupUid): Boolean {
        val group = groups[groupUid]
        if (group?.isOnline != true) return false
        logger.debug("onGroupUsersChanged($groupUid)")
        
        // Track which users have just connected or disconnected from the group
        val newPeers = engine.getGroupPeerList(groupUid)
        group.peers.forEach { (uid, peer) ->
            val newPeer = newPeers[uid] ?: return@forEach
            if (peer.isOnline != newPeer.isOnline && !group.isPeerBlocked(peer.uid)) {
                messageRepo.onGroupPeerConnection(groupUid, peer.uid, peer.isOnline)
            }
        }
        
        return updateGroup(groupUid) { it.copy(peers = newPeers) }
    }
    
    fun onGroupPeerJoin(groupUid: GroupUid, peer: UserData, isSelf: Boolean): Boolean {
        val group = groups[groupUid] ?: return false
        logger.debug("onGroupPeerJoin($groupUid, ${peer.uid})")
        if (group.peers[peer.uid]?.isOnline == false && !group.isPeerBlocked(peer.uid)) {
            messageRepo.onGroupPeerConnection(groupUid, peer.uid, isOnline = true)
        }
        return updateGroup(groupUid) {
            val peers = it.peers.toMutableMap()
            peers[peer.uid] = initGroupPeer(it, peer, isSelf)
            it.copy(peers = peers)
        }
    }
    
    fun onGroupPeerExit(groupUid: GroupUid, peerUid: UserUid): Boolean {
        val group = groups[groupUid] ?: return false
        logger.debug("onGroupPeerExit($groupUid, $peerUid)")
        if (group.peers[peerUid]?.isOnline == true && !group.isPeerBlocked(peerUid)) {
            messageRepo.onGroupPeerConnection(groupUid, peerUid, isOnline = false)
        }
        return updateGroupPeer(groupUid, peerUid) {
            it.copy(
                protocol = ConnectionProtocol.NONE,
                lastOnline = currentTimeMillis(),
            )
        }
    }

    fun onGroupPeerName(groupUid: GroupUid, peerUid: UserUid, name: String): Boolean {
        if (groups[groupUid]?.isOnline != true) return false
        logger.debug("onGroupUserName($groupUid, $peerUid, $name)")
        return updateGroupPeer(groupUid, peerUid) { peer ->
            if (peer.role != ChatGroupRole.OBSERVER && !isPeerBlocked(groupUid, peerUid)) {
                messageRepo.onGroupUserName(groupUid, peerUid, peer.name, name)
            }
            peer.copy(name = name)
        }
    }
    
    fun onGroupPeerStatus(groupUid: GroupUid, peerUid: UserUid, status: UserStatus): Boolean {
        logger.debug("onGroupPeerStatus($groupUid, $peerUid, $status)")
        return updateGroupPeer(groupUid, peerUid) { it.copy(status = status) }
    }

    fun onGroupPeerClientVersion(groupUid: GroupUid, peerUid: UserUid, version: ChatClientVersion): Boolean {
        logger.debug("onGroupPeerClientVersion($groupUid, $peerUid, $version)")
        return updateGroupPeer(groupUid, peerUid) { it.copy(clientVersion = version) }
    }

    fun onGroupPeerKicked(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid) {
        logger.debug("onGroupPeerKicked($groupUid, $sourceUid, $targetUid)")
        if (targetUid == selfProfileRepo.getSelfUid(groupUid)) {
            onGroupErrorState(
                groupUid = groupUid,
                error = ChatGroupErrorType.KICKED,
            )
        }
        updateGroupPeer(groupUid, targetUid) { it.copy(protocol = ConnectionProtocol.NONE) }
    }

    fun onGroupPeerRoleChanged(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid, role: ChatGroupRole) {
        logger.debug("onGroupPeerRoleChanged($groupUid, $sourceUid, $targetUid, $role)")
        updateGroupPeer(groupUid, targetUid) { it.copy(role = role) }
    }
    
    fun sendFriendRequest(name: String, address: String, message: String): UserData? {
        logger.debug("sendFriendRequest($name, $address, $message)")
        val friendUid = engine.sendFriendRequest(address, message) ?: return null
        val newFriend = UserData(
            uid = friendUid,
            name = name,
        )
        friends[friendUid] = newFriend
        return newFriend
    }
    
    fun acceptFriendRequest(friendRequest: FriendRequest): UserData? {
        logger.debug("acceptFriendRequest($friendRequest)")
        friendRequests.remove(friendRequest.uid)
        val friendUid = engine.acceptFriendRequest(friendRequest.uid) ?: return null
        val newFriend = UserData(
            uid = friendUid,
            name = "(Pending Friend)",
        )
        friends[friendUid] = newFriend
        return newFriend
    }
    
    fun declineFriendRequest(friendRequest: FriendRequest) {
        friendRequests.remove(friendRequest.uid)
    }
    
    fun removeFriend(friendUid: UserUid) {
        logger.debug("removeFriend($friendUid)")
        engine.removeFriend(friendUid)
        friends.remove(friendUid)
    }
    
    fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData? {
        logger.debug("createGroup($title, $type)")
        return engine.createGroup(title, type, visibility)?.also { group -> groups[group.uid] = group }
    }
    
    fun setGroupTitle(groupUid: GroupUid, title: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupTitle($groupUid, $title)")
        if (title != group.title && engine.setGroupTitle(groupUid, title)) {
            groups[groupUid] = group.copy(title = title)
        }
    }
    
    fun setGroupDescription(groupUid: GroupUid, description: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupDescription($groupUid, $description)")
        if (description != group.description && engine.setGroupDescription(groupUid, description)) {
            groups[groupUid] = group.copy(description = description)
        }
    }
    
    fun setGroupPassword(groupUid: GroupUid, password: String?) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupPassword($groupUid, *****)")
        if (password != group.password && engine.setGroupPassword(groupUid, password)) {
            groups[groupUid] = group.copy(password = password)
        }
    }
    
    fun setGroupNickname(groupUid: GroupUid, name: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupNickname($groupUid, $name)")
        val self = group.findSelf(self)
        if (name != self.name && engine.setGroupNickname(groupUid, name)) {
            updateGroupPeer(groupUid, self.uid) { it.copy(name = name) }
        }
    }
    
    fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        if (peer.role == role) return
        logger.debug("setGroupRole($groupUid, $peerUid, $role)")
        val self = group.findSelf(self)
        if (self.hasAuthorityOver(peer) && engine.setGroupRole(groupUid, peerUid, role)) {
            updateGroupPeer(groupUid, peerUid) { it.copy(role = role) }
        } else {
            eventHandler += ErrorDialogEvent(
                message = getStringBlocking(Res.string.group_event_peer_role_change_failed),
            )
        }
    }
    
    fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        logger.debug("kickGroupPeer($groupUid, $peerUid)")
        val self = group.findSelf(self)
        if (self.hasAuthorityOver(peer) && engine.kickGroupPeer(groupUid, peerUid)) {
            onGroupPeerExit(groupUid, peerUid)
        } else {
            eventHandler += ErrorDialogEvent(
                message = getStringBlocking(Res.string.group_event_peer_kick_failed),
            )
        }
    }

    fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        logger.debug("setGroupPeerBlocked($groupUid, $peerUid, $block)")
        if (group.isPeerBlocked(peerUid) != block && engine.setGroupPeerBlocked(groupUid, peerUid, block)) {
            updateGroup(groupUid) { it.setPeerBlock(peer.uid, block) }
        }
    }
    
    fun joinGroup(groupUid: GroupUid, selfNickname: String?, password: String?) {
        engine.joinGroup(
            groupUid = groupUid,
            selfNickname = selfNickname ?: self.name,
            password = password,
        )
    }
    
    fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid) {
        if (!engine.sendGroupInvite(friendUid, groupUid)) {
            logger.error("Failed to send group invite for $groupUid to $friendUid")
        }
    }
    
    fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData? {
        removeGroupInvite(invite)
        val newGroup = engine.acceptGroupInvite(invite) ?: return null
        groups[newGroup.uid] = newGroup
        return newGroup
    }

    fun removeGroupInvite(invite: ChatGroupInvite) {
        groupInvites.remove(invite.cookie)
    }
    
    fun removeGroup(groupUid: GroupUid, partingMessage: String?) {
        logger.debug("removeGroup($groupUid)")
        engine.removeGroup(groupUid, partingMessage)
        groups.remove(groupUid)
    }

    fun setContactMuted(uid: ContactUid, isMuted: Boolean) {
        logger.debug("setContactMuted($uid, $isMuted)")
        when (uid) {
            is UserUid -> if (engine.setContactMuted(uid, isMuted)) {
                updateFriend(uid) { it.copy(isMuted = isMuted) }
            }
            is GroupUid -> if (engine.setContactMuted(uid, isMuted)) {
                updateGroup(uid) { it.copy(isMuted = isMuted) }
            }
            else -> {} // Not applicable
        }
    }
    
    fun getSelfAddress() = engine.getSelfAddress()
    
    fun resetSelfAddress() = engine.resetSelfAddress()
    
    private fun updateFriend(uid: UserUid, block: (UserData) -> UserData): Boolean {
        val oldData = friends[uid]
        val newData = block(friends[uid] ?: UserData(uid))
        friends[uid] = newData
        return oldData != newData
    }

    private fun updateGroup(uid: GroupUid, block: (ChatGroupData) -> ChatGroupData): Boolean {
        val oldData = groups[uid]
        val newData = block(groups[uid] ?: ChatGroupData(
            uid = uid,
            selfUid = self.uid,
        ).also { logger.warn("Failed to find existing group $uid") })
        groups[uid] = newData
        return oldData != newData
    }
    
    private fun updateGroupPeer(
        groupUid: GroupUid,
        peerUid: UserUid,
        block: (UserData) -> UserData,
    ) = updateGroup(groupUid) {
        val peer = it.peers[peerUid] ?: run {
            logger.warn("Failed to find existing peer $peerUid in $groupUid")
            return@updateGroup it.copy(peers = engine.getGroupPeerList(groupUid))
        }
        val peers = it.peers.toMutableMap()
        peers[peerUid] = block(peer)
        it.copy(peers = peers)
    }
    
    private fun initGroupPeer(group: ChatGroupData, peer: UserData, isSelf: Boolean): UserData {        
        return if (isSelf) {
            val avatarPath = avatarRepo.getAvatarPath(self.uid)
            avatarPath?.let {
                avatarRepo.onGroupPeerAvatarReceived(
                    group = group,
                    peer = peer,
                    hash = it.name,
                )
            }
            peer.copy(
                avatarPath = avatarPath,
                clientVersion = ChatClientVersion(),
            )
        } else {
            peer.copy(avatarPath = avatarRepo.getAvatarPath(peer.uid))
        }
    }
}