package nand.apps.chat.repo

import androidx.compose.runtime.*
import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.notification.*
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.util.collect
import org.jetbrains.compose.resources.getString
import seers.composeapp.generated.resources.*

class NotificationRepo(
    private val chatEngine: ChatEngine,
    private val contactRepo: ContactRepo,
    private val messageRepo: ChatMessageRepo,
    private val callRepo: CallRepo,
    private val settingsRepo: SettingsRepo,
) {

    val notificationObserver = snapshotFlow { notifications.toMap() }
    
    private val notifications = mutableStateMapOf<ChatNotificationGroup, ChatNotification>()
    private val settings get() = settingsRepo.settings.notificationSettings
    
    fun start(coroutineScope: CoroutineScope) {
        coroutineScope.collect(messageRepo.onNewMessage) { message ->
            if (!message.isGroupConnectionEvent &&
                !messageRepo.isConversationVisible(message.conversationUid) &&
                message.getConversationContact()?.isMuted != true) {
                val hasSelfMention = message.containsSelfMention()
                if (settings.isMessageNotificationsEnabled && hasSelfMention) {
                    showMessageNotification(message)
                }
            }
        }
        coroutineScope.collect(callRepo.onCallReceived) { call ->
            if (settings.isCallNotificationsEnabled &&
                (getPlatform().isMobile || !messageRepo.isConversationVisible(call.contactUid))) {
                showCallNotification(call)
            }
        }
        coroutineScope.collect(messageRepo.unreadObserver) { unreadStates ->
            val latestMessageNotification = notifications[ChatNotificationGroup.MESSAGES] as? MessageNotification ?: return@collect
            val unreadCount = unreadStates[latestMessageNotification.message.conversationUid] ?: 0
            if (unreadCount == 0) dismiss(ChatNotificationGroup.MESSAGES)
        }
    }
    
    fun dismiss(group: ChatNotificationGroup) {
        notifications.remove(group)
    }
    
    fun dismissAll() {
        notifications.clear()
    }
    
    private fun showNotification(notification: ChatNotification) {
        notifications[notification.group] = notification
    }
    
    private suspend fun showMessageNotification(message: ChatMessage) {
        val conversation = message.getConversationContact()
        val sender = message.getSenderContact().takeIf { it != contactRepo.self } ?: return
        showNotification(
            MessageNotification(
                title = when (conversation) {
                    is UserData -> getString(Res.string.notification_new_message_from_user, conversation.name)
                    is ChatGroupData -> getString(Res.string.notification_new_message_in_group, conversation.title)
                    is FriendRequest -> getString(Res.string.notification_new_friend_request)
                    else -> return
                },
                sender = sender.getName() ?: return,
                avatar = sender.getAvatar() ?: return,
                message = message,
            )
        )
    }
    
    private suspend fun showCallNotification(call: CallData) {
        val sender = contactRepo.getContact(call.contactUid) ?: return
        val senderName = sender.getName() ?: return
        showNotification(
            CallNotification(
                sender = senderName,
                avatar = sender.getAvatar() ?: return,
                contactUid = call.contactUid,
                callUid = call.uid,
            )
        )
    }
    
    private suspend fun ChatContact.getName() = when (this) {
        is UserData -> displayName
        is ChatGroupData -> title
        is FriendRequest -> getString(Res.string.friend_request)
        else -> null
    }
    
    private fun ChatContact.getAvatar() = when (this) {
        is UserData -> avatar
        is ChatGroupData -> avatar
        is FriendRequest -> avatar
        else -> null
    }

    private fun ChatMessage.containsSelfMention(): Boolean {
        return when (conversationUid) {
            is GroupUid -> {
                val group = getGroupContact() ?: return false
                return containsSelfMention(
                    group = group,
                    mentionsRegex = chatEngine.constraints.mentionRegex,
                )
            }
            else -> true
        }
    }

    private fun ChatMessage.getConversationContact() = contactRepo.getContact(conversationUid)

    private fun ChatMessage.getGroupContact() = when {
        conversationUid is GroupUid -> contactRepo.getGroup(conversationUid)
        else -> null
    }

    private fun ChatMessage.getSenderContact() = contactRepo.getContact(
        uid = senderUid,
        userGroup = getGroupContact(),
    )
}