package nand.apps.chat.repo

import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.setValue
import androidx.compose.runtime.snapshotFlow
import kotlinx.coroutines.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.io.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.settings.profile.ProfileSettingsData
import nand.apps.chat.model.settings.profile.ProfileSettingsStateData
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatClientVersion
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import okio.Path
import okio.Path.Companion.toPath
import org.koin.core.Koin

private const val SETTINGS_FILE = "settings.json"

class SelfProfileRepo(
    private val engine: ChatEngine,
    private val backgroundWorker: ChatBackgroundWorker,
    private val fileSystem: ChatFileSystem,
    koin: Koin,
) : ChatEngineCallback {

    private val contactRepo: ContactRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val encryption: ChatEngineEncryption by koin.inject()

    private val fileMutex = Mutex()
    private val json = Json { ignoreUnknownKeys = true }
    private lateinit var profileDirectory: ChatDirectory
    private lateinit var coroutineScope: CoroutineScope

    private val logger = getLogger("SelfProfileRepo")

    var self by mutableStateOf(UserData.BLANK)
        private set
    val selfObserver = snapshotFlow { self }

    var settings by mutableStateOf(ProfileSettingsData())
        private set

    lateinit var profilePath: Path
        private set
    val profileName get() = profilePath.run { parent?.name ?: self.name.toSanitizedFileName() }

    override suspend fun onEngineInit(profile: ChatProfileData) {
        profilePath = profile.filePath.toPath()
        profileDirectory = fileSystem.getProfileDirectory(profile)
        settings = loadSettingsData() ?: getDefaultSettings()
        updateEngineState()
        self = engine.getSelfUserData().copy(clientVersion = ChatClientVersion())
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
        // Nothing to do here
    }

    fun setCurrentAvatarPath(path: Path?) {
        self = self.copy(avatarPath = path)
    }

    fun setAutoAway(isAutoAway: Boolean, actualStatus: UserStatus) {
        updateSettings(
            settings = settings.copy(
                status = if (isAutoAway) UserStatus.AWAY else actualStatus,
                autoAwayStatus = actualStatus,
            )
        )
    }

    fun setSelfStatus(status: UserStatus) {
        updateSettings(settings = settings.copy(status = status))
    }

    fun updateSettings(settings: ProfileSettingsData) {
        val currentState = settings.currentState ?: return
        logger.debug("Updated profile settings")
        this.settings = settings
        self = self.copy(
            name = currentState.name,
            status = settings.status,
            statusMessage = currentState.message,
        )
        updateEngineState()
        avatarRepo.onSelfStatusChanged(self)
        coroutineScope.launch(backgroundWorker.dispatcher) { saveSettingsData() }
    }

    fun onSelfConnectionProtocol(protocol: ConnectionProtocol) {
        logger.debug("onSelfConnectionProtocol($protocol)")
        self = self.copy(protocol = protocol)
    }

    fun getSelfUid(conversationUid: ContactUid? = null): UserUid {
        return when (conversationUid) {
            is GroupUid -> contactRepo.getGroup(conversationUid)?.selfUid ?: self.uid
            else -> self.uid
        }
    }

    private fun updateEngineState() {
        val status = settings.status
        val state = settings.currentState ?: return
        val self = engine.getSelfUserData()
        if (self.name != state.name || self.statusMessage != state.message) {
            engine.setSelfUserData(
                name = state.name,
                statusMessage = state.message,
            )
        }
        if (self.status != status) {
            engine.setSelfStatus(status)
        }
    }

    private suspend fun saveSettingsData() {
        logger.verbose("Saving settings to $SETTINGS_FILE")
        fileMutex.withLock {
            profileDirectory.writeToFile(
                fileName = SETTINGS_FILE,
                data = json.encodeToString(settings).encodeToByteArray(),
                encryption = encryption,
            )
        }
    }

    private suspend fun loadSettingsData(): ProfileSettingsData? {
        if (SETTINGS_FILE !in profileDirectory) return null
        logger.verbose("Loading from $SETTINGS_FILE")
        return try {
            withContext<ProfileSettingsData?>(Dispatchers.IO) {
                fileMutex.withLock {
                    profileDirectory.readFromFile(
                        fileName = SETTINGS_FILE,
                        encryption = encryption,
                    )?.let { json.decodeFromString(it.decodeToString()) }
                }
            }?.let { settings ->
                logger.verbose("Successfully loaded from $SETTINGS_FILE")
                val autoAwayStatus = settings.autoAwayStatus
                if (autoAwayStatus != null && settings.status == UserStatus.AWAY) {
                    settings.copy(
                        status = autoAwayStatus,
                        autoAwayStatus = null,
                    )
                } else {
                    settings
                }
            }
        } catch (e: Exception) {
            logger.error("Failed to parse profile settings", e)
            null
        }
    }

    private fun getDefaultSettings(): ProfileSettingsData {
        val self = engine.getSelfUserData()
        return ProfileSettingsData(
            status = self.status,
            states = mapOf(
                UserStatus.DEFAULT to ProfileSettingsStateData(
                    name = self.name,
                    message = self.statusMessage,
                ),
            )
        )
    }
}