package nand.apps.chat.repo

import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.setValue
import androidx.compose.runtime.snapshotFlow
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.settings.*
import nand.apps.chat.model.settings.filetransfer.FileTransferSettingsData
import nand.apps.chat.model.settings.net.NetworkSettingsData
import nand.apps.chat.model.settings.notifications.NotificationSettingsData

private const val SETTINGS_FILE = "settings.json"

class SettingsRepo(fileSystem: ChatFileSystem) {

    var settings by mutableStateOf(SettingsData())
    var isLoaded = false
        private set

    val observer = snapshotFlow { settings }
    
    private val json = Json { ignoreUnknownKeys = true }
    private val configDir = fileSystem.getConfigDirectory()
    private val logger = getLogger("SettingsRepo")
    
    suspend fun save() {
        logger.debug("Saving to $SETTINGS_FILE")
        withContext(Dispatchers.IO) {
            configDir.writeToFile(
                fileName = SETTINGS_FILE,
                data = json.encodeToString(settings).encodeToByteArray(),
            )
        }
    }
    
    suspend fun load() {
        if (SETTINGS_FILE in configDir) {
            logger.verbose("Loading from $SETTINGS_FILE")
            val settings: SettingsData? = try {
                withContext(Dispatchers.IO) {
                    configDir.readFromFile(SETTINGS_FILE)?.let {
                        json.decodeFromString(it.decodeToString())
                    }
                }
            } catch (e: Exception) {
                logger.error("Failed to parse settings file", e)
                null
            }
            settings?.let {
                this.settings = it
                logger.verbose("Successfully loaded from $SETTINGS_FILE")
            }
        }
        isLoaded = true
    }
    
    fun update(block: SettingsData.() -> Any) {
        settings = settings.run {
            when (val data = block()) {
                is SettingsData -> data
                is GeneralSettingsData -> copy(generalSettings = data)
                is ChatSettingsData -> copy(chatSettings = data)
                is NotificationSettingsData -> copy(notificationSettings = data)
                is EmbedSettingsData -> copy(embedSettings = data)
                is FileTransferSettingsData -> copy(fileSettings = data)
                is NetworkSettingsData -> copy(networkSettings = data)
                is DevSettingsData -> copy(devSettings = data)
                else -> return
            }
        }
    }
}