package nand.apps.chat.repo

import kotlinx.coroutines.*
import nand.apps.chat.audio.AudioFrame
import nand.apps.chat.audio.readToFrame
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.io.ChatDirectory
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.sanitizeFileName
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.call.CallStatus
import nand.apps.chat.model.call.VoicemailStage
import nand.apps.chat.model.file.ChatFileTransfer
import nand.apps.chat.model.file.ChatFileTransferDirection
import nand.apps.chat.model.file.ChatFileTransferState
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.getStringBlocking
import nand.apps.chat.util.sha256
import nand.apps.opus.OpusFileReader
import nand.apps.opus.OpusFileWriter
import okio.Path
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.ExperimentalResourceApi
import org.koin.core.Koin
import seers.composeapp.generated.resources.Res
import seers.composeapp.generated.resources.new_voicemail_message
import kotlin.math.roundToLong

private const val GREETING_MESSAGE_FILE_NAME = "greeting.opus"
private const val AUDIO_FRAME_BUFFER_SIZE = 960 * 2
private const val GREETING_TO_TONE_DELAY = 1000L
private const val VOICEMAIL_LENGTH_CHECK_DELAY = 100L

class VoicemailRepo(
    private val fileSystem: ChatFileSystem,
    private val dateTimeFormatter: DateTimeFormatter,
    koin: Koin,
) : ChatEngineCallback {
    
    private val engine: ChatEngine by koin.inject()
    private val encryption: ChatEngineEncryption by koin.inject()
    private val callRepo: CallRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val settings get() = settingsRepo.settings.callSettings
    private val recordings = mutableMapOf<ContactUid, OpusFileWriter?>()
    
    private lateinit var voicemailDirectory: ChatDirectory
    private var coroutineScope: CoroutineScope? = null
    
    private val logger = getLogger("VoicemailRepo")
    
    fun sendVoicemail(contactUid: ContactUid) {
        coroutineScope?.launch { voicemailLoop(contactUid) }
    }
    
    fun removeMessages(contactUid: ContactUid) {
        val directory = voicemailDirectory.getSubDirectory(contactUid.value)
        if (directory.exists()) {
            coroutineScope?.launch(Dispatchers.IO) { directory.deleteRecursively() }
        }
    }

    override suspend fun onEngineInit(profile: ChatProfileData) {
        voicemailDirectory = fileSystem.getProfileDirectory(profile, "voicemail")
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
        coroutineScope = null
    }

    fun onAudioFrame(contactUid: ContactUid, frame: AudioFrame) {
        if (recordings.containsKey(contactUid)) {
            val fileWriter = recordings[contactUid] ?: run {
                val path = createMessagePath(contactUid)
                fileSystem.makeDirectoriesForFile(path.toString())
                OpusFileWriter(
                    path = path.toString(),
                    sampleRate = frame.sampleRate,
                    channels = frame.channels,
                    bitRate = settings.voicemailAudioBitRate,
                )
            }.also { recordings[contactUid] = it }
            if (fileWriter.channels != frame.channels || fileWriter.sampleRate != frame.sampleRate) {
                logger.error("Received unexpected frame while recording voicemail")
                stopRecording(contactUid)
                callRepo.sendCallControl(contactUid, CallControl.CANCEL)
                return
            }
            recordings[contactUid]?.write(frame.samples)
        }
    }
    
    fun getGreetingAudioPath() = voicemailDirectory.getFilePath(GREETING_MESSAGE_FILE_NAME)

    private suspend fun loadGreetingAudio(): OpusFileReader {
        val greetingData = if (GREETING_MESSAGE_FILE_NAME in voicemailDirectory) {
            voicemailDirectory.readFromFile(
                fileName = GREETING_MESSAGE_FILE_NAME,
                encryption = encryption,
            )
        } else {
            null
        } ?: readSoundEffect("default_greeting")

        return OpusFileReader(greetingData)
    }
    
    private suspend fun voicemailLoop(contactUid: ContactUid) {
        val greeting = loadGreetingAudio()
        val buffer = ShortArray(AUDIO_FRAME_BUFFER_SIZE)
        var stage = VoicemailStage.GREETING
        var recordStartTime = 0L
        while (isVoicemailActive(contactUid)) {
            when (stage) {
                VoicemailStage.GREETING -> {
                    greeting.send(contactUid, buffer)
                    delay(GREETING_TO_TONE_DELAY)
                    stage = VoicemailStage.TONE
                }
                VoicemailStage.TONE -> {
                    sendTone(contactUid, buffer)
                    stage = VoicemailStage.RECORDING
                    recordStartTime = currentTimeMillis()
                    recordings[contactUid] = null
                }
                VoicemailStage.RECORDING -> {
                    val elapsed = currentTimeMillis() - recordStartTime
                    if (elapsed >= settings.voicemailMaxDuration) {
                        stopRecording(contactUid)
                        sendTone(contactUid, buffer)
                        callRepo.sendCallControl(contactUid, CallControl.CANCEL)
                        break
                    } else {
                        delay(VOICEMAIL_LENGTH_CHECK_DELAY)
                    }
                }
            }
        }
        stopRecording(contactUid)
    }

    private fun isVoicemailActive(contactUid: ContactUid) =
        callRepo.getCall(contactUid)?.status == CallStatus.VOICEMAIL
    
    private fun stopRecording(contactUid: ContactUid) {
        recordings.remove(contactUid)?.let { file ->
            file.close()
            val filePath = file.path.toPath()
            val fileName = filePath.name
            val dir = fileSystem.getDirectoryForFile(file.path) ?: return
            if (fileName in dir && encryption.isEncryptionSupported()) {
                // Encryption time
                coroutineScope?.launch(Dispatchers.IO) {
                    dir.encryptFile(
                        fileName = fileName,
                        encryption = encryption,
                    )
                    withContext(Dispatchers.Main) {
                        addVoicemailChatMessage(contactUid, filePath)
                    }
                }
            } else {
                addVoicemailChatMessage(contactUid, filePath)
            }
        }
    }
    
    private fun createMessagePath(contactUid: ContactUid): Path {
        val fileName = dateTimeFormatter.formatDateTime(currentTimeMillis())
            .replace(' ', '_')
            .replace(':', '-')
            .sanitizeFileName()
        return voicemailDirectory.getFilePath("${contactUid.value}/$fileName.opus")
    }
    
    private fun addVoicemailChatMessage(contactUid: ContactUid, messagePath: Path) {
        val friendUid = contactUid as? UserUid ?: return // Voicemail is only supported in DMs
        val fileTransfer = ChatFileTransfer(
            fileUid = FileUid((contactUid.value + messagePath.name).sha256()),
            friendUid = friendUid,
            name = messagePath.name,
            path = messagePath.toString(),
            size = fileSystem.getSize(messagePath.toString()),
            direction = ChatFileTransferDirection.INCOMING,
            state = ChatFileTransferState.COMPLETE,
        )
        fileTransferRepo.addHistoricalTransfer(fileTransfer)        
        messageRepo.addMessage(
            ChatMessage(
                conversationUid = contactUid,
                senderUid = contactUid,
                text = getStringBlocking(Res.string.new_voicemail_message),
                type = ChatMessageType.ACTION,
                fileTransferUid = fileTransfer.uid,
            )
        )
    }
    
    private suspend fun OpusFileReader.send(contactUid: ContactUid, buffer: ShortArray) {
        val startTime = currentTimeMillis()
        var samplesPlayed = 0
        while (isVoicemailActive(contactUid)) {
            val frame = readToFrame(buffer) ?: return
            engine.sendAudioFrame(contactUid, frame)
            samplesPlayed += frame.size
            val curTime = currentTimeMillis()
            val elapsedDuration = ((samplesPlayed / frame.sampleRate.toDouble()) * 1000L).roundToLong()
            val targetClock = startTime + elapsedDuration
            val delayMs = targetClock - curTime
            delay(delayMs)
        }
    }
    
    private suspend fun sendTone(contactUid: ContactUid, buffer: ShortArray) =
        OpusFileReader(readSoundEffect("tone")).send(contactUid, buffer)

    @OptIn(ExperimentalResourceApi::class)
    private suspend fun readSoundEffect(name: String) = Res.readBytes("files/voicemail/$name.opus")
}