package nand.apps.chat.text

import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.input.OffsetMapping
import androidx.compose.ui.text.input.TransformedText
import nand.apps.chat.util.copyStyles

class TransformedTextBuilder(
    private val source: TransformedText,
) {
    private val stringBuilder = AnnotatedString.Builder()
    private val indices = mutableListOf<Pair<Int, Int>>()
    var indexOffset = 0
        private set
    private var lastSourceIndex = 0
    
    val length get() = stringBuilder.length
    
    fun appendSource(start: Int, end: Int) {
        stringBuilder.append(source.text.text, start, end)
        appendIndices(start, end)
    }
    
    fun appendReplacement(replacement: String, start: Int, end: Int) {
        stringBuilder.append(replacement)
        appendIndices(start, end)
        indexOffset += replacement.length - (end - start)
    }
    
    fun appendOther(text: String) {
        stringBuilder.append(text)
        for (i in text.indices) {
            indices += lastSourceIndex to i + indexOffset + lastSourceIndex
        }
        indexOffset += text.length
    }
    
    fun pushStyle(style: SpanStyle) {
        stringBuilder.pushStyle(style)
    }
    
    fun pushFontAnnotation(font: UnicodeFont) {
        stringBuilder.pushStringAnnotation(UNICODE_FONT_TAG, font.name)
    }
    
    fun pushMentionAnnotation(uid: String) {
        stringBuilder.pushStringAnnotation(MENTION_TAG, uid)
    }
    
    fun pop() = stringBuilder.pop()
    
    fun addStyle(
        style: SpanStyle,
        start: Int,
        end: Int,
    ) {
        stringBuilder.addStyle(
            style = style,
            start = start,
            end = end,
        )
    }
    
    fun appendTagPreview(
        previewStyle: SpanStyle,
        start: Int,
        end: Int,
    ) {
        pushStyle(previewStyle)
        pushFontAnnotation(UnicodeFont.DEFAULT)
        appendSource(start, end)
        pop()
        pop()
    }
    
    fun build(): TransformedText {
        for (i in lastSourceIndex..source.text.length) {
            indices += i to length
        }
        val indexMap = indices.toMap()
        val offsetMapping = IndexOffsetMapping(
            indices = indices,
            indexMap = indexMap,
            previousMapping = source.offsetMapping,
        )
        stringBuilder.copyStyles(
            src = source.text,
            indexMap = indexMap,
        )
        return TransformedText(
            text = stringBuilder.toAnnotatedString(),
            offsetMapping = offsetMapping,
        )
    }
    
    private fun appendIndices(start: Int, end: Int) {
        if (start < lastSourceIndex) error("Must add source text linearly")
        for (i in lastSourceIndex..<start) {
            indices += i to lastSourceIndex + indexOffset
        }
        indexOffset -= start - lastSourceIndex
        for (i in start..<end) {
            indices += i to i + indexOffset
        }
        lastSourceIndex = end
    }
}

fun String.toTransformedText(
    offsetMapping: OffsetMapping = OffsetMapping.Identity,
) = TransformedText(
    text = AnnotatedString(this),
    offsetMapping = offsetMapping,
)

fun buildTransformedText(
    source: TransformedText,
    block: TransformedTextBuilder.() -> Unit,
) = TransformedTextBuilder(source).apply(block).build()

fun buildTransformedText(
    source: String,
    block: TransformedTextBuilder.() -> Unit,
) = buildTransformedText(
    source = TransformedText(
        text = AnnotatedString(source),
        offsetMapping = OffsetMapping.Identity,
    ),
    block = block,
)

fun TransformedText.buildUpon(
    block: TransformedTextBuilder.() -> Unit,
) = buildTransformedText(
    source = this,
    block = block,
)

fun TransformedText.copy(
    text: AnnotatedString = this.text,
    offsetMapping: OffsetMapping = this.offsetMapping,
) = TransformedText(
    text = text,
    offsetMapping = offsetMapping,
)