package nand.apps.chat.text.emoji

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.coroutines.withContext
import nand.apps.chat.util.removeIf
import org.jetbrains.compose.resources.ExperimentalResourceApi
import seers.composeapp.generated.resources.*

private const val GROUP_PREFIX = "# group: "

@OptIn(ExperimentalResourceApi::class)
class EmojiRepo {
    
    val groups = mutableMapOf<EmojiGroup, List<EmojiData>>()
    private val emojis = mutableMapOf<String, EmojiData>()
    private val animatedEmojiToCode = mutableMapOf<String, String>()
    
    private val groupMutex = Mutex()
    private val animatedMutex = Mutex()
    
    suspend fun getEmojiData(emoji: String): EmojiData? {
        groupMutex.withLock { if (emojis.isEmpty()) loadGroups() }
        return emojis[emoji]
    }
    
    suspend fun getEmojis(
        group: EmojiGroup,
        filter: String = "",
    ): List<EmojiData> {
        groupMutex.withLock { if (groups.isEmpty()) loadGroups() }
        return if (filter.isNotBlank()) {
            withContext(Dispatchers.IO) { filter(filter) }
        } else {
            groups[group].orEmpty()
        }
    }
    
    suspend fun getAnimatedEmojiSegments(
        text: CharSequence,
    ): List<Pair<String, String?>> {
        animatedMutex.withLock { if (animatedEmojiToCode.isEmpty()) loadAnimatedEmojis() }
        return text.splitByEmoji().map { segment -> segment to animatedEmojiToCode[segment] }
    }

    private suspend fun loadGroups() = withContext(Dispatchers.IO) {
        var currentGroup: EmojiGroup
        var emojis = mutableListOf<EmojiData>()
        Res.readBytes("files/emojis.txt").decodeToString().lineSequence().forEach { line ->
            when {
                line.startsWith(GROUP_PREFIX) || line == "#EOF" -> {
                    currentGroup = EmojiGroup.findByDescription(line.substringAfter(GROUP_PREFIX))
                    emojis = mutableListOf()
                    groups[currentGroup] = emojis
                    return@forEach
                }
                line.isBlank() || line.startsWith("#") -> return@forEach
            }

            val args = line.substringAfter("; ")
            val qualifier = args.substringBefore(' ')
            if (qualifier != "fully-qualified") return@forEach

            val comment = args.substringAfter("# ")
            val emoji = comment.substringBefore(' ')
            if (!emoji.isEmojiString()) return@forEach

            val description = comment.substringAfter(" E").substringAfter(' ').lowercase()

            val emojiData = EmojiData(
                value = emoji,
                description = description,
            )
            emojis += emojiData
            this@EmojiRepo.emojis[emoji] = emojiData
        }
        groups.removeIf { _, value -> value.isEmpty() }
    }
    
    private suspend fun loadAnimatedEmojis() = withContext(Dispatchers.IO) {
        Res.readBytes("files/animated_emojis.txt").decodeToString().lineSequence().forEach { line ->
            val (code, emoji) = line.split('\t')
            animatedEmojiToCode[emoji] = code
        }
    }
    
    private fun filter(filter: String): List<EmojiData> {
        val cleaned = filter.lowercase().trim()
        val results = mutableListOf<EmojiData>()
        groups.forEach { (_, emojis) ->
            results += emojis.filter { emoji ->
                cleaned.length == 1 && emoji.description.startsWith(cleaned) || cleaned.length > 1 && emoji.description.contains(cleaned)
            }
        }
        if (cleaned.length > 1) {
            results.sortByDescending {
                val percentage = cleaned.length.toFloat() / it.description.length
                percentage + if (it.description.startsWith(cleaned)) 1 else 0
            }
        }
        return results
    }
}