package nand.apps.chat.text.emoji

private const val NULL = '\u0000'
private const val EMOJI_SELECTOR = '\uFE0F'
private const val ZWJ = '\u200D'

private enum class EmojiModifier {
    NONE,
    TONE,
    REGION,
    SUB_MAJOR,
    SUB_MINOR,
}

fun CharSequence.isEmojiString(): Boolean {
    var hasEmoji = false
    var last = NULL
    for (c in this) {
        if (c.isWhitespace() || c == EMOJI_SELECTOR || c == ZWJ) {
            last = c
            continue
        }
        if (isEmojiChar(last, c)) {
            last = c
            hasEmoji = true
            continue
        }
        return false
    }
    return hasEmoji
}

// 🤮
fun CharSequence.splitByEmoji(): List<String> {
    val segment = StringBuilder()
    val segments = mutableListOf<String>()
    var last = NULL
    var lastIsEmoji = false
    var lastSegEmoji = false
    var lastMod = EmojiModifier.NONE
    var lastSegMod = EmojiModifier.NONE
    var lastSegLength = 0
    for (c in this) {
        val isEmoji = when {
            c == EMOJI_SELECTOR || c == ZWJ -> lastIsEmoji
            else -> isEmojiChar(last, c)
        }
        val modifier = getModifier(last, c)
        val canJoinLastEmoji = isEmoji && last == ZWJ || (c == EMOJI_SELECTOR || c == ZWJ) && lastIsEmoji
        if (!isEmoji && !lastIsEmoji || last == NULL || last.isHighSurrogate() || canJoinLastEmoji) {
            segment.append(c)
            val canApplyToneModifier = modifier == EmojiModifier.TONE && lastSegMod == EmojiModifier.NONE
            val canApplyRegionModifier = modifier == EmojiModifier.REGION && lastSegMod == EmojiModifier.REGION && lastSegLength == 2
            val canApplySubModifier = modifier == EmojiModifier.SUB_MINOR && (lastSegMod == EmojiModifier.SUB_MAJOR || lastSegMod == EmojiModifier.SUB_MINOR)
            if (lastSegEmoji && (canApplyToneModifier || canApplyRegionModifier || canApplySubModifier)) {
                segment.insert(0, segments.removeLast())
            }
        } else {
            lastSegEmoji = lastIsEmoji
            lastSegMod = lastMod
            lastSegLength = segment.length
            if (segment.isNotEmpty()) {
                segments += segment.toString()
                segment.clear()
            }
            segment.append(c)
        }
        lastIsEmoji = isEmoji
        lastMod = modifier
        last = c
    }
    if (segment.isNotEmpty()) segments += segment.toString()
    return segments
}

private fun isEmojiHighSurrogate(c: Char) = c == '\uDB40' || c in '\uD83C'..'\uD83E'

private fun isUnicodeFont(last: Char, c: Char) = last == '\uD83C' && c in '\uDD00'..'\uDD89'

private fun isEmojiChar(last: Char, c: Char) = c in '\u2190'..'\u21FF' ||
        c in '\u231A'..'\u231B' ||
        c in '\u23E9'..'\u23FA' ||
        c in '\u25A0'..'\u2769' ||
        c in '\u2794'..'\u27BF' ||
        c in '\u2934'..'\u2935' ||
        c in '\u2B05'..'\u2B07' ||
        c in '\u2B1B'..'\u2B1C' ||
        c == '\u2B50' || c == '\u2B55' ||
        isEmojiHighSurrogate(c) ||
        c in '\uDC00'..'\uDFFF' && !isUnicodeFont(last, c) && isEmojiHighSurrogate(last) ||
        c == '\uDB40' || c == '\u2328' || c == '\u23CF'

private fun getModifier(high: Char, low: Char) = when {
    high == '\uD83C' && low in '\uDDE6'..'\uDDFF' -> EmojiModifier.REGION
    high == '\uD83C' && low in '\uDFFB'..'\uDFFF' -> EmojiModifier.TONE
    high == '\uD83C' && low == '\uDFF4' -> EmojiModifier.SUB_MAJOR
    high == '\uDB40' && low in '\uDC62'..'\uDC7F' -> EmojiModifier.SUB_MINOR
    else -> EmojiModifier.NONE
}