package nand.apps.chat.ui.app

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.delay
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.image.initSingletonImageLoader
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.locale.setAppLocale
import nand.apps.chat.log.CHAT_LOGGER_LEVEL
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.monitor.AutoAwayMonitor
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.util.collect
import org.koin.core.Koin

private const val SAVE_INTERVAL = 1000L

class AppViewModel(
    private val autoAwayMonitor: AutoAwayMonitor,
    private val settingsRepo: SettingsRepo,
    private val eventHandler: ChatEventHandler,
    private val fileSystem: ChatFileSystem,
    koin: Koin,
) : ViewModel() {

    val stateData = MutableStateFlow(AppViewState())
    private var autoSaveSettings = SettingsData()
    private val tmpDir = fileSystem.getTempDirectory()
    
    init {
        // Coil initialization
        initSingletonImageLoader(koin)
        
        // Settings listener
        viewModelScope.launch {
            fileSystem.paths.init()
            settingsRepo.load()
            val settings = settingsRepo.settings
            CHAT_LOGGER_LEVEL = settings.devSettings.logLevel
            setAppLocale(settings.generalSettings.locale)
            tmpDir.deleteRecursively()
            stateData.emit(
                AppViewState(
                    isAppLoaded = true,
                    inTutorial = checkTutorialState(),
                    settingsData = settings,
                )
            )
            eventHandler += AppLoadedEvent
            monitorSettings()
        }
        
        // Auto-away monitor
        viewModelScope.launch {
            autoAwayMonitor.checkLoop()
        }
        
        // Event listener
        collect(eventHandler.onEvent, ::onEvent)
    }
    
    private suspend fun monitorSettings() {
        collect(settingsRepo.observer) { settings ->
            stateData.emit(
                stateData.value.copy(
                    inTutorial = !settings.generalSettings.isTutorialComplete,
                    settingsData = settings,
                )
            )
            CHAT_LOGGER_LEVEL = settings.devSettings.logLevel
        }
        autoSaveSettings = settingsRepo.settings
        while (true) {
            delay(SAVE_INTERVAL)
            val latest = settingsRepo.settings
            if (autoSaveSettings != latest) {
                autoSaveSettings = latest
                settingsRepo.save()
            }
        }
    }
    
    fun onUserActivity() = autoAwayMonitor.onUserActivity()
    
    private fun onEvent(event: ChatEvent) {
        when (event) {
            is LoginCompleteEvent -> {
                stateData.tryEmit(stateData.value.copy(isLoggedIn = true))
            }
            is LogoutEvent -> {
                stateData.tryEmit(stateData.value.copy(isLoggedIn = false))
            }
            else -> {}
        }
    }
    
    private fun checkTutorialState(): Boolean {
        return when {
            settingsRepo.settings.generalSettings.isTutorialComplete -> false
            fileSystem.getProfileDirectory().isEmpty() -> true
            else -> {
                settingsRepo.update { generalSettings.copy(isTutorialComplete = true) }
                false
            }
        }
    }
}