package nand.apps.chat.ui.avatar

import androidx.compose.foundation.background
import androidx.compose.foundation.border
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.Icon
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Groups
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.FilterQuality
import androidx.compose.ui.graphics.Shape
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.layout.ContentScale
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import coil3.compose.AsyncImage
import nand.apps.chat.model.avatar.ChatAvatar
import nand.apps.chat.model.avatar.ChatGroupAvatar
import nand.apps.chat.model.avatar.ChatUserAvatar
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.text.emoji.Emoji
import nand.apps.chat.ui.call.CallButton
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.trimToLength

@Composable
fun ChatAvatarBox(
    avatar: ChatAvatar,
    modifier: Modifier = Modifier,
    size: ChatAvatarSize = ChatAvatarSize.MEDIUM,
    shape: Shape = ChatTheme.shapes.small,
    status: UserStatus? = null,
    isOnline: Boolean = false,
    badgeText: String? = null,
    call: CallData? = null,
    isLoading: Boolean = false,
    iconOverlay: ImageVector? = null,
    iconOverlayColor: Color = Color.Unspecified,
) {
    val statusColor = status?.let {
        if (isOnline) {
            when (status) {
                UserStatus.DEFAULT -> ChatTheme.colors.statusOnline
                UserStatus.AWAY -> ChatTheme.colors.statusAway
                UserStatus.BUSY -> ChatTheme.colors.statusBusy
            }
        } else {
            ChatTheme.colors.statusOffline
        }
    }
    val sizeDp = size.getSizeDp()
    Box(
        modifier = modifier
            .size(sizeDp)
            .clip(shape)
            .background(avatar.background)
    ) {
        when (avatar) {
            is ChatUserAvatar -> {
                avatar.path?.let { path ->
                    AsyncImage(
                        model = path.toString(),
                        contentDescription = avatar.contentDescription,
                        filterQuality = FilterQuality.High,
                        contentScale = ContentScale.Crop,
                        modifier = Modifier.matchParentSize(),
                    )
                } ?: EmojiAvatar(
                    emoji = avatar.emoji,
                    size = size,
                )
            }
            is ChatGroupAvatar -> GroupAvatar(
                avatar = avatar,
                size = size,
            )
        }
        if (iconOverlay != null || call?.isInProgress == true) {
            Box(
                modifier = Modifier
                    .background(ChatTheme.colors.backgroundFloating)
                    .matchParentSize(),
            )
            when {
                iconOverlay != null -> Icon(
                    imageVector = iconOverlay,
                    contentDescription = null,
                    tint = iconOverlayColor,
                    modifier = Modifier.align(Alignment.Center),
                )
                call?.isInProgress == true -> CallButton(
                    contactUid = call.contactUid,
                    call = call,
                    isClickable = false,
                    modifier = Modifier.align(Alignment.Center),
                )
            }
        }
        statusColor?.let {
            Box(
                modifier = modifier
                    .size(sizeDp)
                    .clip(shape)
                    .border(
                        width = 2.dp,
                        color = statusColor,
                        shape = shape,
                    )
            )
        }
        badgeText?.let {
            TextLine(
                text = it,
                color = Color.White,
                style = when (size) {
                    ChatAvatarSize.XLARGE -> ChatTheme.typography.body2.copy(lineHeight = 14.sp)
                    else -> ChatTheme.typography.overline.copy(lineHeight = 12.sp)
                },
                modifier = Modifier
                    .clip(CircleShape)
                    .background(Color.Red)
                    .padding(start = ChatTheme.dimens.spacingXSmall)
                    .align(Alignment.BottomEnd),
            )
        }
        if (isLoading) {
            CircularProgressIndicator(
                modifier = Modifier
                    .matchParentSize()
                    .padding(ChatTheme.dimens.spacingSmall)
            )
        }
    }
}

@Composable
private fun BoxScope.EmojiAvatar(emoji: Emoji, size: ChatAvatarSize) {
    val fontSize = when (size) {
        ChatAvatarSize.SMALL -> 21.sp
        ChatAvatarSize.MEDIUM -> 28.sp
        ChatAvatarSize.LARGE -> 36.sp
        ChatAvatarSize.XLARGE -> 48.sp
    }
    Text(
        text = emoji.value,
        color = ChatTheme.colors.onSurface,
        textAlign = TextAlign.Center,
        fontSize = fontSize,
        modifier = Modifier
            .matchParentSize()
            .offset(y = ChatTheme.dimens.spacingSmall),
    )
}

@Composable
private fun BoxScope.GroupAvatar(avatar: ChatGroupAvatar, size: ChatAvatarSize) {
    Icon(
        imageVector = Icons.Default.Groups,
        contentDescription = avatar.contentDescription,
        tint = Color.White,
        modifier = Modifier
            .matchParentSize()
            .padding(ChatTheme.dimens.spacingSmall)
            .appendIf(size != ChatAvatarSize.SMALL) {
                padding(bottom = ChatTheme.dimens.spacingMedium)
            }
    )
    if (size != ChatAvatarSize.SMALL) {
        Text(
            text = avatar.groupTitle.trimToLength(3).trim(),
            style = when (size) {
                ChatAvatarSize.XLARGE -> ChatTheme.typography.body2
                else -> ChatTheme.typography.overline
            },
            color = Color.White,
            maxLines = 1,
            textAlign = TextAlign.Center,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .padding(bottom = ChatTheme.dimens.spacingSmall),
        )
    }
}