package nand.apps.chat.ui.call

import androidx.compose.animation.core.*
import androidx.compose.foundation.layout.size
import androidx.compose.material.Icon
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.rotate
import androidx.compose.ui.draw.scale
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.call.CallStatus
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.repo.CallRepo
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun CallButton(
    contactUid: ContactUid,
    call: CallData?,
    modifier: Modifier = Modifier,
    isClickable: Boolean = true,
    isEnabled: Boolean = true,
) {
    val callRepo: CallRepo by getKoin().inject()
    var isAnimationActive by remember { mutableStateOf(false) }
    val status = call?.status ?: CallStatus.NONE
    LaunchedEffect(status) {
        if (status == CallStatus.INCOMING) {
            callRepo.isCallRingingObserver.collect { isAnimationActive = it }
        }
    }
    val infiniteTransition = rememberInfiniteTransition(label = "Call icon")
    val icon = when (status) {
        CallStatus.ACTIVE -> Icons.Outlined.PhoneInTalk
        CallStatus.PAUSED -> Icons.Outlined.PhonePaused
        CallStatus.OUTGOING -> Icons.Outlined.SettingsPhone
        else -> Icons.Outlined.Call
    }
    val contentDescription = when (status) {
        CallStatus.INCOMING -> stringResource(Res.string.answer_call)
        CallStatus.OUTGOING,
        CallStatus.ACTIVE -> stringResource(Res.string.end_call)
        CallStatus.PAUSED -> stringResource(Res.string.resume_call)
        else -> stringResource(Res.string.start_call)
    }
    val color = when (status) {
        CallStatus.ACTIVE -> ChatTheme.colors.statusOnline
        else -> ChatTheme.colors.onSurface
    }
    val animationModifier = modifier
        .appendIf(isAnimationActive) {
            val rotation by infiniteTransition.animateFloat(
                initialValue = -10f,
                targetValue = 20f,
                animationSpec = infiniteRepeatable(tween(50, easing = LinearEasing), RepeatMode.Reverse),
                label = "Rotate call icon"
            )
            rotate(rotation)
        }
    if (isClickable) {
        ShortcutHandler(
            action = ShortcutAction.START_ANSWER_CALL,
            isEnabled = isEnabled && call?.isInProgress != true,
        ) {
            callRepo.startCall(
                contactUid = contactUid,
                isAudioEnabled = true,
                isVideoEnabled = false, // TODO - Support video calling
            )
        }
        IconButton(
            icon = when {
                isEnabled -> icon
                else -> Icons.Outlined.PhoneDisabled
            },
            contentDescription = when {
                isEnabled -> contentDescription
                else -> null
            },
            isEnabled = isEnabled,
            onClick = {
                when (status) {
                    CallStatus.INCOMING -> callRepo.answerCall(
                        contactUid = contactUid,
                        isAudioEnabled = true,
                        isVideoEnabled = false, // TODO - Support video calling
                    )
                    CallStatus.OUTGOING,
                    CallStatus.ACTIVE -> callRepo.sendCallControl(contactUid, CallControl.CANCEL)
                    CallStatus.PAUSED -> callRepo.sendCallControl(contactUid, CallControl.RESUME)
                    else -> callRepo.startCall(
                        contactUid = contactUid,
                        isAudioEnabled = true,
                        isVideoEnabled = false, // TODO - Support video calling
                    )
                }
            },
            color = when {
                isEnabled -> color
                else -> ChatTheme.colors.onBackground2
            },
            size = ChatTheme.dimens.iconMedium,
            modifier = animationModifier.appendIf(!isEnabled) { scale(-1f, 1f) },
        )
    } else {
        Icon(
            imageVector = icon,
            contentDescription = contentDescription,
            tint = color,
            modifier = Modifier
                .size(ChatTheme.dimens.iconMedium)
                .then(animationModifier),
        )
    }
}