package nand.apps.chat.ui.call

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import nand.apps.chat.model.avatar.ChatAvatar
import nand.apps.chat.model.call.*
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun CallHeader(
    call: CallData?,
    contactName: String,
    modifier: Modifier = Modifier,
) {
    val settings = LocalSettingsProvider.current.callSettings
    AnimatedVisibility(
        visible = call != null && settings.isEnabled && call.isInProgress,
        modifier = modifier,
    ) {
        val padding = when {
            isPortraitMode() -> ChatTheme.dimens.spacingLarge
            else -> ChatTheme.dimens.spacingMedium
        }
        CallHeaderContent(
            call = call,
            contactName = contactName,
            modifier = Modifier
                .fillMaxWidth()
                .padding(horizontal = padding)
                .padding(bottom = padding)
        )
    }
}

@Composable
fun CallHeaderContent(
    call: CallData?,
    contactName: String,
    modifier: Modifier = Modifier,
    avatar: ChatAvatar? = null,
) {
    Column(modifier = modifier) {
        call ?: return@Column
        TextLine(
            text = when (call.status) {
                CallStatus.INCOMING -> stringResource(Res.string.incoming_call_message, contactName)
                CallStatus.OUTGOING -> stringResource(Res.string.outgoing_call_message, contactName)
                CallStatus.ACTIVE -> stringResource(Res.string.active_call_message, contactName)
                CallStatus.PAUSED -> stringResource(Res.string.paused_call_message, contactName)
                CallStatus.VOICEMAIL -> stringResource(Res.string.voicemail_call_message, contactName)
                else -> "" // Won't happen
            },
            style = ChatTheme.typography.body1,
            modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingSmall),
        )
        Row(
            verticalAlignment = Alignment.CenterVertically,
        ) {
            avatar?.let {
                ChatAvatarBox(
                    avatar = avatar,
                    modifier = Modifier.padding(end = ChatTheme.dimens.spacingMedium),
                )
            }
            ActionButtons(call)
        }
    }
}

@Composable
private fun ActionButtons(call: CallData) {
    val actions = call.status.getActions()
    if (actions.isEmpty()) return
    val eventHandler = LocalEventHandler.current
    ShortcutHandler { shortcut ->
        actions.find { it.shortcut == shortcut }?.let { action ->
            eventHandler += action.getEvent(call.contactUid)
            true
        } ?: false
    }
    Row(
        horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
    ) {
        actions.forEachIndexed { index, action ->
            SimpleTextButton(
                text = stringResource(action.shortLabel),
                backgroundColor = when {
                    actions.size == 2 && index == 0 -> ChatTheme.colors.secondary
                    else -> ChatTheme.colors.error
                },
                onClick = { eventHandler += action.getEvent(call.contactUid) },
            )
        }
    }
}