package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.LinearProgressIndicator
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Cancel
import androidx.compose.material.icons.filled.OpenInBrowser
import androidx.compose.material.icons.filled.Pause
import androidx.compose.material.icons.filled.PlayArrow
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import io.ktor.http.*
import nand.apps.chat.io.extension
import nand.apps.chat.model.file.*
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.embed.audio.AudioEmbedRenderer
import nand.apps.chat.ui.embed.image.ImageEmbedRenderer
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.formatFileSize
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun FileTransferRow(
    fileTransfer: ChatFileTransfer,
    onFileTransferControl: (ChatFileControl) -> Unit,
    onOpenFile: () -> Unit,
    onDeleteFile: () -> Unit,
    modifier: Modifier = Modifier,
) = with(fileTransfer) {
    // TODO - Use embed framework for this
    if (state == ChatFileTransferState.COMPLETE) {
        val properPath = path.toPath()
        val embedSettings = LocalSettingsProvider.current.embedSettings
        when {
            embedSettings.isEnabled && embedSettings.isImagesEnabled && contentType?.contentType == "image" -> {
                ImageEmbedRenderer(
                    uri = path,
                    type = contentType.contentSubtype,
                    modifier = modifier,
                )
                return
            }
            properPath.extension == "opus" -> {
                AudioEmbedRenderer(
                    uri = path,
                    type = ContentType.Audio.OPUS,
                    onDeleteFile = onDeleteFile,
                    modifier = modifier,
                )
                return
            }
        }
    }
    FileTransferCard(
        fileTransfer = fileTransfer,
        onFileTransferControl = onFileTransferControl,
        onOpenFile = onOpenFile,
        modifier = modifier,
    )
}

@Composable
private fun FileTransferCard(
    fileTransfer: ChatFileTransfer,
    onFileTransferControl: (ChatFileControl) -> Unit,
    onOpenFile: () -> Unit,
    modifier: Modifier = Modifier,
) = with(fileTransfer) {
    val background = when (state) {
        ChatFileTransferState.CANCELLED -> ChatTheme.colors.error
        ChatFileTransferState.COMPLETE -> ChatTheme.colors.secondary
        else -> ChatTheme.colors.surface2
    }
    Column(
        modifier = modifier
            .fillMaxWidth()
            .clip(ChatTheme.shapes.medium)
            .background(background)
            .padding(ChatTheme.dimens.spacingMedium)
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.fillMaxWidth()
        ) {
            Column(
                modifier = Modifier.weight(1f),
            ) {
                TextLine(
                    text = "$name (${stringResource(fileTransfer.getStateString())})",
                    style = ChatTheme.typography.body2,
                )
                TextLine(
                    text = formatFileSize(size),
                    style = ChatTheme.typography.caption,
                )
            }
            if (isInProgress) {
                IconButton(
                    icon = when (state) {
                        ChatFileTransferState.ACTIVE -> Icons.Default.Pause
                        else -> Icons.Default.PlayArrow
                    },
                    contentDescription = when (state) {
                        ChatFileTransferState.ACTIVE -> stringResource(Res.string.pause_transfer)
                        else -> stringResource(Res.string.resume_transfer)
                    },
                    onClick = {
                        when (state) {
                            ChatFileTransferState.ACTIVE -> onFileTransferControl(ChatFileControl.PAUSE)
                            else -> onFileTransferControl(ChatFileControl.RESUME)
                        }
                    }
                )
                IconButton(
                    icon = Icons.Default.Cancel,
                    contentDescription = stringResource(Res.string.stop_transfer),
                    onClick = { onFileTransferControl(ChatFileControl.CANCEL) }
                )
            } else if (state == ChatFileTransferState.COMPLETE) {
                IconButton(
                    icon = Icons.Default.OpenInBrowser,
                    contentDescription = stringResource(Res.string.open_file),
                    onClick = onOpenFile,
                )
            }
        }
        if (isInProgress) {
            LinearProgressIndicator(
                progress = position.toFloat() / size,
                modifier = Modifier
                    .fillMaxWidth()
                    .padding(top = ChatTheme.dimens.spacingMedium),
            )
        }
    }
}

private fun ChatFileTransfer.getStateString() = when (state) {
    ChatFileTransferState.IDLE -> Res.string.file_state_idle
    ChatFileTransferState.ACTIVE -> {
        if (direction == ChatFileTransferDirection.INCOMING) {
            Res.string.file_state_receiving
        } else {
            Res.string.file_state_sending
        }
    }
    ChatFileTransferState.CANCELLED -> Res.string.file_state_cancelled
    ChatFileTransferState.COMPLETE -> Res.string.file_state_completed
}