package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.input.key.Key
import androidx.compose.ui.input.key.KeyEvent
import androidx.compose.ui.input.key.key
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.unit.DpSize
import androidx.compose.ui.unit.IntOffset
import androidx.compose.ui.window.Popup
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme

@Composable
fun MentionPopup(
    startIndex: Int,
    message: TextFieldValue,
    offset: IntOffset,
    size: DpSize,
    group: ChatGroupData,
    onUpdateMessage: (TextFieldValue) -> Unit,
    onSubmitMessage: () -> Unit,
    keyEvent: KeyEvent? = null,
) {
    var selectedIndex by rememberSaveable { mutableStateOf(0) }
    val search = remember(message.text) {
        when {
            startIndex < message.selection.min -> message.text.substring(startIndex, message.selection.min).lowercase()
            else -> ""
        }
    }
    val contacts = remember(group, search) {
        val peers = group.peers.values.run {
            if (search.length <= 1) sortedBy { it.name.lowercase() } else this
        }
        val results = (peers + group).filter {
            when {
                it.uid == group.selfUid -> false
                search.isNotBlank() -> {
                    val name = it.name.lowercase()
                    when {
                        search.length == 1 -> name.startsWith(search[0])
                        else -> name.contains(search)
                    }
                }
                else -> true
            }
        }.also { selectedIndex = selectedIndex.coerceAtMost(it.lastIndex) }
        if (search.length > 1) {
            results.sortedByDescending { contact ->
                val name = contact.name.lowercase()
                val percentage = search.length.toFloat() / name.length
                percentage + if (name.lowercase().startsWith(search)) 1 else 0
            }
        } else {
            results
        }
    }
    val selectContact: (ChatContact) -> Unit = { contact ->
        val startText = message.text.substring(0, startIndex)
        val endText = message.text.substring(message.selection.max)
        val newMessage = "$startText${contact.uid.value} $endText"
        onUpdateMessage(
            message.copy(
                text = newMessage,
                selection = TextRange(startIndex + contact.uid.value.length + 1),
            )
        )
    }
    LaunchedEffect(keyEvent) {
        when (keyEvent?.key) {
            Key.Enter, Key.NumPadEnter -> {
                if (selectedIndex in contacts.indices) {
                    selectContact(contacts[selectedIndex])
                } else {
                    onSubmitMessage()
                }
            }
            Key.DirectionUp -> {
                selectedIndex = (selectedIndex - 1).coerceAtLeast(0)
            }
            Key.DirectionDown -> {
                selectedIndex = (selectedIndex + 1).coerceAtMost(contacts.lastIndex)
            }
        }
    }
    
    Popup(
        alignment = if (isPortraitMode()) Alignment.BottomCenter else Alignment.BottomStart,
        offset = offset,
    ) {
        Box(
            modifier = Modifier
                .width(size.width)
                .heightIn(max = size.height)
                .clip(ChatTheme.shapes.small)
                .background(ChatTheme.colors.background),
        ) {
            LazyColumn(
                modifier = Modifier
                    .background(ChatTheme.colors.surface)
                    .fillMaxWidth(),
            ) {
                items(
                    count = contacts.size,
                    key = { contacts[it].uid.value },
                ) { index ->
                    val contact = contacts[index]
                    TextLine(
                        text = contact.name,
                        style = ChatTheme.typography.body1,
                        modifier = Modifier
                            .background(
                                if (selectedIndex == index) ChatTheme.colors.surface2 else Color.Unspecified
                            )
                            .fillMaxWidth()
                            .onClick { selectContact(contact) }
                            .padding(ChatTheme.dimens.spacingMedium),
                    )
                }
            }
        }
    }
}

private val ChatContact.name get() = when (this) {
    is UserData -> displayName
    is ChatGroupData -> title
    else -> ""
}