package nand.apps.chat.ui.chat.message

import androidx.compose.animation.core.*
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.draw.blur
import androidx.compose.ui.draw.scale
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.unit.dp
import nand.apps.chat.embed.file.ImageEmbedData
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.selection.DisableSelection
import nand.apps.chat.ui.text.ClickableMessageText
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.isUrl
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun MessageBlock(
    message: ChatMessage,
    displayName: String,
    modifier: Modifier = Modifier,
    isCompact: Boolean = false,
    screenState: ChatScreenState? = null,
    group: ChatGroupData? = screenState?.conversationGroup,
) {
    val settings = LocalSettingsProvider.current
    val constraints = LocalChatEngineConstraints.current
    val isTextStylesEnabled = settings.chatSettings.isTextStylesEnabled
    val themeColors = ChatTheme.colors
    val stylized = remember(displayName, message, isTextStylesEnabled, ChatTheme.colors.isLight) {
        message.toStylizedMessage(
            displayName = displayName,
            params = StylizedTextParameters(
                settings = settings.chatSettings,
                colors = themeColors,
                mentionsRegex = constraints.mentionRegex,
                group = group,
            ),
        )
    }
    
    // Apply text effects to style modifier
    val spoilerState = remember { mutableStateOf(MessageTextEffect.SPOILER in stylized.effects) }
    val infiniteTransition = rememberInfiniteTransition()
    var styleModifier = modifier
    if (settings.chatSettings.isTextEffectsEnabled) {
        for (effect in stylized.effects) {
            styleModifier = styleModifier.run {
                when (effect) {
                    MessageTextEffect.SPOILER -> {
                        if (spoilerState.value) {
                            background(ChatTheme.colors.spoiler).alpha(0f)
                        } else {
                            this
                        }
                    }
                    MessageTextEffect.BLUR -> blur(radius = 2.dp)
                    MessageTextEffect.REVERSE -> scale(scaleX = -1f, scaleY = 1f)
                    MessageTextEffect.FLIP -> scale(scaleX = 1f, scaleY = -1f)
                    MessageTextEffect.ZOOM -> {
                        val transition by infiniteTransition.animateFloat(0.5f, 1f, 400)
                        scale(transition)
                    }
                    MessageTextEffect.FADE -> {
                        val transition by infiniteTransition.animateFloat(0f, 1f, 1000, FastOutSlowInEasing)
                        alpha(transition)
                    }
                    MessageTextEffect.ALERT -> {
                        val transition by infiniteTransition.animateFloat(0f, 1f, 400)
                        alpha(transition)
                    }
                    MessageTextEffect.BOUNCE -> {
                        val transition by infiniteTransition.animateFloat(0f, 20f, 300, FastOutSlowInEasing)
                        offset(x = transition.dp)
                    }
                    MessageTextEffect.SHAKE -> {
                        val transition by infiniteTransition.animateFloat(0f, 10f, 50)
                        offset(y = transition.dp)
                    }
                    else -> this
                }
            }
        }
    }

    val isQrCode = MessageTextEffect.QRCODE in stylized.effects &&
            stylized.isNotBlank() &&
            settings.embedSettings.isEnabled &&
            settings.embedSettings.isImagesEnabled

    // Render the text itself
    Row(
        horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
        modifier = Modifier.fillMaxWidth(),
    ) {
        MessageText(
            message = message,
            stylized = stylized,
            isCompact = isCompact,
            isQrCode = isQrCode,
            spoilerState = spoilerState,
            styleModifier = styleModifier,
        )
        if (message.isEdited && !isCompact) {
            DisableSelection {
                TextLine(
                    text = stringResource(Res.string.message_edited),
                    color = ChatTheme.colors.onBackground2,
                    style = ChatTheme.typography.caption,
                )
            }
        }
    }
    
    // Extra content (file transfers, invites, embeds, etc.)
    screenState?.let {
        if (!isCompact) {
            MessageBlockExtra(
                message = message,
                stylized = stylized,
                screenState = screenState,
                modifier = modifier,
                isQrCode = isQrCode,
            )
        }
    }
}

@Composable
private fun RowScope.MessageText(
    message: ChatMessage,
    stylized: StylizedMessage,
    isCompact: Boolean,
    isQrCode: Boolean,
    spoilerState: MutableState<Boolean>,
    styleModifier: Modifier,
) = Box(modifier = Modifier.weight(1f)) {
    val isNegativeScale = MessageTextEffect.REVERSE in stylized.effects || MessageTextEffect.FLIP in stylized.effects
    val annotatedText = stylized.text.text
    val textSize = if (isPortraitMode() && !isCompact) ChatTheme.typography.body1 else ChatTheme.typography.body2
    val textStyle = textSize.copy(
        color = if (message.isAction) ChatTheme.colors.onBackground2 else ChatTheme.colors.onBackground,
        fontStyle = if (message.isAction) FontStyle.Italic else FontStyle.Normal,
    )
    when {
        isCompact && (isQrCode || stylized.isBlank()) -> {
            TextLine(
                text = when {
                    isQrCode -> stringResource(Res.string.chat_message_placeholder_qr_code)
                    message.fileTransferUid != null -> stringResource(Res.string.chat_message_placeholder_file_transfer)
                    else -> stringResource(Res.string.chat_message_placeholder_other)
                },
                style = textSize.copy(fontStyle = FontStyle.Italic),
                color = ChatTheme.colors.onBackground2,
            )
        }
        isCompact -> {
            TextLine(
                text = annotatedText,
                style = textStyle,
                modifier = styleModifier,
            )
        }
        message.shouldHideText(stylized) -> {
            // Do not display text for single URL image embeds
        }
        stylized.isNotEmpty() && !isQrCode -> {
            // XXX - SelectionContainer throws an IllegalStateException when text uses a negative scale
            DisableSelection(isNegativeScale || spoilerState.value) {
                ClickableMessageText(
                    text = annotatedText,
                    style = textStyle,
                    modifier = styleModifier,
                    onClick = {
                        if (MessageTextEffect.SPOILER in stylized.effects) {
                            spoilerState.value = !spoilerState.value
                        }
                    },
                    showUrlHoverIcon = false, // Not supported unless selection is disabled
                )
            }
        }
    }
}

@Composable
private fun InfiniteTransition.animateFloat(
    initial: Float,
    target: Float,
    duration: Int,
    easing: Easing = LinearEasing,
) = animateFloat(
    initialValue = initial,
    targetValue = target,
    animationSpec = infiniteRepeatable(tween(duration, easing = easing), RepeatMode.Reverse),
)

@Composable
private fun ChatMessage.shouldHideText(stylized: StylizedMessage): Boolean {
    return when {
        stylized.isEmojiMessage -> true
        embedData != null && urls.size == 1 && stylized.isUrl() -> {
            val settings = LocalSettingsProvider.current.embedSettings
            when (embedData) {
                is ImageEmbedData -> settings.isEnabled && settings.isImagesEnabled
                else -> false
            }
        }
        else -> false
    }
}