package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.text.selection.DisableSelection
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupType
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.dialog.PasswordDialog
import nand.apps.chat.ui.embed.EmbedContainer
import nand.apps.chat.ui.embed.qrcode.QrCodeEmbedRenderer
import nand.apps.chat.ui.emoji.EmojiRenderer
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import seers.composeapp.generated.resources.Res
import seers.composeapp.generated.resources.password
import seers.composeapp.generated.resources.set_group_password

@Composable
fun MessageBlockExtra(
    message: ChatMessage,
    stylized: StylizedMessage,
    screenState: ChatScreenState,
    modifier: Modifier = Modifier,
    isQrCode: Boolean,
) {
    val eventHandler = LocalEventHandler.current
    val engineConstraints = LocalChatEngineConstraints.current
    val extraModifier = modifier.padding(top = ChatTheme.dimens.spacingSmall)
    val fileTransfer = screenState.fileTransfers[message.fileTransferUid]
    var isPasswordDialogVisible by rememberSaveable { mutableStateOf(false) }
    when {
        fileTransfer != null -> FileTransferRow(
            fileTransfer = fileTransfer,
            modifier = extraModifier,
            onFileTransferControl = { eventHandler += FileTransferControlEvent(fileTransfer, it) },
            onOpenFile = { eventHandler += OpenFileEvent(fileTransfer) },
            onDeleteFile = { eventHandler += UpdateMessageEvent(message.copy(fileTransferUid = null)) }
        )
        message.friendRequest != null -> AcceptDeclineButtons(
            modifier = extraModifier,
            onAccept = { eventHandler += AcceptFriendRequestEvent(message.friendRequest) },
            onDecline = { eventHandler += DeclineFriendRequestEvent(message.friendRequest) },
        )
        message.groupInvite != null -> AcceptDeclineButtons(
            modifier = extraModifier,
            onAccept = {
                if (message.groupInvite.groupType == ChatGroupType.ADVANCED) {
                    isPasswordDialogVisible = true
                } else {
                    eventHandler += AcceptGroupInviteEvent(message.groupInvite)
                }
            },
            onDecline = { eventHandler += DeclineGroupInviteEvent(message.groupInvite) },
        )
        isQrCode -> QrCodeEmbedRenderer(
            text = stylized.text.text,
            modifier = extraModifier,
        )
        stylized.isEmojiMessage -> EmojiRenderer(
            text = stylized.text.text,
            modifier = extraModifier,
        )
        message.urls.size == 1 && MessageTextEffect.NOEMBED !in stylized.effects -> {
            EmbedContainer(
                url = message.urls.first(),
                message = message,
                modifier = extraModifier,
            )
        }
    }
    if (isPasswordDialogVisible) {
        DisableSelection {
            PasswordDialog(
                title = Res.string.set_group_password,
                label = Res.string.password,
                maxLength = engineConstraints.maxGroupPasswordLength,
                isOptional = true,
                onConfirm = { password ->
                    isPasswordDialogVisible = false
                    message.groupInvite?.let { invite ->
                        eventHandler += AcceptGroupInviteEvent(
                            invite = invite.copy(password = password)
                        )
                    }
                },
                onCancel = { isPasswordDialogVisible = false },
            )
        }
    }
}