package nand.apps.chat.ui.chat.message

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.core.*
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.Send
import androidx.compose.material.icons.filled.Warning
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.rotate
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.input.key.Key
import androidx.compose.ui.input.key.KeyEvent
import androidx.compose.ui.input.key.isShiftPressed
import androidx.compose.ui.input.key.key
import androidx.compose.ui.input.pointer.PointerIcon
import androidx.compose.ui.input.pointer.pointerHoverIcon
import androidx.compose.ui.layout.*
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.VisualTransformation
import androidx.compose.ui.unit.*
import kotlinx.coroutines.delay
import nand.apps.chat.io.getFilePath
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.platform.AppState
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.text.checkMentions
import nand.apps.chat.text.getClipboardFilePath
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.FileTransferConfirmationDialog
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.emoji.EmojiPopup
import nand.apps.chat.ui.file.ChatFileSenderDialog
import nand.apps.chat.ui.modifier.*
import nand.apps.chat.ui.provider.*
import nand.apps.chat.ui.settings.FileTransferUploadConfigRoute
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.insert
import nand.apps.chat.util.insertNewline
import nand.apps.chat.util.takeIfUnderLength
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import kotlin.math.roundToInt
import seers.composeapp.generated.resources.*

// TODO - Create a dedicated view model for this
@Composable
fun MessageInputField(
    screenState: ChatScreenState,
    textStyle: TextStyle,
    onSubmit: (String) -> Unit,
    onCancelEdit: () -> Unit,
    onCycle: (MessageCycleDirection, String) -> Unit,
    onSendFile: (String, String) -> Boolean,
    onReaction: (String?) -> Unit,
    modifier: Modifier = Modifier,
) {
    val constraints = LocalChatEngineConstraints.current
    val settings = LocalSettingsProvider.current
    val navController = LocalNavController.current
    val focusManager = LocalFocusManager.current
    val density = LocalDensity.current
    val eventHandler = LocalEventHandler.current
    val appStateProvider: AppStateProvider by getKoin().inject()
    val isPortrait = isPortraitMode()
    val chatSettings = settings.chatSettings
    val fileSettings = settings.fileSettings
    val colors = ChatTheme.colors

    val state = screenState.inputState
    val isEnabled = screenState.conversationUid !is FriendRequestUid
    val isDirectMessage = screenState.conversationUid is UserUid
    val isObserver = screenState.conversationGroup?.selfRole == ChatGroupRole.OBSERVER
    val isButtonsEnabled = isEnabled && !state.isBusy && state.reactMessage == null && !isObserver
    
    var dragFilePath by remember { mutableStateOf<String?>(null) }
    var isSendSelectorVisible by rememberSaveable { mutableStateOf(false) }
    var isGroupUploadDialogVisible by rememberSaveable { mutableStateOf(false) }
    var fileConfirmPath by rememberSaveable { mutableStateOf<String?>(null) }

    var composablePosition by remember { mutableStateOf(IntOffset.Zero) }
    var textInputPosition by remember { mutableStateOf(IntOffset.Zero) }
    var textInputSize by remember { mutableStateOf(IntSize.Zero) }
    
    var mentionIndex by rememberSaveable { mutableStateOf(-1) }
    var mentionPopupOffset by remember { mutableStateOf(IntOffset.Zero) }
    var mentionPopupSize by remember { mutableStateOf(DpSize.Zero) }
    var mentionKeyEvent by remember { mutableStateOf<KeyEvent?>(null) }
    
    var isEmojisVisible by rememberSaveable { mutableStateOf(false) }
    var emojiPopupSize by remember { mutableStateOf(DpSize.Zero) }
    
    val focusRequester = remember { FocusRequester() }
    val isFileTransferEnabled = isDirectMessage ||
            fileSettings.isGroupUploadEnabled &&
            fileSettings.uploadConfig.url.isNotEmpty()

    val styleParams = StylizedTextParameters(
        settings = chatSettings,
        colors = colors,
        tagPreviewStyle = if (chatSettings.isInputTagPreviewEnabled) textStyle else null,
        mentionsRegex = constraints.mentionRegex,
        group = screenState.conversationGroup,
    )
    var message by remember(state.text) {
        mutableStateOf(
            TextFieldValue(
                text = state.text,
                selection = TextRange(state.text.length),
            )
        )
    }
    var visualMessage by remember(state.text) {
        mutableStateOf(AnnotatedString(state.text).toStylizedInputText(styleParams))
    }
    val updateMessage: (TextFieldValue) -> Unit = { newValue ->
        if (isButtonsEnabled) {
            val modified = newValue
                .takeIfUnderLength(constraints.maxMessageLength, message)
                .checkMentions(
                    original = message,
                    originalTransformed = visualMessage,
                )
            if (message.text != modified.text) {
                visualMessage = AnnotatedString(modified.text).toStylizedInputText(styleParams)
            }
            message = modified
        }
    }
    val submitEvent = {
        if (!state.isBusy && state.reactMessage == null && message.text.isNotEmpty()) {
            onSubmit(message.text)
            updateMessage(TextFieldValue())
        }
    }
    val clearMentionState = {
        mentionIndex = -1
        mentionKeyEvent = null
    }
    
    // Mentions popup
    LaunchedEffect(message) {
        if (message.text.isNotEmpty() &&
            message.selection.length == 0 &&
            message.selection.min > 0 &&
            message.text[message.selection.min - 1] == '@') {
            mentionIndex = message.selection.min
        }
    }
    LaunchedEffect(message.selection, isDirectMessage, settings.chatSettings) {
        if (isDirectMessage || message.selection.min < mentionIndex || !settings.chatSettings.isMentionsEnabled) {
            clearMentionState()
        }
    }
    
    // Focus input field after re-focusing the app on desktop
    LaunchedEffect(Unit) {
        eventHandler.onEvent.collect { event ->
            if (event is AppFocusChangedEvent && event.isFocused && getPlatform().isDesktop) {
                delay(50) // Need to wait a few frames or the focus doesn't work
                if (appStateProvider.currentState == AppState.FOREGROUND) {
                    focusRequester.requestFocus()
                }
            }
        }
    }
    
    // Shortcuts
    ShortcutHandler { action ->
        when (action) {
            ShortcutAction.TOGGLE_EMOJI_POPUP -> {
                isEmojisVisible = !isEmojisVisible
                true
            }
            ShortcutAction.SEND_FILE -> {
                if (isButtonsEnabled) {
                    if (isFileTransferEnabled) {
                        isSendSelectorVisible = true
                    } else {
                        isGroupUploadDialogVisible = true
                    }
                    true
                } else {
                    false
                }
            }
            else -> false
        }
    }
    
    Column(
        horizontalAlignment = Alignment.CenterHorizontally,
        modifier = Modifier.onGloballyPositioned {
            composablePosition = it.boundsInRoot().topLeft.round()
        }
    ) {
        AnimatedVisibility(
            visible = isEmojisVisible || state.reactMessage != null,
        ) {
            EmojiPopup(
                size = emojiPopupSize,
                inputFocusRequester = focusRequester,
                onEmoji = { emoji ->
                    state.reactMessage?.let { onReaction(emoji) } ?: updateMessage(message.insert(emoji))
                },
                onDismiss = {
                    state.reactMessage?.let { onReaction(null) }
                    isEmojisVisible = false
                },
                modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingSmall),
            )
        }
        Row(
            verticalAlignment = Alignment.Bottom,
            modifier = modifier.onExternalDrag(
                enabled = fileSettings.isDragDropEnabled,
                onDragStart = {
                    if (isButtonsEnabled) {
                        dragFilePath = it.data.getFilePath()
                    }
                },
                onDragExit = { dragFilePath = null },
                onDrop = {
                    val filePath = dragFilePath
                    if (fileSettings.isDragDropEnabled && filePath != null) {
                        when {
                            !isFileTransferEnabled -> isGroupUploadDialogVisible = true
                            fileSettings.showFileTransferConfirmation -> fileConfirmPath = filePath
                            else -> onSendFile(filePath, message.text)
                        }
                    }
                    dragFilePath = null
                }
            ).onGloballyPositioned {
                val root = it.findRootCoordinates()
                val maxWidthDp = (it.size.width / density.density).dp
                val maxHeightDp = (root.size.height / density.density).dp
                emojiPopupSize = DpSize(
                    width = maxWidthDp.coerceAtMost(512.dp),
                    height = (maxHeightDp / 2).coerceAtMost(256.dp),
                )
                mentionPopupSize = DpSize(
                    width = maxWidthDp / if (isPortrait) 1 else 3,
                    height = maxHeightDp / 2
                )
            },
        ) {
            val infiniteTransition = rememberInfiniteTransition(label = "File Upload Spinner")
            val iconRotation = if (state.isBusy) {
                infiniteTransition.animateFloat(
                    initialValue = 0f,
                    targetValue = 360f,
                    animationSpec = infiniteRepeatable(tween(1000, easing = LinearEasing), RepeatMode.Restart),
                    label = "Rotate"
                )
            } else {
                null
            }
            InputButton(
                icon = if (state.isBusy) Icons.Outlined.Sync else Icons.Outlined.AddPhotoAlternate,
                contentDescription = Res.string.chat_screen_send_file,
                isEnabled = isButtonsEnabled,
                onClick = {
                    if (isFileTransferEnabled) {
                        isSendSelectorVisible = true
                    } else {
                        isGroupUploadDialogVisible = true
                    }
                },
                modifier = Modifier.padding(end = ChatTheme.dimens.spacingSmall),
                iconModifier = Modifier.rotate(iconRotation?.value ?: 0f),
            )
            SimpleTextField(
                value = message,
                onValueChange = updateMessage,
                hintText = when {
                    state.isBusy -> stringResource(Res.string.chat_screen_input_please_wait)
                    state.isUploadError -> stringResource(Res.string.chat_screen_input_upload_error)
                    !isEnabled -> null
                    dragFilePath != null -> stringResource(Res.string.chat_screen_input_drop_file, dragFilePath?.toPath()?.name.orEmpty())
                    !screenState.isConversationOnline -> stringResource(Res.string.contact_is_offline, screenState.conversationTitle.orEmpty())
                    isObserver -> stringResource(Res.string.group_self_observer_message)
                    else -> stringResource(Res.string.chat_screen_input_type_here)
                },
                isEnabled = isButtonsEnabled,
                isError = state.isUploadError,
                isSingleLine = false,
                textStyle = textStyle,
                visualTransformation = if (chatSettings.isInputTextStylesEnabled) {
                    VisualTransformation { visualMessage }
                } else {
                    VisualTransformation.None
                },
                leadingIcon = if (isEnabled && !screenState.isConversationOnline) {
                    {
                        Icon(
                            imageVector = Icons.Default.Warning,
                            contentDescription = stringResource(Res.string.contact_is_offline, screenState.conversationTitle.orEmpty()),
                            tint = ChatTheme.colors.warning,
                            modifier = Modifier.size(ChatTheme.dimens.iconSmall),
                        )
                    }
                } else null,
                trailingIcon = {
                    Row(
                        modifier = Modifier.padding(end = ChatTheme.dimens.spacingSmall),
                    ) {
                        IconButton(
                            icon = Icons.Outlined.AddReaction,
                            contentDescription = stringResource(Res.string.chat_message_add_emojis),
                            isEnabled = isButtonsEnabled || state.reactMessage != null,
                            color = when {
                                isEmojisVisible || state.reactMessage != null -> ChatTheme.colors.primary
                                else -> ChatTheme.colors.onBackground2
                            },
                            onClick = {
                                state.reactMessage?.let {
                                    onReaction(null)
                                    isEmojisVisible = false
                                } ?: run {
                                    isEmojisVisible = !isEmojisVisible
                                    focusRequester.requestFocus()
                                }
                            },
                        )
                        state.editMessage?.let {
                            IconButton(
                                icon = Icons.Outlined.Cancel,
                                contentDescription = stringResource(Res.string.cancel),
                                color = ChatTheme.colors.onBackground2,
                                onClick = onCancelEdit,
                            )
                        }
                    }
                },
                onTextLayout = { layout ->
                    val index = when {
                        mentionIndex in 0..visualMessage.text.length -> mentionIndex
                        else -> visualMessage.offsetMapping.originalToTransformed(message.selection.end)
                    }
                    mentionPopupOffset = IntOffset(
                        x = layout.getHorizontalPosition(
                            offset = index,
                            usePrimaryDirection = true,
                        ).toInt(),
                        y = layout.getLineTop(layout.getLineForOffset(index)).roundToInt(),
                    )
                },
                modifier = Modifier
                    .background(if (dragFilePath != null) ChatTheme.colors.surface3 else Color.Unspecified)
                    .weight(1f)
                    .focusRequester(focusRequester)
                    .onGloballyPositioned { coords ->
                        textInputPosition = coords.boundsInRoot().topLeft.round()
                        textInputSize = coords.size
                    }
                    .onKeyDown { event ->
                        when (event.key) {
                            Key.Enter, Key.NumPadEnter -> {
                                when {
                                    mentionIndex > 0 -> mentionKeyEvent = event
                                    event.isShiftPressed -> updateMessage(message.insertNewline())
                                    else -> submitEvent()
                                }
                                true
                            }
                            Key.DirectionUp -> {
                                when {
                                    mentionIndex > 0 -> mentionKeyEvent = event
                                    else -> onCycle(MessageCycleDirection.UP, message.text)
                                }
                                true
                            }
                            Key.DirectionDown -> {
                                when {
                                    mentionIndex > 0 -> mentionKeyEvent = event
                                    else -> onCycle(MessageCycleDirection.DOWN, message.text)
                                }
                                true
                            }
                            Key.Escape -> {
                                when {
                                    mentionIndex > 0 -> clearMentionState()
                                    isEmojisVisible -> isEmojisVisible = false
                                    else -> focusManager.clearFocus()
                                }
                                true
                            }
                            else -> false
                        }
                    }
                    .onPasteKey {
                        val path = getClipboardFilePath()
                        when {
                            path == null || !fileSettings.isCopyPasteEnabled -> false
                            !isFileTransferEnabled -> {
                                isGroupUploadDialogVisible = true
                                true
                            }
                            fileSettings.showFileTransferConfirmation -> {
                                fileConfirmPath = path.toString()
                                true
                            }
                            else -> onSendFile(path.toString(), message.text)
                        }
                    },
            )
            AnimatedVisibility(
                visible = message.text.isNotEmpty(),
            ) {
                InputButton(
                    icon = when {
                        state.editMessage != null -> Icons.Outlined.DoneOutline
                        else -> Icons.AutoMirrored.Default.Send
                    },
                    contentDescription = Res.string.chat_screen_send_message,
                    isEnabled = isButtonsEnabled,
                    onClick = submitEvent,
                    modifier = Modifier.padding(start = ChatTheme.dimens.spacingSmall),
                )
            }
        }
    }
    when {
        isGroupUploadDialogVisible -> SimpleDialog(
            title = Res.string.group_file_upload_config_dialog_title,
            confirmLabel = Res.string.open_settings,
            onConfirm = {
                isGroupUploadDialogVisible = false
                navController.navigate(FileTransferUploadConfigRoute)
            },
            onCancel = { isGroupUploadDialogVisible = false },
        ) {
            Text(
                text = stringResource(Res.string.group_file_upload_config_dialog_message),
                style = ChatTheme.typography.body2,
            )
        }
        fileConfirmPath != null -> FileTransferConfirmationDialog(
            path = fileConfirmPath?.toPath() ?: return,
            onConfirm = {
                fileConfirmPath?.let { onSendFile(it, message.text) }
                fileConfirmPath = null
            },
            onCancel = { fileConfirmPath = null },
        )
        isSendSelectorVisible -> ChatFileSenderDialog(
            onSelectFile = { path ->
                isSendSelectorVisible = false
                onSendFile(path, message.text)
            },
            onDismiss = { isSendSelectorVisible = false },
        )
        mentionIndex > 0 -> {
            screenState.conversationGroup?.let { group ->
                MentionPopup(
                    startIndex = mentionIndex,
                    message = message,
                    group = group,
                    size = mentionPopupSize,
                    offset = when {
                        isPortrait -> IntOffset(0, -textInputSize.height)
                        else -> (textInputPosition - composablePosition) + mentionPopupOffset + IntOffset(0, -textInputSize.height)
                    },
                    onUpdateMessage = { newMessage ->
                        clearMentionState()
                        updateMessage(newMessage)
                    },
                    onSubmitMessage = {
                        clearMentionState()
                        submitEvent()
                    },
                    keyEvent = mentionKeyEvent,
                )
            }
        }
    }
}

@Composable
private fun InputButton(
    icon: ImageVector,
    contentDescription: StringResource,
    onClick: () -> Unit,
    modifier: Modifier = Modifier,
    iconModifier: Modifier = Modifier,
    isEnabled: Boolean = true,
) = Button(
    colors = ButtonDefaults.buttonColors(
        backgroundColor = ChatTheme.colors.primary,
    ),
    enabled = isEnabled,
    onClick = onClick,
    contentPadding = PaddingValues(ChatTheme.dimens.spacingLarge),
    modifier = modifier
        .size(ChatTheme.dimens.buttonXLarge)
        .pointerHoverIcon(if (isEnabled) PointerIcon.Hand else PointerIcon.Default),
) {
    Icon(
        imageVector = icon,
        tint = ChatTheme.colors.onPrimary,
        contentDescription = stringResource(contentDescription),
        modifier = Modifier
            .fillMaxSize()
            .then(iconModifier),
    )
}