package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.avatar.ChatGroupAvatar
import nand.apps.chat.model.avatar.ChatUserAvatar
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.chat.message.reaction.MessageReactionsRow
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.provider.LocalDateTimeFormatProvider
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun MessageRow(
    message: ChatMessage,
    user: UserData?,
    group: ChatGroupData?,
    isHeaderVisible: Boolean,
    isDateSeparatorVisible: Boolean,
    screenState: ChatScreenState,
    modifier: Modifier = Modifier,
) {
    val groupMessageFallback = stringResource(Res.string.group_message)
    val friendRequestText = stringResource(Res.string.friend_request)
    val unknownSenderText = stringResource(Res.string.unknown_sender)
    val displayName = remember(message, user, group) {
        when {
            message.isGroupMessage -> group?.title ?: groupMessageFallback
            message.friendRequest != null -> friendRequestText
            user != null -> user.displayName
            else -> unknownSenderText
        }
    }
    Column(modifier = modifier) {
        if (isDateSeparatorVisible) {
            MessageDateSeparator(
                timestamp = message.timestamp,
                modifier = Modifier.fillMaxWidth(),
            )
        }
        when {
            group != null && message.isGroupConnectionEvent -> {
                GroupEventMessageRow(
                    message = message,
                    displayName = displayName,
                    modifier = Modifier.fillMaxWidth(),
                )
            }
            else -> MessageRow(
                message = message,
                user = user,
                group = group,
                displayName = displayName,
                isHeaderVisible = isHeaderVisible,
                screenState = screenState,
            )
        }
    }
}

@Composable
private fun MessageRow(
    message: ChatMessage,
    user: UserData?,
    group: ChatGroupData?,
    displayName: String,
    isHeaderVisible: Boolean,
    screenState: ChatScreenState,
) = Row {
    val settings = LocalSettingsProvider.current.chatSettings
    val dateTimeFormatter = LocalDateTimeFormatProvider.current
    val avatarSize = if (isPortraitMode()) ChatAvatarSize.MEDIUM else ChatAvatarSize.SMALL
    
    if (isHeaderVisible && settings.isAvatarsEnabled) {
        val avatar = remember(user, group) {
            when {
                message.isGroupMessage -> group?.avatar ?: ChatGroupAvatar("")
                message.friendRequest != null -> message.friendRequest.avatar
                else -> user?.avatar ?: ChatUserAvatar(displayName)
            }
        }
        ChatAvatarBox(
            avatar = avatar,
            size = avatarSize,
        )
    }
    val rowBackground = when {
        screenState.inputState.editMessage?.uid == message.uid -> ChatTheme.colors.primaryVariant
        else -> Color.Unspecified
    }
    MessageActionMenu(
        message = message,
        screenState = screenState,
        user = user,
        group = group,
        displayName = displayName,
        modifier = Modifier
            .align(Alignment.Top)
            .padding(start = ChatTheme.dimens.spacingMedium + if (isHeaderVisible || !settings.isAvatarsEnabled) 0.dp else avatarSize.getSizeDp())
            .background(rowBackground)
    ) {
        Column {
            if (isHeaderVisible) {
                Box(modifier = Modifier.fillMaxWidth()) {
                    val timestampFormatted = remember(message) {
                        buildString {
                            if (message.isDelivered) append("✓ ")
                            append(dateTimeFormatter.formatTime(message.timestamp))
                        }
                    }
                    val textStyle = if (isPortraitMode()) ChatTheme.typography.body2 else ChatTheme.typography.caption
                    TextLine(
                        text = displayName,
                        style = textStyle.copy(fontWeight = FontWeight.Bold),
                        color = when {
                            message.isGroupMessage -> ChatTheme.colors.statusOnline
                            else -> ChatTheme.colors.onSurface
                        },
                    )
                    TextLine(
                        text = timestampFormatted,
                        style = ChatTheme.typography.caption,
                        color = ChatTheme.colors.onBackground2,
                        modifier = Modifier.align(Alignment.CenterEnd),
                    )
                }
            }
            MessageBlock(
                message = message,
                displayName = displayName,
                screenState = screenState,
            )
            if (settings.isMessageReactionsEnabled) {
                MessageReactionsRow(
                    message = message,
                    selfUid = screenState.selfUid,
                    group = screenState.conversationGroup,
                    modifier = Modifier.padding(top = ChatTheme.dimens.spacingSmall),
                )
            }
        }
    }
}