package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.LazyListState
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import kotlinx.coroutines.launch
import nand.apps.chat.log.getLogger
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.chat.screen.ChatScreenViewModel
import nand.apps.chat.ui.provider.LocalDateTimeFormatProvider
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.scrollbar.VerticalScrollbar
import nand.apps.chat.ui.theme.ChatTheme

private const val DEBUG_SCROLL_BEHAVIOR = false

@Composable
fun MessagesColumn(
    viewModel: ChatScreenViewModel,
    state: ChatScreenState,
    listState: LazyListState = rememberLazyListState(),
    modifier: Modifier = Modifier,
) {
    val coroutineScope = rememberCoroutineScope()
    val chatSettings = LocalSettingsProvider.current.chatSettings
    val dateTimeFormatter = LocalDateTimeFormatProvider.current
    var previousVisibleIndex by remember(state.conversationUid) { mutableStateOf(0) }
    var previousVisibleOffset by remember(state.conversationUid) { mutableStateOf(0) }
    var previousMessagesSize by remember(state.conversationUid) { mutableStateOf(state.messages.size) }
    var lastVisibleIndex by remember { mutableStateOf(0) }
    val currentDate = dateTimeFormatter.formatDate(currentTimeMillis())

    val logger = remember { getLogger("MessagesColumn") }
    
    // Reset scroll to zero whenever we change conversations
    LaunchedEffect(state.conversationUid) {
        listState.scrollToItem(0)
    }

    // Keep the screen scrolled to the bottom when a new message is added
    LaunchedEffect(state.messages.size, listState.firstVisibleItemIndex, listState.firstVisibleItemScrollOffset) {
        if (previousMessagesSize != state.messages.size ||
            previousVisibleIndex != listState.firstVisibleItemIndex ||
            previousVisibleOffset != listState.firstVisibleItemScrollOffset) {

            if (DEBUG_SCROLL_BEHAVIOR) {
                logger.debug("size prev = $previousMessagesSize | cur = ${state.messages.size}")
                logger.debug("index prev = $previousVisibleIndex | cur = ${listState.firstVisibleItemIndex}")
                logger.debug("offset prev = $previousVisibleOffset | cur = ${listState.firstVisibleItemScrollOffset}")
            }
            
            if (previousMessagesSize < state.messages.size && previousVisibleIndex == 0 && previousVisibleOffset == 0) {
                listState.scrollToItem(0)
            }
            
            previousMessagesSize = state.messages.size
            previousVisibleIndex = listState.firstVisibleItemIndex
            previousVisibleOffset = listState.firstVisibleItemScrollOffset
        }
        viewModel.setFirstVisibleMessageIndex(listState.firstVisibleItemIndex)
    }

    // Get top visible row index for the date header
    LaunchedEffect(listState) {
        snapshotFlow { listState.layoutInfo.visibleItemsInfo }.collect { items ->
            lastVisibleIndex = items.lastOrNull()?.index ?: 0
        }
    }
    
    Box(modifier = modifier) {
        // Messages list
        SelectionContainer {
            LazyColumn(
                reverseLayout = true,
                state = listState,
                contentPadding = PaddingValues(vertical = ChatTheme.dimens.spacingLarge),
                modifier = Modifier
                    .fillMaxSize()
                    .padding(horizontal = ChatTheme.dimens.spacingLarge),
            ) {
                items(
                    count = state.messages.size,
                    key = { state.messages[state.getDisplayIndex(it)].uid.value },
                ) { index ->
                    val reverseIndex = state.getDisplayIndex(index)
                    val message = state.messages[reverseIndex]
                    val last = if (reverseIndex == 0) null else state.messages[reverseIndex - 1]
                    val messageDate = dateTimeFormatter.formatDate(message.timestamp)
                    val lastMessageDate = last?.let { dateTimeFormatter.formatDate(it.timestamp) }
                    val isHeaderVisible = !message.shouldMerge(last, chatSettings.messageMergeThreshold)
                    val isDateSeparatorVisible = last != null && messageDate != lastMessageDate
                    val topPadding = when {
                        reverseIndex == 0 -> 0.dp
                        isHeaderVisible -> ChatTheme.dimens.spacingXLarge
                        else -> ChatTheme.dimens.spacingSmall
                    }
                    val user = state.findUser(message)
                    val group = state.findGroup(message)
                    MessageRow(
                        message = message,
                        user = user,
                        group = group,
                        isHeaderVisible = isHeaderVisible || isDateSeparatorVisible,
                        isDateSeparatorVisible = isDateSeparatorVisible,
                        screenState = state,
                        modifier = Modifier
                            .fillMaxWidth()
                            .padding(top = topPadding),
                    )
                }
            }
        }
        VerticalScrollbar(
            listState = listState,
            reverseLayout = true,
            modifier = Modifier.align(Alignment.CenterEnd),
        )

        // Date indicator when viewing messages in the past
        val topMessageDate = remember(state, lastVisibleIndex) {
            state.messages.getOrNull(state.getDisplayIndex(lastVisibleIndex))?.let {
                dateTimeFormatter.formatDate(it.timestamp)
            }
        }
        MessageDateHeader(
            messageDate = topMessageDate ?: currentDate,
            currentDate = currentDate,
            modifier = Modifier.align(Alignment.TopCenter),
        )

        // Unread messages below indicator
        val unreadCount = state.unreadCounts[state.conversationUid]
        UnreadMessagesFooter(
            unreadCount = unreadCount?.total ?: 0,
            onClick = {
                coroutineScope.launch {
                    listState.animateScrollToItem(0)
                }
            },
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .padding(bottom = ChatTheme.dimens.spacingLarge)
        )
    }
}