package nand.apps.chat.ui.chat.message

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.*
import androidx.compose.ui.text.font.FontFamily
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.font.FontSynthesis
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.input.OffsetMapping
import androidx.compose.ui.text.input.TransformedText
import androidx.compose.ui.text.style.BaselineShift
import androidx.compose.ui.text.style.TextDecoration
import androidx.compose.ui.unit.sp
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.settings.ChatSettingsData
import nand.apps.chat.text.*
import nand.apps.chat.text.emoji.isEmojiString
import nand.apps.chat.ui.theme.ChatThemeColors
import nand.apps.chat.util.*

data class StylizedMessage(
    val text: TransformedText,
    val effects: Set<MessageTextEffect> = emptySet(),
) : CharSequence {
    
    constructor(
        text: AnnotatedString,
        effects: Set<MessageTextEffect> = emptySet(),
    ) : this(TransformedText(text, OffsetMapping.Identity), effects)
    
    private val rawText get() = text.text.text

    val isEmojiMessage by lazy { isEmojiString() }
    
    override val length = rawText.length

    override fun get(index: Int) = rawText[index]

    override fun subSequence(startIndex: Int, endIndex: Int) =
        rawText.subSequence(startIndex, endIndex)

    override fun toString() = rawText
}

data class StylizedTextParameters(
    val settings: ChatSettingsData,
    val colors: ChatThemeColors,
    val mentionsRegex: Regex? = null,
    val group: ChatGroupData? = null,
    val tagPreviewStyle: TextStyle? = null,
) {
    
    val urlStyle = SpanStyle(color = colors.primary, textDecoration = TextDecoration.Underline)
}

fun ChatMessage.toStylizedMessage(
    displayName: String,
    params: StylizedTextParameters,
): StylizedMessage = text.toStylizedMessage(
    params = params,
    removeEffectTags = true,
    prefix = if (isAction) displayName else null,
)

fun AnnotatedString.toStylizedInputText(
    params: StylizedTextParameters,
): TransformedText = text.toStylizedMessage(
    params = params,
    removeEffectTags = false,
).text

fun TransformedText.applyStyleTags(
    tagPreviewStyle: SpanStyle? = null,
) = buildTransformedText(this) {
    var start = 0
    var styleStack = 0
    var tagStart = 0
    while (text.indexOf('$', tagStart).also { tagStart = it } != -1) {
        val tagEnd = text.indexOf('$', tagStart + 1)
        if (tagEnd == -1) break
        val code = text.substring(tagStart + 1, tagEnd).lowercase()
        val font = FONT_MAP[code]
        val style = STYLE_MAP[code] ?: COLOR_MAP[code]?.let { SpanStyle(color = it) }
        if (!text.hasUrlAnnotations(tagStart, tagEnd + 1) && (style != null || font != null)) {
            appendSource(start, tagStart)
            tagPreviewStyle?.let {
                appendTagPreview(it, tagStart, tagEnd + 1)
            }
            when {
                style === END_STYLE -> {
                    if (styleStack > 0) {
                        pop()
                        styleStack--
                    }
                }
                font != null -> {
                    pushFontAnnotation(font)
                    styleStack++
                }
                style != null -> {
                    pushStyle(style)
                    styleStack++
                }
            }
            start = tagEnd + 1
        }
        tagStart = tagEnd + 1
    }
    appendSource(start, text.length)
}

private fun TransformedText.addQuoteStyles(
    quoteColor: Color,
): TransformedText {
    if (!text.contains('>')) return this

    val src = this
    val styled = buildAnnotatedString {
        append(src.text)
        var quoteStart = 0
        while (text.indexOf('>', quoteStart).also { quoteStart = it } != -1) {
            val nextNewline = text.indexOf('\n', quoteStart).takeIf { it != -1 } ?: text.length
            if (quoteStart == 0 || text[quoteStart - 1] == '\n') {
                addStyle(SpanStyle(color = quoteColor), quoteStart, nextNewline)
            }
            quoteStart = nextNewline
        }
    }
    return copy(text = styled)
}

fun String.toStylizedMessage(
    params: StylizedTextParameters,
    removeEffectTags: Boolean,
    prefix: String? = null,
): StylizedMessage {
    val (textNoEffects, effects) = MessageTextEffect.parseEffects(this, params.settings.maxTextEffects)
    val text = if (removeEffectTags) textNoEffects else this
    val transformedText = prefix?.let {
        buildTransformedText(text) {
            appendOther("$prefix ")
            appendSource(0, text.length)
        }
    } ?: text.toTransformedText()
    
    val previewStyle = params.tagPreviewStyle?.let { textStyle ->
        SpanStyle(
            color = params.colors.onBackground2,
            fontSize = textStyle.fontSize,
            letterSpacing = textStyle.letterSpacing,
            fontSynthesis = textStyle.fontSynthesis ?: FontSynthesis.None,
            textDecoration = textStyle.textDecoration ?: TextDecoration.None,
            fontStyle = textStyle.fontStyle ?: FontStyle.Normal,
            fontWeight = textStyle.fontWeight ?: FontWeight.Normal,
            baselineShift = textStyle.baselineShift ?: BaselineShift.None,
            textGeometricTransform = textStyle.textGeometricTransform,
        )
    }
    
    var styled = transformedText
        .addUrlAnnotations()
        .addMentions(params)

    if (MessageTextEffect.NOSTYLE !in effects) {
        styled = styled.applyStyleTags(previewStyle)
            .addQuoteStyles(params.colors.quote)
            .addMarkdownStyles()
        if (!params.settings.isTextStylesEnabled) {
            styled = styled.copy(text = styled.text.removeStyles())
        }
    }
    
    return StylizedMessage(
        text = styled
            .applyUrlStyles(params.urlStyle)
            .applyFontAnnotations(),
        effects = effects,
    )
}

private val END_STYLE = SpanStyle()

private val STYLE_MAP = mapOf(
    "str" to SpanStyle(textDecoration = TextDecoration.LineThrough),
    "und" to SpanStyle(textDecoration = TextDecoration.Underline),
    "bold" to SpanStyle(fontWeight = FontWeight.Bold),
    "ita" to SpanStyle(fontStyle = FontStyle.Italic),
    "tiny" to SpanStyle(fontSize = 10.sp),
    "big" to SpanStyle(fontSize = 18.sp),
    "mono" to SpanStyle(fontFamily = FontFamily.Monospace),
    "ser" to SpanStyle(fontFamily = FontFamily.Serif),
    "sub" to SpanStyle(baselineShift = BaselineShift.Subscript, fontSize = 10.sp),
    "sup" to SpanStyle(baselineShift = BaselineShift.Superscript, fontSize = 10.sp),
    "end" to END_STYLE,
)

private val FONT_MAP = mapOf(
    "fra" to UnicodeFont.FRAKTUR,
    "frab" to UnicodeFont.FRAKTUR_BOLD,
    "cur" to UnicodeFont.CURSIVE,
    "cur2" to UnicodeFont.CURSIVE_ALT,
    "dbs" to UnicodeFont.DOUBLE_STRUCK,
    "cir" to UnicodeFont.CIRCLED,
    "ciri" to UnicodeFont.CIRCLED_INVERSE,
    "sqr" to UnicodeFont.SQUARED,
    "sqri" to UnicodeFont.SQUARED_INVERSE,
    "par" to UnicodeFont.PARENTHESES,
    "fw" to UnicodeFont.FULLWIDTH,
    "scap" to UnicodeFont.SMALLCAPS,
    "cjk" to UnicodeFont.CJK,
    "food" to UnicodeFont.FOOD,
)

private val COLOR_MAP = mapOf(
    "red" to Color.Red,
    "gre" to Color.Green,
    "blu" to Color.Blue,
    "yel" to Color.Yellow,
    "cya" to Color.Cyan,
    "mag" to Color.Magenta,
    "ora" to Color(0xFFFF8000),
    "sea" to Color(0xFF00FF80),
    "azu" to Color(0xFF0080FF),
    "ind" to Color(0xFF6A5DFF),
    "vio" to Color(0xFF8F00FF),
    "pnk" to Color(0xFFFF80FF),
    "sal" to Color(0xFFFF8080),
    "but" to Color(0xFFFFFF80),
    "dre" to Color(0xFF800000),
    "dgr" to Color(0xFF008000),
    "dbl" to Color(0xFF000080),
    "hide" to Color.Transparent,
)