package nand.apps.chat.ui.chat.screen

import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.padding
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.outlined.ArrowForwardIos
import androidx.compose.material.icons.outlined.ArrowBackIosNew
import androidx.compose.material.icons.outlined.Groups
import androidx.compose.runtime.Composable
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.call.CallButton
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.chat.message.StylizedTextParameters
import nand.apps.chat.ui.chat.message.toStylizedMessage
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.ClickableMessageText
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun ChatHeader(
    state: ChatScreenState,
    onBackPressed: () -> Unit,
    onTogglePeers: () -> Unit,
    modifier: Modifier = Modifier,
) {
    val isPortrait = isPortraitMode()
    val padding = if (isPortrait) ChatTheme.dimens.spacingLarge else ChatTheme.dimens.spacingMedium
    val settings = LocalSettingsProvider.current
    val isGroupPeersCollapsed = settings.generalSettings.isGroupPeersCollapsed
    val isCallsEnabled = settings.callSettings.isEnabled
    ShortcutHandler(
        action = ShortcutAction.TOGGLE_PEERS,
        isEnabled = !isPortrait,
        onShortcut = onTogglePeers,
    )
    Row(
        verticalAlignment = Alignment.CenterVertically,
        modifier = modifier.padding(
            start = if (isPortrait) 0.dp else padding,
            top = padding,
            bottom = padding,
            end = padding,
        ),
    ) {
        if (isPortrait) {
            IconButton(
                icon = Icons.Outlined.ArrowBackIosNew,
                contentDescription = stringResource(Res.string.back),
                size = ChatTheme.dimens.iconMedium,
                onClick = onBackPressed,
                modifier = Modifier.padding(horizontal = ChatTheme.dimens.spacingSmall),
            )
        }
        val avatarSize = if (isPortrait) ChatAvatarSize.LARGE else ChatAvatarSize.MEDIUM
        when (state.conversationUid) {
            is UserUid -> state.conversationUser?.let {
                ChatAvatarBox(
                    avatar = it.avatar,
                    status = it.status,
                    isOnline = it.isOnline,
                    size = avatarSize,
                )
            }

            is GroupUid -> state.conversationGroup?.let {
                ChatAvatarBox(
                    avatar = it.avatar,
                    status = UserStatus.DEFAULT,
                    isOnline = it.isOnline,
                    size = avatarSize,
                )
            }

            else -> Unit
        }
        val colors = ChatTheme.colors
        val conversationDesc = remember(state, colors, settings.chatSettings) {
            when (val uid = state.conversationUid) {
                is UserUid -> state.conversationUser?.statusMessage
                is GroupUid -> state.conversationGroup?.run {
                    description.ifBlank { getStringBlocking(Res.string.group_chat_peers_online, numPeersOnline, peers.size) }
                }
                is FriendRequestUid -> uid.value
                else -> null
            }?.toStylizedMessage(
                params = StylizedTextParameters(
                    settings = settings.chatSettings,
                    colors = colors,
                ),
                removeEffectTags = true,
            )
        }
        Column(
            modifier = Modifier
                .padding(start = padding)
                .weight(1f),
        ) {
            TextLine(
                text = state.conversationTitle.orEmpty(),
                style = ChatTheme.typography.h6,
            )
            conversationDesc?.text?.run {
                ClickableMessageText(
                    text = text,
                    style = if (isPortrait) ChatTheme.typography.body1 else ChatTheme.typography.body2,
                    color = ChatTheme.colors.onBackground,
                )
            }
        }
        state.conversationUser?.run {
            CallButton(
                contactUid = uid,
                call = state.conversationCall,
                isEnabled = isOnline && isCallsEnabled,
            )
        }
        state.conversationGroup?.run {
            if (ChatGroupFeature.AUDIO_CALLS in features) {
                CallButton(
                    contactUid = uid,
                    call = state.conversationCall,
                    isEnabled = isOnline && isCallsEnabled,
                )
            }
            val showViewPeersButton = isGroupPeersCollapsed || isPortrait
            IconButton(
                icon = when {
                    showViewPeersButton -> Icons.Outlined.Groups
                    else -> Icons.AutoMirrored.Outlined.ArrowForwardIos
                },
                contentDescription = when {
                    showViewPeersButton -> stringResource(Res.string.group_chat_view_peers)
                    else -> stringResource(Res.string.group_chat_hide_peers)
                },
                onClick = onTogglePeers,
                size = ChatTheme.dimens.iconMedium,
            )
        }
    }
}