package nand.apps.chat.ui.chat.screen

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.event.*
import nand.apps.chat.ui.call.CallHeader
import nand.apps.chat.ui.chat.message.MessageInputField
import nand.apps.chat.ui.chat.message.MessagesColumn
import nand.apps.chat.ui.groups.GroupErrorScreen
import nand.apps.chat.ui.groups.GroupPeersColumn
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.theme.ChatTheme

@Composable
fun ChatScreen(
    viewModel: ChatScreenViewModel,
    state: ChatScreenState,
    modifier: Modifier = Modifier,
) = Row(modifier = modifier) {
    
    val eventHandler = LocalEventHandler.current
    val settings = LocalSettingsProvider.current
    val isPortrait = isPortraitMode()
    var isGroupPeersSelected by remember { mutableStateOf(false) }
    val isGroupPeersCollapsed = settings.generalSettings.isGroupPeersCollapsed
    
    BackHandler(
        isEnabled = isPortrait && isGroupPeersSelected,
    ) {
        isGroupPeersSelected = false
    }
    
    Box(
        modifier = Modifier
            .weight(1f)
            .fillMaxHeight()
    ) {
        Column(
            modifier = Modifier.matchParentSize(),
        ) {
            ChatHeader(
                state = state,
                onBackPressed = { eventHandler += ExitConversationEvent },
                onTogglePeers = {
                    if (isPortrait) {
                        isGroupPeersSelected = true
                    } else {
                        eventHandler += UpdateSettingsEvent(
                            settings = settings.copy(
                                generalSettings = settings.generalSettings.copy(
                                    isGroupPeersCollapsed = !isGroupPeersCollapsed,
                                )
                            )
                        )
                    }
                },
                modifier = Modifier.fillMaxWidth(),
            )
            CallHeader(
                call = state.conversationCall,
                contactName = state.conversationTitle.orEmpty(),
                modifier = Modifier.fillMaxWidth(),
            )
            Box(
                modifier = Modifier
                    .background(ChatTheme.colors.onBackground2)
                    .fillMaxWidth()
                    .height(2.dp)
            )
            MessagesColumn(
                viewModel = viewModel,
                state = state,
                modifier = Modifier
                    .fillMaxWidth()
                    .weight(1f),
            )
            MessageInputField(
                screenState = state,
                textStyle = if (isPortrait) ChatTheme.typography.body1 else ChatTheme.typography.body2,
                modifier = Modifier
                    .fillMaxWidth()
                    .padding(horizontal = ChatTheme.dimens.spacingLarge)
                    .padding(bottom = ChatTheme.dimens.spacingLarge),
                onSubmit = viewModel::submitMessage,
                onCycle = viewModel::cycleMessage,
                onSendFile = viewModel::onSendFile,
                onCancelEdit = viewModel::cancelMessageEdit,
                onReaction = viewModel::addReaction,
            )
        }
        state.conversationGroup?.let { group ->
            if (group.error != null) {
                GroupErrorScreen(
                    group = group,
                    error = group.error,
                    modifier = Modifier
                        .background(ChatTheme.colors.backgroundFloating)
                        .matchParentSize(),
                )
            }
        }
    }
    state.conversationGroup?.let { group ->
        AnimatedVisibility(
            visible = isPortrait && isGroupPeersSelected || !isPortrait && !isGroupPeersCollapsed,
        ) {
            GroupPeersColumn(
                state = state,
                group = group,
                sidebarState = if (isPortrait) SidebarState.FULL else SidebarState.DEFAULT,
                onExit = { isGroupPeersSelected = false },
                modifier = Modifier.fillMaxHeight(),
            )
        }
    }
}