package nand.apps.chat.ui.chat.screen

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.isFilePath
import nand.apps.chat.model.event.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.net.FileUploader
import nand.apps.chat.repo.*
import nand.apps.chat.ui.chat.message.MessageCycleDirection
import nand.apps.chat.ui.chat.message.MessageInputState
import nand.apps.chat.ui.util.collect

class ChatScreenViewModel(
    private val messageRepo: ChatMessageRepo,
    private val contactRepo: ContactRepo,
    selfProfileRepo: SelfProfileRepo,
    private val settingsRepo: SettingsRepo,
    private val fileTransferRepo: FileTransferRepo,
    callRepo: CallRepo,
    private val fileUploader: FileUploader,
    private val fileSystem: ChatFileSystem,
    private val eventHandler: ChatEventHandler,
) : ViewModel() {

    val stateData = MutableStateFlow(ChatScreenState())
    
    private val conversationUid get() = stateData.value.conversationUid
    private val editMessage get() = stateData.value.inputState.editMessage
    private val reactMessage get() = stateData.value.inputState.reactMessage
    private val fileSettings get() = settingsRepo.settings.fileSettings
    
    private val messageHistory = mutableListOf<String>()
    private var messageHistoryIndex = 0
    
    init {
        collect(messageRepo.onNewMessage) { message ->
            if (message.conversationUid == conversationUid) {
                addToCurrentConversation(message)
            }
        }
        collect(messageRepo.onMessagesChanged) { conversationUid ->
            if (conversationUid == this.conversationUid) {
                refreshMessages(conversationUid)
            }
        }
        collect(messageRepo.unreadObserver) { unreadCounts ->
            stateData.emit(stateData.value.copy(unreadCounts = unreadCounts))
        }
        collect(selfProfileRepo.selfObserver) { self ->
            stateData.emit(stateData.value.copy(self = self))
        }
        collect(contactRepo.friendObserver) { friends ->
            stateData.emit(stateData.value.copy(friends = friends))
        }
        collect(contactRepo.groupObserver) { groups ->
            stateData.emit(stateData.value.copy(groups = groups))
        }
        collect(fileTransferRepo.observer) { fileTransfers ->
            stateData.emit(stateData.value.copy(fileTransfers = fileTransfers))
        }
        collect(callRepo.callObserver) { calls ->
            stateData.emit(stateData.value.copy(calls = calls))
        }
        collect(callRepo.callActivityObserver) { callActivity ->
            stateData.emit(stateData.value.copy(callActivity = callActivity))
        }
        collect(eventHandler.onEvent, ::onEvent)
    }
    
    private fun onEvent(event: ChatEvent) = when (event) {
        is OpenConversationEvent -> {
            openConversation(event.contact.uid)
        }
        is ExitConversationEvent -> {
            exitConversation(conversationUid)
        }
        is AcceptFriendRequestEvent -> {
            acceptFriendRequest(event.request)
        }
        is DeclineFriendRequestEvent -> {
            exitConversation(event.request.uid)
        }
        is RemoveFriendEvent -> {
            exitConversation(event.friend.uid)
        }
        is RemoveGroupEvent -> {
            exitConversation(event.group.uid)
        }
        is PromptEditMessageEvent -> {
            editMessage(event.message)
        }
        is PromptReactMessageEvent -> {
            reactMessage(event.message)
        }
        else -> Unit
    }
    
    private fun openConversation(conversationUid: ContactUid) {
        if (conversationUid != this.conversationUid) {
            cancelMessageEdit()
            refreshMessages(conversationUid)
        }
    }
    
    private suspend fun addToCurrentConversation(message: ChatMessage) {
        stateData.emit(
            stateData.value.copy(
                messages = stateData.value.messages + message
            )
        )
    }
    
    fun submitMessage(text: String) {
        val conversationUid = conversationUid ?: return
        editMessage?.let { message ->
            val edited = messageRepo.createMessage(text, conversationUid)
            eventHandler += UpdateMessageEvent(
                message = message.copy(
                    text = edited.text,
                    type = edited.type,
                    isEdited = true,
                ),
                localOnly = false,
            )
            editMessage(null)
        } ?: run {
            eventHandler += SendMessageEvent(
                conversationUid = conversationUid,
                text = text,
            )
            if (text.isNotBlank()) {
                messageHistory += text
                messageHistoryIndex = messageHistory.size
            }
            updateInputText("")
        }
    }
    
    fun cancelMessageEdit() {
        editMessage?.let { editMessage(null) }
        reactMessage?.let { reactMessage(null) }
    }
    
    fun cycleMessage(direction: MessageCycleDirection, currentText: String) {
        if (messageHistory.isEmpty() || messageHistoryIndex >= messageHistory.size && currentText.isNotBlank()) return
        messageHistoryIndex = (messageHistoryIndex + direction.value).coerceAtLeast(0).coerceAtMost(messageHistory.size)
        val inputText = if (messageHistoryIndex in messageHistory.indices) messageHistory[messageHistoryIndex] else ""
        updateInputText(inputText)
    }
    
    fun onSendFile(path: String, messageFieldText: String): Boolean {
        if (!path.isFilePath() || !fileSystem.exists(path)) return false
        sendFile(path, messageFieldText)
        return true
    }
    
    fun addReaction(reaction: String?) {
        val message = reactMessage ?: return
        reaction?.let {
            eventHandler += ReactMessageEvent(
                message = message,
                reaction = reaction,
            )
        }
        reactMessage(null)
    }
    
    fun setFirstVisibleMessageIndex(index: Int) {
        messageRepo.onFirstVisibleMessageIndex(index)
    }
    
    private fun sendFile(filePath: String, messageFieldText: String) {
        val conversationUid = conversationUid
        when {
            conversationUid is GroupUid && fileSettings.isGroupUploadEnabled -> {
                updateInputState { it.copy(isBusy = true) }
                viewModelScope.launch {
                    val url = fileUploader.upload(filePath).orEmpty()
                    updateInputState {
                        it.copy(
                            text = when {
                                url.isBlank() -> it.text
                                else -> "$messageFieldText\n".trimStart('\n') + url
                            },
                            isUploadError = url.isBlank(),
                            isBusy = false,
                        )
                    }
                }
            }
            conversationUid is UserUid -> {
                fileTransferRepo.sendFile(filePath, conversationUid)
            }
        }
    }
    
    private fun refreshMessages(conversationUid: ContactUid) {
        val messages = messageRepo.getMessages(conversationUid)
        stateData.tryEmit(
            stateData.value.copy(
                conversationUid = conversationUid,
                messages = messages,
            )
        )
        messageRepo.onConversationChanged(conversationUid)
    }
    
    private fun exitConversation(conversationUid: ContactUid?) {
        if (this.conversationUid == conversationUid) {
            stateData.tryEmit(
                stateData.value.copy(
                    conversationUid = null,
                    inputState = MessageInputState(),
                )
            )
            messageRepo.onConversationChanged(null)
        }
    }
    
    private fun updateInputState(block: (MessageInputState) -> MessageInputState) {
        stateData.tryEmit(
            stateData.value.copy(
                inputState = block(stateData.value.inputState),
            )
        )
    }
    
    private fun updateInputText(text: String) = updateInputState {
        it.copy(
            text = text,
            isUploadError = false,
        )
    }
    
    private fun editMessage(message: ChatMessage?) = updateInputState {
        it.copy(
            text = message?.text?.let { text ->
                when (message.type) {
                    ChatMessageType.ACTION -> "/me $text"
                    else -> text
                }
            }.orEmpty(),
            reactMessage = null,
            editMessage = message?.takeIf { it.text.isNotEmpty() },
        )
    }
    
    private fun reactMessage(message: ChatMessage?) = updateInputState {
        it.copy(
            editMessage = null,
            reactMessage = message,
        )
    }
    
    private fun acceptFriendRequest(request: FriendRequest) {
        val newFriend = contactRepo.getFriend(UserUid(request.uid.value))
        if (newFriend != null) {
            refreshMessages(newFriend.uid)
        } else {
            exitConversation(request.uid)
        }
    }
}