package nand.apps.chat.ui.contacts

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.ui.util.collect

class ContactListViewModel(
    contactRepo: ContactRepo,
    selfProfileRepo: SelfProfileRepo,
) : ViewModel() {

    private var friends = mutableListOf<UserData>()
    private var groups = mutableListOf<ChatGroupData>()
    private var friendRequests = mutableListOf<FriendRequest>()
    private var searchTerms = ""
    
    val stateData = MutableStateFlow(ContactListState())

    init {
        collect(selfProfileRepo.selfObserver) { self ->
            stateData.tryEmit(stateData.value.copy(self = self))
        }
        collect(contactRepo.friendObserver) { friendMap ->
            friends.update(friendMap.values.sortedWith(UserData.CONTACT_LIST_COMPARATOR))
        }
        collect(contactRepo.groupObserver) { groupMap ->
            groups.update(groupMap.values.sortedBy { it.title })
        }
        collect(contactRepo.friendRequestObserver) {
            friendRequests.update(it.values)
        }
    }
    
    fun search(terms: String) {
        if (searchTerms != terms) {
            searchTerms = terms
            update()
        }
    }

    fun openDialog(dialog: ContactDialogType) {
        stateData.tryEmit(stateData.value.copy(dialog = dialog))
    }

    fun dismissDialog() {
        stateData.tryEmit(stateData.value.copy(dialog = null))
    }
    
    private fun update() {
        val contacts = mutableListOf<ChatContact>()
        contacts += friendRequests.filterFriendRequests()
        contacts += groups.filterGroups()
        contacts += friends.filterFriends()
        stateData.tryEmit(stateData.value.copy(contacts = contacts))
    }
    
    private fun <T: ChatContact> MutableList<T>.update(other: Collection<T>) {
        clear()
        addAll(other)
        update()
    }
    
    private fun List<UserData>.filterFriends(): List<ChatContact> {
        if (searchTerms.isBlank()) return this
        
        return filter { user -> user.name.startsWithOrContains(searchTerms) }
    }

    private fun List<ChatGroupData>.filterGroups(): List<ChatContact> {
        if (searchTerms.isBlank()) return this

        return filter { group -> group.title.startsWithOrContains(searchTerms) }
    }

    private fun List<FriendRequest>.filterFriendRequests(): List<ChatContact> {
        if (searchTerms.isBlank()) return this

        return filter { request -> request.message.startsWithOrContains(searchTerms) }
    }
    
    private fun String.startsWithOrContains(terms: String): Boolean {
        return if (terms.length == 1) {
            startsWith(terms.first(), ignoreCase = true)
        } else {
            contains(terms, ignoreCase = true)
        }
    }
}