package nand.apps.chat.ui.contacts

import androidx.compose.foundation.layout.*
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.text.TextStyle
import nand.apps.chat.model.avatar.ChatAvatar
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuArea
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.tooltip.TooltipText

@Composable
fun ContactRow(
    heading: String,
    subHeading: String,
    status: UserStatus,
    isOnline: Boolean,
    avatar: ChatAvatar,
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    headingStyle: TextStyle = getDefaultHeadingStyle(sidebarState),
    contentPadding: PaddingValues = PaddingValues(
        horizontal = ChatTheme.dimens.spacingMedium,
        vertical = ChatTheme.dimens.spacingSmall,
    ),
    unreadCount: UnreadMessageCounts? = null,
    call: CallData? = null,
    iconOverlay: ImageVector? = null,
    iconOverlayColor: Color = Color.Unspecified,
    onClick: (() -> Unit)? = null,
    isActionMenuEnabled: Boolean = false,
    actionMenuItems: () -> List<ActionMenuItem> = { emptyList() },
) {
    ActionMenuArea(
        items = actionMenuItems,
        isEnabled = isActionMenuEnabled,
        onClick = onClick,
        modifier = modifier,
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.padding(contentPadding),
        ) {
            TooltipText(
                tooltip = heading,
                isEnabled = sidebarState == SidebarState.COLLAPSED,
            ) {
                ChatAvatarBox(
                    avatar = avatar,
                    status = status,
                    isOnline = isOnline,
                    iconOverlay = iconOverlay,
                    iconOverlayColor = iconOverlayColor,
                    badgeText = unreadCount?.run {
                        when {
                            directMessages > 0 -> directMessages.toString()
                            groupMessages > 0 -> "٭"
                            else -> null
                        }
                    },
                    call = call,
                    size = when (sidebarState) {
                        SidebarState.FULL -> ChatAvatarSize.LARGE
                        else -> ChatAvatarSize.MEDIUM
                    },
                )
            }
            if (sidebarState != SidebarState.COLLAPSED) {
                Column(
                    modifier = Modifier
                        .fillMaxWidth()
                        .padding(start = ChatTheme.dimens.spacingSmall),
                ) {
                    TextLine(
                        text = heading,
                        style = headingStyle,
                    )
                    TextLine(
                        text = subHeading,
                        style = when (sidebarState) {
                            SidebarState.FULL -> ChatTheme.typography.body2
                            else -> ChatTheme.typography.caption
                        },
                        color = ChatTheme.colors.onBackground,
                    )
                }
            }
        }
    }
}

@Composable
private fun getDefaultHeadingStyle(
    sidebarState: SidebarState,
) = when (sidebarState) {
    SidebarState.FULL -> ChatTheme.typography.h6
    else -> ChatTheme.typography.body1
}