package nand.apps.chat.ui.dialog

import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.hoverable
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.interaction.collectIsHoveredAsState
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.selection.DisableSelection
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Close
import androidx.compose.material.icons.outlined.OpenInBrowser
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.window.Dialog
import androidx.compose.ui.window.DialogProperties
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.provider.LocalUriLauncher
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.isUrl
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun FullscreenPreviewDialog(
    onDismiss: () -> Unit,
    uri: String? = null,
    content: @Composable BoxScope.() -> Unit,
) = DisableSelection {
    Dialog(
        onDismissRequest = onDismiss,
        properties = DialogProperties(usePlatformDefaultWidth = false),
    ) {
        BackHandler(onBack = onDismiss)
        Box(
            modifier = Modifier
                .clickable(
                    onClick = onDismiss,
                    interactionSource = remember { MutableInteractionSource() },
                    indication = null,
                )
                .fillMaxSize(),
        ) {
            Box(
                modifier = Modifier
                    .fillMaxSize(if (isPortraitMode()) 1f else 0.75f)
                    .align(Alignment.Center),
                content = content,
            )
            uri?.let {
                val uriLauncher = LocalUriLauncher.current
                val interactionSource = remember { MutableInteractionSource() }
                val isHovering by interactionSource.collectIsHoveredAsState()
                val contentDescription = when {
                    uri.isUrl() -> stringResource(Res.string.fullscreen_preview_open_url)
                    else -> stringResource(Res.string.fullscreen_preview_open_file)
                }
                val isUriVisible = isHovering || isPortraitMode()
                Row(
                    verticalAlignment = Alignment.CenterVertically,
                    modifier = Modifier
                        .background(if (isUriVisible) Color.Black.copy(alpha = 0.75f) else Color.Unspecified)
                        .fillMaxWidth()
                        .align(Alignment.BottomCenter)
                        .padding(ChatTheme.dimens.spacingMedium),
                ) {
                    TextLine(
                        text = uri,
                        style = ChatTheme.typography.caption,
                        textAlign = TextAlign.Center,
                        color = Color.White,
                        modifier = Modifier
                            .weight(1f)
                            .alpha(if (isUriVisible) 1f else 0f),
                    )
                    IconButton(
                        icon = Icons.Outlined.OpenInBrowser,
                        contentDescription = contentDescription,
                        onClick = { uriLauncher.launchUri(uri) },
                        size = ChatTheme.dimens.iconMedium,
                        color = Color.White,
                        interactionSource = interactionSource,
                        modifier = Modifier.hoverable(interactionSource),
                    )
                }
            }
            IconButton(
                icon = Icons.Default.Close,
                contentDescription = stringResource(Res.string.close_image),
                onClick = onDismiss,
                size = ChatTheme.dimens.iconLarge,
                color = Color.White,
                modifier = Modifier
                    .align(Alignment.TopEnd)
                    .padding(ChatTheme.dimens.spacingMedium),
            )
        }
    }
}