package nand.apps.chat.ui.dialog.image

import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.*
import androidx.compose.foundation.layout.*
import androidx.compose.material.Icon
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.BrokenImage
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.draw.drawWithContent
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.geometry.Rect
import androidx.compose.ui.graphics.*
import androidx.compose.ui.graphics.drawscope.Stroke
import androidx.compose.ui.graphics.drawscope.clipRect
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.unit.toSize
import androidx.compose.ui.window.Dialog
import androidx.compose.ui.window.DialogProperties
import coil3.compose.AsyncImagePainter
import coil3.compose.LocalPlatformContext
import coil3.compose.SubcomposeAsyncImage
import coil3.decode.DecodeUtils
import coil3.size.Scale
import nand.apps.chat.image.ImageCompressionParams
import nand.apps.chat.image.SimpleBitmap
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.spinner.ChasingDots
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.toPx
import nand.apps.chat.ui.viewmodel.composableViewModel
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@OptIn(ExperimentalFoundationApi::class)
@Composable
fun ImageCropperDialog(
    title: StringResource,
    imagePath: String,
    maxSize: Int,
    forceSquare: Boolean,
    onConfirm: (SimpleBitmap) -> Unit,
    onCancel: () -> Unit,
    compressionParams: ImageCompressionParams? = null,
    viewModel: ImageCropperViewModel = composableViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    val isPortrait = isPortraitMode()
    val context = LocalPlatformContext.current
    val snackbarHostState = remember { SnackbarHostState() }
    
    var surfaceRect by remember { mutableStateOf(Rect.Zero) }
    var imageRect by remember { mutableStateOf(Rect.Zero) }
    var cropRect by remember { mutableStateOf(Rect.Zero) }
    var dragPos by remember { mutableStateOf<DragPos?>(null) }
    val dragState = rememberDraggable2DState { offset ->
        dragPos?.let { pos ->
            val rect = cropRect.translate(offset)
            val clampedRect = Rect(
                left = rect.left.coerceAtLeast(imageRect.left),
                top = rect.top.coerceAtLeast(imageRect.top),
                right = rect.right.coerceAtMost(imageRect.right),
                bottom = rect.bottom.coerceAtMost(imageRect.bottom),
            )
            cropRect = when (pos) {
                DragPos.TOP_LEFT -> cropRect.copy(top = clampedRect.top, left = clampedRect.left)
                DragPos.BOTTOM_LEFT -> cropRect.copy(bottom = clampedRect.bottom, left = clampedRect.left)
                DragPos.BOTTOM_RIGHT -> cropRect.copy(bottom = clampedRect.bottom, right = clampedRect.right)
                DragPos.TOP_RIGHT -> cropRect.copy(top = clampedRect.top, right = clampedRect.right)
                DragPos.TOP_CENTER -> cropRect.copy(top = clampedRect.top)
                DragPos.BOTTOM_CENTER -> cropRect.copy(bottom = clampedRect.bottom)
                DragPos.LEFT_CENTER -> cropRect.copy(left = clampedRect.left)
                DragPos.RIGHT_CENTER -> cropRect.copy(right = clampedRect.right)
                DragPos.CENTER -> rect.translate(
                    translateX = when {
                        rect.left < imageRect.left -> imageRect.left - rect.left
                        rect.right > imageRect.right -> imageRect.right - rect.right
                        else -> 0f
                    },
                    translateY = when {
                        rect.top < imageRect.top -> imageRect.top - rect.top 
                        rect.bottom > imageRect.bottom -> imageRect.bottom - rect.bottom
                        else -> 0f
                    }
                )
            }
        }
    }

    val dragHandleRadius = when {
        isPortrait -> ChatTheme.dimens.iconMedium
        else -> ChatTheme.dimens.iconSmall
    }.toPx() / 2f
    val dragHandles = remember(cropRect) {
        listOf(
            DragPos.CENTER to cropRect.center,
            DragPos.TOP_LEFT to cropRect.topLeft,
            DragPos.TOP_RIGHT to cropRect.topRight,
            DragPos.BOTTOM_RIGHT to cropRect.bottomRight,
            DragPos.BOTTOM_LEFT to cropRect.bottomLeft,
            DragPos.TOP_CENTER to cropRect.topCenter,
            DragPos.BOTTOM_CENTER to cropRect.bottomCenter,
            DragPos.LEFT_CENTER to cropRect.centerLeft,
            DragPos.RIGHT_CENTER to cropRect.centerRight,
        ).map { (pos, offset) ->
            pos to Rect(center = offset, radius = dragHandleRadius)
        }
    }
    
    val errorMessage = stringResource(Res.string.image_crop_failed)
    LaunchedEffect(state.isError) {
        if (state.isError) {
            snackbarHostState.showSnackbar(errorMessage)
        }
    }
    
    Dialog(
        properties = DialogProperties(
            usePlatformDefaultWidth = !isPortrait,
        ),
        onDismissRequest = onCancel,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingLarge),
            modifier = if (isPortrait) {
                Modifier
                    .background(ChatTheme.colors.background)
                    .fillMaxSize()
                    .padding(vertical = ChatTheme.dimens.spacingLarge)
                    .padding(horizontal = ChatTheme.dimens.spacingXLarge)
            } else {
                Modifier
                    .background(
                        color = ChatTheme.colors.background,
                        shape = ChatTheme.shapes.small,
                    )
                    .padding(ChatTheme.dimens.spacingLarge)
            }
        ) {
            TextLine(
                text = stringResource(title),
                style = ChatTheme.typography.h6,
                modifier = Modifier.run {
                    if (isPortrait) padding(horizontal = ChatTheme.dimens.spacingLarge) else this
                },
            )
            Box(
                contentAlignment = Alignment.Center,
                modifier = Modifier
                    .background(Color.Black)
                    .fillMaxWidth()
                    .run {
                        if (isPortrait) {
                            weight(1f)
                        } else {
                            height(ChatTheme.dimens.imageSmall)
                        }
                    }
                    .onSizeChanged { size ->
                        surfaceRect = Rect(offset = Offset.Zero, size = size.toSize())
                    },
            ) {
                val outerColor = Color.Black.copy(alpha = 0.5f)
                val strokeOutlineColor = ChatTheme.colors.background
                val strokeColor = ChatTheme.colors.onBackground2
                var imageState: AsyncImagePainter.State by remember { mutableStateOf(AsyncImagePainter.State.Empty) }
                SubcomposeAsyncImage(
                    model = imagePath,
                    contentDescription = stringResource(title),
                    filterQuality = FilterQuality.High,
                    onState = { state ->
                        if (state is AsyncImagePainter.State.Error) {
                            viewModel.onError()
                            cropRect = Rect.Zero
                        }
                        imageState = state
                    },
                    modifier = Modifier
                        .alpha(if (state.isBusy) 0.5f else 1f)
                        .fillMaxSize()
                        .draggable2D(
                            enabled = !state.isBusy,
                            state = dragState,
                            onDragStarted = { offset ->
                                val touchRect = Rect(center = offset, radius = dragHandleRadius)
                                dragPos = dragHandles.find { (_, rect) -> rect.overlaps(touchRect) }?.first
                            },
                        )
                        .drawWithContent {
                            if (cropRect.isEmpty) {
                                drawContent()
                                return@drawWithContent
                            }
                            val clipRect = if (forceSquare) cropRect.toSquare() else cropRect
                            drawContent()
                            clipRect(
                                left = clipRect.left,
                                top = clipRect.top,
                                right = clipRect.right,
                                bottom = clipRect.bottom,
                                clipOp = ClipOp.Difference,
                            ) {
                                drawRect(color = outerColor)
                            }
                            val radiusX = Offset(dragHandleRadius, 0f)
                            val radiusY = Offset(0f, dragHandleRadius)
                            drawLine(
                                color = strokeColor,
                                start = cropRect.center - radiusY,
                                end = cropRect.center + radiusY,
                                cap = StrokeCap.Round,
                                strokeWidth = 2f,
                            )
                            drawLine(
                                color = strokeColor,
                                start = cropRect.center - radiusX,
                                end = cropRect.center + radiusX,
                                cap = StrokeCap.Round,
                                strokeWidth = 2f,
                            )
                            drawRect(
                                color = strokeOutlineColor,
                                topLeft = cropRect.topLeft,
                                size = cropRect.size,
                                style = Stroke(width = 4f),
                            )
                            drawRect(
                                color = strokeColor,
                                topLeft = cropRect.topLeft,
                                size = cropRect.size,
                                style = Stroke(width = 2f),
                            )
                        },
                ) {
                    when (imageState) {
                        is AsyncImagePainter.State.Success -> {
                            LaunchedEffect(surfaceRect, painter.intrinsicSize) {
                                val imageSize = painter.intrinsicSize.run {
                                    val scale = DecodeUtils.computeSizeMultiplier(
                                        srcWidth = width,
                                        srcHeight = height,
                                        dstWidth = surfaceRect.width,
                                        dstHeight = surfaceRect.height,
                                        scale = Scale.FIT,
                                    )
                                    this * scale
                                }
                                imageRect = Rect(
                                    offset = Offset(
                                        x = (surfaceRect.width - imageSize.width) / 2f,
                                        y = (surfaceRect.height - imageSize.height) / 2f,
                                    ),
                                    size = imageSize,
                                )
                                cropRect = if (forceSquare) imageRect.toSquare() else imageRect
                            }
                            Image(
                                painter = painter,
                                contentDescription = contentDescription,
                            )
                        }

                        is AsyncImagePainter.State.Error -> {
                            Icon(
                                imageVector = Icons.Default.BrokenImage,
                                contentDescription = contentDescription,
                                tint = ChatTheme.colors.error,
                                modifier = Modifier.size(ChatTheme.dimens.avatarXLarge),
                            )
                        }

                        else -> Unit
                    }
                }
                if (state.isBusy || imageState is AsyncImagePainter.State.Loading) {
                    ChasingDots()
                }
                SimpleSnackbarHost(
                    state = snackbarHostState,
                    backgroundColor = ChatTheme.colors.error,
                    modifier = Modifier.align(Alignment.BottomCenter),
                )
            }
            Row(
                horizontalArrangement = Arrangement.SpaceEvenly,
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextButton(
                    text = stringResource(Res.string.cancel),
                    onClick = onCancel,
                )
                SimpleTextButton(
                    text = stringResource(Res.string.confirm),
                    isEnabled = !cropRect.isEmpty,
                    onClick = {
                        viewModel.cropImage(
                            context = context,
                            params = ImageCropperParams(
                                inPath = imagePath,
                                srcRect = Rect(0f, 0f, surfaceRect.width, surfaceRect.height),
                                cropRect = if (forceSquare) cropRect.toSquare() else cropRect,
                                compressionParams = compressionParams,
                            ),
                            size = maxSize,
                            onFinish = onConfirm,
                        )
                    }
                )
            }
        }
    }
}

private fun Rect.toSquare() = Rect(center = center, minDimension / 2)

private enum class DragPos {
    TOP_LEFT,
    TOP_RIGHT,
    BOTTOM_RIGHT,
    BOTTOM_LEFT,
    TOP_CENTER,
    BOTTOM_CENTER,
    LEFT_CENTER,
    RIGHT_CENTER,
    CENTER,
}