package nand.apps.chat.ui.dialog.image

import androidx.compose.ui.geometry.Rect
import androidx.compose.ui.unit.roundToIntRect
import coil3.PlatformContext
import coil3.SingletonImageLoader
import coil3.request.SuccessResult
import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.image.*
import okio.FileSystem
import okio.Path.Companion.toPath
import kotlin.math.roundToInt

class ImageCropperViewModel(
    private val pixelReader: ImagePixelReader,
    private val fileSystem: FileSystem,
) : ViewModel() {

    val stateData = MutableStateFlow(ImageCropperState())

    fun cropImage(
        context: PlatformContext,
        params: ImageCropperParams,
        size: Int,
        onFinish: (SimpleBitmap) -> Unit,
    ) = with(params) {
        viewModelScope.launch {
            stateData.emit(
                ImageCropperState(
                    isBusy = true,
                    isError = false,
                )
            )
            val rectScale = (srcRect.maxDimension / cropRect.maxDimension) * (size / srcRect.maxDimension)
            val srcRect = srcRect.scale(rectScale)
            val cropRect = cropRect.scale(rectScale)
            val request = pixelReader.createImageRequestBuilder(context)
                .data(inPath)
                .size(srcRect.width.roundToInt(), srcRect.height.roundToInt())
                .build()
            val bitmap = withContext(Dispatchers.IO) {
                (SingletonImageLoader.get(context).execute(request) as? SuccessResult)?.image?.let { image ->
                    val bitmapAr = image.width.toFloat() / image.height.toFloat()
                    val srcAr = srcRect.width / srcRect.height
                    val scale = if (srcAr >= bitmapAr) image.height / srcRect.height else image.width / srcRect.width
                    val offsetX = if (srcAr >= bitmapAr) (srcRect.width * scale - image.width) / 2 else 0f
                    val offsetY = if (srcAr >= bitmapAr) 0f else (srcRect.height * scale - image.height) / 2
                    val bitmapRect = cropRect.scale(scale).translate(-offsetX, -offsetY).roundToIntRect()
                    pixelReader.readPixels(image, bitmapRect)?.let { pixels ->
                        SimpleBitmap(
                            width = bitmapRect.width,
                            height = bitmapRect.height,
                            pixels = pixels,
                        )
                    }?.also { bitmap ->
                        compressionParams?.write(bitmap)
                    }
                }
            }
            bitmap?.let { onFinish(bitmap) }
            stateData.emit(
                ImageCropperState(
                    isBusy = false,
                    isError = bitmap == null,
                )
            )
        }
    }

    fun onError() {
        stateData.tryEmit(
            ImageCropperState(
                isBusy = false,
                isError = true,
            )
        )
    }
    
    private fun Rect.scale(scale: Float) = Rect(
        left = left * scale,
        top = top * scale,
        right = right * scale,
        bottom = bottom * scale,
    )
    
    private fun ImageCompressionParams.write(bitmap: SimpleBitmap) {
        val imageBytes = when (format) {
            ImageFormat.PNG -> bitmap.toPngByteArray()
            ImageFormat.JPEG -> bitmap.toJpegByteArray(quality)
        }
        if (imageBytes.isNotEmpty()) {
            try {
                fileSystem.write(outPath.toPath()) {
                    write(imageBytes)
                }
            } catch (ignore: Exception) {
            }
        }
    }
}