package nand.apps.chat.ui.embed.audio

import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.layout.*
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.LinearProgressIndicator
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.layout.onSizeChanged
import io.ktor.http.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.delay
import kotlinx.coroutines.withContext
import nand.apps.chat.audio.AudioHandle
import nand.apps.chat.audio.AudioPlayer
import nand.apps.chat.io.*
import nand.apps.chat.model.audio.AudioPlaybackState
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.PermanentActionDialog
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.sha256
import nand.apps.chat.util.toDurationString
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun AudioEmbedRenderer(
    uri: String,
    type: ContentType,
    modifier: Modifier = Modifier,
    reuseCache: Boolean = true,
    onDeleteFile: (() -> Unit)? = null,
) {
    if (!uri.isFilePath() || type.toString() != "audio/opus") {
        // Not supported yet
        return
    }
    
    val encryption: ChatEngineEncryption by getKoin().inject()
    val fileSystem: ChatFileSystem by getKoin().inject()
    val audioPlayer: AudioPlayer by getKoin().inject()

    val fileName = rememberSaveable(uri) { uri.toPath().name }
    var playablePath by rememberSaveable(uri) { mutableStateOf<String?>(null) }
    var audioHandle by remember(uri) { mutableStateOf<AudioHandle?>(null) }
    var state by rememberSaveable(uri) { mutableStateOf(AudioPlaybackState.STOPPED) }
    var timestamp by rememberSaveable(uri) { mutableStateOf(0.0) }
    var duration by rememberSaveable(uri) { mutableStateOf(0.0) }
    var seekWidth by remember { mutableStateOf(0.0) }
    var isDeleteDialogVisible by rememberSaveable { mutableStateOf(false) }
    
    LaunchedEffect(state) {
        when (state) {
            AudioPlaybackState.LOADING -> {
                playablePath = copyToCache(
                    path = uri,
                    fileSystem = fileSystem,
                    encryption = encryption,
                    reuseCache = reuseCache,
                )
                state = playablePath?.let { AudioPlaybackState.PLAYING } ?: AudioPlaybackState.STOPPED
            }
            AudioPlaybackState.PLAYING -> {
                audioHandle = playablePath?.let {
                    audioPlayer.queueFile(
                        path = it.toPath(),
                        startTime = timestamp,
                    )
                }
                duration = audioHandle?.duration ?: 0.0
                while (timestamp < duration && audioHandle != null) {
                    timestamp = audioHandle?.timestamp ?: 0.0
                    delay(20)
                }
                if (timestamp >= duration) state = AudioPlaybackState.STOPPED
            }
            AudioPlaybackState.PAUSED -> {
                audioHandle?.let { audioPlayer.stopSound(it) }
            }
            AudioPlaybackState.STOPPED -> {
                audioHandle?.let {
                    audioPlayer.stopSound(it)
                    timestamp = 0.0
                }
                audioHandle = null
            }
        }
    }
    
    // Stop the sound when the player is removed
    DisposableEffect(Unit) {
        onDispose { audioHandle?.let { audioPlayer.stopSound(it) } }
    }
    
    Column(
        modifier = modifier
            .fillMaxWidth()
            .clip(ChatTheme.shapes.medium)
            .background(ChatTheme.colors.surface2)
            .padding(horizontal = ChatTheme.dimens.spacingMedium)
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.fillMaxWidth()
        ) {
            TextLine(
                text = fileName,
                style = ChatTheme.typography.body1,
                modifier = Modifier.weight(1f),
            )
            if (state == AudioPlaybackState.LOADING) {
                CircularProgressIndicator(
                    modifier = Modifier.size(ChatTheme.dimens.iconMedium),
                )
            } else {
                IconButton(
                    icon = when (state) {
                        AudioPlaybackState.PLAYING -> Icons.Default.Pause
                        else -> Icons.Default.PlayArrow
                    },
                    contentDescription = when (state) {
                        AudioPlaybackState.PLAYING -> stringResource(Res.string.pause)
                        else -> stringResource(Res.string.play)
                    },
                    onClick = {
                        state = when (state) {
                            AudioPlaybackState.PLAYING -> AudioPlaybackState.PAUSED
                            AudioPlaybackState.PAUSED -> AudioPlaybackState.PLAYING
                            AudioPlaybackState.STOPPED -> {
                                if (playablePath != null) AudioPlaybackState.PLAYING else AudioPlaybackState.LOADING
                            }
                            else -> state
                        }
                    }
                )
            }
            IconButton(
                icon = Icons.Default.Stop,
                contentDescription = stringResource(Res.string.stop),
                onClick = { state = AudioPlaybackState.STOPPED },
            )
            if (onDeleteFile != null) {
                IconButton(
                    icon = Icons.Default.Delete,
                    contentDescription = stringResource(Res.string.delete),
                    onClick = {
                        state = AudioPlaybackState.STOPPED
                        isDeleteDialogVisible = true
                    },
                )
            }
        }
        Row(
            verticalAlignment = Alignment.CenterVertically,
            horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            modifier = Modifier
                .fillMaxWidth()
                .padding(vertical = ChatTheme.dimens.spacingMedium),
        ) {
            TextLine(
                text = if (duration > 0) timestamp.toInt().toDurationString() else "--:--",
                style = ChatTheme.typography.button,
            )
            LinearProgressIndicator(
                progress = when {
                    duration > 0 -> (timestamp / duration).toFloat()
                    else -> 0f
                },
                modifier = Modifier
                    .weight(1f)
                    .onSizeChanged { seekWidth = it.width.toDouble() }
                    .pointerInput(duration, seekWidth) {
                        detectTapGestures(
                            onPress = { offset ->
                                if (duration > 0) {
                                    val seekTime = (offset.x / seekWidth) * duration
                                    audioHandle?.seekTo(seekTime)
                                    timestamp = seekTime
                                }
                            }
                        )
                    },
            )
            TextLine(
                text = if (duration > 0) duration.toInt().toDurationString() else "--:--",
                style = ChatTheme.typography.button,
            )
        }
    }
    
    if (isDeleteDialogVisible) {
        PermanentActionDialog(
            title = Res.string.delete_audio,
            message = stringResource(Res.string.delete_audio_confirmation),
            onConfirm = {
                fileSystem.deleteFile(uri)
                playablePath?.let { fileSystem.deleteFile(it) }
                playablePath = null
                isDeleteDialogVisible = false
                onDeleteFile?.invoke()
            },
            onCancel = {
                isDeleteDialogVisible = false
            },
        )
    }
}

private suspend fun copyToCache(
    path: String,
    fileSystem: ChatFileSystem,
    encryption: ChatFileEncryption,
    reuseCache: Boolean,
): String? {
    val properPath = path.toPath()
    val fileName = properPath.name
    val tmpDir = fileSystem.getTempDirectory("audio")
    val outName = "${path.sha256()}.${properPath.extension}"
    val tmpFilePath = tmpDir.getFilePath(outName).toString()
    if (reuseCache && outName in tmpDir) return tmpFilePath
    
    val audioDir = fileSystem.getDirectoryForFile(path) ?: return null

    val success = withContext(Dispatchers.IO) {
        val audioData = audioDir.readFromFile(
            fileName = fileName,
            encryption = encryption,
        ) ?: return@withContext false

        tmpDir.writeToFile(
            fileName = outName,
            data = audioData,
            safe = false,
        )
    }
    return if (success) return tmpFilePath else null
}