package nand.apps.chat.ui.embed.fourchan

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.FilterQuality
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.TextDecoration
import androidx.compose.ui.text.style.TextOverflow
import androidx.compose.ui.text.withStyle
import androidx.compose.ui.unit.dp
import coil3.compose.AsyncImage
import nand.apps.chat.embed.fourchan.FourChanPostMetadata
import nand.apps.chat.io.IMAGE_EXTENSIONS
import nand.apps.chat.ui.dialog.image.StaticImagePreviewDialog
import nand.apps.chat.ui.embed.image.GifEmbedRenderer
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.formatFileSize
import nand.apps.chat.util.rememberHtmlAnnotatedString

private val COLOR_SUBJECT = Color(0xFFB294BB)
private val COLOR_TEXT = Color(0xFFC5C8C6)
private val COLOR_LINK = Color(0xFF5F89AC)
private val COLOR_BACKGROUND = Color(0xFF282A2E)

private const val MAX_THUMB_SIZE = 125

@Composable
fun FourChanPostMetadata.Renderer(
    modifier: Modifier = Modifier,
) {
    val isImagePreviewSupported = image != null && image.extension.trimStart('.') in IMAGE_EXTENSIONS
    var isImagePreviewVisible by rememberSaveable { mutableStateOf(false) }
    Column(
        modifier = modifier
            .background(COLOR_BACKGROUND)
            .padding(ChatTheme.dimens.spacingSmall),
    ) {
        TextLine(
            text = buildAnnotatedString {
                if (isOriginalPost && thread.subject != null) {
                    withStyle(SpanStyle(color = COLOR_SUBJECT, fontWeight = FontWeight.Bold)) {
                        append("${thread.subject} ")
                    }
                }
                withStyle(SpanStyle(fontWeight = FontWeight.Bold)) {
                    append(name)
                }
                append(" $timestampString No.$id")
            },
            color = COLOR_TEXT,
            style = ChatTheme.typography.caption,
        )
        image?.let { image ->
            val fileName = when {
                image.isSpoiler -> "Spoiler Image"
                else -> image.baseName.run {
                    if (length > 25) {
                        "${substring(0, 25)}(...)"
                    } else {
                        this
                    } + image.extension
                }
            }
            TextLine(
                text = buildAnnotatedString {
                    append("File: ")
                    withStyle(SpanStyle(color = COLOR_LINK, textDecoration = TextDecoration.Underline)) {
                        append(fileName)
                    }
                    append(" (${formatFileSize(image.size, kbSize = 1024)}, ${image.width}x${image.height})")
                },
                color = COLOR_TEXT,
                style = ChatTheme.typography.caption,
            )
        }
        Row(
            modifier = Modifier.padding(start = ChatTheme.dimens.spacingXLarge),
        ) {
            image?.let { image ->
                val aspectRatio = image.aspectRatio
                val thumbWidth = MAX_THUMB_SIZE * (if (aspectRatio > 1f) 1f else aspectRatio)
                val thumbHeight = MAX_THUMB_SIZE / (if (aspectRatio > 1f) aspectRatio else 1f)
                AsyncImage(
                    model = when {
                        image.isSpoiler -> "https://s.4cdn.org/image/spoiler-${thread.board}1.png"
                        else -> "https://i.4cdn.org/${thread.board}/${timestampMicro}s.jpg"
                    },
                    contentDescription = image.fileName,
                    filterQuality = FilterQuality.High,
                    modifier = Modifier
                        .padding(vertical = ChatTheme.dimens.spacingSmall)
                        .size(thumbWidth.dp, thumbHeight.dp)
                        .onClick(isImagePreviewSupported) { isImagePreviewVisible = true },
                )
            }
            Text(
                text = rememberHtmlAnnotatedString(comment) { element ->
                    when (element.tagName()) {
                        "s" -> SpanStyle(color = Color.Transparent, background = Color.Black)
                        else -> when (element.className()) {
                            "quotelink" -> SpanStyle(color = COLOR_LINK, textDecoration = TextDecoration.Underline)
                            "deadlink" -> SpanStyle(color = COLOR_LINK, textDecoration = TextDecoration.LineThrough)
                            "quote" -> SpanStyle(color = colors.quote)
                            else -> null
                        }
                    }
                },
                color = COLOR_TEXT,
                style = ChatTheme.typography.caption,
                maxLines = 5,
                overflow = TextOverflow.Ellipsis,
                modifier = Modifier.padding(
                    top = ChatTheme.dimens.spacingLarge,
                    bottom = ChatTheme.dimens.spacingLarge,
                    start = ChatTheme.dimens.spacingXLarge,
                    end = ChatTheme.dimens.spacingXLarge,
                ),
            )
        }
    }
    if (isImagePreviewVisible && image != null && isImagePreviewSupported) {
        // TODO - Support webms
        val imageUrl = "https://i.4cdn.org/${thread.board}/${timestampMicro}${image.extension}"
        if (image.extension == ".gif") {
            GifEmbedRenderer(
                uri = imageUrl,
                isFullscreenOnly = true,
                onDismissFullscreen = { isImagePreviewVisible = false },
            )
        } else {
            StaticImagePreviewDialog(
                model = imageUrl,
                contentDescription = image.fileName,
                onDismiss = { isImagePreviewVisible = false },
            )
        }
    }
}