package nand.apps.chat.ui.embed.image

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.Canvas
import androidx.compose.foundation.background
import androidx.compose.foundation.hoverable
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.interaction.collectIsHoveredAsState
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Pause
import androidx.compose.material.icons.filled.PlayArrow
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.graphics.ImageBitmap
import androidx.compose.ui.graphics.drawscope.scale
import androidx.compose.ui.unit.IntSize
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.delay
import kotlinx.coroutines.withContext
import kotlinx.datetime.Clock
import nand.apps.chat.image.SimpleBitmap
import nand.apps.chat.image.toComposeBitmap
import nand.apps.chat.image.x
import nand.apps.chat.net.TempFileCache
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.FullscreenPreviewDialog
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.isUrl
import nand.apps.gif.GifImage
import nand.apps.gif.GifNative
import okio.FileSystem
import okio.Path.Companion.toPath
import okio.use
import org.jetbrains.compose.resources.ExperimentalResourceApi
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

private val GIF_SIZE_CONSTRAINT = 1..1024

@OptIn(ExperimentalResourceApi::class)
@Composable
fun GifEmbedRenderer(
    uri: String,
    modifier: Modifier = Modifier,
    isClickable: Boolean = true,
    isFullscreenOnly: Boolean = false,
    onDismissFullscreen: () -> Unit = {},
) {
    val fileSystem: FileSystem by getKoin().inject()
    val tmpCache : TempFileCache by getKoin().inject()
    
    var isPaused by remember(uri) { mutableStateOf(false) }
    var isFullscreen by remember { mutableStateOf(isFullscreenOnly) }
    var gifSize by remember(uri) { mutableStateOf(IntSize.Zero) }
    var frameCount by remember(uri) { mutableStateOf(0) }
    var currentFrame by remember(uri) { mutableStateOf<ImageBitmap?>(null) }
    val interactionSource = remember { MutableInteractionSource() }
    LaunchedEffect(uri) {
        withContext(Dispatchers.IO) {
            val gifData = try {
                when {
                    uri.isUrl() -> tmpCache.readFromUrl(uri)
                    uri.startsWith("res://") -> Res.readBytes(uri.substringAfter("://"))
                    else -> fileSystem.read(uri.toPath()) { readByteArray() }
                }
            } catch (ignore: Exception) {
                null
            } ?: return@withContext
            GifNative.openGif(gifData)?.use { gif ->
                if (gif.width !in GIF_SIZE_CONSTRAINT || gif.height !in GIF_SIZE_CONSTRAINT || gif.frameCount < 1) {
                    // GIF canvas is empty or way too big
                    return@use
                }
                withContext(Dispatchers.Main) {
                    gifSize = gif.width x gif.height
                    frameCount = gif.frameCount
                }
                val buffer = ByteArray(gif.width * gif.height * gif.bytesPerPixel)
                var frameNum = 0
                while (gif.frameCount > 1 || currentFrame == null) {
                    val readStart = Clock.System.now()
                    val delay = gif.readFrame(frameNum, buffer)
                    val bitmap = gif.frameToBitmap(buffer)
                    withContext(Dispatchers.Main) {
                        currentFrame = bitmap
                        while (isPaused) delay(20L) // TODO - Less hacky way to wait for un-pause
                    }
                    val readTime = Clock.System.now() - readStart
                    val delayMs = delay - readTime.inWholeMilliseconds
                    if (delayMs > 0) delay(delayMs)
                    if (++frameNum == gif.frameCount) {
                        frameNum = 0
                    }
                }
            }
        }
    }

    currentFrame?.let { frameBitmap ->
        if (!isFullscreenOnly) {
            val isHovering by interactionSource.collectIsHoveredAsState()
            GifFrameRenderer(
                gifSize = gifSize,
                frameBitmap = frameBitmap,
                isPauseButtonDisplayed = isClickable && !isPortraitMode() && isHovering && frameCount > 1,
                isPaused = isPaused,
                onTogglePause = { isPaused = !isPaused },
                modifier = modifier
                    .onClick(isClickable) { isFullscreen = true }
                    .hoverable(interactionSource),
            )
        }
        if (isFullscreen) {
            FullscreenPreviewDialog(
                uri = uri,
                onDismiss = {
                    isFullscreen = false
                    onDismissFullscreen()
                },
            ) {
                GifFrameRenderer(
                    gifSize = gifSize,
                    frameBitmap = frameBitmap,
                    isPauseButtonDisplayed = true,
                    isPaused = isPaused,
                    onTogglePause = { isPaused = !isPaused },
                    modifier = Modifier
                        .align(Alignment.Center)
                        .aspectRatio(gifSize.width.toFloat() / gifSize.height)
                        .fillMaxSize(),
                )
            }
        }
    }
}

@Composable
private fun GifFrameRenderer(
    gifSize: IntSize,
    frameBitmap: ImageBitmap,
    isPauseButtonDisplayed: Boolean,
    isPaused: Boolean,
    onTogglePause: () -> Unit,
    modifier: Modifier = Modifier,
) {
    val gifWidth = gifSize.width
    val gifHeight = gifSize.height
    Box(
        modifier = modifier,
    ) {
        Canvas(
            modifier = Modifier
                .align(Alignment.Center)
                .aspectRatio(gifWidth.toFloat() / gifHeight, true)
                .fillMaxSize(),
        ) {
            scale(size.width / gifWidth, size.height / gifHeight, Offset.Zero) {
                drawImage(frameBitmap)
            }
        }
        AnimatedVisibility(
            visible = isPauseButtonDisplayed,
            modifier = Modifier
                .align(Alignment.BottomEnd)
                .size(ChatTheme.dimens.buttonLarge)
                .padding(ChatTheme.dimens.spacingMedium),
        ) {
            IconButton(
                icon = if (isPaused) Icons.Default.PlayArrow else Icons.Default.Pause,
                contentDescription = stringResource(Res.string.pause_gif),
                onClick = onTogglePause,
                modifier = Modifier
                    .background(
                        color = ChatTheme.colors.backgroundFloating,
                        shape = CircleShape,
                    )
                    .fillMaxSize(),
            )
        }
    }
}

private fun GifImage.frameToBitmap(buffer: ByteArray): ImageBitmap {
    return SimpleBitmap(
        width = width,
        height = height,
        pixels = buffer,
    ).toComposeBitmap()
}