package nand.apps.chat.ui.embed.image

import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.unit.dp
import coil3.compose.LocalPlatformContext
import coil3.request.ImageRequest
import nand.apps.chat.ui.dialog.image.StaticImagePreviewDialog
import nand.apps.chat.ui.theme.ChatTheme
import kotlin.math.roundToInt

@Composable
fun ImageEmbedRenderer(
    model: Any,
    contentDescription: String?,
    modifier: Modifier = Modifier,
    type: String? = null,
) {
    var maxWidth by remember { mutableStateOf(0f) }
    Box(
        modifier = modifier
            .fillMaxWidth()
            .onSizeChanged { maxWidth = it.width.toFloat() },
    ) {
        val imageSizes = getImageSizes()
        val maxImageSize = imageSizes.getMaxImageSize(maxWidth).roundToInt()
        val maxSizeDp = LocalDensity.current.run { maxImageSize.toDp() }
        val contentModifier = Modifier.sizeIn(
            maxWidth = maxSizeDp,
            maxHeight = maxSizeDp,
        )
        when {
            model is String && type == "gif" -> {
                GifEmbedRenderer(
                    uri = model,
                    modifier = contentModifier,
                )
            }
            else -> {
                StaticImageRenderer(
                    model = model,
                    contentDescription = contentDescription,
                    maxImageSize = maxImageSize,
                    modifier = contentModifier,
                )
            }
        }
    }
}

@Composable
@NonRestartableComposable
fun ImageEmbedRenderer(
    uri: String,
    type: String,
    modifier: Modifier = Modifier,
) = ImageEmbedRenderer(
    model = uri,
    contentDescription = uri,
    type = type,
    modifier = modifier,
)

@Composable
private fun StaticImageRenderer(
    model: Any,
    contentDescription: String?,
    maxImageSize: Int,
    modifier: Modifier,
) {
    val context = LocalPlatformContext.current
    val request = remember(model, maxImageSize) {
        ImageRequest.Builder(context)
            .data(model)
            .size(maxImageSize, maxImageSize)
            .build()
    }
    var isDialogVisible by remember { mutableStateOf(false) }
    StaticImageRenderer(
        model = request,
        contentDescription = contentDescription,
        modifier = modifier,
        onClick = { isDialogVisible = true },
    )
    if (isDialogVisible) {
        StaticImagePreviewDialog(
            model = model,
            contentDescription = contentDescription,
            onDismiss = { isDialogVisible = false },
        )
    }
}

private data class ImageSizes(
    val density: Float,
    val xLargeThreshold: Float,
    val largeThreshold: Float,
    val mediumThreshold: Float,
    val xLargeImageSize: Float,
    val largeImageSize: Float,
    val mediumImageSize: Float,
    val smallImageSize: Float,
) {
    
    fun getMaxImageSize(maxWidth: Float) = when {
        maxWidth > xLargeThreshold -> xLargeImageSize
        maxWidth > largeThreshold -> largeImageSize
        maxWidth > mediumThreshold -> mediumImageSize
        else -> smallImageSize
    }
}

private val IMAGE_THRESH_XLARGE = 1400.dp
private val IMAGE_THRESH_LARGE = 1000.dp
private val IMAGE_THRESH_MEDIUM = 500.dp
private var imageSizes: ImageSizes? = null

@Composable
private fun getImageSizes(): ImageSizes {
    val density = LocalDensity.current
    val densityValue = density.density
    return imageSizes?.takeIf { it.density == densityValue } ?: density.run {
        ImageSizes(
            density = densityValue,
            xLargeThreshold = IMAGE_THRESH_XLARGE.toPx(),
            largeThreshold = IMAGE_THRESH_LARGE.toPx(),
            mediumThreshold = IMAGE_THRESH_MEDIUM.toPx(),
            xLargeImageSize = ChatTheme.dimens.imageXLarge.toPx(),
            largeImageSize = ChatTheme.dimens.imageLarge.toPx(),
            mediumImageSize = ChatTheme.dimens.imageMedium.toPx(),
            smallImageSize = ChatTheme.dimens.imageSmall.toPx(),
        )
    }
}