package nand.apps.chat.ui.embed.youtube

import androidx.compose.foundation.*
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.interaction.collectIsHoveredAsState
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.material.Icon
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Warning
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.FilterQuality
import androidx.compose.ui.layout.ContentScale
import androidx.compose.ui.text.style.TextOverflow
import androidx.compose.ui.unit.dp
import coil3.compose.AsyncImage
import kotlinx.datetime.*
import kotlinx.datetime.format.*
import nand.apps.chat.embed.youtube.video.THUMBNAIL_URL_ERROR
import nand.apps.chat.embed.youtube.video.YoutubeVideoMetadata
import nand.apps.chat.time.formatRelativeDate
import nand.apps.chat.time.getLocalInstant
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalUriLauncher
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.*

private val THUMBNAIL_WIDTH = 240.dp
private val THUMBNAIL_HEIGHT = 135.dp

private val UPCOMING_TIMESTAMP_FORMAT = LocalDateTime.Format {
    monthNumber(Padding.NONE)
    char('/')
    dayOfMonth(Padding.NONE)
    char('/')
    yearTwoDigits(2000)
    chars(", ")
    amPmHour(Padding.NONE)
    char(':')
    minute()
    char(' ')
    amPmMarker("AM", "PM")
}

@Composable
fun YoutubeVideoMetadata.Renderer(
    modifier: Modifier = Modifier,
) {
    val outerPadding = ChatTheme.dimens.spacingMedium
    val errorHover = remember { MutableInteractionSource() }
    val uriLauncher = LocalUriLauncher.current
    Row(
        modifier = modifier
            .fillMaxWidth()
            .height(THUMBNAIL_HEIGHT + outerPadding * 2)
            .clip(ChatTheme.shapes.medium)
            .background(ChatTheme.colors.surface2)
            .onClick { uriLauncher.launchUri(url) }
            .padding(outerPadding)
    ) {
        Box(
            modifier = Modifier
                .size(THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT)
                .clip(ChatTheme.shapes.small)
                .background(ChatTheme.colors.surface3)
        ) {
            AsyncImage(
                model = if (isAvailable) thumbnailUrl else THUMBNAIL_URL_ERROR,
                contentDescription = title,
                contentScale = if (isAvailable) ContentScale.Crop else ContentScale.Fit,
                filterQuality = FilterQuality.High,
                modifier = Modifier.matchParentSize(),
            )
            val durationLabel = when {
                isLiveNow -> "LIVE"
                isUpcoming -> "UPCOMING"
                isMembersOnly -> "Members only"
                lengthSecs > 0 -> lengthSecs.toDurationString()
                else -> null
            }
            val durationBackground = when {
                isLiveNow -> Color(0.8f, 0f, 0f, 0.9f)
                isMembersOnly -> Color(0xFF121212)
                else -> Color(0f, 0f, 0f, 0.6f)
            }
            durationLabel?.let {
                TextLine(
                    text = it,
                    style = ChatTheme.typography.caption,
                    color = if (isMembersOnly) Color(0xFF2BA640) else Color.White,
                    modifier = Modifier
                        .align(Alignment.BottomEnd)
                        .padding(ChatTheme.dimens.spacingSmall)
                        .clip(ChatTheme.shapes.medium)
                        .background(durationBackground)
                        .padding(
                            horizontal = ChatTheme.dimens.spacingSmall,
                            vertical = ChatTheme.dimens.spacingXSmall,
                        ),
                )
            }
        }
        Column(
            modifier = Modifier.padding(start = ChatTheme.dimens.spacingMedium),
        ) {
            val isHoveringError by errorHover.collectIsHoveredAsState()
            TextLine(
                text = if (isAvailable) title else error?.reason.orEmpty(),
                style = ChatTheme.typography.h6,
                modifier = Modifier.onClick { uriLauncher.launchUri(url) },
            )
            if (channel?.name != null) {
                Row(
                    verticalAlignment = Alignment.CenterVertically,
                    modifier = Modifier
                        .padding(vertical = ChatTheme.dimens.spacingSmall)
                        .onClick(isEnabled = channel.url != null) {
                            uriLauncher.launchUri(channel.url.orEmpty())
                        },
                ) {
                    AsyncImage(
                        model = channel.thumbnailUrl,
                        contentDescription = title,
                        contentScale = if (isAvailable) ContentScale.Crop else ContentScale.Fit,
                        filterQuality = FilterQuality.High,
                        modifier = Modifier
                            .size(ChatTheme.dimens.avatarSmall)
                            .clip(CircleShape)
                            .background(ChatTheme.colors.surface3),
                    )
                    TextLine(
                        text = channel.name,
                        style = ChatTheme.typography.body1,
                        modifier = Modifier.padding(start = ChatTheme.dimens.spacingSmall),
                    )
                }
            }
            val description = when {
                isAvailable && !isHoveringError && !description.isNullOrBlank() -> description
                else -> error?.description
            }
            description?.let {
                Text(
                    text = it,
                    style = ChatTheme.typography.body2,
                    maxLines = 2,
                    overflow = TextOverflow.Ellipsis,
                    color = if (isHoveringError) {
                        ChatTheme.colors.onError
                    } else {
                        ChatTheme.colors.onBackground
                    },
                )
            }
            Spacer(modifier = Modifier.weight(1f))
            if (isAvailable) {
                val viewLabel = when {
                    isLiveNow -> formatViewCount(liveBroadcast?.viewers ?: 0) + " watching"
                    isUpcoming -> formatViewCount(liveBroadcast?.viewers ?: 0) + " waiting"
                    viewCount > 0 -> formatViewCount(viewCount) + formatPlural(" view", viewCount)
                    else -> "No views"
                }
                val dateLabel = when {
                    isUpcoming && isLiveContent -> {
                        val startDate = getLocalInstant(liveBroadcast?.startDate ?: 0)
                        "Scheduled for ${startDate.format(UPCOMING_TIMESTAMP_FORMAT)}"
                    }
                    else -> (if (isLiveContent) "Streamed " else "") + formatRelativeDate(publishDate ?: 0L)
                }
                Row(
                    modifier = Modifier.fillMaxWidth(),
                    verticalAlignment = Alignment.Bottom,
                ) {
                    TextLine(
                        text = if (isLiveNow) viewLabel else "$viewLabel • $dateLabel",
                        style = ChatTheme.typography.body2,
                        color = ChatTheme.colors.onSurface,
                        modifier = Modifier.weight(1f),
                    )
                    if (error != null && !isMembersOnly) {
                        Icon(
                            imageVector = Icons.Default.Warning,
                            contentDescription = error.reason,
                            tint = ChatTheme.colors.error,
                            modifier = Modifier
                                .size(ChatTheme.dimens.iconSmall)
                                .hoverable(errorHover),
                        )
                    }
                }
            }
        }
    }
}