package nand.apps.chat.ui.emoji

import androidx.compose.foundation.background
import androidx.compose.foundation.border
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.LazyRow
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.text.BasicText
import androidx.compose.foundation.text.KeyboardActions
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.Surface
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.unit.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.text.emoji.EmojiData
import nand.apps.chat.text.emoji.EmojiGroup
import nand.apps.chat.text.emoji.EmojiRepo
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.modifier.onEnter
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.spinner.ChasingDots
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.tooltip.TooltipText
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import kotlin.math.ceil
import kotlin.math.min
import seers.composeapp.generated.resources.*

@Composable
fun EmojiPopup(
    size: DpSize,
    inputFocusRequester: FocusRequester,
    onEmoji: (String) -> Unit,
    onDismiss: () -> Unit,
    modifier: Modifier = Modifier,
) {
    val repo: EmojiRepo by getKoin().inject()
    val settingsRepo: SettingsRepo by getKoin().inject()
    val settings = settingsRepo.settings.chatSettings
    val focusManager = LocalFocusManager.current
    val searchFocus = remember { FocusRequester() }
    var isLoading by remember { mutableStateOf(true) }
    var currentGroup by rememberSaveable { mutableStateOf(settings.emojiPopupGroup) }
    val emojis = remember { mutableStateListOf<EmojiData>() }
    var isSearchVisible by rememberSaveable { mutableStateOf(settings.isEmojiPopupSearchEnabled) }
    var searchInput by rememberTextFieldValue()

    BackHandler(onBack = onDismiss)
    
    LaunchedEffect(currentGroup, searchInput.text) {
        val results = repo.getEmojis(currentGroup, searchInput.text)
        emojis.clear()
        emojis.addAll(results)
        isLoading = false
    }
    LaunchedEffect(Unit) {
        inputFocusRequester.requestFocus()
    }
    
    Surface(
        modifier = modifier
            .clip(ChatTheme.shapes.small)
            .background(ChatTheme.colors.background)
            .size(size)
    ) {
        val borderModifier = Modifier.border(
            color = ChatTheme.colors.surface,
            width = 2.dp,
            shape = ChatTheme.shapes.small,
        )
        if (isLoading) {
            Box(
                contentAlignment = Alignment.Center,
                modifier = Modifier.fillMaxSize().then(borderModifier),
            ) {
                ChasingDots()
            }
            return@Surface
        }
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            modifier = borderModifier,
        ) {
            Row(
                verticalAlignment = Alignment.CenterVertically,
                modifier = Modifier.height(ChatTheme.dimens.avatarXLarge),
            ) {
                IconButton(
                    icon = Icons.Outlined.Search,
                    contentDescription = stringResource(Res.string.emoji_search_hint),
                    color = when {
                        isSearchVisible -> ChatTheme.colors.primary
                        else -> ChatTheme.colors.onBackground
                    },
                    size = ChatTheme.dimens.iconMedium,
                    onClick = {
                        isSearchVisible = !isSearchVisible
                        settingsRepo.update { settings.copy(isEmojiPopupSearchEnabled = isSearchVisible) }
                    },
                    modifier = Modifier.padding(start = ChatTheme.dimens.spacingMedium),
                )
                if (isSearchVisible) {
                    LaunchedEffect(Unit) {
                        if (isSearchVisible) {
                            searchFocus.requestFocus()
                        }
                    }
                    SimpleTextField(
                        value = searchInput,
                        onValueChange = { searchInput = it },
                        hintText = stringResource(Res.string.emoji_search_hint),
                        keyboardOptions = KeyboardOptions(imeAction = ImeAction.Search),
                        keyboardActions = KeyboardActions(
                            onSearch = { focusManager.clearFocus() },
                        ),
                        modifier = Modifier
                            .fillMaxWidth()
                            .padding(end = ChatTheme.dimens.spacingLarge)
                            .focusRequester(searchFocus)
                            .onEnter {
                                emojis.firstOrNull()?.let { onEmoji(it.value) }
                                inputFocusRequester.requestFocus()
                            },
                    )
                } else {
                    LazyRow(
                        modifier = Modifier.fillMaxWidth(),
                    ) {
                        items(
                            items = repo.groups.keys.toList(),
                            key = { it },
                        ) { group ->
                            IconButton(
                                icon = group.toIcon(),
                                contentDescription = stringResource(group.toStringResource()),
                                color = when {
                                    currentGroup == group -> ChatTheme.colors.primary
                                    else -> ChatTheme.colors.onBackground
                                },
                                size = ChatTheme.dimens.iconMedium,
                                onClick = {
                                    currentGroup = group
                                    searchInput = TextFieldValue()
                                    inputFocusRequester.requestFocus()
                                    settingsRepo.update { settings.copy(emojiPopupGroup = currentGroup) }
                                },
                            )
                        }
                    }
                }
            }
            Box(
                modifier = Modifier
                    .fillMaxWidth()
                    .height(1.dp)
                    .padding(horizontal = ChatTheme.dimens.spacingLarge)
                    .background(ChatTheme.colors.onBackground2)
            )
            EmojiGrid(
                emojis = emojis,
                maxWidth = size.width,
                onClick = { emoji ->
                    onEmoji(emoji)
                    inputFocusRequester.requestFocus()
                },
            )
        }
    }
}

@Composable
private fun EmojiGrid(
    emojis: List<EmojiData>,
    maxWidth: Dp,
    onClick: (String) -> Unit,
) {
    val isPortrait = isPortraitMode()
    val containerPadding = ChatTheme.dimens.spacingLarge * 2
    val emojiSize = if (isPortrait) ChatTheme.dimens.iconXLarge else ChatTheme.dimens.iconMedium
    val emojiPadding = ChatTheme.dimens.spacingMedium
    val emojisPerRow = ((maxWidth - containerPadding + emojiPadding) / (emojiSize + emojiPadding)).toInt()
    val numRows = ceil(emojis.size / emojisPerRow.toFloat()).toInt()
    val emojiStyle = (if (isPortrait) ChatTheme.typography.h4 else ChatTheme.typography.h5).copy(
        color = ChatTheme.colors.onSurface
    )
    LazyColumn(
        horizontalAlignment = Alignment.CenterHorizontally,
        verticalArrangement = Arrangement.spacedBy(emojiPadding),
        contentPadding = PaddingValues(vertical = emojiPadding),
        modifier = Modifier.fillMaxWidth(),
    ) {
        items(
            count = numRows,
        ) { row ->
            Row(
                horizontalArrangement = Arrangement.spacedBy(emojiPadding),
            ) {
                val start = emojisPerRow * row
                val end = min(start + emojisPerRow, emojis.size)
                for (i in start..<end) {
                    val emoji = emojis[i]
                    Box(
                        contentAlignment = Alignment.Center,
                        modifier = Modifier
                            .size(emojiSize)
                            .onClick { onClick(emoji.value) },
                    ) {
                        TooltipText(
                            tooltip = emoji.description,
                            modifier = Modifier.fillMaxSize(),
                        ) {
                            BasicText(
                                text = emoji.value,
                                style = emojiStyle,
                            )
                        }
                    }
                }
            }
        }
    }
}

@Composable
private fun EmojiGroup.toIcon() = when (this) {
    EmojiGroup.SMILEYS -> Icons.Outlined.EmojiEmotions
    EmojiGroup.PEOPLE -> Icons.Outlined.EmojiPeople
    EmojiGroup.NATURE -> Icons.Outlined.EmojiNature
    EmojiGroup.FOOD -> Icons.Outlined.EmojiFoodBeverage
    EmojiGroup.TRAVEL -> Icons.Outlined.EmojiTransportation
    EmojiGroup.ACTIVITIES -> Icons.Outlined.EmojiEvents
    EmojiGroup.OBJECTS -> Icons.Outlined.EmojiObjects
    EmojiGroup.SYMBOLS -> Icons.Outlined.EmojiSymbols
    EmojiGroup.FLAGS -> Icons.Outlined.EmojiFlags
    EmojiGroup.OTHER -> Icons.Outlined.QuestionMark
}

private fun EmojiGroup.toStringResource() = when (this) {
    EmojiGroup.SMILEYS -> Res.string.emoji_group_smileys
    EmojiGroup.PEOPLE -> Res.string.emoji_group_people
    EmojiGroup.NATURE -> Res.string.emoji_group_nature
    EmojiGroup.FOOD -> Res.string.emoji_group_food
    EmojiGroup.TRAVEL -> Res.string.emoji_group_travel
    EmojiGroup.ACTIVITIES -> Res.string.emoji_group_activities
    EmojiGroup.OBJECTS -> Res.string.emoji_group_objects
    EmojiGroup.SYMBOLS -> Res.string.emoji_group_symbols
    EmojiGroup.FLAGS -> Res.string.emoji_group_flags
    EmojiGroup.OTHER -> Res.string.emoji_group_other
}