package nand.apps.chat.ui.emoji

import androidx.compose.foundation.Image
import androidx.compose.foundation.layout.ExperimentalLayoutApi
import androidx.compose.foundation.layout.FlowRow
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.text.BasicText
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.TextStyle
import io.github.alexzhirkevich.compottie.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import nand.apps.chat.embed.EmbedPreviewFactory
import nand.apps.chat.text.emoji.EmojiRepo
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.theme.ChatTheme
import org.koin.compose.getKoin

private const val MAX_EMOJIS = 5 // Want to keep this low for network traffic and render performance

@Composable
fun EmojiRenderer(
    text: AnnotatedString,
    modifier: Modifier = Modifier,
) {
    val repo: EmojiRepo by getKoin().inject()    
    val segments = remember { mutableStateListOf<Pair<String, String?>>() }
    val isAnimatedEmojisEnabled = LocalSettingsProvider.current.chatSettings.isAnimatedEmojisEnabled
    LaunchedEffect(text) {
        val emojiSegments = repo.getAnimatedEmojiSegments(text)
        val numEmojis = emojiSegments.count { it.first.isNotBlank() }
        segments.clear()
        if (numEmojis <= MAX_EMOJIS) {
            segments += emojiSegments
        }
    }

    val textStyle = TextStyle(
        color = ChatTheme.colors.onSurface,
        fontSize = ChatTheme.dimens.emojiFontSize,
        lineHeight = ChatTheme.dimens.emojiFontSize,
    )
    if (isAnimatedEmojisEnabled && segments.isNotEmpty()) {
        AnimatedEmojiRenderer(
            segments = segments,
            textStyle = textStyle,
            modifier = modifier,
        )
    } else {
        Text(
            text = text,
            style = textStyle,
        )
    }
}

@OptIn(ExperimentalLayoutApi::class)
@Composable
private fun AnimatedEmojiRenderer(
    segments: List<Pair<String, String?>>,
    textStyle: TextStyle,
    modifier: Modifier,
) {
    val embedFactory: EmbedPreviewFactory by getKoin().inject()
    FlowRow(
        modifier = modifier,
    ) {
        segments.forEach { segment ->
            val (emoji, code) = segment
            code?.let {
                val composition = rememberLottieComposition {
                    LottieCompositionSpec.Url(
                        url = "https://fonts.gstatic.com/s/e/notoemoji/latest/${code.lowercase().replace(' ', '_')}/lottie.json",
                        format = LottieAnimationFormat.Json,
                        request = { url -> embedFactory.client.get(url).body() },
                    )
                }
                if (composition.isComplete) {
                    Image(
                        painter = rememberLottiePainter(
                            composition = composition.value,
                            iterations = Compottie.IterateForever
                        ),
                        contentDescription = emoji,
                        modifier = Modifier.size(ChatTheme.dimens.emojiImageSize),
                    )
                    return@forEach
                }
            }
            BasicText(
                text = emoji,
                style = textStyle,
            )
        }
    }
}