package nand.apps.chat.ui.file

import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.text.KeyboardActions
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.DisabledVisible
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.unit.dp
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.ConfirmationDialog
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.modifier.onEnter
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.spinner.ChasingDots
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.toPx
import nand.apps.chat.ui.viewmodel.composableViewModel
import org.jetbrains.compose.resources.stringResource
import org.koin.core.parameter.parametersOf
import seers.composeapp.generated.resources.*

private val PREVIEW_COLUMN_WIDTH_SMALL = 220.dp
private val PREVIEW_COLUMN_WIDTH_LARGE = 350.dp
private val PREVIEW_COLUMN_WIDTH_THRESHOLD = 1280.dp

private val SHORTCUTS_WIDTH = 160.dp

private val FILE_SIZE_WIDTH = 80.dp
private val FILE_DATE_WIDTH = 100.dp

@OptIn(ExperimentalFoundationApi::class)
@Composable
fun ChatFilePicker(
    options: ChatFilePickerOptions,
    onConfirm: (String) -> Unit,
    onCancel: () -> Unit,
    modifier: Modifier = Modifier,
    snackbarHostState: SnackbarHostState? = null,
    viewModel: ChatFilePickerViewModel = composableViewModel(
        parameters = { parametersOf(options) },
    ),
) {
    val state by viewModel.stateData.collectAsState()
    val focusManager = LocalFocusManager.current
    var textInput by rememberTextFieldValue(options.fileName.orEmpty())

    // Responsive design
    var isPortrait by remember { mutableStateOf(false) }
    var previewColumnWidth by remember { mutableStateOf(0.dp) }
    
    LaunchedEffect(Unit) {
        viewModel.refresh()
        viewModel.actionData.collect { action ->
            when (action) {
                is ConfirmFileAction -> onConfirm(action.path)
                is CancelFilePickerAction -> onCancel()
            }
        }
    }
    
    BackHandler {
        viewModel.onBackButton()
    }
    
    val confirmFile: () -> Unit = {
        state.selectedFile?.let { viewModel.confirmFile(it) } ?: viewModel.confirmFile(textInput.text)
    }
    
    val previewColumnWidthThresholdPx = PREVIEW_COLUMN_WIDTH_THRESHOLD.toPx()
    Box(
        modifier = modifier.onSizeChanged { size ->
            val width = size.width
            val aspectRatio = width.toFloat() / size.height
            isPortrait = aspectRatio < 1f
            previewColumnWidth = when {
                isPortrait -> 0.dp
                width < previewColumnWidthThresholdPx -> PREVIEW_COLUMN_WIDTH_SMALL
                else -> PREVIEW_COLUMN_WIDTH_LARGE
            }
        },
    ) {
        val isFileSelectable = when {
            options.isOpenDialog -> state.selectedFile?.let { file ->
                !file.isDirectory || options.mode != ChatFilePickerMode.FILES_ONLY
            } ?: false
            else -> textInput.text.isNotBlank() && (options.extensions == null ||
                    textInput.text.substringAfterLast('.').lowercase() in options.extensions)
        }
        Column {
            Row(
                verticalAlignment = Alignment.CenterVertically,
                modifier = Modifier
                    .background(ChatTheme.colors.surface2)
                    .padding(ChatTheme.dimens.spacingSmall)
                    .fillMaxWidth(),
            ) {
                ChatFileDirectoryRow(
                    directory = state.directory,
                    textStyle = ChatTheme.typography.body1,
                    onClick = { viewModel.setDirectory(it) },
                    modifier = Modifier
                        .weight(1f)
                        .padding(start = ChatTheme.dimens.spacingMedium),
                )
                IconButton(
                    icon = Icons.Default.DisabledVisible,
                    color = if (state.showHiddenFiles) ChatTheme.colors.icon else ChatTheme.colors.onBackground2,
                    contentDescription = null,
                    size = ChatTheme.dimens.iconMedium,
                    onClick = viewModel::toggleHiddenFiles,
                )
            }
            SimpleTextField(
                value = textInput,
                hintText = stringResource(Res.string.file_picker_filter_hint),
                onValueChange = {
                    textInput = it
                    if (options.isOpenDialog) viewModel.search(it.text)
                },
                keyboardOptions = KeyboardOptions(
                    imeAction = if (options.isOpenDialog) ImeAction.Search else ImeAction.Done,
                ),
                keyboardActions = when {
                    options.isOpenDialog -> KeyboardActions(
                        onSearch = {
                            viewModel.confirmFile(textInput.text)
                            focusManager.clearFocus()
                        },
                    )
                    else -> KeyboardActions(
                        onDone = {
                            viewModel.confirmFile(textInput.text)
                            focusManager.clearFocus()
                        },
                    )
                },
                modifier = Modifier
                    .fillMaxWidth()
                    .onEnter { viewModel.confirmFile(textInput.text) },
            )
            Row(
                modifier = Modifier
                    .fillMaxWidth()
                    .weight(1f)
            ) {
                ChatFileShortcutsColumn(
                    iconsOnly = isPortrait,
                    isEnabled = !state.isLoading,
                    modifier = Modifier
                        .background(ChatTheme.colors.surface)
                        .fillMaxHeight()
                        .appendIf(!isPortrait) { width(SHORTCUTS_WIDTH) },
                    onSelectDirectory = { viewModel.setDirectory(it) },
                )
                val rowVerticalPadding = if (isPortrait) ChatTheme.dimens.spacingMedium else ChatTheme.dimens.spacingSmall
                val fileDateWidth = (if (isPortrait) 0.dp else FILE_DATE_WIDTH)
                LazyColumn(
                    modifier = Modifier.weight(1f),
                ) {
                    stickyHeader {
                        ChatFileSortOptionsRow(
                            sortType = state.sortType,
                            fileSizeWidth = FILE_SIZE_WIDTH,
                            fileDateWidth = fileDateWidth,
                            onSelected = viewModel::setSortType,
                            modifier = Modifier.background(ChatTheme.colors.background),
                            optionModifier = Modifier.padding(
                                horizontal = ChatTheme.dimens.spacingMedium,
                                vertical = rowVerticalPadding,
                            ),
                        )
                    }
                    items(
                        items = state.files,
                        key = { it.name },
                    ) { file ->
                        ChatFileRow(
                            file = file,
                            color = if (state.selectedFile == file) ChatTheme.colors.onSurface else ChatTheme.colors.onBackground,
                            fileSizeWidth = FILE_SIZE_WIDTH,
                            fileDateWidth = fileDateWidth,
                            modifier = Modifier
                                .fillMaxWidth()
                                .clickable(enabled = !state.isLoading) {
                                    viewModel.selectFile(file)
                                }
                                .background(if (state.selectedFile == file) ChatTheme.colors.primaryVariant else Color.Unspecified)
                                .padding(
                                    horizontal = ChatTheme.dimens.spacingMedium,
                                    vertical = rowVerticalPadding,
                                ),
                        )
                    }
                }
                state.selectedFile?.let { file ->
                    ChatFilePreviewColumn(
                        file = file,
                        modifier = Modifier.background(ChatTheme.colors.surface)
                            .width(previewColumnWidth)
                            .fillMaxHeight()
                            .padding(ChatTheme.dimens.spacingMedium),
                    )
                }
            }
            if (isPortrait) {
                ChatFileActionButtonRow(
                    onConfirm = confirmFile,
                    onCancel = onCancel,
                    isConfirmEnabled = isFileSelectable,
                    options = options,
                )
            }
        }
        if (!isPortrait) {
            Surface(
                modifier = Modifier
                    .width(previewColumnWidth)
                    .background(ChatTheme.colors.background)
                    .align(Alignment.BottomEnd),
            ) {
                ChatFileActionButtonRow(
                    onConfirm = confirmFile,
                    onCancel = onCancel,
                    isConfirmEnabled = isFileSelectable,
                    options = options,
                )
            }
        }
        if (state.isLoading) {
            ChasingDots(
                color = ChatTheme.colors.primaryVariant,
                modifier = Modifier
                    .align(Alignment.Center)
                    .matchParentSize()
            )
        }
        snackbarHostState?.let {
            SimpleSnackbarHost(
                state = snackbarHostState,
                backgroundColor = ChatTheme.colors.error,
                modifier = Modifier.align(Alignment.BottomCenter),
            )
        }
    }
    
    state.overwriteFile?.let { file ->
        ConfirmationDialog(
            title = Res.string.file_picker_overwrite_title,
            message = stringResource(Res.string.file_picker_overwrite_message, file.name),
            onConfirm = { viewModel.confirmFile(file.name, isOverwriteAllowed = true) },
            onCancel = viewModel::dismissOverwriteDialog,
        )
    }
}
