package nand.apps.chat.ui.file

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.AttachFile
import androidx.compose.material.icons.outlined.Image
import androidx.compose.material.icons.outlined.PhotoCamera
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.compose.ui.window.Dialog
import nand.apps.chat.camera.CameraLauncher
import nand.apps.chat.image.ImageCompressionParams
import nand.apps.chat.image.ImageFormat
import nand.apps.chat.io.ChatFileBrowser
import nand.apps.chat.io.ChatFilePaths
import nand.apps.chat.io.nameWithoutExtension
import nand.apps.chat.model.event.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.dialog.image.ImageCropperDialog
import nand.apps.chat.ui.menu.ActionMenuRow
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import okio.FileSystem
import okio.Path
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

private const val MAX_PHOTO_SIZE = 1024

@Composable
fun ChatFileSenderDialog(
    onSelectFile: (String) -> Unit,
    onDismiss: () -> Unit,
) {
    val settingsRepo: SettingsRepo by getKoin().inject()
    val filePaths: ChatFilePaths by getKoin().inject()
    val fileSystem: FileSystem by getKoin().inject()
    val fileBrowser: ChatFileBrowser by getKoin().inject()
    val cameraLauncher: CameraLauncher by getKoin().inject()
    val eventHandler = LocalEventHandler.current
    
    var isFilePickerVisible by rememberSaveable {
        mutableStateOf(!cameraLauncher.isSupported &&
                !fileBrowser.isBrowserSupported() &&
                !fileBrowser.isGallerySupported())
    }
    var cropImagePath by rememberSaveable { mutableStateOf("") }
    
    if (cameraLauncher.isSupported) {
        LaunchedEffect(Unit) {
            eventHandler.onEvent.collect { event ->
                if (event is TakePhoto) {
                    cropImagePath = event.path.toString()
                }
            }
        }
    }

    val confirmFile: (Path) -> Unit = { path ->
        settingsRepo.update {
            chatSettings.copy(defaultSendDirectory = path.parent?.toString())
        }
        onSelectFile(path.toString())
    }
    
    when {
        isFilePickerVisible -> {
            val defaultDir = settingsRepo.settings.chatSettings.defaultSendDirectory ?: filePaths.picturesPath.toString()
            ChatFilePickerDialog(
                options = ChatFilePickerOptions(directory = defaultDir),
                onConfirm = { confirmFile(it.toPath()) },
                onCancel = onDismiss,
            )
        }
        cropImagePath.isNotBlank() -> {
            val outPath = remember(cropImagePath) {
                cropImagePath.toPath().run {
                    parent?.resolve("$nameWithoutExtension-cropped.jpg")?.toString()
                } ?: cropImagePath
            }
            ImageCropperDialog(
                title = Res.string.crop_image,
                imagePath = cropImagePath,
                maxSize = MAX_PHOTO_SIZE,
                forceSquare = false,
                compressionParams = ImageCompressionParams(
                    outPath = outPath,
                    format = ImageFormat.JPEG,
                    quality = 0.9f,
                ),
                onConfirm = {
                    if (cropImagePath != outPath) {
                        fileSystem.delete(cropImagePath.toPath())
                    }
                    if (fileSystem.exists(outPath.toPath())) {
                        onSelectFile(outPath)
                    } else {
                        onDismiss()
                    }
                },
                onCancel = onDismiss,
            )
        }
        else -> Dialog(
            onDismissRequest = onDismiss,
        ) {
            Column(
                modifier = Modifier
                    .background(
                        color = ChatTheme.colors.background,
                        shape = ChatTheme.shapes.small,
                    )
                    .fillMaxWidth()
            ) {
                ActionMenuRow(
                    icon = Icons.Outlined.AttachFile,
                    label = stringResource(Res.string.chat_screen_send_file),
                    onClick = {
                        if (fileBrowser.isBrowserSupported()) {
                            fileBrowser.launchBrowser { path ->
                                path?.let { confirmFile(it) }
                            }
                        } else {
                            isFilePickerVisible = true
                        }
                    },
                )
                if (fileBrowser.isGallerySupported()) {
                    ActionMenuRow(
                        icon = Icons.Outlined.Image,
                        label = stringResource(Res.string.chat_screen_send_image),
                        onClick = {
                            fileBrowser.launchGallery { path ->
                                cropImagePath = path?.toString().orEmpty()
                            }
                        },
                    )
                }
                if (cameraLauncher.isSupported) {
                    ActionMenuRow(
                        icon = Icons.Outlined.PhotoCamera,
                        label = stringResource(Res.string.chat_screen_take_photo),
                        onClick = {
                            val dir = filePaths.picturesPath.resolve(".tmp")
                            fileSystem.createDirectories(dir)
                            cameraLauncher.launchForPhoto(dir.resolve("${currentTimeMillis()}.jpg"))
                        },
                    )
                }
            }
        }
    }
}