package nand.apps.chat.ui.friends

import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.PhotoCamera
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardCapitalization
import androidx.compose.ui.text.input.KeyboardType
import androidx.compose.ui.text.input.TextFieldValue
import nand.apps.chat.camera.CameraLauncher
import nand.apps.chat.model.event.*
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.friends.AddFriendDialogAction.*
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.SimpleTextFieldButton
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.remove
import nand.apps.chat.util.takeIfUnderLength
import org.jetbrains.compose.resources.stringArrayResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun AddFriendDialog(
    onDismiss: () -> Unit,
    viewModel: AddFriendDialogViewModel = composableViewModel(),
) {
    var name by rememberTextFieldValue()
    var address by rememberTextFieldValue()
    var message by rememberTextFieldValue(
        initialValue = stringArrayResource(Res.array.add_friend_dialog_message_defaults)
            .random()
            .replace("%s", viewModel.selfName),
    )
    var hasAddressError by remember { mutableStateOf(false) }
    val errorSnackbars = remember { SnackbarHostState() }
    val successSnackbars = remember { SnackbarHostState() }
    val constraints = LocalChatEngineConstraints.current
    val eventHandler = LocalEventHandler.current
    val cameraLauncher: CameraLauncher by getKoin().inject()

    val genericErrorMessage = stringResource(Res.string.add_friend_dialog_error_message)
    val qrCodeErrorMessage = stringResource(Res.string.add_friend_dialog_qr_photo_error)
    val qrCodeSuccessMessage = stringResource(Res.string.add_friend_dialog_qr_photo_success)

    val addressRegex = remember(constraints) {
        Regex("[${constraints.uidCharacterRange}]+")
    }
    val addressRegexInv = remember(constraints) {
        Regex("[^${constraints.uidCharacterRange}]+")
    }
    
    // View model action handler
    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is ExitDialog -> onDismiss()
                is ShowError -> {
                    hasAddressError = true
                    errorSnackbars.showSnackbar(genericErrorMessage)
                }
            }
        }
    }
    
    // Listen for QR code updates from camera
    if (getPlatform().isMobile) {
        LaunchedEffect(Unit) {
            eventHandler.onEvent.collect { event ->
                if (event is TakePhotoQrCode) {
                    val code = event.code.substringAfter(':') // qTox puts "tox:" before the address in QR code
                    if (code.matches(addressRegex) && code.length == constraints.addressLength) {
                        address = TextFieldValue(code)
                        successSnackbars.showSnackbar(qrCodeSuccessMessage)
                    } else {
                        errorSnackbars.showSnackbar(qrCodeErrorMessage)
                    }
                }
            }
        }
    }
    
    val defaultName = stringResource(Res.string.friend_request)
    SimpleDialog(
        title = Res.string.add_friend_dialog_title,
        isConfirmEnabled = addressRegex.matches(address.text) && message.text.isNotBlank(),
        onConfirm = {
            viewModel.addFriend(
                name = name.text.takeIf { it.isNotBlank() } ?: defaultName,
                address = address.text,
                message = message.text,
            )
        },
        onCancel = onDismiss,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
        ) {
            SimpleTextField(
                value = name,
                onValueChange = { name = it.takeIfUnderLength(constraints.maxNameLength, name) },
                labelText = stringResource(Res.string.add_friend_dialog_name_label),
                keyboardOptions = KeyboardOptions(
                    autoCorrect = false,
                    imeAction = ImeAction.Next,
                ),
                modifier = Modifier.fillMaxWidth(),
            )
            Row(
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextField(
                    value = address,
                    onValueChange = {
                        address = it
                            .takeIfUnderLength(constraints.addressLength, address)
                            .remove(addressRegexInv)
                    },
                    labelText = stringResource(Res.string.add_friend_dialog_address_label),
                    isError = hasAddressError,
                    keyboardOptions = KeyboardOptions(
                        autoCorrect = false,
                        capitalization = KeyboardCapitalization.Characters,
                        keyboardType = KeyboardType.Ascii,
                        imeAction = ImeAction.Next,
                    ),
                    modifier = Modifier.weight(1f),
                )
                if (cameraLauncher.isSupported) {
                    SimpleTextFieldButton(
                        buttonIcon = Icons.Outlined.PhotoCamera,
                        buttonDescription = stringResource(Res.string.add_friend_dialog_qr_photo),
                        onClick = { cameraLauncher.launchForQrCode() },
                        modifier = Modifier.padding(
                            start = ChatTheme.dimens.spacingMedium,
                            top = ChatTheme.dimens.spacingMedium
                        ),
                    )
                }
            }
            SimpleTextField(
                value = message,
                onValueChange = { message = it.takeIfUnderLength(constraints.maxFriendRequestLength, message) },
                labelText = stringResource(Res.string.add_friend_dialog_message_label),
                modifier = Modifier.fillMaxWidth(),
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbars,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
        SimpleSnackbarHost(
            state = successSnackbars,
            backgroundColor = ChatTheme.colors.secondary,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
    }
}