package nand.apps.chat.ui.friends

import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import nand.apps.chat.model.call.getActions
import nand.apps.chat.model.event.*
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.contacts.ContactRow
import nand.apps.chat.ui.dialog.ContactProfileDialog
import nand.apps.chat.ui.dialog.PermanentActionDialog
import nand.apps.chat.ui.groups.SendGroupInviteDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun FriendContactRow(
    friend: UserData,
    chatState: ChatScreenState,
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    unreadCount: UnreadMessageCounts? = null,
) {
    val eventHandler = LocalEventHandler.current
    var isRemovalDialogVisible by rememberSaveable { mutableStateOf(false) }
    var isGroupInviteDialogVisible by rememberSaveable { mutableStateOf(false) }
    var isProfileDialogVisible by rememberSaveable { mutableStateOf(false) }
    var isClearMessagesDialogVisible by rememberSaveable { mutableStateOf(false) }
    ContactRow(
        heading = friend.displayName,
        subHeading = friend.statusMessage,
        avatar = friend.avatar,
        status = friend.status,
        isOnline = friend.isOnline,
        sidebarState = sidebarState,
        unreadCount = unreadCount,
        call = chatState.calls[friend.uid],
        iconOverlay = when {
            friend.isMuted -> Icons.Default.NotificationsPaused
            else -> null
        },
        iconOverlayColor = ChatTheme.colors.icon,
        onClick = { eventHandler += OpenConversationEvent(friend) },
        modifier = modifier,
        isActionMenuEnabled = true,
        actionMenuItems = {
            mutableListOf<ActionMenuItem>().apply {
                chatState.calls[friend.uid]?.takeIf { it.isInProgress }?.let { call ->
                    call.status.getActions().forEach { action ->
                        this += ActionMenuItem(
                            icon = action.icon,
                            label = getStringBlocking(action.label),
                            onClick = { eventHandler += action.getEvent(call.contactUid) },
                        )
                    }
                }
                if (friend.isOnline && chatState.groups.isNotEmpty()) {
                    this += ActionMenuItem(
                        icon = Icons.Default.GroupAdd,
                        label = getStringBlocking(Res.string.invite_to_group),
                        onClick = { isGroupInviteDialogVisible = true },
                    )
                }
                this += ActionMenuItem(
                    icon = Icons.Default.PersonRemove,
                    label = getStringBlocking(Res.string.remove_friend),
                    onClick = { isRemovalDialogVisible = true },
                )
                this += ActionMenuItem(
                    icon = Icons.Default.ContactPage,
                    label = getStringBlocking(Res.string.view_contact_profile),
                    onClick = { isProfileDialogVisible = true },
                )
                if (friend.isMuted) {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsActive,
                        label = getStringBlocking(Res.string.unmute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(friend.uid, isMuted = false)
                        },
                    )
                } else {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsPaused,
                        label = getStringBlocking(Res.string.mute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(friend.uid, isMuted = true)
                        },
                    )
                }
                this += ActionMenuItem(
                    icon = Icons.Default.ClearAll,
                    label = getStringBlocking(Res.string.clear_messages),
                    onClick = { isClearMessagesDialogVisible = true }
                )
            }
        }
    )
    when {
        isRemovalDialogVisible -> PermanentActionDialog(
            title = Res.string.remove_friend,
            message = stringResource(Res.string.remove_friend_confirmation, friend.displayName),
            onConfirm = {
                isRemovalDialogVisible = false
                eventHandler += RemoveFriendEvent(friend)
            },
            onCancel = { isRemovalDialogVisible = false },
        )
        isGroupInviteDialogVisible -> SendGroupInviteDialog(
            friend = friend,
            groups = chatState.groups.values.toList(),
            onConfirm = { group ->
                isGroupInviteDialogVisible = false
                eventHandler += SendGroupInviteEvent(friend, group)
            },
            onCancel = { isGroupInviteDialogVisible = false },
        )
        isProfileDialogVisible -> ContactProfileDialog(
            user = friend,
            isSelf = false,
            onDismiss = { isProfileDialogVisible = false },
        )
        isClearMessagesDialogVisible -> PermanentActionDialog(
            title = Res.string.clear_messages,
            message = stringResource(Res.string.clear_messages_confirmation),
            onConfirm = {
                isClearMessagesDialogVisible = false
                eventHandler += ClearMessagesEvent(friend.uid)
            },
            onCancel = { isClearMessagesDialogVisible = false },
        )
    }
}