package nand.apps.chat.ui.groups

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.*
import androidx.compose.material.Checkbox
import androidx.compose.material.SnackbarHostState
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import nand.apps.chat.model.group.ChatGroupType
import nand.apps.chat.model.group.ChatGroupVisibility
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.groups.CreateGroupDialogAction.*
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.SimpleTextRadioButton
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.takeIfUnderLength
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun CreateGroupDialog(
    onDismiss: () -> Unit,
    viewModel: CreateGroupDialogViewModel = composableViewModel(),
) {
    var title by rememberTextFieldValue()
    var type by rememberSaveable { mutableStateOf(ChatGroupType.TEXT_ONLY) }
    var visibility by rememberSaveable { mutableStateOf(ChatGroupVisibility.PRIVATE) }
    var hasError by remember { mutableStateOf(false) }
    val snackbarHostState = remember { SnackbarHostState() }
    val constraints = LocalChatEngineConstraints.current

    val errorMessage = stringResource(Res.string.create_group_dialog_error_message)
    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is ExitDialog -> onDismiss()
                is ShowError -> {
                    hasError = true
                    snackbarHostState.showSnackbar(errorMessage)
                }
            }
        }
    }
    
    SimpleDialog(
        title = Res.string.create_group_dialog_title,
        isConfirmEnabled = title.text.isNotBlank(),
        onConfirm = {
            viewModel.createGroup(
                title = title.text,
                type = type,
                visibility = when (type) {
                    ChatGroupType.ADVANCED -> visibility
                    else -> ChatGroupVisibility.PRIVATE
                },
            )
        },
        onCancel = onDismiss,
    ) {
        Column {
            SimpleTextField(
                value = title,
                onValueChange = { title = it.takeIfUnderLength(constraints.maxGroupNameLength, title) },
                labelText = stringResource(Res.string.create_group_dialog_title_label),
                modifier = Modifier.fillMaxWidth(),
            )
            SimpleTextRadioButton(
                text = stringResource(Res.string.create_group_dialog_type_text),
                selected = type == ChatGroupType.TEXT_ONLY,
                onClick = { type = ChatGroupType.TEXT_ONLY },
            )
            SimpleTextRadioButton(
                text = stringResource(Res.string.create_group_dialog_type_av),
                selected = type == ChatGroupType.AV,
                onClick = { type = ChatGroupType.AV },
            )
            SimpleTextRadioButton(
                text = stringResource(Res.string.create_group_dialog_type_advanced),
                selected = type == ChatGroupType.ADVANCED,
                onClick = { type = ChatGroupType.ADVANCED },
            )
            AnimatedVisibility(
                visible = type == ChatGroupType.ADVANCED,
            ) {
                Row(verticalAlignment = Alignment.CenterVertically) {
                    Checkbox(
                        checked = visibility == ChatGroupVisibility.PUBLIC,
                        onCheckedChange = { checked ->
                            visibility = if (checked) {
                                ChatGroupVisibility.PUBLIC
                            } else {
                                ChatGroupVisibility.PRIVATE
                            }
                        },
                    )
                    TextLine(
                        text = stringResource(Res.string.create_group_dialog_visibility_public),
                        style = ChatTheme.typography.body2,
                    )
                }
            }
        }
        SimpleSnackbarHost(
            state = snackbarHostState,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
    }
}