package nand.apps.chat.ui.groups

import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.Logout
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.call.getActions
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.contacts.ContactRow
import nand.apps.chat.ui.dialog.PermanentActionDialog
import nand.apps.chat.ui.dialog.EditContactDialog
import nand.apps.chat.ui.dialog.PasswordDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
@NonRestartableComposable
fun GroupContactRow(
    group: ChatGroupData,
    selfPeer: UserData,
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    unreadCount: UnreadMessageCounts? = null,
    call: CallData? = null,
) {
    val eventHandler = LocalEventHandler.current
    val constraints = LocalChatEngineConstraints.current
    var dialogType by rememberSaveable { mutableStateOf<GroupRowDialog?>(null) }
    val renameGroupLabel = stringResource(Res.string.rename_group)
    val setDescLabel = stringResource(Res.string.set_group_description)
    val setPasswordLabel = stringResource(Res.string.set_group_password)
    val leaveGroupLabel = stringResource(Res.string.leave_group)
    ContactRow(
        heading = group.title.ifBlank { stringResource(Res.string.group_untitled) },
        subHeading = group.description.ifBlank {
            stringResource(Res.string.group_chat_peers_online, group.numPeersOnline, group.peers.size)
        },
        status = UserStatus.DEFAULT,
        isOnline = group.isOnline,
        iconOverlay = when {
            group.error != null -> Icons.Default.ErrorOutline
            group.isMuted -> Icons.Default.Snooze
            else -> null
        },
        iconOverlayColor = when {
            group.error != null -> ChatTheme.colors.warning
            else -> ChatTheme.colors.icon
        },
        avatar = group.avatar,
        unreadCount = unreadCount,
        sidebarState = sidebarState,
        onClick = { eventHandler += OpenConversationEvent(group) },
        modifier = modifier,
        call = call,
        isActionMenuEnabled = true,
        actionMenuItems = {
            mutableListOf<ActionMenuItem>().apply {
                call?.takeIf { it.isInProgress }?.let { call ->
                    call.status.getActions().forEach { action ->
                        this += ActionMenuItem(
                            icon = action.icon,
                            label = getStringBlocking(action.label),
                            onClick = { eventHandler += action.getEvent(call.contactUid) },
                        )
                    }
                }
                if (group.isTitleEditable) {
                    this += ActionMenuItem(
                        icon = Icons.Default.Edit,
                        label = renameGroupLabel,
                        onClick = { dialogType = GroupRowDialog.RENAME },
                    )
                }
                if (group.isDescEditable && selfPeer.role != ChatGroupRole.OBSERVER) {
                    this += ActionMenuItem(
                        icon = Icons.Default.Topic,
                        label = setDescLabel,
                        onClick = { dialogType = GroupRowDialog.DESCRIPTION },
                    )
                }
                if (ChatGroupFeature.PASSWORD in group.features && selfPeer.role == ChatGroupRole.FOUNDER) {
                    this += ActionMenuItem(
                        icon = Icons.Default.Password,
                        label = setPasswordLabel,
                        onClick = { dialogType = GroupRowDialog.PASSWORD },
                    )
                }
                if (group.isMuted) {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsActive,
                        label = getStringBlocking(Res.string.unmute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(group.uid, isMuted = true)
                        },
                    )
                } else {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsPaused,
                        label = getStringBlocking(Res.string.mute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(group.uid, isMuted = false)
                        },
                    )
                }
                this += ActionMenuItem(
                    icon = Icons.Default.ClearAll,
                    label = getStringBlocking(Res.string.clear_messages),
                    onClick = { dialogType = GroupRowDialog.CLEAR_MESSAGES },
                )
                this += ActionMenuItem(
                    icon = Icons.AutoMirrored.Default.Logout,
                    label = leaveGroupLabel,
                    onClick = { dialogType = GroupRowDialog.REMOVE },
                )
            }
        }
    )
    when (dialogType) {
        GroupRowDialog.RENAME -> EditContactDialog(
            title = Res.string.rename_group,
            label = Res.string.group_title,
            maxLength = constraints.maxGroupNameLength,
            initialValue = group.title,
            onConfirm = { title ->
                dialogType = null
                eventHandler += RenameGroupEvent(group, title)
            },
            onCancel = { dialogType = null },
        )
        GroupRowDialog.DESCRIPTION -> EditContactDialog(
            title = Res.string.set_group_description,
            label = Res.string.group_description,
            maxLength = constraints.maxGroupDescLength,
            initialValue = group.description,
            onConfirm = { desc ->
                dialogType = null
                eventHandler += SetGroupDescriptionEvent(group, desc)
            },
            onCancel = { dialogType = null },
        )
        GroupRowDialog.PASSWORD -> PasswordDialog(
            title = Res.string.set_group_password,
            label = Res.string.group_password,
            maxLength = constraints.maxGroupPasswordLength,
            isOptional = true,
            initialPassword = group.password.orEmpty(),
            onConfirm = { password ->
                dialogType = null
                eventHandler += SetGroupPasswordEvent(group, password)
            },
            onCancel = { dialogType = null },
        )
        GroupRowDialog.CLEAR_MESSAGES -> PermanentActionDialog(
            title = Res.string.clear_messages,
            message = stringResource(Res.string.clear_messages_confirmation),
            onConfirm = {
                dialogType = null
                eventHandler += ClearMessagesEvent(group.uid)
            },
            onCancel = { dialogType = null },
        )
        GroupRowDialog.REMOVE -> PermanentActionDialog(
            title = Res.string.leave_group,
            message = stringResource(Res.string.leave_group_confirmation, group.title),
            onConfirm = {
                dialogType = null
                // TODO - Add optional parting message input to dialog
                eventHandler += RemoveGroupEvent(
                    group = group,
                )
            },
            onCancel = { dialogType = null },
        )
        else -> {}
    }
}

private enum class GroupRowDialog {
    RENAME,
    DESCRIPTION,
    PASSWORD,
    CLEAR_MESSAGES,
    REMOVE,
}