package nand.apps.chat.ui.groups

import androidx.compose.foundation.layout.*
import androidx.compose.material.Icon
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Error
import androidx.compose.material.icons.outlined.Lock
import androidx.compose.material.icons.outlined.SentimentVeryDissatisfied
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.style.TextAlign
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupErrorType
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.dialog.PasswordDialog
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun GroupErrorScreen(
    group: ChatGroupData,
    error: ChatGroupErrorType,
    modifier: Modifier,
) {
    val eventHandler = LocalEventHandler.current
    val engineConstraints = LocalChatEngineConstraints.current
    var isPasswordDialogVisible by rememberSaveable { mutableStateOf(false) }
    Box(
        contentAlignment = Alignment.Center,
        modifier = modifier,
    ) {
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
        ) {
            Icon(
                imageVector = when (error) {
                    ChatGroupErrorType.INVALID_PASSWORD -> Icons.Outlined.Lock
                    ChatGroupErrorType.KICKED -> Icons.Outlined.SentimentVeryDissatisfied
                    else -> Icons.Outlined.Error
                },
                contentDescription = null,
                tint = ChatTheme.colors.icon,
                modifier = Modifier.size(ChatTheme.dimens.iconXLarge),
            )
            Text(
                text = when (error) {
                    ChatGroupErrorType.PEER_LIMIT -> stringResource(Res.string.group_error_peer_limit_title)
                    ChatGroupErrorType.INVALID_PASSWORD -> stringResource(Res.string.group_error_invalid_password_title)
                    ChatGroupErrorType.KICKED -> stringResource(Res.string.group_error_kicked_title)
                    ChatGroupErrorType.UNKNOWN -> stringResource(Res.string.group_error_unknown_title)
                },
                color = ChatTheme.colors.onSurface,
                style = ChatTheme.typography.h6,
                textAlign = TextAlign.Center,
            )
            Text(
                text = when (error) {
                    ChatGroupErrorType.PEER_LIMIT -> stringResource(Res.string.group_error_peer_limit_desc)
                    ChatGroupErrorType.INVALID_PASSWORD -> stringResource(Res.string.group_error_invalid_password_desc)
                    ChatGroupErrorType.KICKED -> stringResource(Res.string.group_error_kicked_desc)
                    ChatGroupErrorType.UNKNOWN -> stringResource(Res.string.group_error_unknown_desc)
                },
                color = ChatTheme.colors.onBackground,
                style = ChatTheme.typography.body1,
                textAlign = TextAlign.Center,
            )
            Row(
                horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                verticalAlignment = Alignment.CenterVertically,
            ) {
                SimpleTextButton(
                    text = stringResource(Res.string.leave_group),
                    onClick = { eventHandler += RemoveGroupEvent(group) },
                )
                if (error == ChatGroupErrorType.INVALID_PASSWORD) {
                    SimpleTextButton(
                        text = stringResource(Res.string.set_group_password),
                        onClick = { isPasswordDialogVisible = true },
                    )
                }
            }
        }
    }
    if (isPasswordDialogVisible) {
        PasswordDialog(
            title = Res.string.set_group_password,
            label = Res.string.password,
            maxLength = engineConstraints.maxGroupPasswordLength,
            onConfirm = { password ->
                isPasswordDialogVisible = false
                eventHandler += JoinGroupEvent(
                    group = group.copy(password = password),
                )
            },
            onCancel = { isPasswordDialogVisible = false },
        )
    }
}