package nand.apps.chat.ui.groups

import androidx.compose.foundation.background
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.VolumeUp
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.user.UserData
import nand.apps.chat.time.formatRelativeDate
import nand.apps.chat.ui.chat.screen.ChatScreenState
import nand.apps.chat.ui.contacts.ContactRow
import nand.apps.chat.ui.dialog.ConfirmationDialog
import nand.apps.chat.ui.dialog.ContactProfileDialog
import nand.apps.chat.ui.dialog.EditContactDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun GroupPeerRow(
    group: ChatGroupData,
    peer: UserData,
    selfPeer: UserData,
    sidebarState: SidebarState,
    chatState: ChatScreenState,
    modifier: Modifier = Modifier,
) {
    val constraints = LocalChatEngineConstraints.current
    val eventHandler = LocalEventHandler.current
    val isSelf = peer.uid == selfPeer.uid
    val user = remember(peer, selfPeer, chatState) {
        when {
            isSelf -> chatState.self
            else -> chatState.friends[peer.uid] ?: peer
        }.copy(
            name = peer.name,
            role = peer.role,
            protocol = peer.protocol,
        )
    }
    val roleName = user.role.getDisplayString()
    val subheading = remember(group, user) {
        when {
            group.isPeerBlocked(user.uid) -> getStringBlocking(Res.string.peer_blocked_subheading)
            !user.isOnline && user.lastOnline > 0L -> formatRelativeDate(user.lastOnline)
            group.hasRoles -> roleName
            else -> ""
        }
    }
    var dialog by rememberSaveable { mutableStateOf<GroupPeerDialogType?>(null) }
    ContactRow(
        heading = user.displayName,
        subHeading = subheading,
        status = user.status,
        isOnline = user.isOnline,
        avatar = user.avatar,
        iconOverlay = when {
            chatState.callActivity[user.uid] == true -> Icons.AutoMirrored.Default.VolumeUp
            else -> null
        },
        sidebarState = sidebarState,
        modifier = Modifier
            .alpha(if (user.isOnline) 1f else 0.5f)
            .appendIf(group.isPeerBlocked(peer.uid)) {
                background(ChatTheme.colors.error.copy(alpha = 0.5f))
            }
            .then(modifier),
        isActionMenuEnabled = true,
        actionMenuItems = {
            mutableListOf<ActionMenuItem>().apply {
                this += ActionMenuItem(
                    icon = Icons.Default.ContactPage,
                    label = getStringBlocking(Res.string.view_contact_profile),
                    onClick = { dialog = GroupPeerDialogType.VIEW_PROFILE },
                )
                if (group.isOnline) {
                    if (isSelf && ChatGroupFeature.EDIT_NICKNAME in group.features) {
                        this += ActionMenuItem(
                            icon = Icons.Default.Edit,
                            label = getStringBlocking(Res.string.set_group_nickname),
                            onClick = { dialog = GroupPeerDialogType.SET_NICKNAME }
                        )
                    }
                    if (!isSelf && ChatGroupFeature.ROLES in group.features && selfPeer.hasAuthorityOver(user)) {
                        this += ActionMenuItem(
                            icon = Icons.Default.SupervisorAccount,
                            label = getStringBlocking(Res.string.set_role),
                            onClick = { dialog = GroupPeerDialogType.SET_ROLE },
                        )
                        this += ActionMenuItem(
                            icon = Icons.Default.DeleteForever,
                            label = getStringBlocking(Res.string.kick_peer),
                            onClick = { dialog = GroupPeerDialogType.KICK },
                        )
                    }
                }
                if (!isSelf) {
                    if (group.isPeerBlocked(peer.uid)) {
                        this += ActionMenuItem(
                            icon = Icons.Default.WavingHand,
                            label = getStringBlocking(Res.string.unblock_peer),
                            onClick = { eventHandler += SetGroupPeerBlockedEvent(group, peer, block = false) },
                        )
                    } else {
                        this += ActionMenuItem(
                            icon = Icons.Default.Block,
                            label = getStringBlocking(Res.string.block_peer),
                            onClick = { dialog = GroupPeerDialogType.BLOCK },
                        )
                    }
                }
            }
        }
    )
    when (dialog) {
        GroupPeerDialogType.SET_NICKNAME -> EditContactDialog(
            title = Res.string.set_group_nickname,
            label = Res.string.nickname,
            maxLength = constraints.maxNameLength,
            initialValue = selfPeer.name,
            onConfirm = { name ->
                dialog = null
                eventHandler += SetGroupNicknameEvent(group, name)
            },
            onCancel = { dialog = null },
        )
        GroupPeerDialogType.VIEW_PROFILE -> ContactProfileDialog(
            user = user,
            isSelf = isSelf,
            onDismiss = { dialog = null },
        )
        GroupPeerDialogType.SET_ROLE -> SetGroupRoleDialog(
            group = group,
            peer = peer,
            selfPeer = selfPeer,
            onDismiss = { dialog = null },
        )
        GroupPeerDialogType.KICK -> ConfirmationDialog(
            title = Res.string.kick_peer,
            message = stringResource(Res.string.kick_peer_confirmation, peer.name),
            onConfirm = {
                dialog = null
                eventHandler += KickGroupPeerEvent(group, peer)
            },
            onCancel = { dialog = null },
        )
        GroupPeerDialogType.BLOCK -> ConfirmationDialog(
            title = Res.string.block_peer,
            message = stringResource(Res.string.block_peer_confirmation, peer.name),
            onConfirm = {
                dialog = null
                eventHandler += SetGroupPeerBlockedEvent(group, peer, block = true)
            },
            onCancel = { dialog = null },
        )
        else -> Unit
    }
}