package nand.apps.chat.ui.login

import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.FolderOpen
import androidx.compose.material.icons.filled.Settings
import androidx.compose.material.icons.outlined.Info
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.ui.navigation.SettingsRoute
import nand.apps.chat.ui.profile.selector.ProfileSelectorDialog
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.settings.AboutScreenRoute
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.snackbar.show
import nand.apps.chat.ui.text.SimplePasswordField
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.rememberGridBrush
import nand.apps.chat.util.takeIfUnderLength
import org.jetbrains.compose.resources.painterResource
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun LoginScreen(
    onLoginComplete: () -> Unit,
    modifier: Modifier = Modifier,
    viewModel: LoginScreenViewModel = koinKmpViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    var name by rememberTextFieldValue(state.username)
    var password by rememberTextFieldValue()
    val focusManager = LocalFocusManager.current
    val engineConstraints = LocalChatEngineConstraints.current
    val navController = LocalNavController.current
    val nameFocus = remember { FocusRequester() }
    val passwordFocus = remember { FocusRequester() }
    val snackbarHostState = remember { SnackbarHostState() }
    val filePickerSnackbar = remember { SnackbarHostState() }
    
    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is UpdateName -> {
                    name = name.copy(text = action.name)
                }
                is ShowLoginError -> {
                    snackbarHostState.showLoginError(action.error)
                }
                is ShowInvalidFileError -> {
                    filePickerSnackbar.show(Res.string.login_error_please_select_valid_profile)
                }
                is ShowInvalidNameError -> {
                    snackbarHostState.show(Res.string.login_error_invalid_profile_name)
                }
                is ShowImportFailedError -> {
                    snackbarHostState.show(Res.string.login_error_import_failed)
                }
                is FinishLogin -> {
                    onLoginComplete()
                    navController.navigate(MainRoute)
                }
            }
        }
    }
    LaunchedEffect(Unit) {
        viewModel.init()
    }
    Box(
        contentAlignment = Alignment.Center,
        modifier = modifier
            .background(
                brush = rememberGridBrush(
                    size = ChatTheme.dimens.iconXLarge,
                    backgroundColor = ChatTheme.colors.background,
                    foregroundColor = ChatTheme.colors.surface,
                )
            )
            .padding(ChatTheme.dimens.spacingMedium),
    ) {
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            modifier = Modifier.padding(ChatTheme.dimens.spacingSmall),
        ) {
            if (getPlatform().isDesktop) {
                LaunchedEffect(Unit) {
                    when {
                        name.text.isEmpty() -> nameFocus.requestFocus()
                        password.text.isEmpty() -> passwordFocus.requestFocus()
                    }
                }
            }
            Image(
                painter = painterResource(Res.drawable.ic_seers),
                contentDescription = APP_NAME,
                modifier = Modifier
                    .size(ChatTheme.dimens.avatarXLarge)
                    .padding(bottom = ChatTheme.dimens.spacingMedium),
            )
            TextLine(
                text = APP_NAME,
                style = ChatTheme.typography.h4,
                color = ChatTheme.colors.onSurface,
                modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingMedium),
            )
            Column(
                modifier = Modifier
                    .background(
                        color = ChatTheme.colors.background,
                        shape = ChatTheme.shapes.small,
                    )
                    .padding(bottom = ChatTheme.dimens.spacingMedium)
                    .padding(horizontal = ChatTheme.dimens.spacingMedium)
            ) {
                SimpleTextField(
                    value = name,
                    onValueChange = { name = it.takeIfUnderLength(engineConstraints.maxNameLength, name) },
                    labelText = stringResource(Res.string.username),
                    trailingIcon = {
                        IconButton(
                            icon = Icons.Filled.FolderOpen,
                            contentDescription = stringResource(Res.string.open_profile),
                            isEnabled = !state.isLoginBusy,
                            onClick = { viewModel.openProfileSelector() },
                        )
                    },
                    onSubmit = { focusManager.moveFocus(FocusDirection.Down) },
                    keyboardOptions = KeyboardOptions(
                        autoCorrect = false,
                        imeAction = ImeAction.Next,
                    ),
                    isEnabled = !state.isLoginBusy,
                    modifier = Modifier.focusRequester(nameFocus),
                )
                SimplePasswordField(
                    password = password,
                    onValueChange = { password = it },
                    onToggleVisibility = { passwordFocus.requestFocus() },
                    onSubmit = {
                        focusManager.clearFocus()
                        viewModel.performLogin(name.text, password.text)
                    },
                    isEnabled = !state.isLoginBusy,
                    modifier = Modifier
                        .padding(top = ChatTheme.dimens.spacingSmall)
                        .focusRequester(passwordFocus),
                )
            }
            SimpleTextButton(
                text = stringResource(Res.string.login),
                isEnabled = name.text.isNotBlank() && !state.isLoginBusy,
                isLoading = state.isLoginBusy,
                onClick = { viewModel.performLogin(name.text, password.text) },
                modifier = Modifier.padding(top = ChatTheme.dimens.spacingMedium),
            )
        }
        IconButton(
            icon = Icons.Outlined.Info,
            contentDescription = stringResource(Res.string.about_screen_title),
            size = ChatTheme.dimens.iconMedium,
            isEnabled = !state.isLoginBusy,
            onClick = { navController.navigate(AboutScreenRoute) },
            modifier = Modifier.align(Alignment.BottomStart),
        )
        IconButton(
            icon = Icons.Default.Settings,
            contentDescription = stringResource(Res.string.settings),
            size = ChatTheme.dimens.iconMedium,
            isEnabled = !state.isLoginBusy,
            onClick = { navController.navigate(SettingsRoute) },
            modifier = Modifier.align(Alignment.BottomEnd),
        )
        SimpleSnackbarHost(
            state = snackbarHostState,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
    }
    
    when {
        state.isProfileSelectorVisible -> ProfileSelectorDialog(
            onProfileSelected = viewModel::selectProfile,
            onDismiss = viewModel::dismissProfileSelector,
        )
        state.isRenameRequired -> SimpleDialog(
            title = Res.string.login_profile_rename_required_title,
            onConfirm = { viewModel.confirmRename(name.text) },
            onCancel = viewModel::cancelProfileImport,
        ) {
            Column(
                verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            ) {
                Text(
                    text = stringResource(Res.string.login_profile_rename_required_message, state.username),
                    color = ChatTheme.colors.onSurface,
                )
                SimpleTextField(
                    value = name,
                    onValueChange = { name = it },
                    labelText = stringResource(Res.string.username),
                    onSubmit = { viewModel.confirmRename(it.text) },
                    keyboardOptions = KeyboardOptions(autoCorrect = false),
                )
            }
        }
        state.isImportRequired -> SimpleDialog(
            title = Res.string.login_profile_import_required_title,
            onConfirm = viewModel::confirmProfileImport,
            onCancel = viewModel::cancelProfileImport,
        ) {
            Text(
                text = stringResource(Res.string.login_profile_import_required_message),
                color = ChatTheme.colors.onSurface,
            )
        }
        state.isPasswordWarningVisible -> {
            SimpleDialog(
                title = Res.string.login_new_profile_password_empty_title,
                onConfirm = { viewModel.performLogin(name.text, password.text) },
                onCancel = viewModel::dismissNewProfileDialog,
            ) {
                Text(
                    text = stringResource(Res.string.login_new_profile_password_empty_message),
                    color = ChatTheme.colors.onSurface,
                )
            }
        }
        state.isProfileNew -> {
            var confirmedPassword by rememberTextFieldValue()
            val isConfirmEnabled = password.text.isEmpty() || password.text == confirmedPassword.text
            SimpleDialog(
                title = Res.string.login_new_profile_title,
                isConfirmEnabled = isConfirmEnabled,
                onConfirm = { viewModel.performLogin(name.text, password.text) },
                onCancel = viewModel::dismissNewProfileDialog,
            ) {
                Column(
                    verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                ) {
                    Text(
                        text = stringResource(Res.string.login_new_profile_message, name.text),
                        color = ChatTheme.colors.onSurface,
                    )
                    if (password.text.isNotEmpty()) {
                        SimplePasswordField(
                            password = confirmedPassword,
                            label = Res.string.login_new_profile_confirm_password,
                            onValueChange = { confirmedPassword = it },
                            onSubmit = {
                                if (isConfirmEnabled) {
                                    viewModel.performLogin(name.text, password.text)
                                }
                            },
                        )
                    }
                }
            }
        }
    }
}