package nand.apps.chat.ui.login

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.mock.MockChatEngine
import nand.apps.chat.io.*
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.profile.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.util.fifoSharedFlow

private const val MOCK_PROFILE_NAME = "MOCK"

class LoginScreenViewModel(
    private val fileSystem: ChatFileSystem,
    private val engine: ChatEngine,
    private val settingsRepo: SettingsRepo,
    private val credentialsStore: ProfileCredentialsStore,
) : ViewModel() {

    private val profileDirectory = fileSystem.getProfileDirectory()
    private val defaultUsername = settingsRepo.settings.generalSettings.loginUsername
    
    val stateData = MutableStateFlow(LoginScreenState(username = defaultUsername))
    val actionData = fifoSharedFlow<LoginScreenAction>()
    
    suspend fun init() {
        when {
            // Check the engine type instead of MOCK_ENGINE to align with initKoin(mockEngine = true)
            engine is MockChatEngine -> {
                stateData.emit(stateData.value.copy(isLoginBusy = true))
                engine.initialize(
                    ChatProfileData(
                        filePath = fileSystem.getProfileDirectory("$MOCK_PROFILE_NAME/$DEFAULT_PROFILE_FILE_NAME").path.toString(),
                        username = MOCK_PROFILE_NAME,
                    )
                )
                stateData.emit(stateData.value.copy(isLoginBusy = false))
                actionData.emit(FinishLogin)
            }
            defaultUsername.isNotBlank() -> {
                val cachedPassword = credentialsStore.loadPassword()
                if (!cachedPassword.isNullOrBlank()) {
                    performLogin(
                        username = defaultUsername,
                        password = cachedPassword,
                        isAutomatic = true,
                    )
                }
            }
        }
    }
    
    fun performLogin(
        username: String,
        password: String,
        isAutomatic: Boolean = false,
    ) {
        if (!username.isValidFileName()) {
            if (isAutomatic) {
                stateData.tryEmit(stateData.value.copy(username = ""))
            } else {
                actionData.tryEmit(ShowInvalidNameError)
            }
            return
        }
        val dir = profileDirectory.getSubDirectory(username)
        val profile = dir.getPrimaryProfile() ?: dir.getFilePath(DEFAULT_PROFILE_FILE_NAME)
        if (profile.name !in dir) {
            if (isAutomatic) return
            when {
                !stateData.value.isProfileNew -> {
                    stateData.tryEmit(stateData.value.copy(isProfileNew = true))
                    return
                }
                password.isEmpty() && !stateData.value.isPasswordWarningVisible -> {
                    stateData.tryEmit(stateData.value.copy(isPasswordWarningVisible = true))
                    return
                }
            }
        }
        viewModelScope.launch {
            stateData.emit(
                stateData.value.copy(
                    isLoginBusy = true,
                    isProfileNew = false,
                    isPasswordWarningVisible = false,
                )
            )
            val result = engine.initialize(
                ChatProfileData(
                    filePath = profile.toString(),
                    username = username,
                    password = password,
                    networkSettings = settingsRepo.settings.networkSettings,
                )
            )
            if (result == ChatEngineInitResult.SUCCESS) {
                settingsRepo.update { generalSettings.copy(loginUsername = username) }
                credentialsStore.savePassword(password)
                stateData.emit(stateData.value.copy(isLoginBusy = false))
                actionData.emit(FinishLogin)
            } else {
                stateData.emit(
                    stateData.value.copy(
                        isProfileNew = false,
                        isLoginBusy = false,
                    )
                )
                if (!isAutomatic) {
                    actionData.emit(ShowLoginError(result))
                }
            }
        }
    }

    fun openProfileSelector() {
        stateData.tryEmit(stateData.value.copy(isProfileSelectorVisible = true))
    }

    fun dismissProfileSelector() {
        stateData.tryEmit(stateData.value.copy(isProfileSelectorVisible = false))
    }
    
    fun dismissNewProfileDialog() {
        stateData.tryEmit(
            stateData.value.copy(
                isProfileNew = false,
                isPasswordWarningVisible = false,
            )
        )
    }
    
    // TODO - Remove username from file path in favor of hash or something
    fun selectProfile(path: String) {
        val selection = fileSystem.selectProfile(path) ?: run {
            actionData.tryEmit(ShowInvalidFileError)
            return
        }
        actionData.tryEmit(UpdateName(selection.username))
        stateData.tryEmit(
            stateData.value.copy(
                username = selection.username,
                isProfileSelectorVisible = false,
                filePath = selection.filePath.toString(),
                isImportRequired = selection.isImportRequired,
            )
        )
    }
    
    fun cancelProfileImport() {
        actionData.tryEmit(UpdateName(defaultUsername))
        stateData.tryEmit(
            stateData.value.copy(
                filePath = null,
                username = defaultUsername,
                isImportRequired = false,
                isRenameRequired = false,
            )
        )
    }
    
    fun confirmProfileImport() {
        val (result, outputFile) = fileSystem.importProfile(
            username = stateData.value.username,
            filePath = stateData.value.filePath.orEmpty(),
        )
        when (result) {
            ProfileImportResult.IMPORT_SUCCESSFUL -> {
                stateData.tryEmit(
                    stateData.value.copy(
                        filePath = outputFile.toString(),
                        isImportRequired = false,
                    )
                )
            }
            ProfileImportResult.PROFILE_EXISTS -> {
                stateData.tryEmit(stateData.value.copy(isRenameRequired = true))
            }
            ProfileImportResult.INVALID_FILE -> {
                cancelProfileImport()
                actionData.tryEmit(ShowInvalidFileError)
            }
        }
    }
    
    fun confirmRename(username: String) {
        if (!username.isValidFileName()) {
            actionData.tryEmit(ShowInvalidNameError)
            return
        }
        stateData.tryEmit(
            stateData.value.copy(
                username = username,
                isRenameRequired = false,
            )
        )
        confirmProfileImport()
    }
}