package nand.apps.chat.ui.main

import androidx.compose.animation.*
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.ui.chat.screen.ChatScreen
import nand.apps.chat.ui.chat.screen.ChatScreenPlaceholder
import nand.apps.chat.ui.chat.screen.ChatScreenViewModel
import nand.apps.chat.ui.dialog.ConfirmationDialog
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.navigation.LoginRoute
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun MainScreen(
    modifier: Modifier = Modifier,
    viewModel: ChatScreenViewModel,
) {
    val eventHandler = LocalEventHandler.current
    val chatScreenState by viewModel.stateData.collectAsState()
    val isConversationOpen = chatScreenState.conversationUid != null
    var isLogoutConfirmationShown by rememberSaveable { mutableStateOf(false) }

    BackHandler {
        if (isConversationOpen) {
            eventHandler += ExitConversationEvent
        } else {
            isLogoutConfirmationShown = true
        }
    }
    
    ShortcutHandler(ShortcutAction.LOGOUT) {
        isLogoutConfirmationShown = true
    }
    
    Row(modifier = modifier) {
        AnimatedVisibility(
            visible = !isPortraitMode() || !isConversationOpen,
        ) {
            MainSidebar(
                chatState = chatScreenState,
            )
        }
        AnimatedVisibility(
            visible = isConversationOpen,
            enter = EnterTransition.None,
        ) {
            ChatScreen(
                viewModel = viewModel,
                state = chatScreenState,
                modifier = Modifier.fillMaxSize(),
            )
        }
        if (!isConversationOpen && !isPortraitMode()) {
            ChatScreenPlaceholder(
                state = chatScreenState,
                modifier = Modifier.fillMaxSize(),
            )
        }
    }
    if (isLogoutConfirmationShown) {
        val navController = LocalNavController.current
        ConfirmationDialog(
            title = Res.string.logout_confirmation_dialog_title,
            message = stringResource(Res.string.logout_confirmation_dialog_message),
            onConfirm = {
                isLogoutConfirmationShown = false
                eventHandler += LogoutEvent
                navController.navigate(LoginRoute, reset = true)
            },
            onCancel = { isLogoutConfirmationShown = false },
        )
    }
}