package nand.apps.chat.ui.profile

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.Button
import androidx.compose.material.ButtonDefaults
import androidx.compose.material.Icon
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.QrCode2
import androidx.compose.material.icons.outlined.Lock
import androidx.compose.material.icons.outlined.LockOpen
import androidx.compose.material.icons.outlined.Settings
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.draw.clip
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.input.pointer.PointerIcon
import androidx.compose.ui.input.pointer.pointerHoverIcon
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import nand.apps.chat.io.IMAGE_EXTENSIONS
import nand.apps.chat.model.avatar.ChatUserAvatar
import nand.apps.chat.model.settings.profile.ProfileSettingsStateData
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.ConfirmationDialog
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.dialog.image.ImageCropperDialog
import nand.apps.chat.ui.file.ChatFilePickerDialog
import nand.apps.chat.ui.file.ChatFilePickerOptions
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.settings.ProfileSettingsRoute
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.tooltip.TooltipText
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.takeIfUnderLength
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

// XXX - Keep this low to avoid exceeding 64k size limit
private const val MAX_AVATAR_SIZE = 128

@Composable
fun SelfProfileDialog(
    onDismiss: () -> Unit,
    viewModel: SelfProfileDialogViewModel = composableViewModel()
) {
    val state by viewModel.stateData.collectAsState()
    val settings = state.settings
    val profileState = settings.currentState ?: ProfileSettingsStateData(name = "", message = "")
    var name by rememberTextFieldValue(
        initialValue = profileState.name,
        key = settings,
    )
    var message by rememberTextFieldValue(
        initialValue = profileState.message,
        key = settings,
    )
    val avatar = remember(name.text, state.avatarPath) {
        ChatUserAvatar(
            username = name.text,
            path = state.avatarPath,
        )
    }
    val navController = LocalNavController.current
    val engineConstraints = LocalChatEngineConstraints.current
    val focusManager = LocalFocusManager.current
    val focusRequester = remember { FocusRequester() }
    
    LaunchedEffect(Unit) {
        focusRequester.requestFocus()
    }
    
    val onConfirm = {
        viewModel.applyStateChange(name.text, message.text)
        viewModel.persistSettings()
        onDismiss()
    }

    SimpleDialog(
        title = Res.string.your_profile,
        isConfirmEnabled = name.text.isNotBlank(),
        onConfirm = onConfirm,
        onCancel = onDismiss,
        titleEndContent = {
            IconButton(
                icon = Icons.Outlined.Settings,
                contentDescription = stringResource(Res.string.profile_settings),
                size = ChatTheme.dimens.iconMedium,
                onClick = {
                    onConfirm()
                    navController.navigate(ProfileSettingsRoute)
                },
            )
        },
    ) {
        Row(
            modifier = Modifier.fillMaxWidth(),
        ) {
            Column(
                horizontalAlignment = Alignment.CenterHorizontally,
                verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
                modifier = Modifier.padding(top = ChatTheme.dimens.spacingMedium),
            ) {
                ChatAvatarBox(
                    avatar = avatar,
                    size = ChatAvatarSize.XLARGE,
                    status = settings.status,
                    isOnline = true,
                    isLoading = state.isAvatarBusy,
                    modifier = Modifier
                        .alpha(if (state.isAvatarBusy) 0.5f else 1f)
                        .onClick(
                            isEnabled = !state.isAvatarBusy,
                            onClick = viewModel::openFilePicker,
                        ),
                )
                Button(
                    colors = ButtonDefaults.buttonColors(
                        backgroundColor = ChatTheme.colors.primary,
                    ),
                    onClick = viewModel::openAddressDialog,
                    contentPadding = PaddingValues(ChatTheme.dimens.spacingSmall),
                    modifier = Modifier
                        .size(ChatTheme.dimens.buttonLarge)
                        .pointerHoverIcon(PointerIcon.Hand),
                ) {
                    Icon(
                        imageVector = Icons.Default.QrCode2,
                        tint = ChatTheme.colors.onPrimary,
                        contentDescription = stringResource(Res.string.copy_address),
                        modifier = Modifier.size(ChatTheme.dimens.iconLarge)
                    )
                }
            }
            Column(
                modifier = Modifier
                    .padding(start = ChatTheme.dimens.spacingMedium)
                    .weight(1f),
            ) {
                SimpleTextField(
                    value = name,
                    labelText = stringResource(Res.string.username),
                    textStyle = ChatTheme.typography.body1,
                    onValueChange = { name = it.takeIfUnderLength(engineConstraints.maxNameLength, name) },
                    onSubmit = { focusManager.moveFocus(FocusDirection.Next) },
                    keyboardOptions = KeyboardOptions(
                        autoCorrect = false,
                        imeAction = ImeAction.Next,
                    ),
                    modifier = Modifier
                        .fillMaxWidth()
                        .focusRequester(focusRequester),
                )
                SimpleTextField(
                    value = message,
                    onValueChange = { message = it.takeIfUnderLength(engineConstraints.maxStatusMessageLength, message) },
                    labelText = stringResource(Res.string.status_message),
                    textStyle = ChatTheme.typography.body2,
                    onSubmit = {
                        if (name.text.isNotBlank()) onConfirm()
                    },
                    modifier = Modifier.fillMaxWidth(),
                )
                Row(
                    verticalAlignment = Alignment.CenterVertically,
                    modifier = Modifier.padding(top = ChatTheme.dimens.spacingMedium),
                ) {
                    SelfProfileStatusSelector(
                        status = settings.status,
                        onSelected = {
                            viewModel.applyStateChange(name.text, message.text)
                            viewModel.setStatus(it)
                        },
                        modifier = Modifier
                            .clip(ChatTheme.shapes.small)
                            .background(ChatTheme.colors.surface)
                    )
                    val lockTooltip = when {
                        settings.isLockedToDefaultState -> stringResource(Res.string.unlock_profile_state_desc)
                        else -> stringResource(Res.string.lock_profile_state_desc)
                    }
                    TooltipText(lockTooltip) {
                        IconButton(
                            icon = when {
                                settings.isLockedToDefaultState -> Icons.Outlined.Lock
                                else -> Icons.Outlined.LockOpen
                            },
                            contentDescription = lockTooltip,
                            size = ChatTheme.dimens.iconMedium,
                            color = when {
                                settings.isLockedToDefaultState -> ChatTheme.colors.onBackground2
                                else -> ChatTheme.colors.icon
                            },
                            onClick = {
                                if (settings.showLockWarning) {
                                    viewModel.showLockWarning()
                                } else {
                                    viewModel.applyStateChange(name.text, message.text)
                                    viewModel.toggleStateLock()
                                }
                            },
                        )
                    }
                }
            }
        }
    }

    val tempAvatarPath = state.tempAvatarPath
    val filePickerDirectory = state.filePickerDirectory
    when {
        tempAvatarPath != null -> ImageCropperDialog(
            title = Res.string.crop_avatar,
            imagePath = tempAvatarPath,
            maxSize = MAX_AVATAR_SIZE,
            forceSquare = true,
            onConfirm = { viewModel.createAvatar(it) },
            onCancel = viewModel::closeAvatarDialog,
        )
        filePickerDirectory != null -> ChatFilePickerDialog(
            options = ChatFilePickerOptions(
                directory = filePickerDirectory,
                extensions = IMAGE_EXTENSIONS,
            ),
            onConfirm = viewModel::selectAvatar,
            onCancel = viewModel::closeFilePicker,
        )
        state.isAddressVisible -> SelfProfileAddressDialog(
            onClose = viewModel::closeAddressDialog
        )
        state.isLockConfirmationVisible -> ConfirmationDialog(
            title = when {
                settings.isLockedToDefaultState -> Res.string.unlock_profile_state
                else -> Res.string.lock_profile_state
            },
            message = when {
                settings.isLockedToDefaultState -> stringResource(Res.string.unlock_profile_state_confirmation_message)
                else -> stringResource(Res.string.lock_profile_state_confirmation_message)
            },
            onConfirm = { dontAskAgain ->
                viewModel.dismissLockWarning(showWarningAgain = !dontAskAgain)
                viewModel.applyStateChange(name.text, message.text)
                viewModel.toggleStateLock()
            },
            onCancel = { viewModel.dismissLockWarning(showWarningAgain = true) },
            isDontAskAgainEnabled = true,
        )
    }
}