package nand.apps.chat.ui.profile

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.image.SimpleBitmap
import nand.apps.chat.io.ChatFileBrowser
import nand.apps.chat.io.ChatFilePaths
import nand.apps.chat.model.settings.profile.ProfileSettingsData
import nand.apps.chat.model.settings.profile.ProfileSettingsStateData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.SelfProfileRepo

class SelfProfileDialogViewModel(
    private val selfProfileRepo: SelfProfileRepo,
    private val avatarRepo: ChatAvatarRepo,
    private val filePaths: ChatFilePaths,
    private val fileBrowser: ChatFileBrowser,
) : ViewModel() {

    val stateData = MutableStateFlow(
        SelfProfileDialogState(
            settings = selfProfileRepo.settings,
            avatarPath = avatarRepo.getSelfAvatarPath(),
        )
    )
    private val settings get() = stateData.value.settings
    
    fun persistSettings() {
        selfProfileRepo.updateSettings(settings)
    }

    fun applyStateChange(name: String, message: String) {
        val status = when {
            settings.isLockedToDefaultState -> UserStatus.DEFAULT
            else -> settings.status
        }
        val defaultName = settings.currentState?.name ?: selfProfileRepo.profileName
        val profileState = ProfileSettingsStateData(
            name = name.takeIf { it.isNotBlank() } ?: defaultName,
            message = message,
        )
        updateSettings {
            copy(
                states = settings.states.toMutableMap().apply {
                    this[status] = profileState
                },
            )
        }
    }
    
    fun setStatus(status: UserStatus) = updateSettings {
        copy(status = status)
    }
    
    fun openAddressDialog() = updateState {
        copy(isAddressVisible = true)
    }

    fun closeAddressDialog() = updateState {
        copy(isAddressVisible = false)
    }
    
    fun openFilePicker() = when {
        fileBrowser.isGallerySupported() -> fileBrowser.launchGallery { path ->
            path?.let { selectAvatar(it.toString()) }
        }
        else -> updateState {
            copy(filePickerDirectory = filePaths.picturesPath.toString())
        }
    }
    
    fun closeFilePicker() = updateState {
        copy(filePickerDirectory = null)
    }

    fun toggleStateLock() {
        val currentState = settings.currentState ?: return
        val isLockedToDefaultState = !settings.isLockedToDefaultState
        updateSettings {
            copy(
                isLockedToDefaultState = isLockedToDefaultState,
                states = when {
                    isLockedToDefaultState -> mapOf(UserStatus.DEFAULT to currentState)
                    else -> UserStatus.entries.associateWith { currentState }
                },
            )
        }
    }

    fun showLockWarning() = updateState { copy(isLockConfirmationVisible = true) }

    fun dismissLockWarning(showWarningAgain: Boolean) = updateState {
        copy(
            isLockConfirmationVisible = false,
            settings = settings.copy(showLockWarning = showWarningAgain),
        )
    }
    
    fun selectAvatar(path: String) = updateState {
        copy(tempAvatarPath = path, filePickerDirectory = null)
    }
    
    fun createAvatar(bitmap: SimpleBitmap) {
        updateState { copy(tempAvatarPath = null, isAvatarBusy = true) }
        viewModelScope.launch {
            val status = if (settings.isLockedToDefaultState) UserStatus.DEFAULT else settings.status
            avatarRepo.setSelfAvatar(
                bitmap = bitmap,
                status = status,
            )
            updateState {
                copy(
                    isAvatarBusy = false,
                    avatarPath = avatarRepo.getSelfAvatarPath(status),
                )
            }
        }
    }
    
    fun closeAvatarDialog() = updateState {
        copy(tempAvatarPath = null)
    }
    
    private fun updateState(block: SelfProfileDialogState.() -> SelfProfileDialogState) {
        stateData.tryEmit(stateData.value.block())
    }

    private fun updateSettings(block: ProfileSettingsData.() -> ProfileSettingsData) = updateState {
        val newSettings = settings.block()
        val avatarStatus = if (newSettings.isLockedToDefaultState) UserStatus.DEFAULT else newSettings.status
        copy(
            settings = newSettings,
            avatarPath = avatarRepo.getSelfAvatarPath(avatarStatus),
        )
    }
}