package nand.apps.chat.ui.profile

import androidx.compose.foundation.layout.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.contacts.ContactRow
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.contacts.ContactDialogType
import nand.apps.chat.ui.friends.AddFriendDialog
import nand.apps.chat.ui.groups.CreateGroupDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.navigation.SettingsRoute
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun SelfProfileRow(
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    viewModel: SelfProfileRowViewModel = koinKmpViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    val navController = LocalNavController.current
    val buttonSize = when (sidebarState) {
        SidebarState.FULL -> ChatTheme.dimens.buttonXLarge
        else -> ChatTheme.dimens.buttonLarge
    }
    val iconSize = when (sidebarState) {
        SidebarState.FULL -> ChatTheme.dimens.iconMedium
        else -> ChatTheme.dimens.iconSmall
    }
    val isLightMode = ChatTheme.colors.isLight
    ShortcutHandler { action ->
        when (action) {
            ShortcutAction.OPEN_SELF_PROFILE -> {
                viewModel.openDialog(ContactDialogType.SELF_PROFILE)
                true
            }
            ShortcutAction.ADD_FRIEND -> {
                viewModel.openDialog(ContactDialogType.ADD_FRIEND)
                true
            }
            ShortcutAction.ADD_GROUP -> {
                viewModel.openDialog(ContactDialogType.CREATE_GROUP)
                true
            }
            ShortcutAction.TOGGLE_THEME -> {
                viewModel.toggleTheme(isLightMode)
                true
            }
            ShortcutAction.OPEN_SETTINGS -> {
                navController.navigate(SettingsRoute)
                true
            }
            else -> false
        }
    }
    Column(
        horizontalAlignment = Alignment.CenterHorizontally,
        modifier = modifier,
    ) {
        ContactRow(
            heading = state.self.displayName,
            subHeading = state.self.statusMessage,
            avatar = state.self.avatar,
            status = state.self.status,
            isOnline = state.self.isOnline,
            sidebarState = sidebarState,
            contentPadding = PaddingValues(ChatTheme.dimens.spacingMedium),
            onClick = { viewModel.openDialog(ContactDialogType.SELF_PROFILE) },
            isActionMenuEnabled = true,
            actionMenuItems = {
                mutableListOf<ActionMenuItem>().apply {
                    this += ActionMenuItem(
                        icon = Icons.Default.AccountCircle,
                        label = getStringBlocking(Res.string.status_online),
                        onClick = { viewModel.setStatus(UserStatus.DEFAULT) },
                    )
                    this += ActionMenuItem(
                        icon = Icons.Default.ModeStandby,
                        label = getStringBlocking(Res.string.status_away),
                        onClick = { viewModel.setStatus(UserStatus.AWAY) },
                    )
                    this += ActionMenuItem(
                        icon = Icons.Default.DoNotDisturbOnTotalSilence,
                        label = getStringBlocking(Res.string.status_busy),
                        onClick = { viewModel.setStatus(UserStatus.BUSY) },
                    )
                }
            }
        )
        Row(
            verticalAlignment = Alignment.CenterVertically,
            horizontalArrangement = Arrangement.SpaceEvenly,
            modifier = Modifier
                .height(buttonSize)
                .appendIf(sidebarState != SidebarState.COLLAPSED) { fillMaxWidth() },
        ) {
            val buttonModifier = Modifier.size(buttonSize)
            if (sidebarState != SidebarState.COLLAPSED) {
                IconButton(
                    icon = Icons.Default.PersonAddAlt1,
                    contentDescription = stringResource(Res.string.add_friend_dialog_title),
                    size = iconSize,
                    modifier = buttonModifier,
                    onClick = { viewModel.openDialog(ContactDialogType.ADD_FRIEND) },
                )
                IconButton(
                    icon = Icons.Default.GroupAdd,
                    contentDescription = stringResource(Res.string.create_group_dialog_title),
                    size = iconSize,
                    modifier = buttonModifier,
                    onClick = { viewModel.openDialog(ContactDialogType.CREATE_GROUP) },
                )
                IconButton(
                    icon = if (isLightMode) Icons.Default.DarkMode else Icons.Default.LightMode,
                    contentDescription = stringResource(Res.string.settings_general_theme_toggle),
                    size = iconSize,
                    modifier = buttonModifier,
                    onClick = { viewModel.toggleTheme(isLightMode) },
                )
            }
            IconButton(
                icon = Icons.Default.Settings,
                contentDescription = stringResource(Res.string.settings),
                size = iconSize,
                modifier = buttonModifier,
                onClick = { navController.navigate(SettingsRoute) },
            )
        }
    }
    when (state.dialog) {
        ContactDialogType.SELF_PROFILE -> SelfProfileDialog(
            onDismiss = viewModel::dismissDialog,
        )
        ContactDialogType.ADD_FRIEND -> AddFriendDialog(
            onDismiss = viewModel::dismissDialog,
        )
        ContactDialogType.CREATE_GROUP -> CreateGroupDialog(
            onDismiss = viewModel::dismissDialog,
        )
        else -> Unit
    }
}