package nand.apps.chat.ui.profile.auth

import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.size
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.text.SimplePasswordField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun ProfileAuthenticationDialog(
    onAccepted: () -> Unit,
    onCancel: () -> Unit,
    viewModel: ProfileAuthenticationDialogViewModel = composableViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    if (!state.isAuthRequired) {
        onAccepted()
        return
    }
    var password by rememberTextFieldValue()
    val passwordFocus = remember { FocusRequester() }
    SimpleDialog(
        title = Res.string.authentication_dialog_title,
        isConfirmEnabled = !state.isBusy && password.text.isNotEmpty(),
        onConfirm = { viewModel.submit(password.text) },
        onCancel = onCancel,
    ) {
        Box {
            Column(
                verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
            ) {
                Text(
                    text = when {
                        state.isErrorShown -> stringResource(Res.string.authentication_dialog_error)
                        else -> stringResource(Res.string.authentication_dialog_message)
                    },
                    color = when {
                        state.isErrorShown -> ChatTheme.colors.error
                        else -> ChatTheme.colors.onBackground
                    },
                    style = ChatTheme.typography.body2,
                )
                SimplePasswordField(
                    password = password,
                    onValueChange = { password = it },
                    onSubmit = { viewModel.submit(password.text) },
                    isError = state.isErrorShown,
                    isEnabled = !state.isBusy,
                    modifier = Modifier.focusRequester(passwordFocus),
                )
                LaunchedEffect(Unit) { passwordFocus.requestFocus() }
            }
            if (state.isBusy) {
                CircularProgressIndicator(
                    modifier = Modifier
                        .size(ChatTheme.dimens.iconMedium)
                        .align(Alignment.Center)
                )
            }
        }
    }
}