package nand.apps.chat.ui.profile.password

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.material.Text
import androidx.compose.runtime.*
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.navigation.LoginRoute
import nand.apps.chat.ui.profile.ProfileMigrationBusyDialog
import nand.apps.chat.ui.profile.ProfileMigrationFailedDialog
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.text.SimplePasswordField
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun LoginPasswordDialog(
    onCancel: () -> Unit,
    viewModel: LoginPasswordDialogViewModel = composableViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    var currentPassword by rememberTextFieldValue()
    var newPassword by rememberTextFieldValue()
    val navController = LocalNavController.current

    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is NavToLoginScreenAction -> navController.navigate(LoginRoute, reset = true)
            }
        }
    }

    when {
        state.isBusy -> {
            ProfileMigrationBusyDialog(message = Res.string.profile_edit_password_busy_message)
            return
        }
        state.migrationFailed -> {
            ProfileMigrationFailedDialog()
            return
        }
    }

    SimpleDialog(
        title = Res.string.settings_profile_login_password_title,
        isConfirmEnabled = newPassword.text.isNotEmpty() && currentPassword.text != newPassword.text,
        onConfirm = { viewModel.submit(currentPassword.text, newPassword.text) },
        onCancel = onCancel,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
        ) {
            AnimatedVisibility(
                visible = state.isPasswordIncorrect
            ) {
                TextLine(
                    text = stringResource(Res.string.authentication_dialog_error),
                    color = ChatTheme.colors.error,
                    style = ChatTheme.typography.body2,
                )
            }
            SimplePasswordField(
                password = currentPassword,
                onValueChange = { currentPassword = it },
                label = Res.string.current_password,
                isError = state.isPasswordIncorrect,
            )
            SimplePasswordField(
                password = newPassword,
                onValueChange = { newPassword = it },
                label = Res.string.new_password,
            )
        }
    }

    if (state.isConfirmationVisible) {
        SimpleDialog(
            title = Res.string.profile_edit_password_confirmation_title,
            onConfirm = { viewModel.submit(currentPassword.text, newPassword.text) },
            onCancel = onCancel,
        ) {
            Text(
                text = stringResource(Res.string.profile_edit_password_confirmation_message),
                color = ChatTheme.colors.onSurface,
                style = ChatTheme.typography.body1,
            )
        }
    }
}