package nand.apps.chat.ui.profile.password

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.chat.model.event.*
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileEncryptionMux
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.profile.ProfileAuthenticator
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.util.fifoSharedFlow

class LoginPasswordDialogViewModel(
    private val selfProfileRepo: SelfProfileRepo,
    private val avatarRepo: ChatAvatarRepo,
    private val fileSystem: ChatFileSystem,
    private val eventHandler: ChatEventHandler,
    private val authenticator: ProfileAuthenticator,
    private val encryption: ChatEngineEncryption,
) : ViewModel() {
    
    val stateData = MutableStateFlow(LoginPasswordDialogState())
    val actionData = fifoSharedFlow<LoginPasswordDialogAction>()
    
    fun submit(
        currentPassword: String,
        newPassword: String,
    ) {
        val isPasswordCorrect = runBlocking { authenticator.isPasswordCorrect(currentPassword) }
        if (!isPasswordCorrect) {
            stateData.tryEmit(stateData.value.copy(isPasswordIncorrect = true))
            return
        }
        
        if (stateData.value.isConfirmationVisible) {
            viewModelScope.launch {
                migrateProfile(
                    currentPassword = currentPassword,
                    newPassword = newPassword,
                )
            }
        } else {
            stateData.tryEmit(stateData.value.copy(isConfirmationVisible = true))
        }
    }
    
    private suspend fun migrateProfile(
        currentPassword: String,
        newPassword: String,
    ) {
        val profileFileName = selfProfileRepo.profilePath.name
        val profileDir = selfProfileRepo.profilePath.parent?.let { fileSystem.getDirectory(it.toString()) } ?: return
        
        stateData.emit(
            stateData.value.copy(
                isPasswordIncorrect = false,
                isConfirmationVisible = false,
                isBusy = true,
            )
        )

        eventHandler += LogoutEvent
         
        val success = withContext(Dispatchers.IO) {
            profileDir.readFromFile(profileFileName)?.let { profileData ->
                val encryption = ChatFileEncryptionMux(
                    encryption = encryption.getFileEncryption(key = newPassword, cipher = null),
                    decryption = encryption.getFileEncryption(key = currentPassword, cipher = profileData),
                )
                profileDir.encrypt(encryption)
                avatarRepo.avatarCacheDirectory.encrypt(encryption)
            } ?: false
        }
        
        if (success) {
            actionData.emit(NavToLoginScreenAction)
        } else {
            stateData.emit(
                stateData.value.copy(
                    isBusy = false,
                    migrationFailed = true,
                )
            )
        }
    }
}