package nand.apps.chat.ui.profile.username

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import nand.apps.chat.io.isValidFileName
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.navigation.LoginRoute
import nand.apps.chat.ui.profile.ProfileMigrationBusyDialog
import nand.apps.chat.ui.profile.ProfileMigrationFailedDialog
import nand.apps.chat.ui.profile.auth.ProfileAuthenticationDialog
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.takeIfUnderLength
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun LoginUsernameDialog(
    onCancel: () -> Unit,
    viewModel: LoginUsernameDialogViewModel = composableViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    var name by rememberTextFieldValue(state.username)
    var errorText by rememberSaveable { mutableStateOf("") }
    val engineConstraints = LocalChatEngineConstraints.current
    val navController = LocalNavController.current

    val invalidNameError = stringResource(Res.string.login_error_invalid_profile_name)
    val existingNameError = stringResource(Res.string.profile_username_already_exists)
    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is ShowInvalidNameError -> errorText = invalidNameError
                is ShowExistingNameError -> errorText = existingNameError
                is NavToLoginScreenAction -> navController.navigate(LoginRoute, reset = true)
            }
        }
    }
    
    when {
        state.isAuthRequired -> {
            ProfileAuthenticationDialog(
                onAccepted = viewModel::onAuthenticationAccepted,
                onCancel = onCancel,
            )
            return
        }
        state.isBusy -> {
            ProfileMigrationBusyDialog(message = Res.string.profile_edit_username_busy_message)
            return
        }
        state.migrationFailed -> {
            ProfileMigrationFailedDialog()
            return
        }
    }
    
    SimpleDialog(
        title = Res.string.settings_profile_login_username_title,
        isConfirmEnabled = name.text.isValidFileName(),
        onConfirm = { viewModel.submit(name.text) },
        onCancel = onCancel,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
        ) {
            AnimatedVisibility(
                visible = errorText.isNotEmpty(),
            ) {
                Text(
                    text = errorText,
                    color = ChatTheme.colors.error,
                    style = ChatTheme.typography.body2,
                )
            }
            SimpleTextField(
                value = name,
                labelText = stringResource(Res.string.username),
                isError = errorText.isNotEmpty(),
                onValueChange = {
                    name = it.takeIfUnderLength(engineConstraints.maxNameLength, name)
                    errorText = ""
                },
                onSubmit = { viewModel.submit(name.text) },
            )
        }
    }
    
    if (state.isConfirmationVisible) {
        SimpleDialog(
            title = Res.string.profile_edit_username_confirmation_title,
            isConfirmEnabled = name.text.isValidFileName(),
            onConfirm = { viewModel.submit(name.text) },
            onCancel = onCancel,
        ) {
            Text(
                text = stringResource(Res.string.profile_edit_username_confirmation_message, name.text),
                color = ChatTheme.colors.onSurface,
                style = ChatTheme.typography.body1,
            )
        }
    }
}