package nand.apps.chat.ui.profile.username

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.model.event.*
import nand.apps.chat.io.ChatDirectory
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.isValidFileName
import nand.apps.chat.profile.ProfileAuthenticator
import nand.apps.chat.repo.AVATAR_CACHE_DIRECTORY
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.util.fifoSharedFlow

class LoginUsernameDialogViewModel(
    private val selfProfileRepo: SelfProfileRepo,
    private val avatarRepo: ChatAvatarRepo,
    private val fileSystem: ChatFileSystem,
    private val eventHandler: ChatEventHandler,
    authenticator: ProfileAuthenticator,
) : ViewModel() {
    
    val stateData = MutableStateFlow(
        LoginUsernameDialogState(
            username = selfProfileRepo.profileName,
            isAuthRequired = authenticator.isAuthenticationRequired(),
        )
    )
    val actionData = fifoSharedFlow<LoginUsernameDialogAction>()
    
    fun onAuthenticationAccepted() {
        stateData.tryEmit(stateData.value.copy(isAuthRequired = false))
    }
    
    fun submit(username: String) {
        val profileDir = selfProfileRepo.profilePath.parent?.let { fileSystem.getDirectory(it.toString()) } ?: return
        if (!username.isValidFileName() || !profileDir.exists()) {
            actionData.tryEmit(ShowInvalidNameError)
            return
        }
        
        val outputDir = fileSystem.getProfileDirectory(username)
        if (outputDir.exists()) {
            actionData.tryEmit(ShowExistingNameError)
            return
        }
        
        if (stateData.value.isConfirmationVisible) {
            viewModelScope.launch {
                migrateAvatarCache(username)
                migrateProfile(profileDir, outputDir)
            }
        } else {
            stateData.tryEmit(stateData.value.copy(isConfirmationVisible = true))
        }
    }
    
    private suspend fun migrateAvatarCache(newUsername: String) {
        val inputDir = avatarRepo.avatarCacheDirectory
        val outputDir = fileSystem.getCacheDirectory("$AVATAR_CACHE_DIRECTORY/$newUsername")
        if (inputDir.exists()) migrateDirectory(inputDir, outputDir)
    }
    
    private suspend fun migrateProfile(inputDir: ChatDirectory, outputDir: ChatDirectory) {
        stateData.emit(
            stateData.value.copy(
                isConfirmationVisible = false,
                isBusy = true,
            )
        )
        eventHandler += LogoutEvent
        
        if (migrateDirectory(inputDir, outputDir)) {
            actionData.emit(NavToLoginScreenAction)
        } else {
            stateData.emit(
                stateData.value.copy(
                    isBusy = false,
                    migrationFailed = true,
                )
            )
        }
    }
    
    private suspend fun migrateDirectory(
        inputDir: ChatDirectory,
        outputDir: ChatDirectory,
    ) = withContext(Dispatchers.IO) {
        outputDir.deleteRecursively() && inputDir.moveTo(outputDir, deleteOutputOnFailure = true)
    }
}