package nand.apps.chat.ui.settings.call

import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.Composable
import androidx.compose.ui.Modifier
import nand.apps.chat.ui.settings.SettingsColumn
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.settings.dialog.BitRateSettingDialog
import nand.apps.chat.ui.settings.dialog.VoicemailGreetingDialog
import nand.apps.chat.ui.settings.row.SettingsCheckboxRow
import nand.apps.chat.ui.settings.row.SettingsDialogRow
import nand.apps.chat.ui.settings.row.SettingsInputDialogRow
import nand.apps.chat.ui.settings.row.SettingsSeparator
import nand.apps.chat.util.REGEX_NON_DIGITS
import nand.apps.chat.util.clamp
import nand.apps.chat.util.toIntOrZero
import seers.composeapp.generated.resources.*

private val AUDIO_BIT_RATE_OPTIONS = listOf(16, 32, 64)

@Composable
fun CallSettingsScreen(
    viewModel: SettingsViewModel,
    modifier: Modifier = Modifier,
) = SettingsColumn(
    title = Res.string.settings_calls_title,
    modifier = modifier,
) {
    val state = viewModel.settings.callSettings
    SettingsCheckboxRow(
        title = Res.string.settings_calls_enable_title,
        description = Res.string.settings_calls_enable_desc,
        icon = Icons.Outlined.Call,
        isChecked = state.isEnabled,
        onCheckedChange = { checked ->
            viewModel.update(state.copy(isEnabled = checked))
        }
    )
    SettingsInputDialogRow(
        title = Res.string.settings_calls_max_rings_title,
        description = Res.string.settings_calls_max_rings_desc,
        label = Res.string.settings_calls_max_rings_label,
        icon = Icons.Outlined.PhoneInTalk,
        value = state.maxRings.toString(),
        regex = REGEX_NON_DIGITS,
        isEnabled = state.isEnabled,
        onConfirm = { rings ->
            viewModel.update(state.copy(maxRings = rings.toIntOrZero().coerceAtLeast(1)))
        }
    )
    SettingsSeparator(
        text = Res.string.settings_calls_voicemail,
    )
    SettingsCheckboxRow(
        title = Res.string.settings_calls_enable_voicemail_title,
        description = Res.string.settings_calls_enable_voicemail_desc,
        icon = Icons.Outlined.Voicemail,
        isChecked = state.isVoicemailEnabled,
        onCheckedChange = { checked ->
            viewModel.update(state.copy(isVoicemailEnabled = checked))
        }
    )
    SettingsDialogRow(
        title = Res.string.settings_calls_voicemail_greeting_title,
        description = Res.string.settings_calls_voicemail_greeting_desc,
        icon = Icons.Outlined.WavingHand,
        isEnabled = state.isVoicemailEnabled,
        isLoginRequired = true,
    ) { onConfirm ->
        VoicemailGreetingDialog(
            recordBitRate = state.voicemailAudioBitRate,
            onConfirm = onConfirm,
        )
    }
    SettingsInputDialogRow(
        title = Res.string.settings_calls_max_voicemail_duration_title,
        description = Res.string.settings_calls_max_voicemail_duration_desc,
        label = Res.string.settings_calls_max_voicemail_duration_desc,
        icon = Icons.Outlined.HourglassEmpty,
        value = state.voicemailMaxDuration.toString(),
        regex = REGEX_NON_DIGITS,
        isEnabled = state.isVoicemailEnabled,
        onConfirm = { seconds ->
            viewModel.update(state.copy(voicemailMaxDuration = seconds.toIntOrZero().clamp(5, 600) * 1000))
        }
    )
    SettingsDialogRow(
        title = Res.string.settings_calls_voicemail_bit_rate_title,
        description = Res.string.settings_calls_voicemail_bit_rate_desc,
        icon = Icons.Outlined.GraphicEq,
        isEnabled = state.isVoicemailEnabled,
    ) { onConfirm ->
        BitRateSettingDialog(
            title = Res.string.settings_calls_voicemail_bit_rate_title,
            kbpsOptions = AUDIO_BIT_RATE_OPTIONS,
            defaultKbps = state.voicemailAudioBitRate / 1000,
            onConfirm = { kbps ->
                viewModel.update(state.copy(voicemailAudioBitRate = kbps * 1000))
                onConfirm()
            },
            onCancel = onConfirm,
        )
    }
}