package nand.apps.chat.ui.settings.dialog

import androidx.compose.foundation.Canvas
import androidx.compose.foundation.gestures.awaitEachGesture
import androidx.compose.foundation.gestures.awaitFirstDown
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.graphics.Brush
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.StrokeCap
import androidx.compose.ui.input.pointer.pointerInput
import nand.apps.chat.model.settings.theme.ChatThemeSetting
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.clamp
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

private val HUE_RANGE = 0f..360f

@Composable
fun ColorSettingDialog(
    viewModel: SettingsViewModel,
    onConfirm: () -> Unit,
) {
    val state = viewModel.settings.generalSettings
    val settingsRepo: SettingsRepo by getKoin().inject()
    var hue by rememberSaveable { mutableStateOf(state.appThemeHue) }
    val hueBrush = remember { generateHueGradient(12) }
    SimpleDialog(
        title = Res.string.settings_general_theme_color_title,
        titleStyle = ChatTheme.typography.body1,
        confirmLabel = null,
        cancelLabel = Res.string.confirm,
        onCancel = { onConfirm() },
    ) {
        Column(
            modifier = Modifier.fillMaxWidth(),
        ) {
            val strokeColor = ChatTheme.colors.onBackground2
            Row(
                horizontalArrangement = Arrangement.SpaceEvenly,
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextButton(
                    text = stringResource(Res.string.settings_general_theme_dark),
                    onClick = { viewModel.update(state.copy(appTheme = ChatThemeSetting.DARK)) },
                )
                SimpleTextButton(
                    text = stringResource(Res.string.reset),
                    onClick = {
                        hue = -1f
                        viewModel.update(state.copy(appThemeHue = -1f))
                    },
                )
                SimpleTextButton(
                    text = stringResource(Res.string.settings_general_theme_light),
                    onClick = { viewModel.update(state.copy(appTheme = ChatThemeSetting.LIGHT)) },
                )
            }
            Canvas(
                modifier = Modifier
                    .fillMaxWidth()
                    .height(ChatTheme.dimens.iconXLarge)
                    .padding(top = ChatTheme.dimens.spacingMedium)
                    .pointerInput(Unit) {
                        awaitEachGesture {
                            val event = awaitFirstDown()
                            hue = ((event.position.x / size.width.toFloat()) * HUE_RANGE.endInclusive).clamp(HUE_RANGE)
                            settingsRepo.update { generalSettings.copy(appThemeHue = hue) }
                        }
                    }
            ) {
                drawRect(hueBrush)
                if (hue in HUE_RANGE) {
                    val x = (hue / HUE_RANGE.endInclusive) * size.width
                    drawLine(
                        color = strokeColor,
                        start = Offset(x, 0f),
                        end = Offset(x, size.height),
                        cap = StrokeCap.Round,
                        strokeWidth = 2f,
                    )
                }
            }
        }
    }
}

private fun generateHueGradient(steps: Int): Brush {
    val interval = HUE_RANGE.endInclusive / steps.toFloat()
    val colors = mutableListOf<Color>()
    for (step in 0..steps) {
        colors += Color.hsv(
            hue = (interval * step).clamp(HUE_RANGE),
            saturation = 1f,
            value = 1f,
        )
    }
    return Brush.horizontalGradient(colors)
}