package nand.apps.chat.ui.settings.dialog

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardType
import nand.apps.chat.model.event.*
import nand.apps.chat.model.net.NetworkProxyType
import nand.apps.chat.model.settings.net.NetworkProxySettingsData
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.navigation.LoginRoute
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalLoginState
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.SimpleTextRadioButton
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.takeIfNumber
import nand.apps.chat.util.takeIfValidHost
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun NetworkProxySettingDialog(
    onUpdateProxy: (NetworkProxySettingsData) -> Unit,
    onDismiss: () -> Unit,
) {
    val initialState = LocalSettingsProvider.current.networkSettings.proxy ?: NetworkProxySettingsData()
    var type by rememberSaveable { mutableStateOf(initialState.type) }
    var host by rememberTextFieldValue(initialState.host.orEmpty())
    var port by rememberTextFieldValue(initialState.port.takeIf { it > 0 }?.toString().orEmpty())
    var isConfirmationDialogVisible by rememberSaveable { mutableStateOf(false) }
    
    val focusManager = LocalFocusManager.current
    val eventHandler = LocalEventHandler.current
    val navController = LocalNavController.current
    val isLoggedIn = LocalLoginState.current
    
    SimpleDialog(
        title = Res.string.settings_network_proxy_title,
        onConfirm = {
            onUpdateProxy(
                NetworkProxySettingsData(
                    type = type,
                    host = host.text.takeIf { it.isNotBlank() },
                    port = port.text.toIntOrNull() ?: 0,
                )
            )
            if (isLoggedIn) {
                isConfirmationDialogVisible = true
            } else {
                onDismiss()
            }
        },
        onCancel = onDismiss,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
        ) {
            Row(
                verticalAlignment = Alignment.CenterVertically,
            ) {
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_network_proxy_type_none),
                    selected = type == NetworkProxyType.NONE,
                    onClick = { type = NetworkProxyType.NONE },
                )
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_network_proxy_type_http),
                    selected = type == NetworkProxyType.HTTP,
                    onClick = { type = NetworkProxyType.HTTP },
                )
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_network_proxy_type_socks),
                    selected = type == NetworkProxyType.SOCKS,
                    onClick = { type = NetworkProxyType.SOCKS },
                )
            }
            AnimatedVisibility(
                visible = type != NetworkProxyType.NONE
            ) {
                Row(
                    horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                    verticalAlignment = Alignment.CenterVertically,
                ) {
                    SimpleTextField(
                        value = host,
                        onValueChange = { host = it.takeIfValidHost(host) },
                        labelText = stringResource(Res.string.settings_network_proxy_host),
                        keyboardOptions = KeyboardOptions(
                            autoCorrect = false,
                            imeAction = ImeAction.Next,
                        ),
                        onSubmit = { focusManager.moveFocus(FocusDirection.Next) },
                        modifier = Modifier.weight(0.7f),
                    )
                    SimpleTextField(
                        value = port,
                        onValueChange = { port = it.takeIfNumber(port, maxDigits = 5) },
                        labelText = stringResource(Res.string.settings_network_proxy_port),
                        keyboardOptions = KeyboardOptions(
                            keyboardType = KeyboardType.Number,
                            imeAction = ImeAction.Done,
                        ),
                        onSubmit = { focusManager.clearFocus() },
                        modifier = Modifier.weight(0.3f),
                    )
                }
            }
        }
    }
    if (isConfirmationDialogVisible) {
        SimpleDialog(
            title = Res.string.settings_network_proxy_title,
            confirmLabel = Res.string.yes,
            cancelLabel = Res.string.no,
            onConfirm = {
                onDismiss()
                eventHandler += LogoutEvent
                navController.navigate(LoginRoute, reset = true)
            },
            onCancel = onDismiss,
        ) {
            Text(
                text = stringResource(Res.string.settings_network_proxy_relog_message),
                color = ChatTheme.colors.onSurface,
                style = ChatTheme.typography.body1,
            )
        }
    }
}