package nand.apps.chat.ui.settings.dialog

import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.runtime.Composable
import androidx.compose.runtime.remember
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.style.TextAlign
import nand.apps.chat.time.DateFormatOrder
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.provider.LocalDateTimeFormatProvider
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.text.SimpleTextCheckbox
import nand.apps.chat.ui.text.SimpleTextRadioButton
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun TimeSettingDialog(
    viewModel: SettingsViewModel,
    onConfirm: () -> Unit,
) {
    val state = viewModel.settings.generalSettings
    val dateTimeFormatter = LocalDateTimeFormatProvider.current
    val time = remember { currentTimeMillis() }
    val timeFormat = state.timeFormat
    val dateFormat = state.dateFormat
    SimpleDialog(
        title = Res.string.settings_general_time_display_title,
        titleStyle = ChatTheme.typography.body1,
        confirmLabel = null,
        cancelLabel = Res.string.confirm,
        onCancel = { onConfirm() },
    ) {
        Column {
            TextLine(
                text = dateTimeFormatter.formatDateTime(time),
                style = ChatTheme.typography.body1,
                textAlign = TextAlign.Center,
                modifier = Modifier.fillMaxWidth(),
            )
            Row {
                SimpleTextCheckbox(
                    text = stringResource(Res.string.settings_general_time_include_seconds),
                    isChecked = state.timeFormat.includeSeconds,
                    onCheckedChange = { checked ->
                        viewModel.update(state.copy(timeFormat = timeFormat.copy(includeSeconds = checked)))
                    },
                    modifier = Modifier.weight(1f),
                )
                SimpleTextCheckbox(
                    text = stringResource(Res.string.settings_general_time_include_ampm),
                    isChecked = state.timeFormat.includeAmPm,
                    onCheckedChange = { checked ->
                        viewModel.update(state.copy(timeFormat = timeFormat.copy(includeAmPm = checked)))
                    },
                    modifier = Modifier.weight(1f),
                )
            }
            Row {
                SimpleTextCheckbox(
                    text = stringResource(Res.string.settings_general_time_dash_separator),
                    isChecked = state.dateFormat.useDashSeparator,
                    onCheckedChange = { checked ->
                        viewModel.update(state.copy(dateFormat = dateFormat.copy(useDashSeparator = checked)))
                    },
                    modifier = Modifier.weight(1f),
                )
                SimpleTextCheckbox(
                    text = stringResource(Res.string.settings_general_time_full_year),
                    isChecked = state.dateFormat.useFullYear,
                    onCheckedChange = { checked ->
                        viewModel.update(state.copy(dateFormat = dateFormat.copy(useFullYear = checked)))
                    },
                    modifier = Modifier.weight(1f),
                )
            }
            Row {
                DateFormatOrder.entries.forEach { order ->
                    SimpleTextRadioButton(
                        text = stringResource(
                            when (order) {
                                DateFormatOrder.YMD -> Res.string.settings_general_time_year_month_day
                                DateFormatOrder.MDY -> Res.string.settings_general_time_month_day_year
                                DateFormatOrder.DMY -> Res.string.settings_general_time_day_month_year
                            }
                        ),
                        selected = state.dateFormat.order == order,
                        onClick = {
                            viewModel.update(state.copy(dateFormat = dateFormat.copy(order = order)))
                        },
                    )
                }
            }
        }
    }
}