package nand.apps.chat.ui.settings.dialog

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Stop
import androidx.compose.material.icons.outlined.Mic
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import io.ktor.http.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import nand.apps.chat.audio.AudioRecorder
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.file.OPUS
import nand.apps.chat.repo.VoicemailRepo
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.embed.audio.AudioEmbedRenderer
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.toDurationString
import nand.apps.opus.OpusFileWriter
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun VoicemailGreetingDialog(
    recordBitRate: Int,
    onConfirm: () -> Unit,
) {
    val voicemailRepo: VoicemailRepo by getKoin().inject()
    val audioRecorder: AudioRecorder by getKoin().inject()
    val fileSystem: ChatFileSystem by getKoin().inject()
    val encryption: ChatEngineEncryption by getKoin().inject()
    
    val greetingPath = rememberSaveable { voicemailRepo.getGreetingAudioPath().toString() }
    var isPathValid by rememberSaveable { mutableStateOf(fileSystem.exists(greetingPath)) }
    var isRecorderActive by rememberSaveable { mutableStateOf(false) }
    var sharedRecorderActive by rememberSaveable { mutableStateOf(false) }
    var startRecordTimeMs by rememberSaveable { mutableStateOf(0L) }
    var recordDurationSecs by rememberSaveable { mutableStateOf(0) }
    var fileWriter by remember { mutableStateOf<OpusFileWriter?>(null) }
    
    LaunchedEffect(isRecorderActive) {
        if (isRecorderActive) {
            isPathValid = false
            audioRecorder.onFrame.collect { frame ->
                if (fileWriter == null) {
                    fileSystem.makeDirectoriesForFile(greetingPath)
                    fileSystem.deleteFile(greetingPath)
                    fileWriter = OpusFileWriter(
                        path = greetingPath,
                        sampleRate = frame.sampleRate,
                        channels = frame.channels,
                        bitRate = recordBitRate,
                    )
                }
                fileWriter?.write(frame.samples)
                recordDurationSecs = ((currentTimeMillis() - startRecordTimeMs) / 1000).toInt()
            }
        } else {
            fileWriter?.run {
                close()
                if (encryption.isEncryptionSupported()) {
                    withContext(Dispatchers.IO) {
                        val dir = fileSystem.getDirectoryForFile(greetingPath)
                        val fileName = greetingPath.toPath().name
                        val bytes = dir?.readFromFile(fileName) ?: return@withContext
                        dir.writeToFile(
                            fileName = fileName,
                            data = bytes,
                            encryption = encryption,
                        )
                    }
                }
                isPathValid = fileSystem.exists(greetingPath)
            }
            fileWriter = null
        }
    }
    
    SimpleDialog(
        title = Res.string.settings_calls_voicemail_greeting_title,
        titleStyle = ChatTheme.typography.body1,
        confirmLabel = null,
        cancelLabel = Res.string.confirm,
        isCancelEnabled = fileWriter == null,
        onCancel = onConfirm,
    ) {
        Column {
            Row(
                horizontalArrangement = Arrangement.Center,
                verticalAlignment = Alignment.CenterVertically,
                modifier = Modifier.fillMaxWidth()
            ) {
                IconButton(
                    icon = when {
                        isRecorderActive -> Icons.Filled.Stop
                        else -> Icons.Outlined.Mic
                    },
                    contentDescription = stringResource(
                        when {
                            isRecorderActive -> Res.string.stop_recording
                            else -> Res.string.record_voicemail_greeting
                        }
                    ),
                    onClick = {
                        if (isRecorderActive) {
                            isRecorderActive = false
                            if (!sharedRecorderActive) audioRecorder.stop()
                        } else {
                            sharedRecorderActive = audioRecorder.isActive
                            if (!sharedRecorderActive) audioRecorder.start()
                            recordDurationSecs = 0
                            startRecordTimeMs = currentTimeMillis()
                            isRecorderActive = audioRecorder.isActive
                        }
                    },
                )
                TextLine(
                    text = when {
                        isRecorderActive -> recordDurationSecs.toDurationString()
                        else -> "--:--"
                    },
                    style = ChatTheme.typography.body1,
                    modifier = Modifier.padding(start = ChatTheme.dimens.spacingMedium),
                )
            }
            AnimatedVisibility(
                visible = isPathValid,
            ) {
                AudioEmbedRenderer(
                    uri = greetingPath,
                    type = ContentType.Audio.OPUS,
                    reuseCache = false,
                    onDeleteFile = { isPathValid = false },
                )
            }
        }
    }
}