package nand.apps.chat.ui.settings.shortcut

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.input.key.*
import androidx.compose.ui.text.font.FontFamily
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.settings.shortcut.ShortcutSettingsData
import nand.apps.chat.model.settings.shortcut.toShortcutKeyCombo
import nand.apps.chat.platform.ChatPlatform
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.modifier.toShortcutCombo
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.shortcut.KeyEventHandler
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.toolbar.SimpleNavToolbar
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun ShortcutSettingsScreen(
    viewModel: SettingsViewModel,
    modifier: Modifier = Modifier,
) = Column(modifier = modifier) {
    val settings = LocalSettingsProvider.current.shortcutSettings
    val actions = remember {
        ShortcutAction.entries.filter { it.isSupported(getPlatform()) }
    }
    var selectedAction by rememberSaveable { mutableStateOf<ShortcutAction?>(null) }
    KeyEventHandler(selectedAction != null) { event ->
        if (event.type == KeyEventType.KeyUp) {
            selectedAction?.let { action ->
                val combo = event.toShortcutCombo()
                when {
                    combo != null -> {
                        viewModel.setShortcutKey(action, combo)
                        selectedAction = null
                    }
                    event.key == Key.Backspace -> {
                        viewModel.setShortcutKey(action, null)
                        selectedAction = null
                    }
                }
            }
        }
        true
    }
    SimpleNavToolbar(
        text = Res.string.settings_shortcuts_title,
    )
    LazyColumn(
        contentPadding = PaddingValues(
            vertical = when {
                isPortraitMode() -> 0.dp
                else -> ChatTheme.dimens.spacingSmall
            },
        ),
        modifier = Modifier.fillMaxSize(),
    ) {
        items(
            items = actions,
            key = { it },
        ) { action ->
            ShortcutRow(
                action = action,
                isSelected = selectedAction == action,
                settings = settings,
                modifier = Modifier
                    .background(
                        when {
                            selectedAction == action -> ChatTheme.colors.secondaryVariant
                            else -> Color.Unspecified
                        }
                    )
                    .onClick { selectedAction = action },
            )
        }
    }
}

@Composable
private fun ShortcutRow(
    action: ShortcutAction,
    isSelected: Boolean,
    settings: ShortcutSettingsData,
    modifier: Modifier = Modifier,
) = Row(
    verticalAlignment = Alignment.CenterVertically,
    modifier = modifier
        .fillMaxWidth()
        .padding(
            horizontal = ChatTheme.dimens.spacingXLarge,
            vertical = if (isPortraitMode()) ChatTheme.dimens.spacingLarge else ChatTheme.dimens.spacingMedium,
        ),
) {
    TextLine(
        text = stringResource(action.toStringResource()),
        style = ChatTheme.typography.body1,
        modifier = Modifier.weight(1f),
    )
    val combo = settings.shortcuts[action]?.toShortcutKeyCombo()?.toPlatformString()
    TextLine(
        text = when {
            isSelected -> stringResource(Res.string.shortcut_selected)
            else -> combo ?: stringResource(Res.string.shortcut_unassigned)
        },
        style = ChatTheme.typography.body2.copy(fontFamily = FontFamily.Monospace),
        color = when {
            isSelected || combo != null -> ChatTheme.colors.onSurface
            else -> ChatTheme.colors.onBackground2
        },
        textAlign = TextAlign.End,
    )
}

private fun ShortcutAction.toStringResource() = when (this) {
    ShortcutAction.ADD_FRIEND -> Res.string.add_friend_dialog_title
    ShortcutAction.ADD_GROUP -> Res.string.create_group_dialog_title
    ShortcutAction.OPEN_SETTINGS -> Res.string.shortcut_open_settings
    ShortcutAction.OPEN_SELF_PROFILE -> Res.string.shortcut_open_self_profile
    ShortcutAction.TOGGLE_THEME -> Res.string.settings_general_theme_toggle
    ShortcutAction.TOGGLE_CONTACTS -> Res.string.shortcut_toggle_contacts
    ShortcutAction.TOGGLE_PEERS -> Res.string.shortcut_toggle_peers
    ShortcutAction.TOGGLE_EMOJI_POPUP -> Res.string.shortcut_toggle_emojis
    ShortcutAction.SEND_FILE -> Res.string.chat_screen_send_file
    ShortcutAction.START_ANSWER_CALL -> Res.string.shortcut_start_answer_call
    ShortcutAction.PAUSE_RESUME_CALL -> Res.string.shortcut_pause_resume_call
    ShortcutAction.END_CALL -> Res.string.shortcut_end_call
    ShortcutAction.LOGOUT -> Res.string.shortcut_log_out
    ShortcutAction.NAVIGATE_BACK -> Res.string.shortcut_navigate_back
    ShortcutAction.HIDE_WINDOW -> Res.string.shortcut_hide_window
    ShortcutAction.TOGGLE_FULLSCREEN -> Res.string.shortcut_toggle_fullscreen
    ShortcutAction.TOGGLE_DEV_CONSOLE -> Res.string.shortcut_toggle_dev_console
    ShortcutAction.ZOOM_IN -> Res.string.shortcut_zoom_in
    ShortcutAction.ZOOM_OUT -> Res.string.shortcut_zoom_out
    ShortcutAction.ZOOM_RESET -> Res.string.shortcut_zoom_reset
}

private fun ShortcutAction.isSupported(platform: ChatPlatform) = when (this) {
    ShortcutAction.HIDE_WINDOW,
    ShortcutAction.TOGGLE_DEV_CONSOLE,
    ShortcutAction.TOGGLE_FULLSCREEN -> platform.isDesktop
    else -> true
}