package nand.apps.chat.ui.text

import androidx.compose.foundation.background
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.interaction.collectIsFocusedAsState
import androidx.compose.foundation.layout.defaultMinSize
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.text.BasicTextField
import androidx.compose.foundation.text.KeyboardActions
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.SolidColor
import androidx.compose.ui.semantics.semantics
import androidx.compose.ui.text.TextLayoutResult
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.VisualTransformation
import nand.apps.chat.ui.modifier.onEnter
import nand.apps.chat.ui.theme.ChatTheme

@OptIn(ExperimentalMaterialApi::class)
@Composable
fun SimpleTextField(
    value: TextFieldValue,
    modifier: Modifier = Modifier,
    textStyle: TextStyle = ChatTheme.typography.body2,
    hintText: String? = null,
    labelText: String? = null,
    isEnabled: Boolean = true,
    isSingleLine: Boolean = true,
    isError: Boolean = false,
    visualTransformation: VisualTransformation = VisualTransformation.None,
    keyboardOptions: KeyboardOptions = KeyboardOptions.Default,
    keyboardActions: KeyboardActions = KeyboardActions(),
    leadingIcon: @Composable (() -> Unit)? = null,
    trailingIcon: @Composable (() -> Unit)? = null,
    onValueChange: (TextFieldValue) -> Unit = {},
    onSubmit: (TextFieldValue) -> Unit = {},
    onTextLayout: (TextLayoutResult) -> Unit = {},
) {
    val interactionSource = remember { MutableInteractionSource() }
    val isFocused by interactionSource.collectIsFocusedAsState()
    val labelColor = when {
        isError -> ChatTheme.colors.error
        isFocused -> Color.Unspecified
        else -> ChatTheme.colors.onBackground2
    }
    val colors = TextFieldDefaults.outlinedTextFieldColors()
    val mergedTextStyle = textStyle.merge(TextStyle(color = ChatTheme.colors.onSurface))
    val placeholder = createTextLine(hintText, textStyle)
    val label = createTextLine(labelText, textStyle, labelColor)
    val shape = TextFieldDefaults.OutlinedTextFieldShape
    BasicTextField(
        value = value,
        modifier = if (label != null) {
            modifier
                .semantics(mergeDescendants = true) {}
                .padding(top = ChatTheme.dimens.spacingMedium)
        } else {
            modifier
        }
            .background(colors.backgroundColor(isEnabled).value, shape)
            .defaultMinSize(
                minWidth = TextFieldDefaults.MinWidth,
                minHeight = TextFieldDefaults.MinHeight
            )
            .onEnter { if (isEnabled) onSubmit(value) },
        onValueChange = { if (isEnabled) onValueChange(it) },
        // TODO - https://youtrack.jetbrains.com/issue/CMP-1587
        //enabled = isEnabled,
        readOnly = false,
        textStyle = mergedTextStyle,
        cursorBrush = SolidColor(colors.cursorColor(isError).value),
        visualTransformation = visualTransformation,
        keyboardOptions = keyboardOptions,
        keyboardActions = keyboardActions,
        interactionSource = interactionSource,
        singleLine = isSingleLine,
        maxLines = if (isSingleLine) 1 else Int.MAX_VALUE,
        minLines = 1,
        onTextLayout = onTextLayout,
        decorationBox = @Composable { innerTextField ->
            TextFieldDefaults.OutlinedTextFieldDecorationBox(
                value = value.text,
                visualTransformation = visualTransformation,
                innerTextField = innerTextField,
                placeholder = placeholder,
                label = label,
                leadingIcon = leadingIcon,
                trailingIcon = trailingIcon,
                singleLine = isSingleLine,
                enabled = isEnabled,
                isError = isError,
                interactionSource = interactionSource,
                colors = colors,
                border = {
                    TextFieldDefaults.BorderBox(
                        isEnabled,
                        isError,
                        interactionSource,
                        colors,
                        shape
                    )
                }
            )
        }
    )
}

@Composable
private fun createTextLine(
    text: String?,
    textStyle: TextStyle,
    color: Color = ChatTheme.colors.onBackground2
): @Composable (() -> Unit)? = text?.let {
    {
        TextLine(
            text = text,
            style = textStyle,
            color = color,
        )
    }
}

// TODO - Fix annoying bug on mobile where tapping on the text field moves the cursor to the start by default
@Composable
fun rememberTextFieldValue(
    initialValue: String = "",
    key: Any = Unit,
) = remember(key) {
    mutableStateOf(
        TextFieldValue(
            text = initialValue,
            selection = TextRange(initialValue.length),
        )
    )
}