package nand.apps.chat.ui.theme

import androidx.compose.material.Colors
import androidx.compose.material.darkColors
import androidx.compose.material.lightColors
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.compositeOver
import nand.apps.chat.util.clamp

data class ChatThemeColors(
    val material: Colors,
    val surface2: Color,
    val surface3: Color,
    val backgroundFloating: Color,
    val onBackground2: Color,
    val icon: Color,
    val statusOnline: Color,
    val statusAway: Color,
    val statusBusy: Color,
    val statusOffline: Color,
    val quote: Color,
    val spoiler: Color,
    val warning: Color,
    val mention: Color,
    val onMention: Color,
) {
    val primary: Color get() = material.primary
    val primaryVariant: Color get() = material.primaryVariant
    val secondary: Color get() = material.secondary
    val secondaryVariant: Color get() = material.secondaryVariant
    val background: Color get() = material.background
    val surface: Color get() = material.surface
    val error: Color get() = material.error
    val onPrimary: Color get() = material.onPrimary
    val onSecondary: Color get() = material.onSecondary
    val onBackground: Color get() = material.onBackground
    val onSurface: Color get() = material.onSurface
    val onError: Color get() = material.onError
    val isLight: Boolean get() = material.isLight
}

private val BACKGROUND_LIGHT = Color(0xFFCCCCCC)
private val BACKGROUND_DARK = Color(0xFF121212)
private val PRIMARY_COLOR = Color(0xFF339999)
private val SECONDARY_COLOR = Color(0xFF459933)

val LightChatColors = generateLightTheme()

val DarkChatColors = generateDarkTheme()

fun generateLightTheme(
    background: Color = BACKGROUND_LIGHT,
    maxAlpha: Float = 1f,
) = ChatThemeColors(
    material = lightColors(
        primary = PRIMARY_COLOR,
        primaryVariant = PRIMARY_COLOR.copy(alpha = 0.5f),
        secondary = SECONDARY_COLOR,
        secondaryVariant = SECONDARY_COLOR.copy(alpha = 0.5f),
        onSecondary = Color(0xFF003300),
        background = background,
        onBackground = onLight(0.85f, background),
        surface = onLight(0.1f, background),
        onSurface = onLight(maxAlpha, background),
        error = Color(0xFF993333),
        onError = Color(0xFFFF5555),
    ),
    surface2 = onLight(0.2f, background),
    surface3 = onLight(0.3f, background),
    backgroundFloating = background.copy(alpha = 0.75f),
    onBackground2 = onLight(0.7f, background),
    icon = onLight(0.8666667f, background),
    statusOnline = Color(0xFF39A627),
    statusAway = Color(0xFFCB9600),
    statusBusy = Color(0xFFCE3233),
    statusOffline = Color.DarkGray,
    quote = Color(0xFF789922),
    spoiler = Color(0xFF222222),
    warning = Color(0xFFA76E00),
    mention = Color(0x33CC7400),
    onMention = Color(0xFFCC7400),
)

fun generateDarkTheme(
    background: Color = BACKGROUND_DARK,
    maxAlpha: Float = 1f,
) = ChatThemeColors(
    material = darkColors(
        primary = PRIMARY_COLOR,
        primaryVariant = PRIMARY_COLOR.copy(alpha = 0.25f),
        onPrimary = Color.White,
        secondary = SECONDARY_COLOR,
        secondaryVariant = Color.Green.copy(alpha = 0.1f),
        onSecondary = Color(0xFF80FF80),
        background = background,
        onBackground = onDark(0.75f, background),
        surface = onDark(0.025f, background),
        onSurface = onDark(maxAlpha, background),
        error = Color(0xFF993333),
        onError = Color(0xFFFF5555),
    ),
    surface2 = onDark(0.05f, background),
    surface3 = onDark(0.1f, background),
    backgroundFloating = background.copy(alpha = 0.75f),
    onBackground2 = onDark(0.5f, background),
    icon = onDark(maxAlpha, background),
    statusOnline = Color(0xFF80FF80),
    statusAway = Color.Yellow,
    statusBusy = Color.Red,
    statusOffline = Color.DarkGray,
    quote = Color(0xFFB5BD68),
    spoiler = Color.Gray,
    warning = Color(0xFFFFA700),
    mention = Color(0x33FF9100),
    onMention = Color(0xFFFF9100),
)

fun generateLightTheme(hue: Float) = generateLightTheme(
    background = Color.hsv(
        hue = hue,
        saturation = 0.3f,
        value = 0.8f,
    ),
    maxAlpha = 0.95f,
)

fun generateDarkTheme(hue: Float) = generateDarkTheme(
    background = Color.hsv(
        hue = hue,
        saturation = 1f,
        value = 0.14f,
    ),
    maxAlpha = 0.95f,
)

private fun onDark(
    alpha: Float,
    background: Color,
) = onBackground(
    color = Color.White,
    background = background,
    alpha = alpha,
)

private fun onLight(
    alpha: Float,
    background: Color,
) = onBackground(
    color = Color.Black,
    background = background,
    alpha = alpha,
)

private fun onBackground(
    color: Color,
    background: Color,
    alpha: Float,
) = color.copy(alpha = alpha).compositeOver(background)

private fun Color.multiply(other: Color, factor: Float = 1f): Color {
    return Color(
        red = (red * other.red * factor).clamp(0f..1f),
        green = (green * other.green * factor).clamp(0f..1f),
        blue = (blue * other.blue * factor).clamp(0f..1f),
        alpha = alpha,
    )
}