package nand.apps.chat.ui.tutorial

import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.style.TextAlign
import kotlinx.coroutines.launch
import nand.apps.chat.io.PROFILE_EXTENSIONS
import nand.apps.chat.io.ZIP_EXTENSIONS
import nand.apps.chat.io.nameWithoutExtension
import nand.apps.chat.profile.ProfileBrowser
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.file.ChatFilePickerDialog
import nand.apps.chat.ui.file.ChatFilePickerMode
import nand.apps.chat.ui.file.ChatFilePickerOptions
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun TutorialImportProfileScreen(viewModel: TutorialViewModel) {
    val navController = LocalNavController.current
    val profileBrowser: ProfileBrowser by getKoin().inject()
    val profiles = remember { viewModel.getDefaultProfiles() }
    var isFilePickerVisible by rememberSaveable { mutableStateOf(false) }
    val errorSnackbar = remember { SnackbarHostState() }
    val profileErrorMessage = stringResource(Res.string.login_error_please_select_valid_profile)
    val coroutineScope = rememberCoroutineScope()
    val onSelect: (String) -> Unit = { path ->
        if (viewModel.selectProfile(path)) {
            navController.navigate(TutorialPasswordRoute)
        } else {
            coroutineScope.launch { errorSnackbar.showSnackbar(profileErrorMessage) }
        }
    }
    Box(
        modifier = Modifier.fillMaxSize(),
    ) {
        TutorialStepColumn(
            title = stringResource(Res.string.tutorial_profile_import_select),
            modifier = Modifier.align(Alignment.Center),
        ) {
            Text(
                text = stringResource(Res.string.tutorial_profile_import_desc),
                color = ChatTheme.colors.onBackground,
                style = ChatTheme.typography.body2,
            )
            if (profiles.isNotEmpty()) {
                LazyColumn(
                    horizontalAlignment = Alignment.CenterHorizontally,
                    modifier = Modifier.appendIf(profiles.size > 3) { weight(1f) },
                ) {
                    items(
                        items = profiles,
                        key = { it.name },
                    ) { path ->
                        TextLine(
                            text = path.nameWithoutExtension,
                            style = ChatTheme.typography.body1,
                            textAlign = TextAlign.Center,
                            modifier = Modifier
                                .onClick { onSelect(path.toString()) }
                                .padding(ChatTheme.dimens.spacingLarge),
                        )
                    }
                }
            }
            SimpleTextButton(
                text = stringResource(Res.string.tutorial_profile_import_select_browse),
                onClick = {
                    if (profileBrowser.isSupported()) {
                        profileBrowser.launch(
                            onResult = { profile -> profile?.filePath?.let { onSelect(it) } }
                        )
                    } else {
                        isFilePickerVisible = true
                    }
                },
            )
            SimpleTextButton(
                text = stringResource(Res.string.cancel),
                onClick = { navController.navigate(TutorialUsernameRoute) },
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbar,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .fillMaxWidth()
                .padding(ChatTheme.dimens.spacingLarge)
        )
    }
    if (isFilePickerVisible) {
        val directory = remember { viewModel.getDefaultImportDirectory() }
        ChatFilePickerDialog(
            options = ChatFilePickerOptions(
                directory = directory,
                extensions = PROFILE_EXTENSIONS + ZIP_EXTENSIONS,
                mode = ChatFilePickerMode.FILES_AND_DIRECTORIES,
            ),
            onConfirm = onSelect,
            onCancel = { isFilePickerVisible = false },
            snackbarHostState = errorSnackbar,
        )
    }
}