package nand.apps.chat.ui.tutorial

import androidx.compose.animation.*
import androidx.compose.animation.core.*
import androidx.compose.animation.core.Spring.StiffnessMediumLow
import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.BasicText
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.BlurredEdgeTreatment
import androidx.compose.ui.draw.blur
import androidx.compose.ui.draw.clip
import androidx.compose.ui.draw.rotate
import androidx.compose.ui.unit.dp
import kotlinx.coroutines.delay
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.locale.ChatLocale
import nand.apps.chat.locale.isAppLocaleSupported
import nand.apps.chat.locale.setAppLocale
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.painterResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun TutorialIntroScreen() {
    val navController = LocalNavController.current
    val settingsRepo: SettingsRepo by getKoin().inject()
    val isLogoVisible = remember { MutableTransitionState(false) }
    var isTextVisible by remember { mutableStateOf(false) }
    val transition = updateTransition(isLogoVisible)
    val infiniteTransition = rememberInfiniteTransition("RidgeRotate")
    val blur by transition.animateFloat(
        transitionSpec = { spring(stiffness = StiffnessMediumLow * 2) },
        targetValueByState = { if (it) 0f else 10f },
    )
    val ridgeRotation by infiniteTransition.animateFloat(
        initialValue = 0f,
        targetValue = 360f,
        animationSpec = infiniteRepeatable(tween(10000, easing = LinearEasing), RepeatMode.Restart),
        label = "RidgeRotate",
    )
    val textRotation by infiniteTransition.animateFloat(
        initialValue = 360f,
        targetValue = 0f,
        animationSpec = infiniteRepeatable(tween(10000, easing = LinearEasing), RepeatMode.Restart),
        label = "TextRotate",
    )
    LaunchedEffect(Unit) {
        delay(200)
        isLogoVisible.targetState = true
        delay(1000)
        isTextVisible = true
    }
    Box(
        contentAlignment = Alignment.Center,
        modifier = Modifier.fillMaxSize(),
    ) {
        val logoSize = ChatTheme.dimens.imageThumbnail
        AnimatedVisibility(
            visibleState = isLogoVisible,
            enter = scaleIn() + fadeIn(),
            exit = scaleOut() + fadeOut(),
            modifier = Modifier
                .size(logoSize)
                .blur(blur.dp, edgeTreatment = BlurredEdgeTreatment.Unbounded),
        ) {
            Image(
                painter = painterResource(Res.drawable.ic_seers_no_ridges),
                contentDescription = null,
            )
            Image(
                painter = painterResource(Res.drawable.ic_seers_ridges),
                contentDescription = null,
                modifier = Modifier.rotate(ridgeRotation),
            )
        }
        AnimatedVisibility(
            visible = isTextVisible,
            enter = fadeIn(),
            exit = fadeOut(),
            modifier = Modifier.size((logoSize.value * 1.35).dp),
        ) {
            Image(
                painter = painterResource(Res.drawable.ic_seers_welcome_text),
                contentDescription = stringResource(Res.string.tutorial_welcome_message, APP_NAME),
                modifier = Modifier.rotate(textRotation),
            )
        }
        AnimatedVisibility(
            visible = isTextVisible && isAppLocaleSupported(),
            enter = fadeIn() + slideInVertically { it / 2 },
            modifier = Modifier.align(Alignment.BottomCenter),
        ) {
            Column(
                horizontalAlignment = Alignment.CenterHorizontally,
                modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingXLarge * if (isPortraitMode()) 2 else 1),
            ) {
                TextLine(
                    text = stringResource(Res.string.tutorial_choose_language),
                    style = ChatTheme.typography.h6,
                )
                Row(
                    horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                    modifier = Modifier.padding(top = ChatTheme.dimens.spacingMedium),
                ) {
                    ChatLocale.entries.forEach { locale ->
                        BasicText(
                            text = locale.emoji,
                            style = ChatTheme.typography.h3,
                            modifier = Modifier
                                .clip(ChatTheme.shapes.small)
                                .background(ChatTheme.colors.background)
                                .onClick {
                                    setAppLocale(locale)
                                    settingsRepo.update { settingsRepo.settings.generalSettings.copy(locale = locale) }
                                    navController.navigate(TutorialAppUsageRoute)
                                }
                                .padding(ChatTheme.dimens.spacingSmall),
                        )
                    }
                }
            }
        }
        AnimatedVisibility(
            visible = isTextVisible && !isAppLocaleSupported(),
            enter = fadeIn() + slideInVertically { it / 2 },
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .padding(bottom = ChatTheme.dimens.spacingXLarge),
        ) {
            SimpleTextButton(
                text = stringResource(Res.string.tutorial_continue),
                onClick = { navController.navigate(TutorialAppUsageRoute) },
            )
        }
    }
}