package nand.apps.chat.ui.tutorial

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Surface
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.style.TextAlign
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.login.showLoginError
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.settings.dialog.NetworkProxySettingDialog
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.spinner.ChasingDots
import nand.apps.chat.ui.text.SimpleTextCheckbox
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun TutorialNetworkSettingsScreen(
    viewModel: TutorialViewModel,
) {
    val navController = LocalNavController.current
    val state by viewModel.stateData.collectAsState()
    var isUdpEnabled by rememberSaveable { mutableStateOf(true) }
    var isIpv6Enabled by rememberSaveable { mutableStateOf(true) }
    var isLanEnabled by rememberSaveable { mutableStateOf(true) }
    var isProxyDialogVisible by rememberSaveable { mutableStateOf(false) }
    val errorSnackbar = remember { SnackbarHostState() }
    Box(
        modifier = Modifier.fillMaxSize(),
    ) {
        TutorialStepColumn(
            title = stringResource(Res.string.tutorial_network_settings_title),
            modifier = Modifier.align(Alignment.Center),
        ) {
            LaunchedEffect(Unit) {
                viewModel.loginResult.collect { result ->
                    if (result == ChatEngineInitResult.SUCCESS) {
                        viewModel.completeTutorial()
                        navController.navigate(MainRoute, reset = true)
                    } else {
                        errorSnackbar.showLoginError(result)
                    }
                }
            }
            Text(
                text = stringResource(Res.string.tutorial_network_settings_desc),
                color = ChatTheme.colors.onBackground,
                style = ChatTheme.typography.body2,
                textAlign = TextAlign.Center,
                modifier = Modifier.padding(
                    vertical = ChatTheme.dimens.spacingMedium,
                    horizontal = ChatTheme.dimens.spacingXLarge,
                ),
            )
            SimpleTextCheckbox(
                text = stringResource(Res.string.settings_network_udp_enabled_title),
                isChecked = isUdpEnabled,
                onCheckedChange = { isUdpEnabled = it },
            )
            SimpleTextCheckbox(
                text = stringResource(Res.string.settings_network_ipv6_enabled_title),
                isChecked = isIpv6Enabled,
                onCheckedChange = { isIpv6Enabled = it },
            )
            SimpleTextCheckbox(
                text = stringResource(Res.string.settings_network_lan_enabled_title),
                isChecked = isLanEnabled,
                onCheckedChange = { isLanEnabled = it },
            )
            SimpleTextButton(
                text = stringResource(Res.string.settings_network_proxy_title),
                onClick = { isProxyDialogVisible = true },
            )
            SimpleTextButton(
                text = stringResource(Res.string.tutorial_complete),
                onClick = {
                    viewModel.updateNetworkSettings(
                        isUdpEnabled = isUdpEnabled,
                        isIpv6Enabled = isIpv6Enabled,
                        isLanEnabled = isLanEnabled,
                    )
                    viewModel.performLogin()
                },
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbar,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .fillMaxWidth()
                .padding(ChatTheme.dimens.spacingLarge)
        )
        if (state.isLoggingIn) {
            Surface(
                modifier = Modifier
                    .background(ChatTheme.colors.backgroundFloating)
                    .matchParentSize()
            ) {
                ChasingDots()
            }
        }
    }
    if (isProxyDialogVisible) {
        NetworkProxySettingDialog(
            onUpdateProxy = { viewModel.updateProxySettings(it) },
            onDismiss = { isProxyDialogVisible = false },
        )
    }
}