package nand.apps.chat.ui.tutorial

import androidx.compose.foundation.layout.*
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.style.TextAlign
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.login.showLoginError
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.text.SimplePasswordField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun TutorialPasswordScreen(viewModel: TutorialViewModel) {
    val state by viewModel.stateData.collectAsState()
    val navController = LocalNavController.current
    val focusManager = LocalFocusManager.current
    var password by rememberTextFieldValue()
    var confirmedPassword by rememberTextFieldValue()
    val passwordFocus = remember { FocusRequester() }
    val errorSnackbar = remember { SnackbarHostState() }
    Box(
        modifier = Modifier.fillMaxSize(),
    ) {
        TutorialStepColumn(
            title = stringResource(Res.string.tutorial_enter_password),
            modifier = Modifier.align(Alignment.Center),
        ) {
            // Account data may not have a password (data is unencrypted) but we want to discourage this
            // So blank passwords are only prevented when creating a new account
            val isPasswordValid = !viewModel.isNewAccount || password.text.isNotBlank() && confirmedPassword.text == password.text
            LaunchedEffect(Unit) {
                if (getPlatform().isDesktop) {
                    passwordFocus.requestFocus()
                }
                viewModel.authResult.collect { success ->
                    if (success) {
                        navController.navigate(TutorialAppSettingsRoute, reset = true)
                    } else {
                        errorSnackbar.showLoginError(ChatEngineInitResult.INVALID_CREDENTIALS)
                    }
                }
            }
            val performAuthCheck = { if (isPasswordValid) viewModel.performAuthCheck(password.text) }
            SimplePasswordField(
                password = password,
                onValueChange = { password = it },
                onToggleVisibility = { passwordFocus.requestFocus() },
                onSubmit = {
                    if (viewModel.isNewAccount) focusManager.moveFocus(FocusDirection.Next) else performAuthCheck() 
                },
                isEnabled = !state.isLoggingIn,
                modifier = Modifier.focusRequester(passwordFocus),
            )
            if (viewModel.isNewAccount) {
                SimplePasswordField(
                    password = confirmedPassword,
                    label = Res.string.login_new_profile_confirm_password,
                    onValueChange = { confirmedPassword = it },
                    onSubmit = { performAuthCheck() },
                    isEnabled = password.text.isNotBlank() && !state.isLoggingIn,
                )
            }
            Text(
                text = stringResource(Res.string.tutorial_enter_password_desc),
                color = ChatTheme.colors.onBackground,
                style = ChatTheme.typography.body2,
                textAlign = TextAlign.Center,
                modifier = Modifier.padding(
                    vertical = ChatTheme.dimens.spacingMedium,
                    horizontal = ChatTheme.dimens.spacingXLarge,
                ),
            )
            SimpleTextButton(
                text = stringResource(Res.string.tutorial_continue),
                isEnabled = isPasswordValid && !state.isLoggingIn,
                onClick = { performAuthCheck() },
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbar,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .fillMaxWidth()
                .padding(ChatTheme.dimens.spacingLarge)
        )
        if (state.isLoggingIn) {
            CircularProgressIndicator(
                modifier = Modifier.align(Alignment.Center),
            )
        }
    }
}