package nand.apps.chat.ui.tutorial

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.ChatSettingsData
import nand.apps.chat.model.settings.EmbedSettingsData
import nand.apps.chat.model.settings.call.CallSettingsData
import nand.apps.chat.model.settings.filetransfer.FileTransferSettingsData
import nand.apps.chat.model.settings.net.NetworkProxySettingsData
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.profile.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.util.fifoSharedFlow

class TutorialViewModel(
    private val fileSystem: ChatFileSystem,
    private val settingsRepo: SettingsRepo,
    private val credentialsStore: ProfileCredentialsStore,
    private val authenticator: ProfileAuthenticator,
    private val engine: ChatEngine,
    private val eventHandler: ChatEventHandler,
) : ViewModel() {

    val stateData = MutableStateFlow(TutorialState())
    val authResult = fifoSharedFlow<Boolean>()
    val loginResult = fifoSharedFlow<ChatEngineInitResult>()
    
    val username get() = stateData.value.username
    val password get() = stateData.value.password
    val isNewAccount get() = stateData.value.isNewAccount
    val usedAppBefore get() = stateData.value.usedAppBefore
    val settings get() = settingsRepo.settings
    
    fun updateState(
        username: String = this.username,
        usedAppBefore: Boolean = this.usedAppBefore,
        isNewAccount: Boolean = this.isNewAccount,
    ) {
        stateData.tryEmit(
            stateData.value.copy(
                username = username,
                isNewAccount = isNewAccount,
                usedAppBefore = usedAppBefore,
            )
        )
    }
    
    fun hasExistingProfiles() = fileSystem.directoryHasProfiles(fileSystem.paths.toxPath)
    
    fun getDefaultImportDirectory(): String {
        return when {
            hasExistingProfiles() -> fileSystem.paths.toxPath
            else -> fileSystem.paths.downloadPath
        }.toString()
    }
    
    fun getDefaultProfiles() = fileSystem.getDirectory(getDefaultImportDirectory()).getProfiles()
    
    fun selectProfile(path: String): Boolean {
        val selection = fileSystem.selectProfile(path) ?: return false
        if (selection.isImportRequired) {
            val (result, _) = fileSystem.importProfile(
                username = selection.username,
                filePath = selection.filePath.toString(),
            )
            if (result == ProfileImportResult.INVALID_FILE) {
                return false
            }
        }
        updateState(
            username = selection.username,
            isNewAccount = false,
        )
        return true
    }
    
    fun performAuthCheck(password: String) {
        viewModelScope.launch {
            stateData.emit(stateData.value.copy(isLoggingIn = true))
            val success = authenticator.isLoginCorrect(getProfileData(password))
            stateData.emit(
                stateData.value.copy(
                    isLoggingIn = false,
                    password = if (success) password else null,
                )
            )
            if (success) {
                settingsRepo.update { generalSettings.copy(loginUsername = username) }
                credentialsStore.savePassword(password)
            }
            authResult.emit(success)
        }
    }
    
    fun performLogin() {
        viewModelScope.launch {
            stateData.emit(stateData.value.copy(isLoggingIn = true))
            val result = engine.initialize(getProfileData(password.orEmpty()))
            stateData.emit(stateData.value.copy(isLoggingIn = false))
            if (result == ChatEngineInitResult.SUCCESS) {
                eventHandler += LoginCompleteEvent
            }
            loginResult.emit(result)
        }
    }
    
    fun completeTutorial() {
        settingsRepo.update { generalSettings.copy(isTutorialComplete = true) }
    }
    
    fun updateAppSettings(setting: TutorialAppSetting) {
        settingsRepo.update {
            when (setting) {
                TutorialAppSetting.MAXIMUM -> {
                    // Nothing to do - the default settings are already "maximum"
                    this
                }
                TutorialAppSetting.COMPATIBILITY -> copy(
                    chatSettings = ChatSettingsData(
                        isTextStylesEnabled = false,
                        isInputTextStylesEnabled = false,
                        isInputTagPreviewEnabled = false,
                        isTextEffectsEnabled = false,
                        isMentionsEnabled = false,
                        isMessageEditsEnabled = false,
                        isMessageReactionsEnabled = false,
                    )
                )
                TutorialAppSetting.SECURITY -> copy(
                    chatSettings = ChatSettingsData(
                        isMessageHistoryEnabled = false,
                        isAnimatedEmojisEnabled = false,
                    ),
                    embedSettings = EmbedSettingsData(
                        isEnabled = false,
                        isImagesEnabled = false,
                    ),
                    fileSettings = FileTransferSettingsData(
                        isDragDropEnabled = false,
                        isCopyPasteEnabled = false,
                        isGroupUploadEnabled = false,
                    ),
                )
                TutorialAppSetting.MINIMAL -> copy(
                    chatSettings = ChatSettingsData(
                        isTextStylesEnabled = false,
                        isInputTextStylesEnabled = false,
                        isInputTagPreviewEnabled = false,
                        isTextEffectsEnabled = false,
                        isAvatarsEnabled = false,
                        isMessageHistoryEnabled = false,
                        isAnimatedEmojisEnabled = false,
                        isMentionsEnabled = false,
                        isMessageEditsEnabled = false,
                        isMessageReactionsEnabled = false,
                        isPeerNameMessagesEnabled = false,
                    ),
                    callSettings = CallSettingsData(
                        isEnabled = false,
                        isVoicemailEnabled = false,
                    ),
                    embedSettings = EmbedSettingsData(
                        isEnabled = false,
                        isImagesEnabled = false,
                    ),
                    fileSettings = FileTransferSettingsData(
                        isDragDropEnabled = false,
                        isCopyPasteEnabled = false,
                        isGroupUploadEnabled = false,
                    ),
                )
            }
        }
    }
    
    fun updateNetworkSettings(
        isUdpEnabled: Boolean,
        isIpv6Enabled: Boolean,
        isLanEnabled: Boolean,
    ) {
        settingsRepo.update {
            networkSettings.copy(
                isUdpEnabled = isUdpEnabled,
                isIPv6Enabled = isIpv6Enabled,
                isLanEnabled = isLanEnabled,
            )
        }
    }
    
    fun updateProxySettings(proxy: NetworkProxySettingsData) {
        settingsRepo.update { networkSettings.copy(proxy = proxy) }
    }
    
    private fun getProfileData(password: String): ChatProfileData {
        val dir = fileSystem.getProfileDirectory(username)
        val profile = dir.getPrimaryProfile() ?: dir.getFilePath(DEFAULT_PROFILE_FILE_NAME)
        return ChatProfileData(
            filePath = profile.toString(),
            username = username,
            password = password,
            networkSettings = settings.networkSettings,
        )
    }
}