package nand.apps.chat.util

import androidx.compose.runtime.Composable
import androidx.compose.runtime.remember
import androidx.compose.ui.text.*
import androidx.compose.ui.text.font.FontFamily
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.TransformedText
import androidx.compose.ui.text.style.BaselineShift
import androidx.compose.ui.text.style.TextDecoration
import androidx.compose.ui.unit.sp
import com.fleeksoft.ksoup.Ksoup
import com.fleeksoft.ksoup.nodes.Document
import com.fleeksoft.ksoup.nodes.Element
import com.fleeksoft.ksoup.nodes.TextNode
import io.ktor.client.*
import io.ktor.client.plugins.*
import io.ktor.client.request.*
import io.ktor.client.statement.*
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.JsonObject
import nand.apps.chat.text.copy
import nand.apps.chat.ui.theme.ChatThemeScope

private val URL_REGEX = Regex("https?://([\\w.\\-]{1,255}\\.[a-z]{1,32})\\b[-\\w()@$:;%+.,~#?&/=']*")
private val HOST_CHARS_REGEX = Regex("[\\w.\\-:]{1,255}")

fun CharSequence.isUrl() = URL_REGEX.matches(this)

fun CharSequence.getHostName() = URL_REGEX.matchEntire(this)?.groupValues?.getOrNull(1)

fun CharSequence.findUrls() = URL_REGEX.findAll(this).map { it.value }.toList()

fun TextFieldValue.takeIfValidHost(fallback: TextFieldValue) =
    if (text.isEmpty() || HOST_CHARS_REGEX.matches(text)) this else fallback

@OptIn(ExperimentalTextApi::class)
fun TransformedText.addUrlAnnotations() = copy(
    text = buildAnnotatedString {
        append(text)
        URL_REGEX.findAll(text).forEach { result ->
            val start = result.range.first
            val end = result.range.last + 1
            addUrlAnnotation(UrlAnnotation(result.value), start, end)
        }
    },
)

@OptIn(ExperimentalTextApi::class)
fun TransformedText.applyUrlStyles(urlStyle: SpanStyle) = copy(
    text = buildAnnotatedString {
        append(text)
        text.getUrlAnnotations().forEach {
            addStyle(urlStyle, it.start, it.end)
        }
    },
)

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.getUrlAnnotations() =
    getUrlAnnotations(0, length)

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.hasUrlAnnotations(start: Int, end: Int) =
    getUrlAnnotations(start, end).isNotEmpty()

fun createDefaultHttpClient() = HttpClient {
    install(UserAgent) {
        // Generic web browser user agent to help avoid bot filters
        agent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Safari/537.36"
    }
}

suspend fun HttpClient.getHtmlDocument(url: String): Document? {
    return try {
        Ksoup.parse(get(url).bodyAsText())
    } catch (ignore: Exception) {
        null
    }
}

suspend fun HttpClient.getJsonObject(url: String): JsonObject? {
    return try {
        Json.decodeFromString(get(url).bodyAsText())
    } catch (ignore: Exception) {
        null
    }
}

@Composable
fun rememberHtmlAnnotatedString(
    htmlString: String,
    onElement: ChatThemeScope.(Element) -> SpanStyle? = { null },
) = ChatThemeScope {
    remember(htmlString) {
        buildHtmlAnnotatedString(htmlString, this, onElement)
    }
}

private fun buildHtmlAnnotatedString(
    htmlString: String,
    scope: ChatThemeScope,
    onElement: ChatThemeScope.(Element) -> SpanStyle? = { null },
) = buildAnnotatedString {
    parseElement(Ksoup.parse(htmlString).body(), scope, onElement)
}

private fun AnnotatedString.Builder.parseElement(
    element: Element,
    scope: ChatThemeScope,
    onElement: ChatThemeScope.(Element) -> SpanStyle?,
) {
    element.childNodes().forEach { node ->
        when (node) {
            is TextNode -> {
                val text = node.text()
                append(text)
            }
            is Element -> when (node.tagName()) {
                "br" -> append("\n")
                else -> withStyle(scope.onElement(node) ?: scope.getDefaultHtmlStyle(node.tagName())) {
                    parseElement(node, scope, onElement)
                }
            }
        }
    }
}

private fun ChatThemeScope.getDefaultHtmlStyle(tagName: String) = when (tagName) {
    "b" -> SpanStyle(fontWeight = FontWeight.Bold)
    "i", "em" -> SpanStyle(fontStyle = FontStyle.Italic)
    "s" -> SpanStyle(textDecoration = TextDecoration.LineThrough)
    "sub" -> SpanStyle(
        baselineShift = BaselineShift.Subscript,
        fontSize = 10.sp,
    )
    "sup" -> SpanStyle(
        baselineShift = BaselineShift.Superscript,
        fontSize = 10.sp,
    )
    "code" -> SpanStyle(
        fontFamily = FontFamily.Monospace,
        background = colors.surface3,
    )
    else -> SpanStyle()
}