package nand.apps.chat.util

import androidx.compose.ui.text.*
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.util.fastForEach
import kotlinx.coroutines.runBlocking
import nand.apps.chat.text.UNICODE_FONT_TAG
import okio.ByteString.Companion.toByteString
import okio.utf8Size
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.getString
import kotlin.math.pow
import kotlin.math.roundToInt
import kotlin.uuid.ExperimentalUuidApi
import kotlin.uuid.Uuid

private const val THOUSAND = 1000
private const val TEN_THOUSAND = 10_000
private const val MILLION = 1_000_000
private const val TEN_MILLION = 10_000_000
private const val BILLION = 1_000_000_000
private const val TEN_BILLION = 10_000_000_000

val REGEX_DIGITS = Regex("\\d+")
val REGEX_NON_DIGITS = Regex("\\D+")

const val CHARACTER_RANGE_HEX = "0-9A-F"

@OptIn(ExperimentalUuidApi::class)
fun randomUid() = Uuid.random().toString()

fun TextFieldValue.insert(text: String) = TextFieldValue(
    text = this.text.substring(0..<selection.start) + text + this.text.substring(selection.end..<this.text.length),
    selection = TextRange(selection.end + text.length)
)

fun TextFieldValue.insertNewline() = insert("\n")

fun TextFieldValue.replace(regex: Regex, replacement: String) =
    copy(text = text.replace(regex, replacement))

fun TextFieldValue.remove(regex: Regex) =
    copy(text = text.replace(regex, ""))

fun TextFieldValue.takeIfUnderLength(length: Int, fallback: TextFieldValue): TextFieldValue {
    val maxPossibleLength = text.length * 3
    return when {
        maxPossibleLength > length && text.utf8Size() > length -> fallback
        else -> this
    }
}

fun TextFieldValue.takeIfNumber(
    fallback: TextFieldValue,
    maxDigits: Int = Int.MAX_VALUE,
) = if (text.isEmpty() || REGEX_DIGITS.matches(text) && text.length <= maxDigits) this else fallback

fun String.trimToLength(length: Int) =
    if (this.length > length) substring(0, length) else this

fun String.sha256() = toString().encodeToByteArray().toByteString().sha256().hex().uppercase()

fun String?.toIntOrZero() = this?.toIntOrNull() ?: 0

fun String?.toLongOrZero() = this?.toLongOrNull() ?: 0L

fun String.decodeHex(): ByteArray {
    check(length % 2 == 0) { "Must have an even length" }
    return chunked(2)
        .map { it.toInt(16).toByte() }
        .toByteArray()
}

fun getStringBlocking(resource: StringResource, vararg formatArgs: Any) =
    runBlocking { getString(resource, *formatArgs) }

fun Double.roundToDecimal(decimalPlaces: Int): Float {
    val exp = 10.0.pow(decimalPlaces)
    return (this * exp).roundToInt() / exp.toFloat()
}

fun Double.roundToDecimalString(decimalPlaces: Int) =
    roundToDecimal(decimalPlaces).toString().trimEnd('.', '0')

fun formatFileSize(sizeBytes: Long, kbSize: Int = 1000): String {
    val kb = sizeBytes / kbSize.toDouble()
    val mb = kb / kbSize.toDouble()
    val gb = mb / kbSize.toDouble()
    return when {
        sizeBytes < kbSize -> "$sizeBytes B"
        kb < kbSize -> "${kb.roundToInt()} KB"
        mb < kbSize -> "${mb.roundToDecimalString(if (mb >= 100) 1 else 2)} MB"
        else -> "${gb.roundToDecimalString(if (gb >= 100) 1 else 2)} GB"
    }
}

fun formatViewCount(count: Long) = when {
    count < THOUSAND -> count.toString()
    count < MILLION -> (count / THOUSAND.toDouble()).roundToDecimalString(if (count >= TEN_THOUSAND) 0 else 1) + "K"
    count < BILLION -> (count / MILLION.toDouble()).roundToDecimalString(if (count >= TEN_MILLION) 0 else 1) + "M"
    else -> (count / BILLION.toDouble()).roundToDecimalString(if (count >= TEN_BILLION) 0 else 1) + "B"
}

fun formatPlural(word: String, quantity: Int) =
    if (quantity == 1) word else "${word}s"

fun formatPlural(word: String, quantity: Long) =
    if (quantity == 1L) word else "${word}s"

fun Int.toDurationString(): String {
    var seconds = this
    var minutes = seconds / 60
    val hours = minutes / 60
    seconds -= minutes * 60
    minutes -= hours * 60
    return buildString {
        if (hours > 0) {
            append(hours)
            append(if (minutes < 10) ":0$minutes" else ":$minutes")
        } else {
            append(minutes)
        }
        append(if (seconds < 10) ":0$seconds" else ":$seconds")
    }
}

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.Builder.copyStyles(
    src: AnnotatedString,
    indexMap: Map<Int, Int>,
) {
    // Needs more copy-pasting
    src.spanStyles.fastForEach { range ->
        addStyle(
            style = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.paragraphStyles.fastForEach { range ->
        addStyle(
            style = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getTtsAnnotations(0, src.length).fastForEach { range ->
        addTtsAnnotation(
            ttsAnnotation = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getUrlAnnotations().fastForEach { range ->
        addUrlAnnotation(
            urlAnnotation = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getStringAnnotations(0, src.length).fastForEach { range ->
        addStringAnnotation(
            tag = range.tag,
            annotation = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
}

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.removeStyles() = buildAnnotatedString {
    append(text)
    getTtsAnnotations(0, length).fastForEach { range ->
        addTtsAnnotation(
            ttsAnnotation = range.item,
            start = range.start,
            end = range.end,
        )
    }
    getUrlAnnotations().fastForEach { range ->
        addUrlAnnotation(
            urlAnnotation = range.item,
            start = range.start,
            end = range.end,
        )
    }
    getStringAnnotations(0, length).fastForEach { range ->
        if (range.tag == UNICODE_FONT_TAG) return@fastForEach
        addStringAnnotation(
            tag = range.tag,
            annotation = range.item,
            start = range.start,
            end = range.end,
        )
    }
}

private fun indexError(index: Int): Nothing {
    error("Missing index: $index")
}