package nand.apps.chat.io

import nand.apps.chat.platform.ChatPlatform
import kotlin.test.Test
import kotlin.test.assertEquals

class FileNameTests {
    
    @Test
    fun `Check valid file names for Linux`() {
        assertFileNameValidity(
            platform = ChatPlatform.LINUX,
            fileNames = listOf("test.txt", ".test", "test", "test_txt", "test test", "test: test"),
            isValidExpected = true,
        )
        assertFileNameValidity(
            platform = ChatPlatform.LINUX,
            fileNames = listOf("test.", "", " ", "test ", ".", "../", ". . . ."),
            isValidExpected = false,
        )
    }
    
    @Test
    fun `Check valid file names for Windows`() {
        assertFileNameValidity(
            platform = ChatPlatform.WINDOWS,
            fileNames = listOf("test.txt", ".test", "test", "test_txt", "test test", "TEST"),
            isValidExpected = true,
        )
        assertFileNameValidity(
            platform = ChatPlatform.WINDOWS,
            fileNames = listOf("test.", "", " ", "test ", "te\\st", ".", "../", ". . . .", "test: test", "CON", "lpt3", "aux.txt", "<test>"),
            isValidExpected = false,
        )
    }
    
    @Test
    fun `Sanitize file names for Linux`() {
        assertSanitizedFileNames(
            platform = ChatPlatform.LINUX,
            fileNames = listOf(
                "test.txt" to "test.txt",
                "test" to "test",
                " " to UNTITLED_FILE,
                "." to UNTITLED_FILE,
                ". . . . . ." to UNTITLED_FILE,
                "test." to "test",
                "test " to "test",
                ".test" to ".test",
                "/test" to "_test",
                "test/" to "test_",
                "\\test" to "\\test",
                "test:test" to "test:test",
            ),
        )
    }

    @Test
    fun `Sanitize file names for Windows`() {
        assertSanitizedFileNames(
            platform = ChatPlatform.WINDOWS,
            fileNames = listOf(
                "lpt" to "lpt",
                "LPT9" to UNTITLED_FILE,
                "COM3.txt" to "$UNTITLED_FILE.txt",
                "<test>" to "_test_",
                "test..." to "test",
                "..." to UNTITLED_FILE,
                "NUL. ." to UNTITLED_FILE,
                "\\test/" to "_test_",
                "test:test" to "test_test",
            ),
        )
    }
    
    private fun assertFileNameValidity(
        platform: ChatPlatform,
        fileNames: List<String>,
        isValidExpected: Boolean,
    ) = fileNames.forEach { fileName ->
        assertEquals(
            expected = isValidExpected,
            actual = fileName.isValidFileName(platform),
            message = when (isValidExpected) {
                true -> "\"$fileName\" is expected to be valid on $platform"
                else -> "\"$fileName\" isn't expected to be a valid on $platform"
            },
        )
    }
    
    private fun assertSanitizedFileNames(
        platform: ChatPlatform,
        fileNames: List<Pair<String, String>>,
    ) = fileNames.forEach { (input, expected) ->
        val actual = input.sanitizeFileName(platform)
        assertEquals(
            expected = expected,
            actual = actual,
            message = "$input -> $actual (expected $expected)",
        )
    }
}