package nand.apps.chat.message

import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.test.runTest
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.settings.ChatSettingsData
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.*
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.navigation.MainRoute
import org.koin.core.Koin
import org.koin.core.context.stopKoin
import kotlin.random.Random
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test
import kotlin.test.assertEquals

private const val MAX_MESSAGES = 10

class ChatMessageHistoryTests {

    private lateinit var koin: Koin
    private lateinit var engine: ChatEngine
    private lateinit var avatarRepo: ChatAvatarRepo
    private lateinit var contactRepo: ContactRepo
    private lateinit var selfProfileRepo: SelfProfileRepo
    private lateinit var messageRepo: ChatMessageRepo
    private lateinit var settingsRepo: SettingsRepo
    private val profile = ChatProfileData(
        username = "TEST",
        filePath = "",
    )

    @BeforeTest
    fun setup() {
        koin = initKoinTest().koin.apply {
            engine = get()
            avatarRepo = get()
            contactRepo = get()
            selfProfileRepo = get()
            messageRepo = get()
            settingsRepo = get()
        }
    }
    
    @AfterTest
    fun teardown() {
        stopKoin()
        selfProfileRepo.onEngineStop()
        contactRepo.onEngineStop()
        messageRepo.onEngineStop()
    }
    
    @Test
    fun `Friend message history is maintained between encode and decode`() {
        messageHistoryTest(getFriendList().random())
    }
    
    @Test
    fun `Group message history is maintained between encode and decode`() {
        messageHistoryTest(getGroupList().find { ChatGroupFeature.EDIT_TITLE in it.features }!!)
    }
    
    @Test
    fun `Advanced group message history is maintained between encode and decode`() {
        messageHistoryTest(getGroupList().find { ChatGroupFeature.ROLES in it.features }!!)
    }
    
    @Test
    fun `Message history is trimmed when exceeding max`() {
        messageHistoryTest(
            contact = getFriendList().random(),
            numberMessages = MAX_MESSAGES + 10,
        )
    }
    
    @Test
    fun `Message history edited message`() {
        messageHistoryTest(
            contact = getFriendList().random(),
            editMessageIndex = 5,
        )
    }

    @Test
    fun `Message history removed message`() {
        messageHistoryTest(
            contact = getFriendList().random(),
            removeMessageIndex = 5,
        )
    }
    
    private fun getFriendList() = runBlocking { engine.getFriendList() }
    
    private fun getGroupList() = runBlocking { engine.getGroupList() }
    
    private fun messageHistoryTest(
        contact: ChatContact,
        numberMessages: Int = MAX_MESSAGES,
        editMessageIndex: Int = -1,
        removeMessageIndex: Int = -1,
    ) = runTest {
        avatarRepo.onEngineInit(profile)
        contactRepo.onEngineInit(profile)
        messageRepo.onEngineInit(profile)
        selfProfileRepo.onEngineInit(profile)
        messageRepo.onEngineStart(backgroundScope)
        settingsRepo.update { copy(chatSettings = ChatSettingsData(maxMessages = MAX_MESSAGES)) }
        val selfUid = engine.getSelfUserData().uid
        val messages = MutableList(numberMessages) { index ->
            ChatMessage(
                conversationUid = contact.uid,
                senderUid = when (contact) {
                    is UserData -> if (Random.Default.nextBoolean()) contact.uid else selfUid
                    is ChatGroupData -> contact.peers.values.random().uid
                    else -> error("Unsupported contact")
                },
                text = "Chat message $index",
                timestamp = currentTimeMillis() + index * 1000,
                type = if (index == 0) ChatMessageType.ACTION else ChatMessageType.NORMAL,
                isDelivered = contact is UserData,
            )
        }
        messageRepo.onNavRoute(MainRoute)
        messageRepo.onConversationChanged(contact.uid)
        messages.forEach { messageRepo.addMessage(it) }
        if (editMessageIndex != -1) {
            messageRepo.updateMessage(
                messages[editMessageIndex].copy(
                    text = "Edited message",
                    isEdited = true,
                ),
                localOnly = true,
            )
        }
        if (removeMessageIndex != -1) {
            messageRepo.removeMessage(
                message = messages[removeMessageIndex],
                localOnly = true,
            )
            messages.removeAt(removeMessageIndex)
        }
        contactRepo.onEngineStop()
        messageRepo.onEngineStop()
        runBlocking {
            contactRepo.onEngineInit(profile)
            messageRepo.onEngineInit(profile)
        }
        val actualMessages = messageRepo.getMessages(contact.uid)
        if (MAX_MESSAGES >= numberMessages) {
            assertEquals(messages.size, actualMessages.size, "Messages length isn't equal")
            messages.forEachIndexed { index, message ->
                val expected = message.run {
                    if (editMessageIndex == index) {
                        copy(
                            text = "Edited message",
                            isEdited = true,
                        )
                    } else {
                        this
                    }
                }
                assertEquals(expected, actualMessages[index], "Message at $index isn't equal")
            }
        } else {
            assertEquals(MAX_MESSAGES, actualMessages.size, "Messages length isn't equal")
            for (i in 0..<MAX_MESSAGES) {
                val expected = messages[numberMessages - MAX_MESSAGES + i]
                val actual = actualMessages[i]
                assertEquals(expected, actual, "Message at $i isn't equal")
            }
        }
    }
}