package nand.apps.chat.net.packet

import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.test.runTest
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.net.packet.ChatPacketHeader
import nand.apps.chat.model.net.packet.ChatPacketType
import nand.apps.chat.model.net.packet.ClientVersionPacketData
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.repo.ContactRepo
import org.koin.core.Koin
import org.koin.core.context.stopKoin
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test
import kotlin.test.assertEquals

class ChatPacketTests {
    
    private lateinit var koin: Koin
    private lateinit var engine: ChatEngine
    private lateinit var converter: ChatPacketConverter
    private lateinit var packetRepo: ChatPacketRepo
    private lateinit var contactRepo: ContactRepo
    
    @BeforeTest
    fun setup() {
        koin = initKoinTest().koin.apply {
            engine = get()
            converter = get()
            packetRepo = get()
            contactRepo = get()
        }
        runBlocking {
            contactRepo.onEngineInit(ChatProfileData("", ""))
        }
    }
    
    @AfterTest
    fun teardown() {
        stopKoin()
        contactRepo.onEngineStop()
    }
    
    @Test
    fun `Version packet serialized and deserialized`() = runTest {
        packetRepo.onEngineStart(backgroundScope)
        val payload = ClientVersionPacketData(APP_NAME, 1, 2, 3)
        val packets = converter.serializePackets(payload)
        val friend = runBlocking { engine.getFriendList().random() }
        for (packet in packets) {
            packetRepo.onFriendPacketReceived(
                friendUid = friend.uid,
                packet = packet,
            )?.join()
        }
        assertEquals(
            expected = payload.clientVersion,
            actual = contactRepo.getFriend(friend.uid)?.clientVersion,
        )
    }
    
    @Test
    fun `Version packet from constructed JSON string`() {
        val headerBytes = "{\"uid\":\"123456\",\"type\":\"CLIENT_VERSION\",\"pos\":0,\"len\":123}".encodeToByteArray().run {
            copyOf(size + 1)
        }
        val payloadBytes = "{\"name\":\"Seers Lodge\",\"major\":1,\"minor\":2,\"patch\":3}".encodeToByteArray()
        val packet = headerBytes + payloadBytes
        
        assertEquals(
            expected = ChatPacketHeader(
                uid = "123456",
                type = ChatPacketType.CLIENT_VERSION,
                pos = 0,
                len = 123,
            ),
            actual = converter.getHeader(packet),
        )
        assertEquals(
            expected = ClientVersionPacketData(
                name = APP_NAME,
                major = 1,
                minor = 2,
                patch = 3,
            ),
            actual = converter.getPayloadComplete(ChatPacketType.CLIENT_VERSION, payloadBytes)
        )
    }
}