package nand.apps.chat.text

import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.ExperimentalTextApi
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.input.TransformedText
import androidx.compose.ui.util.fastForEachIndexed
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.settings.ChatSettingsData
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.ui.chat.message.*
import nand.apps.chat.ui.theme.DarkChatColors
import nand.apps.chat.util.getUrlAnnotations
import kotlin.test.assertEquals as kAssertEquals

val TestChatColors = DarkChatColors
private val MENTIONS_REGEX = Regex("@[A-Z0-9]{64}")

@OptIn(ExperimentalTextApi::class)
fun assertEquals(
    expected: AnnotatedString,
    actual: AnnotatedString,
) {
    kAssertEquals(
        expected = expected.text,
        actual = actual.text,
        message = "Text is not equal"
    )
    kAssertEquals(
        expected = expected.spanStyles.size,
        actual = actual.spanStyles.size,
        message = "Number of span styles isn't equal",
    )
    expected.spanStyles.fastForEachIndexed { index, expectedRange ->
        val actualRange = actual.spanStyles[index]
        kAssertEquals(
            expected = expectedRange.item,
            actual = actualRange.item,
            message = "Span style #$index isn't equal",
        )
        kAssertEquals(
            expected = expectedRange.start,
            actual = actualRange.start,
            message = "Start index isn't equal for span style #$index",
        )
        kAssertEquals(
            expected = expectedRange.end,
            actual = actualRange.end,
            message = "End index isn't equal for span style #$index",
        )
    }
    val expectedStringAnnotations = expected.getStringAnnotations(0, expected.length)
    val actualStringAnnotations = actual.getStringAnnotations(0, actual.length)
    kAssertEquals(
        expected = expectedStringAnnotations.size,
        actual = actualStringAnnotations.size,
        message = "Number of string annotations isn't equal",
    )
    expectedStringAnnotations.fastForEachIndexed { index, expectedRange ->
        val actualRange = actualStringAnnotations[index]
        kAssertEquals(
            expected = expectedRange.tag,
            actual = actualRange.tag,
            message = "Annotation tag #$index isn't equal",
        )
        kAssertEquals(
            expected = expectedRange.item,
            actual = actualRange.item,
            message = "Annotation #$index isn't equal",
        )
        kAssertEquals(
            expected = expectedRange.start,
            actual = actualRange.start,
            message = "Start index isn't equal for annotation #$index",
        )
        kAssertEquals(
            expected = expectedRange.end,
            actual = actualRange.end,
            message = "End index isn't equal for annotation #$index",
        )
    }
    val expectedUrls = expected.getUrlAnnotations()
    val actualUrls = actual.getUrlAnnotations()
    kAssertEquals(
        expected = expectedUrls.size,
        actual = actualUrls.size,
        message = "Number of URL annotations isn't equal",
    )
    expectedUrls.fastForEachIndexed { index, expectedRange ->
        val actualRange = actualUrls[index]
        kAssertEquals(
            expected = expectedRange.tag,
            actual = actualRange.tag,
            message = "URL tag #$index isn't equal",
        )
        kAssertEquals(
            expected = expectedRange.item,
            actual = actualRange.item,
            message = "URL #$index isn't equal",
        )
        kAssertEquals(
            expected = expectedRange.start,
            actual = actualRange.start,
            message = "Start index isn't equal for URL #$index",
        )
        kAssertEquals(
            expected = expectedRange.end,
            actual = actualRange.end,
            message = "End index isn't equal for URL #$index",
        )
    }
    kAssertEquals(
        expected = expected,
        actual = actual,
        message = "Annotated string isn't equal",
    )
}

fun assertEquals(
    expected: TransformedText,
    actual: TransformedText,
) = assertEquals(
    expected = expected.text,
    actual = actual.text,
)

fun assertEquals(
    expected: AnnotatedString,
    actual: TransformedText,
) = assertEquals(
    expected = expected,
    actual = actual.text,
)

fun assertEquals(
    expected: StylizedMessage,
    actual: StylizedMessage,
) {
    assertEquals(
        expected = expected.text,
        actual = actual.text,
    )
    kAssertEquals(
        expected = expected.effects,
        actual = actual.effects,
    )
}

fun buildMarkdownText(text: String) =
    text.toTransformedText().addMarkdownStyles()

fun buildFontTagText(text: String) =
    parseStyleTags(text).applyFontAnnotations()

fun parseStyleTags(text: String) =
    text.toTransformedText().applyStyleTags()

fun AnnotatedString.Builder.addFontAnnotation(font: UnicodeFont, start: Int, end: Int) =
    addStringAnnotation(UNICODE_FONT_TAG, font.name, start, end)

fun buildStylizedMessage(
    text: String,
    username: String = "",
    isStylesEnabled: Boolean = true,
    group: ChatGroupData? = null,
) = ChatMessage(
    conversationUid = UserUid(""),
    senderUid = UserUid(""),
    text = text,
    type = if (username.isNotBlank()) ChatMessageType.ACTION else ChatMessageType.NORMAL,
).toStylizedMessage(
    displayName = username,
    params = StylizedTextParameters(
        settings = ChatSettingsData(isTextStylesEnabled = isStylesEnabled),
        colors = TestChatColors,
        mentionsRegex = MENTIONS_REGEX,
        group = group,
    ),
)

fun buildStylizedInputText(
    text: String,
    includeTagPreview: Boolean,
    group: ChatGroupData? = null,
) = AnnotatedString(text).toStylizedInputText(
    params = StylizedTextParameters(
        colors = TestChatColors,
        settings = ChatSettingsData(),
        tagPreviewStyle = if (includeTagPreview) TextStyle.Default else null,
        mentionsRegex = MENTIONS_REGEX,
        group = group,
    ),
)