package nand.apps.chat.text.message

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.*
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.font.FontSynthesis
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.BaselineShift
import androidx.compose.ui.text.style.TextDecoration
import nand.apps.chat.text.*
import kotlin.test.Test

@OptIn(ExperimentalTextApi::class)
class ChatInputTextTests {
    
    private val previewStyle = SpanStyle(
        color = TestChatColors.onBackground2,
        fontSize = TextStyle.Default.fontSize,
        letterSpacing = TextStyle.Default.letterSpacing,
        fontSynthesis = FontSynthesis.None,
        textDecoration = TextDecoration.None,
        fontStyle = FontStyle.Normal,
        fontWeight = FontWeight.Normal,
        baselineShift = BaselineShift.None,
    )
    
    @Test
    fun `Input text with tag preview`() = assertEquals(
        expected = buildAnnotatedString {
            append("This text is \$gre\$Green\$end\$")
            addPreviewStyle(13, 18)
            addStyle(SpanStyle(color = Color.Green), 18, 28)
            addPreviewStyle(23, 28)
        },
        actual = buildStylizedInputText(
            text = "This text is \$gre\$Green\$end\$",
            includeTagPreview = true,
        ),
    )

    @Test
    fun `Input text without tag preview`() = assertEquals(
        expected = buildAnnotatedString {
            append("This text is Green")
            addStyle(SpanStyle(color = Color.Green), 13, 18)
        },
        actual = buildStylizedInputText(
            text = "This text is \$gre\$Green\$end\$",
            includeTagPreview = false,
        ),
    )

    @Test
    fun `Input text with overlapping tag previews`() = assertEquals(
        expected = buildAnnotatedString {
            append("This is \$bold\$bold text that's also \$red\$red\$end\$.")
            addPreviewStyle(8, 14)
            addStyle(SpanStyle(fontWeight = FontWeight.Bold), 14, 50)
            addPreviewStyle(36, 41)
            addStyle(SpanStyle(color = Color.Red), 41, 49)
            addPreviewStyle(44, 49)
        },
        actual = buildStylizedInputText(
            text = "This is \$bold\$bold text that's also \$red\$red\$end\$.",
            includeTagPreview = true,
        ),
    )

    @Test
    fun `Input text with font tag previews`() = assertEquals(
        expected = buildAnnotatedString {
            append("Check out this \$cur\$\uD835\uDCB8\uD835\uDC5C\uD835\uDC5C\uD835\uDCC1 \uD835\uDCBB\uD835\uDC5C\uD835\uDCC3\uD835\uDCC9\$end\$.")
            addPreviewStyle(15, 20)
            addFontAnnotation(UnicodeFont.CURSIVE, 20, 42)
            addPreviewStyle(37, 42)
        },
        actual = buildStylizedInputText(
            text = "Check out this \$cur\$cool font\$end\$.",
            includeTagPreview = true,
        ),
    )
    
    @Test
    fun `Input text with stylized URL`() = assertEquals(
        expected = buildAnnotatedString {
            append("Go to https://en.wikipedia.org/ for more info")
            addUrlAnnotation(UrlAnnotation("https://en.wikipedia.org/"), 6, 31)
            addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 6, 31)
        },
        actual = buildStylizedInputText(
            text = "Go to https://en.wikipedia.org/ for more info",
            includeTagPreview = true,
        ),
    )

    @Test
    fun `Input text with styles disabled`() = assertEquals(
        expected = AnnotatedString("/nostyle \$red\$Some text and \$cur\$fonts\$end\$ and **markdown**"),
        actual = buildStylizedInputText(
            text = "/nostyle \$red\$Some text and \$cur\$fonts\$end\$ and **markdown**",
            includeTagPreview = true,
        ),
    )
    
    private fun AnnotatedString.Builder.addPreviewStyle(start: Int, end: Int) {
        addStyle(previewStyle, start, end)
        addFontAnnotation(UnicodeFont.DEFAULT, start, end)
    }
}