package nand.apps.chat.text.message

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.*
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.BaselineShift
import androidx.compose.ui.unit.sp
import nand.apps.chat.text.*
import nand.apps.chat.ui.chat.message.MessageTextEffect
import nand.apps.chat.ui.chat.message.StylizedMessage
import kotlin.test.Test

class StylizedMessageTests {
    
    @Test
    fun `Regular message with no styling or effects`() = assertEquals(
        expected = StylizedMessage(
            text = AnnotatedString("This is a regular message."),
        ),
        actual = buildStylizedMessage("This is a regular message."),
    )

    @Test
    fun `Message with multiple text colors`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("This is a message with red and blue text.")
                addStyle(SpanStyle(color = Color.Red), 23, 26)
                addStyle(SpanStyle(color = Color.Blue), 31, 35)
            },
        ),
        actual = buildStylizedMessage("This is a message with \$red\$red\$end\$ and \$blu\$blue\$end\$ text."),
    )

    @Test
    fun `Quote message with red text`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append(">Red quote text")
                addStyle(SpanStyle(color = Color.Red), 1, 15)
                addStyle(SpanStyle(color = TestChatColors.quote), 0, 15)
            },
        ),
        actual = buildStylizedMessage(">\$red\$Red quote text"),
    )

    @Test
    fun `Superscript action message`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("NAND is tiny")
                addStyle(SpanStyle(baselineShift = BaselineShift.Superscript, fontSize = 10.sp), 8, 12)
            },
        ),
        actual = buildStylizedMessage(
            text = "is \$sup\$tiny",
            username = "NAND",
        ),
    )

    @Test
    fun `Message containing unicode fonts and markdown 1`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("\uD835\uDD99\uD835\uDD8A\uD835\uDD98\uD835\uDD99ⓣⓔⓢⓣ\uD835\uDD99\uD835\uDD8A\uD835\uDD98\uD835\uDD99")
                addFontAnnotation(UnicodeFont.FRAKTUR_BOLD, 0, 20)
                addFontAnnotation(UnicodeFont.CIRCLED, 8, 12)
                addStyle(SpanStyle(fontStyle = FontStyle.Italic), 12, 20)
            },
        ),
        actual = buildStylizedMessage("\$frab\$test\$cir\$test\$end\$*test*"),
    )

    @Test
    fun `Message containing unicode fonts and markdown 2`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("\uD835\uDD31t")
                addFontAnnotation(UnicodeFont.FRAKTUR, 0, 2)
                addStyle(SpanStyle(fontWeight = FontWeight.Bold), 2, 3)
            },
        ),
        actual = buildStylizedMessage("\$fra\$t\$end\$**t**"),
    )

    @Test
    fun `Message containing unicode fonts and markdown 3`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("\uD835\uDCFD\uD835\uDCEE\uD835\uDCFC\uD835\uDCFD")
                addFontAnnotation(UnicodeFont.CURSIVE_ALT, 0, 8)
                addStyle(SpanStyle(fontWeight = FontWeight.Bold), 0, 8)
            },
        ),
        actual = buildStylizedMessage("\$cur2\$**test**\$end\$"),
    )

    @Test
    fun `Message with styles and markdown with styles removed`() = assertEquals(
        expected = StylizedMessage(
            text = AnnotatedString("Some text and fonts and markdown"),
        ),
        actual = buildStylizedMessage(
            text = "\$red\$Some text and \$cur\$fonts\$end\$ and **markdown**",
            isStylesEnabled = false,
        ),
    )

    @Test
    fun `Message with styles and markdown with style processing disabled`() = assertEquals(
        expected = StylizedMessage(
            text = AnnotatedString("\$red\$Some text and \$cur\$fonts\$end\$ and **markdown**"),
            effects = setOf(MessageTextEffect.NOSTYLE),
        ),
        actual = buildStylizedMessage(
            text = "/nostyle \$red\$Some text and \$cur\$fonts\$end\$ and **markdown**",
        ),
    )

    @Test
    fun `Message containing a couple text effects and styles`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("EARTHQUAKE")
                addStyle(SpanStyle(fontWeight = FontWeight.Bold), 0, 10)
                addStyle(SpanStyle(color = Color.Red), 0, 10)
                addStyle(SpanStyle(fontSize = 18.sp), 0, 10)
            },
            effects = setOf(MessageTextEffect.SHAKE, MessageTextEffect.ZOOM),
        ),
        actual = buildStylizedMessage("/shake /zoom \$red\$\$big\$**EARTHQUAKE**"),
    )

    @Test
    fun `Action message with a text effect`() = assertEquals(
        expected = StylizedMessage(
            text = AnnotatedString("Jimmy is bouncing"),
            effects = setOf(MessageTextEffect.BOUNCE),
        ),
        actual = buildStylizedMessage(
            "/bounce is bouncing",
            username = "Jimmy",
        ),
    )
}