package nand.apps.chat.text.message

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.ExperimentalTextApi
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.UrlAnnotation
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.TextDecoration
import nand.apps.chat.text.*
import nand.apps.chat.ui.chat.message.StylizedMessage
import kotlin.test.Test

@OptIn(ExperimentalTextApi::class)
class UrlMessageTests {
    
    @Test
    fun `Message containing stylized URL`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("Check out this site: https://www.example.com/")
                addUrlAnnotation(UrlAnnotation("https://www.example.com/"), 21, 45)
                addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 21, 45)
            },
        ),
        actual = buildStylizedMessage("Check out this site: https://www.example.com/"),
    )

    @Test
    fun `Message with style tags within URLs are ignored`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("Here's a URL: https://www.example.com/\$cya\$/page with tags")
                addUrlAnnotation(UrlAnnotation("https://www.example.com/\$cya\$/page"), 14, 48)
                addStyle(SpanStyle(color = Color.Red), 9, 58)
                addStyle(SpanStyle(color = Color.Yellow), 49, 58)
                addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 14, 48)
            }
        ),
        actual = buildStylizedMessage("Here's a \$red\$URL: https://www.example.com/\$cya\$/page \$yel\$with tags"),
    )

    @Test
    fun `Message with markdown within URLs are ignored`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("URL with underscores: https://www.example.com/Thing_page_bla with tags")
                addUrlAnnotation(UrlAnnotation("https://www.example.com/Thing_page_bla"), 22, 60)
                addStyle(SpanStyle(fontWeight = FontWeight.Bold), 9, 20)
                addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 22, 60)
            }
        ),
        actual = buildStylizedMessage("URL with **underscores**: https://www.example.com/Thing_page_bla with tags"),
    )

    @Test
    fun `Message with URLs and unicode fonts`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("Cool font \uD835\uDCF1\uD835\uDCFD\uD835\uDCFD\uD835\uDCF9\uD835\uDCFC://\uD835\uDD00\uD835\uDD00\uD835\uDD00.\uD835\uDCEE\uD835\uDD01\uD835\uDCEA\uD835\uDCF6\uD835\uDCF9\uD835\uDCF5\uD835\uDCEE.\uD835\uDCEC\uD835\uDCF8\uD835\uDCF6/")
                addUrlAnnotation(UrlAnnotation("https://www.example.com/"), 10, 52)
                addFontAnnotation(UnicodeFont.CURSIVE_ALT, 10, 52)
                addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 10, 52)
            }
        ),
        actual = buildStylizedMessage("Cool font \$cur2\$https://www.example.com/"),
    )

    @Test
    fun `Message with styles while setting disabled`() = assertEquals(
        expected = StylizedMessage(
            text = buildAnnotatedString {
                append("Only styling for URLs like https://youtube.com/")
                addUrlAnnotation(UrlAnnotation("https://youtube.com/"), 27, 47)
                addStyle(SpanStyle(color = TestChatColors.primary, textDecoration = TextDecoration.Underline), 27, 47)
            }
        ),
        actual = buildStylizedMessage(
            text = "**Only styling** for \$yel\$URLs\$end\$ like https://youtube.com/",
            isStylesEnabled = false,
        ),
    )
}