package nand.apps.chat.audio

import nand.apps.chat.log.getLogger
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.ushl
import javax.sound.sampled.AudioSystem
import javax.sound.sampled.DataLine
import javax.sound.sampled.TargetDataLine
import kotlin.math.roundToInt

// Buffer lengths in milliseconds that Tox supports
// The qTox default is 20ms, however certain operating systems don't handle <33ms properly
// See https://gitlab.com/nand-apps/seers-lodge/-/issues/86
private val VALID_BUFFER_LENGTHS_MS = listOf(2.5f, 5f, 10f, 20f, 40f, 60f)

class DesktopAudioRecorder : DesktopAudioLine(), AudioRecorder {
    
    private val logger = getLogger("AudioRecorder")

    override val isActive get() = isRunning
    override val onFrame = fifoSharedFlow<AudioFrame>()

    override fun start() {
        if (!isRunning) {
            isRunning = true
            val t = Thread(this, "AudioRecorder")
            t.isDaemon = true
            t.priority = Thread.MAX_PRIORITY
            t.start()
        }
    }

    override fun stop() {
        isRunning = false
    }

    override fun run() {
        try {
            val speakerInfo = DataLine.Info(TargetDataLine::class.java, audioFormat)
            val recorder = AudioSystem.getLine(speakerInfo) as TargetDataLine
            val maxBufferSize = recorder.bufferSize
            val minBufferSize = maxBufferSize / 16
            val validBufferSizes = VALID_BUFFER_LENGTHS_MS.map { sizeMs ->
                ((sizeMs / 1000f) * sampleRate).roundToInt() * channels * 2
            }
            val bufferSize = validBufferSizes.find { it >= minBufferSize }
                ?: error("Cannot find small enough buffer size! Min = $minBufferSize")
            recorder.open(audioFormat, bufferSize)
            recorder.start()

            val data = ByteArray(bufferSize)
            while (isRunning) {
                val numRead = recorder.read(data, 0, bufferSize)
                if (numRead == bufferSize) {
                    onFrame.tryEmit(createFrame(data))
                }
            }

            recorder.drain()
            recorder.close()
        } catch (e: Exception) {
            logger.error("Something catastrophic happened", e)
        }
    }
    
    private fun createFrame(bytes: ByteArray): AudioFrame {
        val samples = ShortArray(bytes.size / 2)
        var i = 0
        for (b in bytes.indices step 2) {
            var s = (bytes[b + 1] ushl 8) or (bytes[b] ushl 0)
            if (s > Short.MAX_VALUE) s -= 65536
            samples[i++] = s.toShort()
        }
        return AudioFrame(
            samples = samples,
            sampleRate = sampleRate,
            channels = channels,
        )
    }
}