package nand.apps.chat.image

import androidx.compose.ui.graphics.asSkiaBitmap
import androidx.compose.ui.graphics.toAwtImage
import androidx.compose.ui.graphics.toComposeImageBitmap
import coil3.ImageLoader
import coil3.asImage
import coil3.decode.*
import coil3.fetch.SourceFetchResult
import coil3.request.Options
import coil3.size.Precision
import com.sksamuel.scrimage.ImmutableImage
import okio.use
import org.jetbrains.skia.Image

/**
 * Alternate image loader that's meant to override the default SkiaImageLoader which
 * does a pretty crappy job of resampling images. This image loader trades off decode
 * performance for image quality.
 */
class ScrimageDecoder(
    private val source: ImageSource,
    private val options: Options,
) : Decoder {

    override suspend fun decode(): DecodeResult {
        val bytes = source.source().use { it.readByteArray() }
        val srcImage = try {
            // Load image using Scrimage
            ImmutableImage.loader().fromBytes(bytes)
        } catch (e: Exception) {
            // Otherwise fallback to Skia
            ImmutableImage.wrapAwt(Image.makeFromEncoded(bytes).toComposeImageBitmap().toAwtImage())
        }
        
        // Mostly copied from SkiaImageDecoder.decode()
        val srcWidth = srcImage.width
        val srcHeight = srcImage.height
        val dstSize = options.sizePx(srcWidth x srcHeight)
        var multiplier = DecodeUtils.computeSizeMultiplier(
            srcWidth = srcWidth,
            srcHeight = srcHeight,
            dstWidth = dstSize.width,
            dstHeight = dstSize.height,
            scale = options.scale,
        )

        if (options.precision == Precision.INEXACT) {
            multiplier = multiplier.coerceAtMost(1.0)
        }

        val dstWidth = (multiplier * srcWidth).toInt()
        val dstHeight = (multiplier * srcHeight).toInt()

        val scaled = srcImage.scaleTo(dstWidth, dstHeight)
        val dstBitmap = scaled.awt().toComposeImageBitmap().asSkiaBitmap()
        val isSampled = dstBitmap.width < srcWidth || dstBitmap.height < srcHeight

        return DecodeResult(
            image = dstBitmap.asImage(),
            isSampled = isSampled,
        )
    }
    
    class Factory : Decoder.Factory {

        override fun create(result: SourceFetchResult, options: Options, imageLoader: ImageLoader): Decoder {
            return ScrimageDecoder(result.source, options)
        }
    }
}