package nand.apps.chat.ui

import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.awt.ComposeWindow
import androidx.compose.ui.unit.dp
import androidx.compose.ui.window.*
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.BuildConfig.APP_VERSION
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.event.*
import nand.apps.chat.monitor.DesktopUserActivityMonitor
import nand.apps.chat.ui.app.App
import nand.apps.chat.ui.dispatcher.DesktopKeyEventDispatcher
import nand.apps.chat.ui.event.DesktopEventHandler
import nand.apps.chat.ui.notification.DesktopNotificationHost
import nand.apps.chat.ui.notification.UnreadMessageTracker
import nand.apps.chat.ui.window.*
import org.koin.compose.getKoin
import java.awt.GraphicsEnvironment
import java.util.prefs.Preferences

const val DEFAULT_WINDOW_TITLE = "$APP_NAME v$APP_VERSION"
private const val PREF_WINDOW_WIDTH = "windowWidth"
private const val PREF_WINDOW_HEIGHT = "windowHeight"
private const val PREF_WINDOW_MAXIMIZED = "windowMaximized"

// TODO - Break up this function
@Composable
fun ApplicationScope.DesktopApp() {

    // Start chat engine once login has been completed
    var isAppLoaded by rememberSaveable { mutableStateOf(false) }
    var isShuttingDown by rememberSaveable { mutableStateOf(false) }
    
    // Window size
    val screenDevice = GraphicsEnvironment.getLocalGraphicsEnvironment().defaultScreenDevice
    val maxWindowBounds = screenDevice.getMaxWindowBounds()
    val prefs = Preferences.userNodeForPackage(javaClass)
    val windowState = rememberWindowState(
        width = prefs.getInt(PREF_WINDOW_WIDTH, 780).coerceAtMost(maxWindowBounds.width).dp,
        height = prefs.getInt(PREF_WINDOW_HEIGHT, 450).coerceAtMost(maxWindowBounds.height).dp,
        placement = when {
            prefs.getBoolean(PREF_WINDOW_MAXIMIZED, false) -> WindowPlacement.Maximized
            else -> WindowPlacement.Floating
        },
        position = WindowPosition(Alignment.Center),
    )
    var isWindowVisible by remember { mutableStateOf(true) }
    LaunchedEffect(windowState.size, windowState.placement) {
        prefs.putInt(PREF_WINDOW_WIDTH, windowState.size.width.value.toInt())
        prefs.putInt(PREF_WINDOW_HEIGHT, windowState.size.height.value.toInt())
        prefs.putBoolean(PREF_WINDOW_MAXIMIZED, windowState.placement == WindowPlacement.Maximized)
    }

    // Unread notifications
    var unreadMessageCount by remember { mutableStateOf(0) }
    UnreadMessageTracker(
        onUnreadMessage = { unreadMessageCount++ },
        onReadMessages = { unreadMessageCount = 0 },
    )
    
    // Auto-away detection
    DesktopUserActivityMonitor()

    var windowTitle by remember { mutableStateOf(DEFAULT_WINDOW_TITLE) }
    var composeWindow by remember { mutableStateOf<ComposeWindow?>(null) }

    DesktopEventHandler(
        onWindowTitle = { windowTitle = it },
        onAppLoaded = { isAppLoaded = true },
        onAppFocused = { event ->
            when {
                !isWindowVisible -> isWindowVisible = true
                windowState.isMinimized -> windowState.isMinimized = false
                composeWindow?.isActive == false -> composeWindow?.toFront()
                isWindowVisible && event.toggle -> isWindowVisible = false
            }
        },
    )

    val koin = getKoin()
    val keyEventDispatcher = remember { DesktopKeyEventDispatcher() }
    Window(
        title = (if (unreadMessageCount > 0) "*" else "") + windowTitle,
        icon = windowIconPainter(unreadMessageCount > 0),
        onCloseRequest = {
            if (isAppLoaded) {
                val engine: ChatEngine by koin.inject()
                engine.stop()
            }
            isShuttingDown = true
        },
        state = windowState,
        visible = isWindowVisible,
        onPreviewKeyEvent = rememberShortcutListener(
            windowState = windowState,
            keyEventDispatcher = keyEventDispatcher,
        ),
    ) {
        composeWindow = window
        if (isShuttingDown) {
            DesktopShutdownScreen()
        } else {
            keyEventDispatcher.CompositionLocalProvider {
                App()
            }
        }
    }
    
    if (isAppLoaded) {
        ChatWidget()
        DesktopNotificationHost(composeWindow)
        WindowFocusEventHandler(
            window = composeWindow,
            windowState = windowState,
            isWindowVisible = isWindowVisible,
        )
    }
    
    LaunchedEffect(isShuttingDown) {
        if (isShuttingDown) {
            exitApplication()
        }
    }
}
