package nand.apps.chat.ui.dev

import androidx.compose.foundation.*
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.text.BasicTextField
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.foundation.window.WindowDraggableArea
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Close
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.drawWithContent
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.geometry.Size
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.RectangleShape
import androidx.compose.ui.graphics.SolidColor
import androidx.compose.ui.graphics.drawscope.Stroke
import androidx.compose.ui.graphics.drawscope.clipRect
import androidx.compose.ui.input.key.Key
import androidx.compose.ui.input.key.key
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.font.FontFamily
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.window.WindowScope
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import nand.apps.chat.dev.DevCommandLogType
import nand.apps.chat.dev.DevConsoleLine
import nand.apps.chat.dev.DevConsoleRepo
import nand.apps.chat.ui.chat.message.MessageCycleDirection
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.modifier.onEnter
import nand.apps.chat.ui.modifier.onKeyDown
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import org.koin.compose.getKoin

private val FRAME_BG_COLOR = Color(160, 160, 160, 128)
private val TRANSPARENT_BLACK = Color(0, 0, 0, 128)
private val BORDER_SELECTION = Color(0, 0, 0, 196)

@Composable
fun DevConsoleScreen(
    onClose: () -> Unit,
    modifier: Modifier = Modifier,
    viewModel: DevConsoleViewModel = koinKmpViewModel(),
    windowScope: WindowScope? = null,
) = ChatTheme {

    val listState = rememberLazyListState()
    var input by rememberTextFieldValue()
    
    val repo: DevConsoleRepo by getKoin().inject()
    val lines by repo.lineObserver.run {
        produceState(emptyList<DevConsoleLine>(), this) {
            collect {
                value = it
                listState.scrollToItem(it.size)
            }
        }
    }
    
    val outerPadding = ChatTheme.dimens.spacingLarge + ChatTheme.dimens.spacingSmall
    val textStyle = ChatTheme.typography.caption.copy(
        color = ChatTheme.colors.onSurface,
        fontFamily = FontFamily.Monospace,
    )
    
    val focusRequester = remember { FocusRequester() }
    val isWindowVisible = windowScope?.window?.isVisible ?: true
    val coroutineScope = rememberCoroutineScope()
    coroutineScope.launch {
        if (isWindowVisible) {
            // XXX - We have to wait a couple frames or else requestFocus() doesn't work
            delay(10L)
            focusRequester.requestFocus()
        }
    }
    
    LaunchedEffect(Unit) {
        viewModel.inputAction.collect { text ->
            input = TextFieldValue(text = text, selection = TextRange(text.length))
        }
    }
    
    Box(
        modifier = modifier.background(FRAME_BG_COLOR),
    ) {
        windowScope?.WindowDraggableArea(
            modifier = Modifier
                .fillMaxWidth()
                .height(ChatTheme.dimens.buttonMedium),
        )
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            modifier = Modifier.padding(outerPadding),
        ) {
            Row(
                modifier = Modifier.fillMaxWidth()
            ) {
                TextLine(
                    text = "Console",
                    style = textStyle,
                    color = ChatTheme.colors.onSurface,
                    modifier = Modifier.weight(1f),
                )
                IconButton(
                    icon = Icons.Outlined.Close,
                    contentDescription = "Close",
                    color = ChatTheme.colors.onBackground,
                    onClick = onClose,
                    modifier = Modifier.size(ChatTheme.dimens.iconSmall),
                )
            }
            Box(
                modifier = Modifier
                    .background(TRANSPARENT_BLACK)
                    .fillMaxWidth()
                    .weight(1f)
                    .drawPixelBoxBorder(),
            ) {
                SelectionContainer {
                    LazyColumn(
                        state = listState,
                        contentPadding = PaddingValues(vertical = ChatTheme.dimens.spacingSmall),
                        modifier = Modifier
                            .fillMaxSize()
                            .padding(
                                start = ChatTheme.dimens.spacingSmall,
                                end = ChatTheme.dimens.spacingXLarge,
                            ),
                    ) {
                        items(
                            items = lines,
                        ) { line ->
                            TextLine(
                                text = line.text,
                                style = textStyle,
                                color = when (line.type) {
                                    DevCommandLogType.LOG -> ChatTheme.colors.onBackground2
                                    DevCommandLogType.WARNING -> ChatTheme.colors.warning
                                    DevCommandLogType.ERROR -> ChatTheme.colors.onError
                                    DevCommandLogType.HIGHLIGHT -> ChatTheme.colors.statusOnline
                                    DevCommandLogType.DEFAULT -> ChatTheme.colors.onSurface
                                }
                            )
                        }
                    }
                }
                VerticalScrollbar(
                    adapter = rememberScrollbarAdapter(listState),
                    style = defaultScrollbarStyle().copy(
                        shape = RectangleShape,
                        unhoverColor = FRAME_BG_COLOR,
                        hoverColor = FRAME_BG_COLOR,
                        thickness = outerPadding,
                    ),
                    modifier = Modifier
                        .align(Alignment.CenterEnd)
                        .fillMaxHeight(),
                )
            }
            Row {
                Box(
                    modifier = Modifier
                        .background(TRANSPARENT_BLACK)
                        .weight(1f)
                        .height(ChatTheme.dimens.iconSmall)
                        .drawPixelBoxBorder(),
                ) {
                    BasicTextField(
                        value = input,
                        onValueChange = { input = it },
                        textStyle = textStyle,
                        cursorBrush = SolidColor(ChatTheme.colors.onBackground),
                        singleLine = true,
                        modifier = Modifier
                            .focusRequester(focusRequester)
                            .matchParentSize()
                            .padding(ChatTheme.dimens.spacingSmall)
                            .onEnter { viewModel.submitInput(input.text) }
                            .onKeyDown { event ->
                                when (event.key) {
                                    Key.DirectionUp -> {
                                        viewModel.cycleHistory(MessageCycleDirection.UP)
                                        true
                                    }
                                    Key.DirectionDown -> {
                                        viewModel.cycleHistory(MessageCycleDirection.DOWN)
                                        true
                                    }
                                    else -> false
                                }
                            },
                    )
                }
                Box(
                    contentAlignment = Alignment.Center,
                    modifier = Modifier
                        .padding(start = ChatTheme.dimens.spacingLarge)
                        .width(ChatTheme.dimens.avatarXLarge)
                        .height(ChatTheme.dimens.iconSmall)
                        .border(Dp.Hairline, Color.Black)
                        .clickable { viewModel.submitInput(input.text) }
                        .drawWithContent {
                            drawContent()
                            clipRect {
                                val stroke = Stroke(Stroke.HairlineWidth)
                                drawRect(
                                    color = Color.White,
                                    topLeft = Offset(1f, 1f),
                                    size = size,
                                    style = stroke,
                                )
                                drawRect(
                                    color = BORDER_SELECTION,
                                    size = Size(size.width - 1f, size.height - 1f),
                                    style = stroke,
                                )
                            }
                        },
                ) {
                    TextLine(
                        text = "Submit",
                        style = textStyle,
                    )
                }
            }
        }
    }
}

private fun Modifier.drawPixelBoxBorder() = drawWithContent {
    drawContent()
    clipRect {
        val stroke = Stroke(Stroke.HairlineWidth)
        drawRect(
            color = BORDER_SELECTION,
            size = Size(size.width + 1f, size.height + 1f),
            style = stroke,
        )
        drawRect(
            color = Color.White,
            topLeft = Offset(-1f, -1f),
            style = stroke,
        )
    }
}