package nand.apps.chat.ui.dev

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.launch
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.BuildConfig.APP_VERSION
import nand.apps.chat.audio.AudioPlayer
import nand.apps.chat.dev.DevCommandDefaultHandlers
import nand.apps.chat.dev.DevCommandLogType
import nand.apps.chat.dev.DevConsoleLine
import nand.apps.chat.dev.DevConsoleRepo
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.theme.ChatThemeSetting
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.chat.message.MessageCycleDirection
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.monitor.AutoAwayMonitor
import nand.apps.chat.util.fifoSharedFlow
import okio.Path.Companion.toPath
import java.io.PrintStream

private const val WARNING_TEXT = "WARN: "

class DevConsoleViewModel(
    private val repo: DevConsoleRepo,
    private val engine: ChatEngine,
    private val settingsRepo: SettingsRepo,
    private val autoAwayMonitor: AutoAwayMonitor,
    private val eventHandler: ChatEventHandler,
    private val audioPlayer: AudioPlayer,
    private val handlers: DevCommandDefaultHandlers,
) : ViewModel() {
    
    private val defaultOut = System.out
    private val defaultErr = System.err
    private val logger = getLogger("DevConsole")

    private val history = mutableListOf<String>()
    private var historyIndex = 0
    private var settings get() = settingsRepo.settings
        set(value) { settingsRepo.settings = value }
    
    val inputAction = fifoSharedFlow<String>()

    init {
        System.setOut(PrintStreamRedirect(defaultOut))
        System.setErr(PrintStreamRedirect(defaultErr))
    }
    
    override fun onCleared() {
        super.onCleared()
        System.setOut(defaultOut)
        System.setErr(defaultErr)
    }
    
    fun submitInput(text: String) {
        printConsole("] $text")
        text.split(';').forEach { handleCommand(it) }
        history += text
        historyIndex = history.size
        inputAction.tryEmit("")
    }
    
    fun cycleHistory(direction: MessageCycleDirection) {
        if (history.isEmpty()) return
        historyIndex = (historyIndex + direction.value).coerceAtLeast(0).coerceAtMost(history.size)
        val inputText = if (historyIndex in history.indices) history[historyIndex] else ""
        inputAction.tryEmit(inputText)
    }
    
    private fun handleCommand(text: String) {
        val rawArgs = text.split(' ')
        val command = rawArgs[0]
        val args = rawArgs.subList(1, rawArgs.size)
        val fullArgs = if (rawArgs.size > 1) text.substringAfter(' ') else ""
        when (command) {
            "clear" -> repo.clearLines()
            "warn" -> logger.warn(fullArgs)
            "error" -> logger.error(fullArgs)
            "log" -> logger.debug(fullArgs)
            "get_friends" -> dumpFriendList()
            "get_groups" -> dumpGroupList()
            "self_addr" -> printConsole(engine.getSelfAddress())
            "auto_away" -> autoAwayMonitor.forceAutoAway()
            "play_sound" -> playSound(fullArgs)
            "play_audio" -> playAudio(fullArgs)
            "version" -> {
                printConsole("$APP_NAME v$APP_VERSION")
                printConsole("Chat engine: ${engine.version}")
            }
            "theme" -> onThemeCommand(args)
            "error_dialog" -> eventHandler += ErrorDialogEvent(fullArgs)
            else -> {
                for (handler in handlers.values) {
                    try {
                        if (handler.onCommand(command, args, fullArgs)) return
                    } catch (e: IllegalStateException) {
                        printConsole(e.message, lineType = DevCommandLogType.ERROR)
                        return
                    }
                }
                printConsole("Unknown command: \"$command\"")
            }
        }
    }
    
    private fun printConsole(
        text: String,
        lineType: DevCommandLogType = DevCommandLogType.DEFAULT,
    ) {
        val line = DevConsoleLine(
            text = text,
            type = lineType,
        )
        repo.appendLine(line)
    }
    
    private fun printConsole(
        value: Any?,
        lineType: DevCommandLogType = DevCommandLogType.DEFAULT,
    ) = printConsole(value.toString(), lineType)
    
    private fun printConsole(user: UserData) =
        printConsole(
            text = user.toString(),
            lineType = when {
                !user.isOnline -> DevCommandLogType.DEFAULT
                user.status == UserStatus.AWAY -> DevCommandLogType.WARNING
                user.status == UserStatus.BUSY -> DevCommandLogType.ERROR
                else -> DevCommandLogType.HIGHLIGHT
            }
        )
    
    private fun dumpFriendList() {
        viewModelScope.launch {
            engine.getFriendList().forEach { printConsole(it) }
            printConsole("")
        }
    }
    
    private fun dumpGroupList() {
        viewModelScope.launch {
            engine.getGroupList().forEach { group ->
                val peers = engine.getGroupPeerList(group.uid)
                printConsole(group, DevCommandLogType.HIGHLIGHT)
                peers.forEach { (_, peer) -> printConsole(peer) }
                printConsole("")
            }
        }
    }
    
    private fun onThemeCommand(args: List<String>) {
        if (args.isEmpty()) {
            printConsole(settings.generalSettings.appTheme)
            return
        }
        ChatThemeSetting.entries.find { it.name == args[0].uppercase() }?.let { theme ->
            settings = settings.copy(generalSettings = settings.generalSettings.copy(appTheme = theme))
        } ?: printConsole("Unknown theme: \"${args[0]}\"")
    }
    
    private fun playSound(name: String) {
        audioPlayer.queueSound(name) ?: printConsole("Failed to play sound: $name", lineType = DevCommandLogType.ERROR)
    }
    
    private fun playAudio(path: String) {
        audioPlayer.queueFile(path.toPath()) ?: printConsole("Failed to play audio file: $path", lineType = DevCommandLogType.ERROR)
    }
    
    private inner class PrintStreamRedirect(src: PrintStream) : PrintStream(src) {
        
        private val isError = src == defaultErr
        
        override fun println(x: Any?) {
            x?.toString()?.appendToConsole()
            super.println(x)
        }

        override fun println(s: String?) {
            s?.appendToConsole()
            super.println(s)
        }
        
        private fun String.appendToConsole() {
            val isWarning = !isError && contains(" $WARNING_TEXT")
            printConsole(
                text = if (isWarning) replace(WARNING_TEXT, "") else this,
                lineType = when {
                    isError -> DevCommandLogType.ERROR
                    isWarning -> DevCommandLogType.WARNING
                    else -> DevCommandLogType.LOG
                }
            )
        }
    }
}