package nand.apps.chat.ui.menu

import androidx.compose.foundation.*
import androidx.compose.foundation.interaction.MutableInteractionSource
import androidx.compose.foundation.interaction.collectIsHoveredAsState
import androidx.compose.foundation.layout.*
import androidx.compose.material.Icon
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.ExperimentalComposeUiApi
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.shadow
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.focus.FocusManager
import androidx.compose.ui.geometry.Rect
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.input.InputMode
import androidx.compose.ui.input.InputModeManager
import androidx.compose.ui.input.key.Key
import androidx.compose.ui.input.key.KeyEventType
import androidx.compose.ui.input.key.key
import androidx.compose.ui.input.key.type
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.platform.LocalInputModeManager
import androidx.compose.ui.unit.dp
import androidx.compose.ui.window.Popup
import androidx.compose.ui.window.PopupProperties
import androidx.compose.ui.window.rememberPopupPositionProviderAtPosition
import nand.apps.chat.ui.selection.DisableSelection
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme

actual fun isCustomActionMenu() = true

@Composable
actual fun ActionMenuContent(
    items: () -> List<ActionMenuItem>,
    isEnabled: Boolean,
    modifier: Modifier,
    content: @Composable () -> Unit,
) {
    val state = remember { ContextMenuState() }
    val highlightColor = if (state.status is ContextMenuState.Status.Open) ChatTheme.colors.primaryVariant else Color.Unspecified
    Box(
        modifier = modifier
            .background(highlightColor)
            .contextMenuOpenDetector(state, isEnabled),
        propagateMinConstraints = true,
    ) {
        content()
        DisableSelection {
            LocalContextMenuRepresentation.current.Representation(
                state = state,
                items = {
                    val menuItems = items()
                    if (menuItems.isEmpty()) {
                        state.status = ContextMenuState.Status.Closed
                    }
                    menuItems.map { ContextMenuIconRow(it) }
                },
            )
        }
    }
}

@OptIn(ExperimentalFoundationApi::class)
private fun Modifier.contextMenuOpenDetector(
    state: ContextMenuState,
    enabled: Boolean = true
): Modifier = this.contextMenuOpenDetector(
    key = state,
    enabled = enabled && (state.status is ContextMenuState.Status.Closed),
) { pointerPosition ->
    state.status = ContextMenuState.Status.Open(Rect(pointerPosition, 0f))
}

class ContextMenuIconRepresentation : ContextMenuRepresentation {

    @OptIn(ExperimentalComposeUiApi::class)
    @Composable
    override fun Representation(state: ContextMenuState, items: () -> List<ContextMenuItem>) {
        val status = state.status
        if (status is ContextMenuState.Status.Open) {
            var focusManager: FocusManager? by mutableStateOf(null)
            var inputModeManager: InputModeManager? by mutableStateOf(null)

            Popup(
                properties = PopupProperties(focusable = true),
                popupPositionProvider = rememberPopupPositionProviderAtPosition(status.rect.center),
                onDismissRequest = {
                    state.status = ContextMenuState.Status.Closed
                },
                onKeyEvent = {
                    if (it.type == KeyEventType.KeyDown) {
                        when (it.key) {
                            Key.DirectionDown  -> {
                                inputModeManager?.requestInputMode(InputMode.Keyboard)
                                focusManager?.moveFocus(FocusDirection.Next)
                                true
                            }
                            Key.DirectionUp -> {
                                inputModeManager?.requestInputMode(InputMode.Keyboard)
                                focusManager?.moveFocus(FocusDirection.Previous)
                                true
                            }
                            else -> false
                        }
                    } else {
                        false
                    }
                },
            ) {
                focusManager = LocalFocusManager.current
                inputModeManager = LocalInputModeManager.current
                Column(
                    modifier = Modifier
                        .shadow(8.dp)
                        .background(ChatTheme.colors.primary)
                        .padding(vertical = ChatTheme.dimens.spacingSmall)
                        .width(IntrinsicSize.Max)
                        .verticalScroll(rememberScrollState())

                ) {
                    items().forEach { item ->
                        MenuItemContent(
                            itemHoverColor = ChatTheme.colors.surface3,
                            onClick = {
                                state.status = ContextMenuState.Status.Closed
                                item.onClick()
                            }
                        ) {
                            if (item is ContextMenuIconRow) {
                                Icon(
                                    imageVector = item.action.icon,
                                    contentDescription = item.label,
                                    tint = ChatTheme.colors.onPrimary,
                                )
                            }
                            TextLine(
                                text = item.label,
                                style = ChatTheme.typography.body2,
                                color = ChatTheme.colors.onPrimary,
                            )
                        }
                    }
                }
            }
        }
    }
}

private class ContextMenuIconRow(
    val action: ActionMenuItem,
) : ContextMenuItem(action.label, action.onClick)

@Composable
private fun MenuItemContent(
    itemHoverColor: Color,
    onClick: () -> Unit,
    content: @Composable RowScope.() -> Unit
) {
    val interactionSource = remember { MutableInteractionSource() }
    val isHovered by interactionSource.collectIsHoveredAsState()
    Row(
        modifier = Modifier
            .clickable(onClick = onClick)
            .hoverable(interactionSource)
            .background(if (isHovered) itemHoverColor else Color.Transparent)
            .fillMaxWidth()
            .sizeIn(
                minWidth = 112.dp,
                maxWidth = 280.dp,
                minHeight = ChatTheme.dimens.iconMedium,
            )
            .padding(
                horizontal = ChatTheme.dimens.spacingMedium,
            ),
        verticalAlignment = Alignment.CenterVertically,
        horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
        content = content,
    )
}