package nand.apps.chat.ui.notification

import androidx.compose.animation.*
import androidx.compose.animation.core.MutableTransitionState
import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.LazyItemScope
import androidx.compose.foundation.lazy.items
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Close
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.awt.ComposeWindow
import androidx.compose.ui.draw.clip
import androidx.compose.ui.unit.DpRect
import androidx.compose.ui.unit.DpSize
import androidx.compose.ui.unit.dp
import androidx.compose.ui.window.DialogWindow
import androidx.compose.ui.window.WindowPosition
import androidx.compose.ui.window.rememberDialogState
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import kotlinx.coroutines.delay
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.audio.AudioPlayer
import nand.apps.chat.model.notification.CallNotification
import nand.apps.chat.model.notification.ChatNotificationGroup
import nand.apps.chat.model.notification.MessageNotification
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dispatcher.DesktopKeyEventDispatcher
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.ChatAppSurface
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.window.getMaxWindowBoundsDp
import nand.apps.chat.ui.window.windowIconPainter
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*
import java.awt.GraphicsEnvironment

private val DIALOG_WIDTH = 384.dp
private val DIALOG_ROW_HEIGHT = 96.dp
private val DIALOG_ROW_SPACING = 4.dp
private val DIALOG_TOP_MARGIN = 40.dp
private const val SHRINK_DIALOG_DELAY = 250L

@Composable
fun DesktopNotificationHost(
    window: ComposeWindow?,
    viewModel: DesktopNotificationViewModel = koinKmpViewModel(),
) {
    val state by viewModel.state.collectAsState()
    val settingsRepo: SettingsRepo by getKoin().inject()
    val audioPlayer: AudioPlayer by getKoin().inject()
    val scale = settingsRepo.settings.generalSettings.densityScale
    val notificationSettings = settingsRepo.settings.notificationSettings
    val keyEventDispatcher = remember { DesktopKeyEventDispatcher() }
   
    // Notification visibility transitions
    var isDialogVisible by remember { mutableStateOf(false) }
    val isGroupVisible = remember { ChatNotificationGroup.entries.associateWith { MutableTransitionState(false) } }
    var cachedVisibleCount by remember { mutableStateOf(0) }
    val screenDevice = window?.graphicsConfiguration?.device ?: GraphicsEnvironment.getLocalGraphicsEnvironment().defaultScreenDevice
    val screenBounds = screenDevice.getMaxWindowBoundsDp()
    val dialogState = rememberDialogState(
        size = getDialogSize(
            numRows = 1,
            scale = scale,
        ),
        position = getDialogPosition(
            screenBounds = screenBounds,
            scale = scale,
        ),
    )
    
    LaunchedEffect(Unit) {
        viewModel.actionData.collect { action ->
            when (action) {
                is NotificationUpdate -> {
                    action.show.forEach { isGroupVisible[it]?.targetState = true }
                    action.dismiss.forEach { isGroupVisible[it]?.targetState = false }
                    if (action.show.isNotEmpty()) {
                        isDialogVisible = true
                        val dialogSize = getDialogSize(
                            numRows = isGroupVisible.values.filter { it.isVisible }.size,
                            scale = scale,
                        )
                        if (dialogSize.height > dialogState.size.height) dialogState.size = dialogSize
                        if (notificationSettings.isNotificationSoundEnabled) {
                            audioPlayer.queueSound("notification")
                        }
                    }
                }
            }
        }
    }
    
    // Re-position notification dialog whenever the chat window is moved to another screen
    // or the screen size is changed
    LaunchedEffect(screenBounds) {
        dialogState.position = getDialogPosition(
            screenBounds = screenBounds,
            scale = scale,
        )
    }

    val visibleStates = isGroupVisible.mapNotNull { (group, visibleState) ->
        if (visibleState.isVisible) group to visibleState else null
    }
    val visibleCount = isGroupVisible.count { it.value.isVisible }
    LaunchedEffect(visibleCount) {
        val oldVisibleCount = cachedVisibleCount
        cachedVisibleCount = visibleCount
        when {
            visibleCount == 0 -> isDialogVisible = false
            visibleCount < oldVisibleCount -> {
                delay(SHRINK_DIALOG_DELAY)
                val dialogSize = getDialogSize(
                    numRows = visibleCount,
                    scale = scale,
                )
                if (dialogSize != dialogState.size) dialogState.size = dialogSize
            }
        }
    }

    // Using the 'visible' param below causes some strange inconsistent behavior between platforms
    // when toggling visibility of these dialog windows, such as displaying them as a regular window.
    if (!isDialogVisible) return
    
    DialogWindow(
        state = dialogState,
        title = stringResource(Res.string.notification_host_title, APP_NAME),
        icon = windowIconPainter(hasUnreadMessages = true),
        transparent = true,
        undecorated = true,
        resizable = false,
        focusable = false,
        alwaysOnTop = true,
        onPreviewKeyEvent = keyEventDispatcher::onKeyEvent,
        onCloseRequest = {},
    ) {
        ChatAppSurface(
            settings = settingsRepo.settings,
            includeBackground = false,
        ) {
            keyEventDispatcher.CompositionLocalProvider {
                LazyColumn(
                    verticalArrangement = Arrangement.spacedBy(DIALOG_ROW_SPACING),
                    modifier = Modifier.fillMaxSize(),
                ) {
                    items(
                        items = visibleStates,
                        key = { it.first },
                    ) { (group, visibleState) ->
                        NotificationContainer(
                            state = state,
                            group = group,
                            visibleState = visibleState,
                            viewModel = viewModel,
                        )
                    }
                }
            }
        }
    }
}

@OptIn(ExperimentalFoundationApi::class)
@Composable
private fun LazyItemScope.NotificationContainer(
    state: DesktopNotificationState,
    group: ChatNotificationGroup,
    visibleState: MutableTransitionState<Boolean>,
    viewModel: DesktopNotificationViewModel,
) {
    val notification = state.notifications[group]
    AnimatedVisibility(
        visibleState = visibleState,
        enter = fadeIn() + slideInHorizontally { it },
        exit = fadeOut() + slideOutHorizontally { it },
        modifier = Modifier
            .fillMaxWidth()
            .height(DIALOG_ROW_HEIGHT)
            .animateItemPlacement(),
    ) {
        Box(
            modifier = Modifier
                .clip(ChatTheme.shapes.small)
                .background(ChatTheme.colors.backgroundFloating)
                .fillMaxSize()
                .onClick(group == ChatNotificationGroup.MESSAGES) {
                    notification?.let(viewModel::onNotificationClick)
                },
        ) {
            Box(
                modifier = Modifier
                    .fillMaxSize()
                    .padding(ChatTheme.dimens.spacingMedium),
            ) {
                when (notification) {
                    is MessageNotification -> MessageNotificationRenderer(
                        notification = notification,
                        group = state.groups[notification.message.conversationUid],
                    )

                    is CallNotification -> {
                        state.calls[notification.contactUid]?.takeIf { it.isInProgress }
                            ?.let { call ->
                                CallNotificationRenderer(
                                    notification = notification,
                                    call = call,
                                )
                            } ?: viewModel.onNotificationDismiss(notification)
                    }

                    else -> Unit
                }
            }
            IconButton(
                icon = Icons.Outlined.Close,
                contentDescription = stringResource(Res.string.close),
                modifier = Modifier.align(Alignment.TopEnd),
                onClick = { notification?.let(viewModel::onNotificationDismiss) },
            )
        }
    }
}

private val MutableTransitionState<Boolean>.isVisible get() = currentState || targetState

private fun getDialogSize(numRows: Int, scale: Float): DpSize{
    val rowHeight = DIALOG_ROW_HEIGHT * numRows
    val spacing = DIALOG_ROW_SPACING * (numRows - 1).coerceAtLeast(0)
    return DpSize(
        width = DIALOG_WIDTH * scale,
        height = (rowHeight + spacing) * scale,
    )
}

private fun getDialogPosition(screenBounds: DpRect, scale: Float) = WindowPosition(
    x = screenBounds.right - (DIALOG_WIDTH * scale),
    y = screenBounds.top + DIALOG_TOP_MARGIN,
)