package nand.apps.chat.ui.window

import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.Density
import androidx.compose.ui.unit.DpSize
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.window.DialogWindow
import androidx.compose.ui.window.WindowPosition
import androidx.compose.ui.window.rememberDialogState
import kotlinx.coroutines.delay
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.notifications.ChatWidgetSettingsData
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.painterResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

private val WIDGET_BASE_SIZE = 36.dp

@Composable
fun ChatWidget() {
    var isVisible by rememberSaveable { mutableStateOf(false) }
    var unreadDirectMessages by rememberSaveable { mutableStateOf(0) }
    var unreadGroupMessages by rememberSaveable { mutableStateOf(0) }
    val messageRepo: ChatMessageRepo by getKoin().inject()
    val settingsRepo: SettingsRepo by getKoin().inject()
    val eventHandler: ChatEventHandler by getKoin().inject()
    val settings = settingsRepo.settings.notificationSettings.widgetSettings
    val size = WIDGET_BASE_SIZE * settings.size.scale
    val state = rememberDialogState(
        width = size,
        height = size,
        position = WindowPosition(settings.placement.toAlignment()),
    )
    
    // Update widget state when settings are changed
    LaunchedEffect(Unit) {
        var oldState: ChatWidgetSettingsData? = null
        settingsRepo.observer.collect {
            with(it.notificationSettings.widgetSettings) {
                if (oldState != this) {
                    isVisible = isEnabled
                    if (isEnabled) {
                        val sizeDp = WIDGET_BASE_SIZE * this.size.scale
                        state.size = DpSize(sizeDp, sizeDp)
                        state.position = WindowPosition(placement.toAlignment())
                    }
                    oldState = this
                }
            }
        }
    }
    
    LaunchedEffect(Unit) {
        messageRepo.unreadObserver.collect { unreadStates ->
            unreadDirectMessages = unreadStates.values.sumOf { it.directMessages }
            unreadGroupMessages = unreadStates.values.sumOf { it.groupMessages }
        }
    }

    val defaultDensity = LocalDensity.current

    if (!isVisible) return
    
    DialogWindow(
        state = state,
        title = stringResource(Res.string.chat_widget_title, APP_NAME),
        transparent = true,
        undecorated = true,
        resizable = false,
        focusable = false,
        alwaysOnTop = true,
        onCloseRequest = {},
    ) {
        // Flash the widget icon
        var isHighlighted by remember { mutableStateOf(false) }
        val unreadCount = remember(settings, unreadDirectMessages, unreadGroupMessages) {
            unreadDirectMessages + when {
                settings.isGroupUnreadEnabled && settings.isGroupUnreadNumberShown -> unreadGroupMessages
                else -> 0
            }
        }
        LaunchedEffect(unreadCount > 0 && settings.isFlashEnabled) {
            while (unreadCount > 0 && settings.isFlashEnabled) {
                isHighlighted = !isHighlighted
                delay(1000)
            }
            isHighlighted = false
        }
        ChatTheme {
            CompositionLocalProvider(
                LocalDensity provides Density(defaultDensity.density * settings.size.scale)
            ) {
                Box(
                    modifier = Modifier
                        .clip(ChatTheme.shapes.medium)
                        .background(ChatTheme.colors.backgroundFloating)
                        .clickable {
                            eventHandler += RequestAppFocusEvent(toggle = true)
                        }
                        .fillMaxSize()
                        .padding(5.dp),
                ) {
                    Image(
                        painter = painterResource(
                            when {
                                isHighlighted -> Res.drawable.ic_seers_highlighted
                                else -> Res.drawable.ic_seers
                            }
                        ),
                        contentDescription = stringResource(Res.string.chat_widget_content_description),
                        modifier = Modifier.fillMaxSize(),
                    )
                    val badgeText = when {
                        unreadCount > 0 -> unreadCount.toString()
                        unreadGroupMessages > 0 && settings.isGroupUnreadEnabled -> "٭"
                        else -> ""
                    }
                    if (badgeText.isNotEmpty()) {
                        TextLine(
                            text = badgeText,
                            color = Color.White,
                            style = ChatTheme.typography.overline.copy(lineHeight = 0.sp),
                            textAlign = TextAlign.Center,
                            modifier = Modifier
                                .clip(CircleShape)
                                .background(Color.Red)
                                .padding(start = (1.5).dp)
                                .align(Alignment.BottomEnd),
                        )
                    }
                }
            }
        }
    }
}