package nand.apps.chat.ui.window

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.verticalScroll
import androidx.compose.material.*
import androidx.compose.runtime.Composable
import androidx.compose.runtime.NonRestartableComposable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.input.pointer.PointerIcon
import androidx.compose.ui.input.pointer.pointerHoverIcon
import androidx.compose.ui.platform.LocalClipboardManager
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.TextUnit
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.window.WindowPosition
import androidx.compose.ui.window.WindowState
import androidx.compose.ui.window.singleWindowApplication
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.net.DesktopUriLauncher
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.theme.DarkChatColors

private const val EMOJI = "\uD83E\uDD2F"
private const val ISSUES_URL = "https://gitlab.com/nand-apps/seers-lodge/-/issues"

fun appCrashWindow(
    error: Throwable,
) = singleWindowApplication(
    title = "$APP_NAME Crashed! $EMOJI",
    state = WindowState(
        width = 1000.dp,
        height = 650.dp,
        position = WindowPosition.Aligned(Alignment.Center),
    ),
) {
    val errorText = error.stackTraceToString().replace("\t", "    ")
    val clipboardManager = LocalClipboardManager.current
    Surface(
        modifier = Modifier
            .background(DarkChatColors.background)
            .fillMaxSize()
            .padding(horizontal = 16.dp)
    ) {
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            modifier = Modifier
                .background(DarkChatColors.background)
                .fillMaxWidth()
                .verticalScroll(rememberScrollState())
                .padding(vertical = 16.dp),
        ) {
            CenteredText(
                text = EMOJI,
                fontSize = 72.sp,
            )
            CenteredText(
                text = "$APP_NAME has crashed!",
                fontSize = 36.sp,
            )
            CenteredText(
                text = "My sincerest apologies! I screwed up big time.",
                fontSize = 20.sp,
            )
            CenteredText(
                text = "The wife's not gonna let me hear the end of this...",
                fontSize = 12.sp,
                color = DarkChatColors.onBackground,
                modifier = Modifier.padding(top = 8.dp),
            )
            CenteredText(
                text = "The offending error is displayed below. Please screenshot or copy it and report here:",
                fontSize = 16.sp,
                modifier = Modifier.padding(top = 24.dp),
            )
            CenteredText(
                text = ISSUES_URL,
                fontSize = 18.sp,
                color = DarkChatColors.primary,
                modifier = Modifier
                    .padding(top = 4.dp)
                    .onClick { DesktopUriLauncher().launchUri(ISSUES_URL) },
            )
            Button(
                colors = ButtonDefaults.buttonColors(
                    backgroundColor = DarkChatColors.primary,
                ),
                onClick = {
                    clipboardManager.setText(AnnotatedString(errorText))
                },
                modifier = Modifier
                    .pointerHoverIcon(PointerIcon.Hand)
                    .padding(top = 16.dp),
            ) {
                CenteredText(
                    text = "Copy Error to Clipboard",
                    fontSize = 16.sp,
                )
            }
            Text(
                text = errorText,
                color = DarkChatColors.onBackground,
                modifier = Modifier.padding(top = 8.dp),
            )
        }
    }
}

@NonRestartableComposable
@Composable
private fun CenteredText(
    text: String,
    fontSize: TextUnit,
    modifier: Modifier = Modifier,
    color: Color = DarkChatColors.onSurface,
) = Text(
    text = text,
    fontSize = fontSize,
    color = color,
    textAlign = TextAlign.Center,
    modifier = modifier,
)