package nand.apps.chat.ui.window

import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.input.key.*
import androidx.compose.ui.window.WindowPlacement
import androidx.compose.ui.window.WindowState
import kotlinx.coroutines.delay
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.dev.DevConsoleWindow
import nand.apps.chat.ui.dispatcher.DesktopKeyEventDispatcher
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.ui.modifier.toShortcutAction
import org.koin.compose.getKoin

@Composable
fun rememberShortcutListener(
    windowState: WindowState,
    keyEventDispatcher: DesktopKeyEventDispatcher,
): (KeyEvent) -> Boolean {
    val eventHandler: ChatEventHandler by getKoin().inject()
    val settingsRepo: SettingsRepo by getKoin().inject()
    var previousWindowPlacement by remember { mutableStateOf(WindowPlacement.Floating) }
    var nextWindowPlacement by remember { mutableStateOf(WindowPlacement.Floating) }
    var isDevConsoleVisible by rememberSaveable { mutableStateOf(false) }

    DevConsoleWindow(
        isVisible = isDevConsoleVisible,
        windowSize = windowState.size,
        onClose = {
            isDevConsoleVisible = false
            eventHandler += RequestAppFocusEvent()
        },
    )
    
    LaunchedEffect(nextWindowPlacement) {
        if (windowState.placement == WindowPlacement.Fullscreen && nextWindowPlacement == WindowPlacement.Maximized) {
            // XXX - Transitioning directly from fullscreen to maximized has no effect,
            // so we need to do this ugly hack
            windowState.placement = WindowPlacement.Floating
            delay(100L)
        }
        windowState.placement = nextWindowPlacement
    }
    
    return { event ->
        keyEventDispatcher.onKeyEvent(event) || if (event.type == KeyEventType.KeyDown) {
            when (val action = event.toShortcutAction(settingsRepo.settings.shortcutSettings)) {
                ShortcutAction.NAVIGATE_BACK -> {
                    keyEventDispatcher.navigateBack()
                    true
                }
                ShortcutAction.TOGGLE_DEV_CONSOLE -> {
                    if (isDevConsoleVisible || settingsRepo.settings.devSettings.isDevConsoleEnabled) {
                        isDevConsoleVisible = !isDevConsoleVisible
                    }
                    true
                }
                ShortcutAction.TOGGLE_FULLSCREEN -> {
                    nextWindowPlacement = when (windowState.placement) {
                        WindowPlacement.Fullscreen -> {
                            if (previousWindowPlacement != WindowPlacement.Fullscreen) {
                                previousWindowPlacement
                            } else {
                                // Safeguard against getting stuck in fullscreen
                                WindowPlacement.Maximized
                            }
                        }
                        else -> {
                            previousWindowPlacement = windowState.placement
                            WindowPlacement.Fullscreen
                        }
                    }
                    true
                }
                ShortcutAction.HIDE_WINDOW -> {
                    eventHandler += RequestAppFocusEvent(toggle = true)
                    true
                }
                else -> action?.let { keyEventDispatcher.onShortcutAction(it) } ?: false
            }
        } else {
            false
        }
    }
}