# Build Instructions

### Requirements
- [Git LFS](https://github.com/git-lfs/git-lfs#installing)
- [Java JDK 21](https://openjdk.org/projects/jdk/21/)
- [Gradle](https://docs.gradle.org/current/userguide/installation.html)
- [CMake](https://cmake.org/download/) (3.16.3 or above)
- [Android SDK command line tools](https://developer.android.com/studio)
    - You'll need `sdkmanager` specifically, unless you already have Android Studio installed

Install Git LFS and clone the repo:
```shell
git clone https://gitlab.com/nand-apps/seers-lodge
cd seers-lodge
```

**Note:** Regardless of whether you're building the Android app or not, you'll need to define an Android home directory.
You can simply create a `local.properties` file in the root of the project and define the location using the `sdk.dir` property.

```properties
sdk.dir=<your home directory>/Android
```

You'll also need to accept the Android SDK licenses using `sdkmanager`, which is part of the command line tools mentioned in the requirements section.
```shell
sdkmanager --licenses --sdk_root=<Android home directory>
```

## Desktop

There are a few different tasks that need to be run to build the desktop app.

Typically, you'll need to:
- Install required build packages using your system's package manager (i.e. APT, Pacman, Brew, etc.)
- Run the `buildDesktopNativeLibraries` and `buildDesktopJniLibraries` tasks in Gradle to generate the native libraries the app requires
- Run one of the `packageRelease*` tasks to generate an installer

Each package task is unique to your operating system, but if you only want to run the app you can do:
```shell
./gradlew desktopRun -DmainClass=MainKt
```

**See below for build instructions specific to each platform.**

### Linux

The most straightforward of the three platforms. First install the following packages, if you don't have them already.

APT example:
```shell
sudo apt install git git-lfs openjdk-21-jdk autoconf automake libtool make cmake \
         wget tar unzip gcc g++ pkg-config ninja-build yasm dpkg-dev
```

Run these tasks to build the native libraries:

```shell
./gradlew buildDesktopNativeLibraries buildDesktopJniLibraries
```

Then run this task to generate a .deb installer:

```shell
./gradlew packageReleaseDeb
```

The installer can then be found under: `composeApp/build/compose/binaries/main-release/deb/SeersLodge-<version>.deb`

You can also generate an .rpm installer using the `packageReleaseRpm` task. This may require installing the `rpm` package.

### macOS

Install the [Homebrew](https://formulae.brew.sh/) package manager if you don't have it already.

The instructions here are similar to Linux, however the `package` task may fail unless you point Gradle to the [AWS Corretto JDK](https://formulae.brew.sh/cask/corretto@21#default) installed via Brew (see [this issue](https://github.com/JetBrains/compose-multiplatform/issues/3107)).

Then set `org.gradle.java.home` in your `local.properties` to the installed location of this JDK.

```shell
./gradlew buildDesktopNativeLibraries buildDesktopJniLibraries
```

For macOS, run the `packageReleaseDmg` task instead to generate an app installer:

```shell
./gradlew packageReleaseDmg
```

The installer can then be found under: `composeApp/build/compose/binaries/main-release/dmg/SeersLodge-<version>.dmg`

You can also generate a .pkg installer using the `packageReleasePkg` task.

**Note**: The machine you're building on will only produce installers compatible with other machines that use the same
CPU architecture. Newer macs use Apple CPUs (ARM) while older macs use Intel CPUs (x64).

### Windows

This is a bit more involved than the other platforms...

First you need to install [MSYS2](https://www.msys2.org/). This is a package manager and shell that allows you to
execute Unix commands required to build the native libraries. It's recommended to use the `MINGW64` environment for
these instructions, but you can probably use the other environments if you know what you're doing.

Open the `MINGW64.exe` shell in the `C:\msys64` directory and enter the following commands to install the required packages:

```shell
pacman -S git \
    mingw-w64-x86_64-autotools \
    mingw-w64-x86_64-gcc \
    mingw-w64-x86_64-make \
    mingw-w64-x86_64-cmake \
    mingw-w64-x86_64-yasm \
    mingw-w64-x86_64-ninja
```

Then navigate to the `seers-lodge` directory where you cloned the app source code and run the following:

```shell
./gradlew buildDesktopNativeLibraries buildDesktopJniLibraries \
    -PnativeBuildMingwPath=/mingw64 \
    -PnativeBuildBashExe=/usr/bin/bash.exe \
    -PnativeBuildTarget=x86_64-w64-mingw32 \
    -PnativeBuildCygRoot=/ \
    -PnativeBuildCmakeJni=true \
    -PnativeBuildOpenMp=true
```

If you're wondering what all those extra arguments are:
- `-PnativeBuildMingwPath=<path>` Path to your environment's system root (see the prefix column [here](https://www.msys2.org/docs/environments/))
- `-PnativeBuildBashExe=<exe path>` Path to `bash.exe` for executing commands
- `-PnativeBuildTarget=<target>` Build target; Shouldn't have to change this
- `-PnativeBuildCygRoot=<path>` Root of the system; Usually either `/` or `/cygdrive` if using Cygwin
- `-PnativeBuildCmakeJni=<flag>` True to override CMake's default Java configuration
- `-PnativeBuildOpenMp=<flag>` True to link OpenMP to the Soxr library (used for audio resampling)

Once that's finished (without errors hopefully) then just run one more gradle command to generate a .msi installer:

```shell
./gradlew packageReleaseMsi
```

The installer can then be found under: `composeApp/build/compose/binaries/main-release/msi/SeersLodge-<version>.msi`

If you run into problems you can try using the `Build Windows` [CI task](../.gitlab-ci.yml) as a reference. This file
contains commands for building the app and installer "from scratch" in Powershell.

## Android

To build the Android app you'll first need to compile the native libraries using the task below (Linux only):

```shell
./gradlew buildAndroidNativeLibraries
```

Then you can generate the APK by running:

```shell
./gradlew assembleDebug
```

If all goes well, the generated APK can be found under: `composeApp/build/outputs/apk/debug/SeersLodge-<version>.apk`

### Release APK

If you want to generate a release APK you'll need to [generate a keystore](https://developer.android.com/studio/publish/app-signing) and define a few environment variables:

- `ANDROID_KEY_STOREFILE` The path to the keystore (.jks) file
- `ANDROID_KEY_ALIAS` Keystore alias (usually "release")
- `ANDROID_KEYSTORE_PASSWORD` The password to the keystore

Then simply run:
```shell
./gradlew assembleRelease
```